!-----------------------------------------------------------------------
!* This module contains physical constants and mathematical routines.
!-----------------------------------------------------------------------
module PhysMath

!-----------------------------------------------------------------------
!* authors:                
!*
!* Alex Ignacio da Silva        ignacioalex@yahoo.com
!* and
!* Sueli M. Viegas              viegas@iagusp.usp.br
!*
!* Instituto Astronomico e Geofisico
!* Universidade de Sao Paulo
!* Av. Miguel Stefano, 4200
!* 04301-904 Sao Paulo - SP
!* Brazil
!*
!* History:
!* Version    Date (D/M/Y)    Comment
!* -------    ------------    -------
!* 1.0        28/04/2000      original code (A.I. Silva)
!-----------------------------------------------------------------------

use Precision

implicit none

public

!***************************
!* a few useful constants: *
!***************************

! it won't use all the digits, of course!
real(WP), parameter :: pi = 3.141592653589793238462643383279502884197_WP

! Physical constants taken from the Particle Data Group.
! C. Caso et al, 1998: European Physical Journal C3, 1
! (Based on the 1998 CODATA recommended values.)
real(WP), parameter :: k = 1.3806503E-16_WP    ! Boltzmann's constant, in erg/K
real(WP), parameter :: c = 2.99792458E10_WP    ! speed of light, in cm/s
real(WP), parameter :: h = 6.62606876E-27_WP   ! Planck's constant, in erg.s
real(WP), parameter :: m = 9.10938188E-28_WP   ! electron mass, in g
real(WP), parameter :: q = 1.602176462E-19_WP  ! electron charge, in C

contains

! *******************************************************
! *------- swaps any two integer numbers A <-> B -------*
! *******************************************************
subroutine swap(A,B)
integer, intent(inout) :: A, B
integer :: C

C = A
A = B
B = C

end subroutine swap

! ***************************************************************************
! *------- Evaluates a polynomial Y = a0 + a1*X + a2*X^2 + a3*X^3... -------*
! ***************************************************************************
function pol(order,coefficients,X)
real(WP) :: pol
integer, intent(in) :: order    ! order of the polynomial
real(WP), intent(in) :: coefficients(:)    ! (/ a0, a1, a2... /) (1:order+1)
real(WP), intent(in) :: X

integer :: i    ! counter

if (size(coefficients)/=order+1) then
    print *, "Error in calling function pol!"
    print *, "The number of polynomial coefficients supplied does not equal order+1."
    stop
end if

pol = coefficients(order+1)
do i=1, order
    pol = pol*X + coefficients(order-i+1)
end do

end function pol

! *************************************************************
! *------- Interpolates linearly a given data set Y(X) -------*
! *************************************************************
function LInterpol(X,Y,x0)
real(WP) :: LInterpol    ! value of Y(x0) interpolated
real(WP), intent(in) :: X(:)  ! vector containing the abscissa points
real(WP), intent(in) :: Y(:)  ! vector containing the ordinate points
real(WP), intent(in) :: x0    ! value to be interpolated

integer :: npoints   ! number of points supplied
integer :: i,j,k

npoints = size(X)

if (size(Y)/=npoints) then
    print *, "Error in calling function LInterpol!"
    print *, "Number of abscissa points differ from number of ordinate points."
    stop
end if

if (npoints<2) then
    print *, "Error in calling function LInterpol!"
    print *, "Less than 2 points were supplied."
    stop
end if

if (x0>=X(1) .and. x0<=X(npoints)) then
    i = 1
    j = npoints
    do while ( (j-i)>1 )
        k = (i+j) / 2
        if ( X(k)<=x0 ) then
            i = k
        else
            j = k
        end if
    end do
    LInterpol = Y(i) + (x0-X(i))*(Y(j)-Y(i))/(X(j)-X(i))
else
    print *, "Error in calling function LInterpol!"
    print *, "The value to be linearly interpolated lies outside &
                &the bounds allowed by the data set."
    stop
end if

end function LInterpol

end module PhysMath
