module PopRatio
!-----------------------------------------------------------------------
!* authors:    
!*
!* Alex Ignacio da Silva        ignacioalex@yahoo.com
!* and
!* Sueli M. Viegas              viegas@iagusp.usp.br
!*
!* Instituto Astronomico e Geofisico
!* Universidade de Sao Paulo
!* Av. Miguel Stefano, 4200
!* 04301-904 Sao Paulo - SP
!* Brazil
!*
!* This module contains routines to calculate the population ratios of
!* fine-structure levels of a given atom or ion.
!*
!* Atomic data is read from a separate file.
!*
!* History:
!* Version    Date (D/M/Y)    Comment
!* -------    ------------    -------
!* 1.0        28/04/2000      original code (A.I. Silva)
!-----------------------------------------------------------------------

use F90_LAPACK
use Spline
use Precision
use PhysMath
use FieldIntensities

implicit none

public

!***************************
!* a few useful constants: *
!***************************

real(WP), parameter :: constant = 1.0_WP/8.0_WP/pi/h    ! useful for calculating
                                                        ! Einstein's B coefficient

real(WP) :: const2    ! useful for calculating electron collision rates
                      ! It will be initialized.

real(WP), parameter :: T0 = 2.728_WP    ! today's CMBR temperature, in K

!*************
!* new types *
!*************

type Table
    integer :: npoints, level1, level2
    real(WP), pointer :: X(:), Y(:)
end type

type TCollisionalData
    character(len=10) :: partner_name    ! name of collision partner
    integer :: maxlevel
    integer :: ninterpol
    type (Table), pointer :: Values(:)
    real(WP), pointer :: q(:,:)
end type

type Thlevel
    logical :: is_included
    integer :: w    ! statistical weight
    real(WP) :: En  ! energy, in cm^-1
    real(WP), pointer :: Aground(:), Kground(:)
end type

!*************
!* variables *
!*************

integer, private :: c1, c2    ! counters
! (shared by subroutine PopulationRatio and function Rate_Extrapol)

integer :: nlevels    ! number of levels taken into account
integer :: nground    ! number of levels that belong to the ground term
integer :: nhlevels   ! number of higher-energy levels (those connected to the ground term
                                                                        ! by fluorescence)
integer :: npartners    ! number of collision partners
character(len=10) :: species_name    ! name of atom or ion being calculated

integer, pointer :: g(:)     ! statistical weights
real(WP), pointer :: E(:)    ! energies
real(WP), pointer :: A(:,:)  ! transition probabilities
real(WP), pointer :: B(:,:)  ! Einstein's B coefficients
real(WP), pointer :: u(:,:)  ! energy densities
real(WP), pointer :: gam(:,:), Totgam(:,:)    ! indirect excitation rates
real(WP), pointer :: matrix(:,:) ! matrix of the linear system
real(WP), pointer :: indep(:)    ! independent terms of the linear system
real(WP), pointer :: X(:)        ! solution of the linear system

type (Thlevel), pointer :: hlevel(:)
type (TCollisionalData), pointer :: CD(:)

! variables that store the values of T, z, beta, f, n in last call to PopulationRatio
real(WP), private :: Told, zold, betaold, fold
real(WP), pointer, private :: nold(:)

contains

! ************************************************
! *------- skips the comments from a file -------*
! ************************************************
subroutine skip_comments (number)
integer, intent(in) :: number    ! unit number of the file

character :: symbol    ! symbol that indicates the end of comments

read (number,*) symbol
do while (symbol /= "#")
    read (number,*) symbol
end do

end subroutine skip_comments

! ****************************************************************
! *------- reads the atomic data and initialize variables -------*
! ****************************************************************
subroutine InitPopRatio(unit,nblevels,lambdai,lambdaf)
integer, intent(in) :: unit    ! unit number of atomic data file (opened externaly)
integer, intent(in) :: nblevels    ! number of levels to be taken into account
real(WP), optional, intent(in) :: lambdai, lambdaf
                    ! among the higher levels connected to the ground term,
                    ! include only those that are connected via multiplets between lambdai
                    ! and lambdaf (in )

real(WP) :: lambda1, lambda2    ! aliases of lambdai and lambdaf

integer :: i, j, int1, int2, ntransitions 
integer :: nmax
integer :: nread, npartial, ndata, ntread
real(WP) :: tprob, energy, weight
real(WP), pointer :: buffer(:,:)
logical :: morethanone

const2 = h*h/(2.0_WP*pi*m)**1.5_WP/sqrt(k)    ! initialization of the 'constant'

call skip_comments(unit)
read (unit,*) species_name
call skip_comments(unit)
read (unit,*) nground
if (nblevels<nground) nground=nblevels
allocate ( gam(nground,nground) )
allocate ( Totgam(nground,nground) )
call skip_comments(unit)
read (unit,*) nmax
if (nblevels<=nmax) then
    nlevels = nblevels
else
    print *, "The number of levels included can't be higher than ", nmax
    print *, "Include the necessary atomic data in ", species_name, &
                    & "atomic data file if you want to take more levels into account"
    stop
end if
allocate ( g(nlevels) )
allocate ( E(nlevels) )
allocate ( A(nlevels,nlevels) )
allocate ( B(nlevels,nlevels) )
allocate ( u(nlevels,nlevels) )
allocate ( matrix(nlevels-1,nlevels-1) )
allocate ( indep(nlevels-1) )
allocate ( X(nlevels-1) )
call skip_comments(unit)
do i=1, nlevels
    read (unit,*) E(i),g(i)
end do

!-----------------------------------------------------------------------
!* section [1]: Initializes Einstein coefficients.
!-----------------------------------------------------------------------
call skip_comments(unit)
read (unit,*) ntransitions
A(:,:) = 0.0_WP
B(:,:) = 0.0_WP
do i=1, ntransitions
    read (unit,*) int1, int2, tprob
    if ((int1<=nlevels) .and. (int2<=nlevels)) then
        if (int1<int2) then
            A(int2,int1) = tprob
            B(int2,int1) = constant*A(int2,int1)/(E(int2)-E(int1))**3.0_WP    ! E in cm^-1
            B(int1,int2) = g(int2)*B(int2,int1)/g(int1)
        else
            print *, "Error in ", species_name, "atomic data file:"
            print *, "j>=i while reading transition probabilities Aij!! :", int1, int2, tprob
            stop
        end if
    end if
end do

!-----------------------------------------------------------------------
!* section [2]: Selects the relevant higher-levels to account for fluorescence.
!-----------------------------------------------------------------------
if (.not. present(lambdai)) then
    lambda1 = 0.0_WP
else
    lambda1 = lambdai
end if
if (.not. present(lambdaf)) then
    lambda2 = huge(1.0_WP)
else
    lambda2 = lambdaf
end if
if (lambda2<lambda1) then
    write (*,*) "Error in calling INITPOPRATIO: lambdaf < lambdai."
    stop
end if
call skip_comments(unit)
nhlevels = 0
read (unit,*) ntransitions
if (ntransitions>1) then    ! if there are less than 2 transitions listed, skips this block
int1 = nground + 1
int2 = 0 
do while (int1>nground .and. int2<ntransitions)
    read (unit,*) energy, weight, int1, tprob
    int2 = int2 + 1
end do
if (int2<ntransitions) then    ! proceeds only if there is some transition left to be read
allocate ( buffer(ntransitions-int2+1,nground+2) )
buffer(:,:) = 0.0_WP
buffer(1,1) = energy
buffer(1,2) = weight
buffer(1,int1+2) = tprob
morethanone = .false.
do i=1, ntransitions - int2
    read (unit,*) energy, weight, int1, tprob
    if (int1<=nground) then
        buffer(nhlevels+2,:) = 0.0_WP
        buffer(nhlevels+2,1) = energy
        buffer(nhlevels+2,2) = weight
        buffer(nhlevels+2,int1+2) = tprob
        if ( buffer(nhlevels+2,1) == buffer(nhlevels+1,1) ) then
            buffer(nhlevels+1,int1+2) = buffer(nhlevels+2,int1+2)
            morethanone = .true.
        else
            if ( (1.0E8_WP/(buffer(nhlevels+1,1)-E(1))>=lambda1) .and. &
                &(1.0E8_WP/(buffer(nhlevels+1,1)-E(nground))<=lambda2) .and. morethanone ) then
                nhlevels = nhlevels + 1
                morethanone = .false.
            else
                do j=1, nground+2
                    buffer(nhlevels+1,j) = buffer(nhlevels+2,j)
                end do
                morethanone = .false.
            end if
        end if
    end if
end do
if ( (1.0E8_WP/(buffer(nhlevels+1,1)-E(1))>=lambda1) .and. &
        &(1.0E8_WP/(buffer(nhlevels+1,1)-E(nground))<=lambda2) .and. morethanone ) &
        & nhlevels = nhlevels + 1
allocate ( hlevel(nhlevels) )
do i=1, nhlevels
    hlevel(i)%En = buffer(i,1)
    hlevel(i)%w = int( buffer(i,2) )
    if (any(E==hlevel(i)%En)) then
        hlevel(i)%is_included = .true.
    else
        hlevel(i)%is_included = .false.
    end if
    allocate ( hlevel(i)%Aground(nground) )
    allocate ( hlevel(i)%Kground(nground) )
    do j=1, nground
        hlevel(i)%Aground(j) = buffer(i,j+2)
    end do
end do
deallocate (buffer)
end if
end if

!-----------------------------------------------------------------------
!* section [3]: Initializes collisional data.
!-----------------------------------------------------------------------
call skip_comments(unit)
read (unit,*) npartners
allocate ( CD(npartners) )
do i=1, npartners
    call skip_comments(unit)
    read (unit,*) CD(i)%partner_name
    read (unit,*) CD(i)%maxlevel
    read (unit,*) CD(i)%ninterpol
    ntransitions = CD(i)%ninterpol
    if (CD(i)%maxlevel>nlevels) CD(i)%maxlevel=nlevels
    allocate ( CD(i)%Values(CD(i)%maxlevel*(CD(i)%maxlevel-1)/2) )
    allocate ( CD(i)%q(CD(i)%maxlevel,CD(i)%maxlevel) )
    nread = 0
    ntread = 0
    do while(nread<ntransitions)
        read (unit,*) npartial, ndata
        allocate ( buffer(2,ndata) )
        read (unit,*) buffer(1,:)
        do j=1, npartial
            read (unit,*) int1, int2, buffer(2,:)
            if (int1<=nlevels .and. int2<=nlevels) then
                CD(i)%Values(ntread+1)%npoints = ndata
                allocate ( CD(i)%Values(ntread+1)%X(ndata) )
                allocate ( CD(i)%Values(ntread+1)%Y(ndata) )
                CD(i)%Values(ntread+1)%X = log10 (buffer(1,:))
                if ( CD(i)%partner_name /= 'electron' ) then
                    CD(i)%Values(ntread+1)%Y = log10 ( buffer(2,:) )
                    CD(i)%Values(ntread+1)%level1 = int1
                    CD(i)%Values(ntread+1)%level2 = int2
                else
                    CD(i)%Values(ntread+1)%Y = buffer(2,:)
                    if (int1<int2) then
                    ! this makes level1>level2, necessary for calculating DE-excitation rates
                        ! ( simply swaps values, since omega(i,j) = omega(j,i) )
                        call swap(int1,int2)
                    end if
                    CD(i)%Values(ntread+1)%level1 = int1
                    CD(i)%Values(ntread+1)%level2 = int2
                end if
                ntread = ntread + 1
            end if
        end do
        deallocate (buffer)
        nread = nread + npartial
    end do
    CD(i)%ninterpol = ntread
end do

Told = -100.0_WP
zold = -100.0_WP
betaold = -100.0_WP
fold = -100.0_WP
allocate ( nold(npartners) )
nold(:) = -100.0_WP

end subroutine InitPopRatio

! ********************************************
! *------- destroys dynamic variables -------*
! ********************************************
subroutine FinishPopRatio

integer :: i, j

deallocate (nold)

do i=npartners, 1, -1
    do j=CD(i)%ninterpol, 1, -1
        deallocate ( CD(i)%Values(j)%Y )
        deallocate ( CD(i)%Values(j)%X )
    end do
    deallocate (CD(i)%q)
    deallocate (CD(i)%Values)
end do
deallocate (CD)

do i=nhlevels, 1, -1
    deallocate (hlevel(i)%Kground)
    deallocate (hlevel(i)%Aground)
end do
deallocate (hlevel)

deallocate (X)
deallocate (indep)
deallocate (matrix)
deallocate (u)
deallocate (B)
deallocate (A)
deallocate (E)
deallocate (g)

deallocate (Totgam)
deallocate (gam)

end subroutine FinishPopRatio

! *******************************************************************************
! *------- calculates the population ratios of the fine-structure levels -------*
! *******************************************************************************
subroutine PopulationRatio(T,z,n,beta,f)
real(WP), intent(in) :: T   ! kinetic temperature, in K
real(WP), intent(in) :: z   ! redshift (enter -1._WP if you don't want
                            !           to include the CMBR)
real(WP), intent(in) :: n(:)    
                        ! array of densities of collision partners, in cm^-3
                        ! (its dimension must be equal to the number of partners
                        !   entered in atomic data file (and the densities must
                        !   be entered in same same order as they appear there))
                        ! If you want to exclude some collision partner, simply
                        ! enter n(i) = 0._WP
real(WP), optional, intent(in) :: beta    
                                    ! parameter that modifies the power-law index of T(z):
                                    ! T(z) = T0 * (1+z)^(1-beta)
real(WP), optional, intent(in) :: f ! factor that multiplies the radiation field
                                    ! in function urad

integer :: i, j, int1, int2, ntransitions
real(WP) :: beta2, f2    ! aliases of beta, f                                
real(WP) :: TheSum

!-----------------------------------------------------------------------
!* section [1]: Sets the aliases.
!-----------------------------------------------------------------------
if (.not. present(beta)) then
    beta2 = 0.0_WP
else
    beta2 = beta
end if
if (.not. present(f)) then
    f2 = 1.0_WP
else
    f2 = f
end if

!-----------------------------------------------------------------------
!* section [2]: Calculates the collision rates.
!-----------------------------------------------------------------------
if ( T/=Told .or. any(nold(:)==0.0_WP .and. n(:)/=0.0_WP) ) then ! if n(i) was 0 in last call, 
                                                        ! the collision rates
                                                        ! were not calculated.
Told = T
nold(:) = n(:)
do c1=1, npartners
    if ( n(c1) /= 0.0_WP ) then
        CD(c1)%q = 0.0_WP
        ntransitions = CD(c1)%ninterpol
        if ( CD(c1)%partner_name /= 'electron' ) then
            do c2=1, ntransitions
                int1 = CD(c1)%Values(c2)%level1
                int2 = CD(c1)%Values(c2)%level2
                if ( (log10(T)<=CD(c1)%Values(c2)%X(CD(c1)%Values(c2)%npoints)) .and. &
                            & (log10(T)>=CD(c1)%Values(c2)%X(1)) ) then
                    CD(c1)%q(int1,int2) = 10.0_WP ** &
                      & Spline3 (CD(c1)%Values(c2)%X,CD(c1)%Values(c2)%Y,log10(T))
                else
                    CD(c1)%q(int1,int2) = Rate_Extrapol(species_name,CD(c1)%partner_name,&
                                                        & int1,int2,T)
                end if
                CD(c1)%q(int2,int1) = g(int1)*exp((E(int2)-E(int1))*c*h/k/T)/g(int2) &
                                            & *CD(c1)%q(int1,int2)    ! with E in cm^-1
            end do
        else
            do c2=1, ntransitions
                int1 = CD(c1)%Values(c2)%level1
                int2 = CD(c1)%Values(c2)%level2
                if ( (log10(T)<=CD(c1)%Values(c2)%X(CD(c1)%Values(c2)%npoints)) .and. &
                            & (log10(T)>=CD(c1)%Values(c2)%X(1)) ) then
                    CD(c1)%q(int1,int2) = const2/sqrt(T)/g(int1) * &
                      & Spline3 (CD(c1)%Values(c2)%X,CD(c1)%Values(c2)%Y,log10(T)) ! int1>int2
                else
                    CD(c1)%q(int1,int2) = const2/sqrt(T)/g(int1)*&
                                &Rate_Extrapol(species_name,CD(c1)%partner_name,int1,int2,T)
                end if
                CD(c1)%q(int2,int1) = g(int1)*exp((E(int2)-E(int1))*c*h/k/T)/g(int2) &
                                            & *CD(c1)%q(int1,int2)    ! with E in cm^-1
            end do
        end if
    end if
end do
end if

!-----------------------------------------------------------------------
!* section [3]: Sets the radiation field intensities.
!-----------------------------------------------------------------------
if(z/=zold .or. beta2/=betaold .or. f2/=fold) then
zold = z
betaold = beta2
fold = f2
u(:,:) = 0.0_WP
if (z/=-1.0_WP) then
    do i=1, nlevels-1
        do j=i+1, nlevels
            u(i,j) = Planck(Tcmbr(z,beta2),(E(j)-E(i))*c)
            u(j,i) = u(i,j)
        end do
    end do
end if
do i=1, nlevels-1
    do j=i+1, nlevels
        u(i,j) = u(i,j) + f2*urad( 1.0E8_WP/(E(j)-E(i)),z )
        u(j,i) = u(i,j)
    end do
end do

!-----------------------------------------------------------------------
!* section [4]: Calculates the indirect excitation rates.
!-----------------------------------------------------------------------
gam(:,:) = 0.0_WP
Totgam(:,:) = 0.0_WP

do i=1, nhlevels
    do j=1, nground
        hlevel(i)%Kground(j) = constant*hlevel(i)%Aground(j)/(hlevel(i)%En-E(j))**3.0_WP &
                                                & *f2*urad( 1.0E8_WP/(hlevel(i)%En-E(j)),z )
    end do
end do

do i=1, nground-1
    do j=i+1, nground
        do int1=1, nhlevels
            TheSum = sum( hlevel(int1)%Aground + hlevel(int1)%Kground )
            Totgam(i,j) = Totgam(i,j) +    hlevel(int1)%w*hlevel(int1)%Kground(i)/g(i) * &
                                & (hlevel(int1)%Aground(j)+hlevel(int1)%Kground(j))/TheSum
            Totgam(j,i) = Totgam(j,i) + hlevel(int1)%w*hlevel(int1)%Kground(j)/g(j) * &
                                & (hlevel(int1)%Aground(i)+hlevel(int1)%Kground(i))/TheSum
            if (.not. hlevel(int1)%is_included) then
                gam(i,j) = gam(i,j) + hlevel(int1)%w*hlevel(int1)%Kground(i)/g(i) * &
                                & (hlevel(int1)%Aground(j)+hlevel(int1)%Kground(j))/TheSum
                gam(j,i) = gam(j,i) + hlevel(int1)%w*hlevel(int1)%Kground(j)/g(j) * &
                                & (hlevel(int1)%Aground(i)+hlevel(int1)%Kground(i))/TheSum
            end if
        end do
    end do
end do

end if

!-----------------------------------------------------------------------
!* section [5]: Solves the linear system.
!-----------------------------------------------------------------------
do i=1, nlevels-1
    indep(i) = -B(1,i+1)*u(1,i+1)
    if ( (i+1)<=nground ) indep(i) = indep(i) - gam(1,i+1)
    do int1=1, npartners
        if ( (i+1)<=CD(int1)%maxlevel ) then
            indep(i) = indep(i) - n(int1)*CD(int1)%q(1,i+1)
        end if
    end do
    do j=1, nlevels-1
        if (j /= i) then
            matrix(i,j) = A(j+1,i+1) + B(j+1,i+1)*u(j+1,i+1)    ! if j<i A(j+1,i+1)=0
            if ( (i+1)<=nground .and. (j+1)<=nground ) matrix(i,j) = matrix(i,j) + gam(j+1,i+1)
            do int1=1, npartners
                if ( (j+1)<=CD(int1)%maxlevel .and. (i+1)<=CD(int1)%maxlevel ) then
                    matrix(i,j) = matrix(i,j) + n(int1)*CD(int1)%q(j+1,i+1)
                end if
            end do
        else
            matrix(i,j) = 0.0_WP
            do int2=1, nlevels
                matrix(i,j) = matrix(i,j) - A(i+1,int2) - B(i+1,int2)*u(i+1,int2)    
                                                    ! if i>=j A(j,i)=0 and if i=j B(i,j) =0
                if ( (i+1)<=nground .and. int2<=nground ) matrix(i,j) = matrix(i,j) - &
                                                                            & gam(i+1,int2)
                do int1=1, npartners
                    if ( (i+1)<=CD(int1)%maxlevel .and. int2<=CD(int1)%maxlevel ) then
                        matrix(i,j) = matrix(i,j) - n(int1)*CD(int1)%q(i+1,int2)
                    end if
                end do
            end do
        end if
    end do
end do

! calls Lapack90 subroutine to solve the linear system:
call LA_GESVX (matrix,indep,X)

end subroutine PopulationRatio

! *************************************************************************************
! *------- returns the largest LTE deviation in the relative population ratios -------*
! *------- of the first nl levels (relative to the LTE population ratio) .     -------*
! *************************************************************************************
function LTEdev(nl,T_LTE)
real(WP) :: LTEdev  ! maximum abs( (pop-pop(LTE)) ) / pop(LTE)
integer, intent(in) :: nl  ! analyse the first nl levels (can't be higher than the number
                           ! of levels being taken into account)
real(WP), intent(in) :: T_LTE  ! temperature of LTE equilibrium, in K .

real(WP) :: popLTE(nl-1)   ! population ratios in LTE conditions.

if ( nl>nlevels ) then
    print *, "Error in calling function LTEdev: nl can't be higher than the number &
                &of levels being taken into account!"
    stop
end if

popLTE(:) = g(2:nl)*exp(-E(2:nl)*c*h/k/T_LTE)/g(1)  ! E in cm^-1

LTEdev = maxval ( abs(X(1:nl-1)-popLTE(:))/popLTE(:) )

end function LTEdev

! ***************************************************************
! *------- returns the energy density of radiation field -------*
! ***************************************************************
function urad(lambda,z)
real(WP) :: urad    ! in erg/cm^3/Hz
real(WP), intent(in) :: lambda    ! in 
real(WP), intent(in) :: z    ! redshift (for calculating the UVB field)

urad = 0.0_WP

! Some examples:

urad = Fgal(lambda)
!if (z/=-1._WP) then
!    urad = urad + Fuvb(lambda,z,3)
!end if
!urad = urad + Fhothalo(lambda,2,2)

!urad = Planck(4.0E3_WP,c/lambda*1.0E8_WP)

end function urad

! *******************************************************************************
! *------- calculates the energy density of a black-body radiation field -------*
! *******************************************************************************
function Planck(T,nu)
real(WP) :: Planck
real(WP), intent(in) :: T    ! in K
real(WP), intent(in) :: nu    ! in Hz

real(WP) :: argument    ! argument of the exponential

argument = h*nu/k/T

! checks whether the argument will not cause overflow in the exponential:
if ( argument <= log(huge(argument)) ) then
    Planck = 8.0_WP*pi*h/c/c/c*nu*nu*nu/(exp(argument)-1.0_WP)
else
    Planck = 0.0_WP
end if

end function Planck

! *******************************************************************
! *------- returns the temperature of the CMBR at redshift z -------*
! *******************************************************************
function Tcmbr(z,beta)
real(WP) :: Tcmbr
real(WP), intent(in) :: z, beta

if (beta==0.0_WP) then 
    Tcmbr = T0*(1.0_WP+z)
else
    Tcmbr = T0*(1.0_WP+z)**(1.0_WP-beta)
end if

end function Tcmbr

! **********************************************************************************
! *------- Extrapolates the collision rate q(l1->l2) (collision strength if -------*
! *------- PartnerName=='electron', gam(l1,l2) ) if the kinetic temperature -------*
! *------- was out of range.                                                -------*
! **********************************************************************************
function Rate_Extrapol(SpeciesName,PartnerName,l1,l2,T)
real(WP) :: Rate_Extrapol
character(len=10), intent(in) :: SpeciesName    ! name of atom or ion beeing calculated
character(len=10), intent(in) :: PartnerName    ! name of the collision partner
integer, intent(in) :: l1, l2    ! l1 -> l2
real(WP), intent(in) :: T    ! kinetic temperature, in K .

if (SpeciesName=='CI' .and. PartnerName=='electron') then
    Rate_Extrapol = CI_gam(l1,l2,T)    ! l1>l2
else if (SpeciesName=='OI' .and. PartnerName=='proton') then
    Rate_Extrapol = OI_proton(l1,l2,T)    ! l1>l2
else
    print *, " Error: T out of range!"
    print *, " T = ", T, " K"
    print *, " Enter a new value in the interval:"
    print *, 10.0_WP**(CD(c1)%Values(c2)%X(1)), " K - ", &
            &10.0_WP**(CD(c1)%Values(c2)%X(CD(c1)%Values(c2)%npoints)), " K"
    write (*,'(" , or extrapolate interpolation connecting levels ",I3,"-",I3,&
            &" by collisions with ",a)') CD(c1)%Values(c2)%level1, &
            &CD(c1)%Values(c2)%level2, CD(c1)%partner_name
    write (*,'(" You may also exclude this collision partner from the calculation &
            &by setting its density to zero, n(",I2,")=0.0_WP .")') c1
    stop
end if

end function Rate_Extrapol

!--------------------------------------------------------------------------------------
!*------- returns the Maxwellian-averaged collision strength gam(l2->l1) for CI -------
!*
!* Source:
!*
!*    C. T. Johnson, P. G. Burke and A. E. Kingston, 1987: J. Phys. B 20, 2553
!*    (fits to the transitions involving the ground term levels 3PJ - 3PJ')
!*
!*    D. Pquignot and S. M. V. Aldrovandi, 1976: A&A 50, 141
!*    (fits to the transitions involving the ground term and levels 1D and 1S)
!--------------------------------------------------------------------------------------
function CI_gam(l1,l2,T)
real(WP) :: CI_gam
integer, intent(in) :: l1,l2    ! l1>l2
real(WP), intent(in) :: T    ! in K

real(WP) :: Johnson1(3,5), Johnson2(3,5)

! coefficients of Johnson's fits for 7.5 <= T <= 1000 K :
! 2->1:
data Johnson1(1,:) / -9.25141_WP, -0.773782_WP, 0.361184_WP, -0.150892E-1_WP, -0.656325E-3_WP /
! 3->1:
data Johnson1(2,:) / -7.69735_WP, -1.30743_WP, 0.697638_WP, -0.111338_WP, 0.705277E-2_WP /
! 3->2:
data Johnson1(3,:) / -7.43870_WP, -0.574430_WP, 0.358264_WP, -0.418166E-1_WP, 0.235272E-2_WP /

! coefficients of Johnson's fits for 1000 < T <= 10000 K :
! 2->1:
data Johnson2(1,:) / 444.600_WP, -227.913_WP, 42.5952_WP, -3.47620_WP, 0.105085_WP /
! 3->1:
data Johnson2(2,:) / 350.609_WP, -187.474_WP, 36.1803_WP, -3.03283_WP, 0.938138E-1_WP /
! 3->2:
data Johnson2(3,:) / 386.186_WP, -202.192_WP, 38.5049_WP, -3.19268_WP, 0.978573E-1_WP /

real(WP) :: Pequignot(3,3)

! coefficients of Pquignot's fits for 500 <= T <= 30000 K :
! 5->4:
data Pequignot(1,:) / 1.003_WP, 0.0_WP, 0.0_WP /
! 5->(1+2+3):
data Pequignot(2,:) / 1.41090_WP, -0.45452_WP, 0.05848_WP /
! 4->(1+2+3):
data Pequignot(3,:) / 1.08540_WP, -0.07507_WP, -0.02150_WP /

! checks if T is within the temperature range of validity of the fittings:
select case(l1)
case (2:3)
    if (T<7.5_WP .or. T>10000._WP) then
        print *, "Error in calling function CI_gam!"
        print *, "T = ", T, "K"
        print *, "The temperature range of validity of the fittings for transitions 3PJ - &
                &3PJ' induced by electrons is: 7.5 K <= T <= 10000 K ."
        print *, "Enter a different value for T or exclude electrons from the calculation &
                &by setting its density to zero."
        stop
    end if
case (4:5)
    if (T<500._WP .or. T>30000._WP) then
        print *, "Error in calling function CI_gam!"
        print *, "T = ", T, "K"
        print *, "The temperature range of validity of the fittings for transitions 3PJ - &
                &1D2 (or 1S0) induced by electrons is: 500 K <= T <= 30000 K ."
        print *, "Enter a different value for T or exclude electrons from the calculation &
                &by setting its density to zero."
        stop
    end if
case default
    print *, "Error in calling function CI_gam!"
    print *, "No fittings available for transitions involving level ", l1, "."
    stop
end select

select case(l1)
case (2)
    if (T<=1000._WP) then
        CI_gam = exp( pol(4,Johnson1(1,:),log(T)) )
    else 
        CI_gam = exp( pol(4,Johnson2(1,:),log(T)) )
    end if
case (3)
    if (l2==1) then
        if (T<=1000._WP) then
            CI_gam = exp( pol(4,Johnson1(2,:),log(T)) )
        else
            CI_gam = exp( pol(4,Johnson2(2,:),log(T)) )
        end if
    else
        if (T<=1000._WP) then
            CI_gam = exp( pol(4,Johnson1(3,:),log(T)) )
        else
            CI_gam = exp( pol(4,Johnson2(3,:),log(T)) )
        end if
    end if
case (4)
    CI_gam = ((l2-1)*2+1)*1.156_WP*T/1.E4_WP*pol(2,Pequignot(3,:),T/1.E4_WP)/9._WP
case (5)
    if (l2==4) then
        CI_gam = 0.2764_WP*sqrt(T/1.E4_WP)*pol(2,Pequignot(1,:),T/1.E4_WP)
    else
        CI_gam = ((l2-1)*2+1)*0.2482_WP*T/1.E4_WP*pol(2,Pequignot(2,:),T/1.E4_WP)/9._WP
    end if
end select

end function CI_gam

!---------------------------------------------------------------------------------
!*------- returns the de-excitation rates by collision with protons for OI -------
!*
!* Source:
!*
!*    Pquignot, D., 1990: A&A 231, 499
!*    Pquignot, D., 1996: A&A 313, 1026 (erratum)
!*    (fits to the transitions involving the ground term levels 3PJ - 3PJ')
!---------------------------------------------------------------------------------
function OI_proton(l1,l2,T)
real(WP) :: OI_proton    ! de-excitation rate, in cm^3 s^-1
integer, intent(in) :: l1, l2    ! upper/lower level, l1>l2
real(WP), intent(in) :: T    ! kinetic temperature, in K

real(WP) :: a(3,2:4), b(3,2:4)    ! the parameters of Pguignot's fits

! 3P1 - 3P2:
data a(1,:) / 1.40E-3_WP,    2.14E-2_WP,    2.78E-1_WP    /
data b(1,:) / 0.90_WP,    1.30_WP,        0.87_WP    /

! 3P0 - 3P2:
data a(2,:) / 1.12E-4_WP,    3.90E-3_WP,    8.25E-2_WP    /
data b(2,:) / 1.60_WP,    1.40_WP,        0.80_WP    /

! 3P0 - 3P1:
data a(3,:) / 3.10E-4_WP,    0._WP,        2.29E-2_WP    /
data b(3,:) / 1.06_WP,    0._WP,        0.69_WP    /

real(WP) gamma    ! the "generalized collision strength" of Pguignot

integer i, j
real(WP) T1, T2    ! temperature limits

if (T>=40._WP .and. T<=2.E4_WP) then
    select case (l1)
    case (2)
        ! 3P1 - 3P2
        i = 1
        T1 = 194.6640056970577_WP
    !    T1 = (a(i,2)/a(i,3)*10._WP**(3._WP*b(i,3)-2._WP*b(i,2)))**(1._WP/(b(i,3)-b(i,2)))
        T2 = 3686.241409029949_WP
    !    T2 = (a(i,3)/a(i,4)*10._WP**(4._WP*b(i,4)-3._WP*b(i,3)))**(1._WP/(b(i,4)-b(i,3)))
        if (T<=T1) then
            j = 2
        else if (T<=T2) then
            j = 3
        else
            j = 4
        end if
    case (3)
        if (l2==1) then
            ! 3P0 - 3P2
            i = 2
            T1 = 511.9563354829943_WP
        !    T1 = (a(i,2)/a(i,3)*10._WP**(3._WP*b(i,3)-2._WP*b(i,2)))**(1._WP/(b(i,3)-b(i,2)))
            T2 = 7510.15503734149_WP
        !    T2 = (a(i,3)/a(i,4)*10._WP**(4._WP*b(i,4)-3._WP*b(i,3)))**(1._WP/(b(i,4)-b(i,3)))
            if (T<=T1) then
                j = 2
            else if (T<=T2) then
                j = 3
            else
                j = 4
            end if
        else
            ! 3P0 - 3P1
            i = 3
            T1 = 2090.255757718991_WP
        !    T1 = (a(i,2)/a(i,4)*10._WP**(4._WP*b(i,4)-2._WP*b(i,2)))**(1._WP/(b(i,4)-b(i,2)))
            if (T<=T1) then
                j = 2
            else
                j = 4
            end if
        end if
    case default
        print *, "Error in calling function OI_proton!"
        print *, "No fittings available for transitions involving level ", l1, "."
        stop
    end select
    gamma = a(i,j)*(T/10._WP**(0.0_WP+j))**b(i,j)
    OI_proton = const2/sqrt(T)/g(l1)*gamma
else
    print *," Error in calling function OI_proton!"
    print *," T = ",T," K"
    print *, " The temperature range of validity of the fittings for transitions 3PJ - &
                &3PJ' induced by protons is: 40 K <= T <= 20000 K ."
    print *, " Enter a different value for T or exclude protons from the calculation &
                &by setting its density to zero."
    stop
end if

end function OI_proton

end module PopRatio
