module Spline
!-----------------------------------------------------------------------
!* authors:    
!*
!* Alex Ignacio da Silva        ignacioalex@yahoo.com
!* and
!* Sueli M. Viegas              viegas@iagusp.usp.br
!*
!* Instituto Astronomico e Geofisico
!* Universidade de Sao Paulo
!* Av. Miguel Stefano, 4200
!* 04301-904 Sao Paulo - SP
!* Brazil
!*
!* This module is a Fortran 90 interface to the cubic spline interpolation
!* routines of PPPACK library:
!*
!*** A practical guide to splines, by C. de Boor ***
!*** (Springer-Verlag, Berlin/Heidelberg, 1978)  ***
!*
!* Note: the original routines underwent slight modifications in order
!*       to support single and double precision.    
!*
!* History:
!* Version    Date (D/M/Y)    Comment
!* -------    ------------    -------
!* 1.0        28/04/2000      original code (A.I. Silva)
!-----------------------------------------------------------------------

use Precision

implicit none

public

interface cubspl

    subroutine scubspl ( tau, c, n, ibcbeg, ibcend )
    use Precision
    real(SP) :: tau(n)
    real(SP) :: c(4,n)
    integer :: n
    integer :: ibcbeg, ibcend
    end subroutine scubspl

    subroutine dcubspl ( tau, c, n, ibcbeg, ibcend )
    use Precision
    real(DP) :: tau(n)
    real(DP) :: c(4,n)
    integer :: n
    integer :: ibcbeg, ibcend
    end subroutine dcubspl

end interface

interface ppvalu

    function sppvalu (break, coef, l, k, x, jderiv )
    use Precision
    real(SP) :: sppvalu
    real(SP) :: break(l+1)
    real(SP) :: coef(k,l)
    integer :: l, k
    real(SP) :: x
    integer :: jderiv
    end function sppvalu

    function dppvalu (break, coef, l, k, x, jderiv )
    use Precision
    real(DP) :: dppvalu
    real(DP) :: break(l+1)
    real(DP) :: coef(k,l)
    integer :: l, k
    real(DP) :: x
    integer :: jderiv
    end function dppvalu

end interface

contains

function Spline3 (X,Y,x0)
real(WP) :: Spline3 ! value of Y(x0) interpolated
real(WP) :: X(:), Y(:)  ! abscissas and ordinates values
real(WP) :: x0  ! value to be interpolated

integer :: npoints  ! number of points supplied
real(WP) :: coefficients(4,size(X))

npoints = size(X)

if (size(Y)/=npoints) then
    print *, "Error in calling function Spline3!"
    print *, "Number of abscissa points differ from number of ordinate points."
    stop
end if

if (npoints<2) then
    print *, "Error in calling function Spline3!"
    print *, "Less than 2 points were supplied."
    stop
end if

if (x0>=X(1) .and. x0<=X(npoints)) then
    ! fits the spline with the not-a-knot condition:
    coefficients(1,:) = Y(:)
    call cubspl (X,coefficients,npoints,0,0)
    ! evaluates the spline:
    Spline3 = ppvalu (X,coefficients(:,1:npoints-1),npoints-1,4,x0,0)
else
    print *, "Error in calling function Spline3!"
    print *, "The value to be interpolated lies outside &
                &the bounds allowed by the data set."
    stop
end if

end function Spline3

end module Spline
