module FieldIntensities
!---------------------------------------------------------------------------------
!* authors:                
!*
!* Alex Ignacio da Silva        ignacioalex@yahoo.com
!* and
!* Sueli M. Viegas              viegas@iagusp.usp.br
!*
!* Instituto Astronomico e Geofisico
!* Universidade de Sao Paulo
!* Av. Miguel Stefano, 4200
!* 04301-904 Sao Paulo - SP
!* Brazil
!*
!* This module contains routines to calculate the radiation field
!* intensities from some models in the literature:
!*
!* 1) Fgal: the UV field of the Galaxy.
!*    Source:
!*    Gondhalekar, P.M., Phillips, A.P. & Wilson, R., 1980: A&A 85, 272
!*
!* 2) Fuvb: the UV background of all QSOs at redshift z.
!*    Source:
!*    Madau, P., Haardt, F. & Rees, M.J., 1998: submitted to ApJ (astro-ph/9809058)
!*
!* 3) Fhothalo: the radiation field of the hot halo model.
!*    Source:
!*    Viegas, S.M. & Friaa, A.C.S., 1995: MNRAS 272, L35
!*
!* The input is the wavelength (in ), and the output is the spectral energy
!* density (in erg/cm^3/Hz).
!*
!* History:
!* Version    Date (D/M/Y)    Comment
!* -------    ------------    -------
!* 1.0        28/04/2000      original code (A.I. Silva)
!---------------------------------------------------------------------------------

use Precision
use PhysMath

implicit none

public

contains

!--------------------------------------------------------------------------------
!*------- returns the energy density of the radiation field of the Galaxy -------
!* Source:
!* Gondhalekar, P.M., Phillips, A.P. & Wilson, R., 1980: A&A 85, 272
!--------------------------------------------------------------------------------
function Fgal(lambda)
real(WP) :: Fgal    ! in erg/cm^3/Hz
real(WP), intent(in) :: lambda    ! in 

! This is the field of Gondhalekar et al. 1980 ( vs. erg cm^-2 s^-1 cm^-1):
! ( This table is interpolated linearly. )
integer, parameter :: nvalues = 64    ! number of values in the table bellow:
real(WP) :: lambdas(nvalues) = (/ 930._WP,    975._WP,    1025._WP,    1075._WP, &
    & 1125._WP,    1175._WP,    1220._WP,    1270._WP,    1325._WP,    1380._WP,    1400._WP, &
    & 1420._WP,    1440._WP,    1460._WP,    1480._WP,    1500._WP,    1520._WP,    1540._WP, &
    & 1560._WP,    1580._WP,    1600._WP,    1620._WP,    1640._WP,    1660._WP,    1680._WP, &
    & 1700._WP,    1720._WP,    1780._WP,    1800._WP,    1820._WP,    1840._WP,    1860._WP, &
    & 1880._WP,    1900._WP,    1920._WP,    1940._WP,    1960._WP,    1980._WP,    2000._WP, &
    & 2020._WP,    2040._WP,    2060._WP,    2080._WP,    2100._WP,    2120._WP,    2180._WP, &
    & 2200._WP,    2220._WP,    2240._WP,    2260._WP,    2280._WP,    2300._WP,    2320._WP, &
    & 2340._WP,    2360._WP,    2380._WP,    2400._WP,    2420._WP,    2440._WP,    2460._WP, &
    & 2480._WP,    2500._WP,    2520._WP,    2740._WP /)
real(WP) :: Fs(nvalues) = (/ 83.1_WP,    148.2_WP,    145.7_WP,    159.3_WP, &
    & 181.6_WP,    166.7_WP,    98.1_WP,    175.4_WP,    177.8_WP,    156.9_WP,    149.4_WP, &
    & 149.4_WP,    154.4_WP,    156.9_WP,    153.1_WP,    150.7_WP,    143.3_WP,    134.6_WP, &
    & 130.9_WP,    134.6_WP,    128.4_WP,    124.7_WP,    128.4_WP,    133.4_WP,    135.9_WP, &
    & 133.4_WP,    126.0_WP,    115.8_WP,    118.2_WP,    114.1_WP,    110.4_WP,    105.1_WP, &
    & 105.5_WP,    101.0_WP,    97.6_WP,    97.2_WP,    97.2_WP,    96.6_WP,    93.1_WP,  &
    & 88.8_WP,    86.9_WP,    84.1_WP,    82.3_WP,    81.4_WP,    78.9_WP,    75.1_WP,  &
    & 73.6_WP,    73.6_WP,    71.4_WP,    72.7_WP,    72.7_WP,    71.4_WP,    69.2_WP,  &
    & 67.2_WP,    65.6_WP,    63.6_WP,    63.7_WP,    64.3_WP,    65.6_WP,    65.2_WP,  &
    & 64.8_WP,    65.0_WP,    62.2_WP,    53.8_WP /)

Fgal = 0.0_WP
if (lambda>=lambdas(1) .and. lambda<=lambdas(nvalues)) then    
    Fgal = lambda*lambda/c/c*1.0E-16_WP*LInterpol(lambdas,Fs,lambda)
end if

end function Fgal

!---------------------------------------------------------------------------------------
!*------- returns the energy density of the ultraviolet background field of QSOs -------
!* Source:
!* Madau, P., Haardt, F. & Rees, M. J., 1998: submitted to ApJ (astro-ph/9809058)
!---------------------------------------------------------------------------------------
function Fuvb(lambda,z,model)
real(WP) :: Fuvb    ! in erg/cm^3/Hz
real(WP), intent(in) :: lambda    ! in 
real(WP), intent(in) :: z    ! redshift at which the UVB is to be calculated
integer, intent(in) :: model    ! model = 1 corresponds to q0 =    0.1
                ! model = 2 corresponds to q0 =    0.5
                ! model = 3 corresponds to q0 = 0.5 (revised in MHR 1998)

! Note that there are very narrow spikes at lambda = 228, 304, 1216.
integer, parameter :: nlambdas = 167    ! lambda values bellow are in 
real(WP) :: lambdas(nlambdas) = (/ 4.303_WP,    4.558_WP,    4.828_WP,    &
    & 5.114_WP,    5.417_WP,    5.737_WP,    6.077_WP,    6.437_WP,    6.818_WP,    7.222_WP,    &
    & 7.650_WP,    8.103_WP,    8.583_WP,    9.091_WP,    9.630_WP,    1.020E1_WP,    1.080E1_WP,    &
    & 1.144E1_WP,    1.212E1_WP,    1.284E1_WP,    1.360E1_WP,    1.441E1_WP,    1.526E1_WP,    1.616E1_WP,    &
    & 1.712E1_WP,    1.813E1_WP,    1.921E1_WP,    2.034E1_WP,    2.155E1_WP,    2.283E1_WP,    2.418E1_WP,    &
    & 2.561E1_WP,    2.713E1_WP,    2.873E1_WP,    3.043E1_WP,    3.224E1_WP,    3.415E1_WP,    3.617E1_WP,    &
    & 3.831E1_WP,    4.058E1_WP,    4.298E1_WP,    4.553E1_WP,    4.822E1_WP,    5.108E1_WP,    5.411E1_WP,    &
    & 5.731E1_WP,    6.070E1_WP,    6.430E1_WP,    6.811E1_WP,    7.214E1_WP,    7.641E1_WP,    8.094E1_WP,    &
    & 8.573E1_WP,    9.081E1_WP,    9.619E1_WP,    1.019E2_WP,    1.079E2_WP,    1.143E2_WP,    1.211E2_WP,    &
    & 1.282E2_WP,    1.358E2_WP,    1.439E2_WP,    1.524E2_WP,    1.614E2_WP,    1.710E2_WP,    1.811E2_WP,    &
    & 1.919E2_WP,    2.032E2_WP,    2.153E2_WP,    2.280E2_WP,    2.28001E2_WP,2.415E2_WP,    2.558E2_WP,    &
    & 2.710E2_WP,    2.870E2_WP,    3.040E2_WP,    3.04001E2_WP,3.220E2_WP,    3.411E2_WP,    3.613E2_WP,    &
    & 3.827E2_WP,    4.053E2_WP,    4.293E2_WP,    4.548E2_WP,    4.817E2_WP,    5.102E2_WP,    5.404E2_WP,    &
    & 5.725E2_WP,    6.064E2_WP,    6.423E2_WP,    6.803E2_WP,    7.206E2_WP,    7.633E2_WP,    8.085E2_WP,    &
    & 8.564E2_WP,    9.071E2_WP,    9.120E2_WP,    9.660E2_WP,    1.023E3_WP,    1.084E3_WP,    1.148E3_WP,    &
    & 1.216E3_WP,    1.21601E3_WP,1.288E3_WP,    1.364E3_WP,    1.445E3_WP,    1.531E3_WP,    1.621E3_WP,    &
    & 1.717E3_WP,    1.819E3_WP,    1.927E3_WP,    2.041E3_WP,    2.162E3_WP,    2.290E3_WP,    2.425E3_WP,    &
    & 2.569E3_WP,    2.721E3_WP,    2.882E3_WP,    3.053E3_WP,    3.234E3_WP,    3.425E3_WP,    3.628E3_WP,    &
    & 3.843E3_WP,    4.071E3_WP,    4.312E3_WP,    4.567E3_WP,    4.838E3_WP,    5.124E3_WP,    5.428E3_WP,    &
    & 5.749E3_WP,    6.090E3_WP,    6.450E3_WP,    6.832E3_WP,    7.237E3_WP,    7.666E3_WP,    8.119E3_WP,    &
    & 8.600E3_WP,    9.110E3_WP,    9.649E3_WP,    1.022E4_WP,    1.083E4_WP,    1.147E4_WP,    1.215E4_WP,    &
    & 1.287E4_WP,    1.363E4_WP,    1.443E4_WP,    1.529E4_WP,    1.620E4_WP,    1.715E4_WP,    1.817E4_WP,    &
    & 1.925E4_WP,    2.039E4_WP,    2.159E4_WP,    2.287E4_WP,    2.423E4_WP,    2.566E4_WP,    2.718E4_WP,    &
    & 2.879E4_WP,    3.050E4_WP,    3.230E4_WP,    3.422E4_WP,    3.624E4_WP,    3.839E4_WP,    4.066E4_WP,    &
    & 4.307E4_WP,    4.562E4_WP,    4.832E4_WP /)

integer, parameter :: nzs = 34
real(WP) :: zs(nzs) = (/ 0.000_WP,    5.922E-2_WP,    1.220E-1_WP,    1.884E-1_WP,    &
    & 2.588E-1_WP,    3.333E-1_WP,    4.123E-1_WP,    4.959E-1_WP,    5.845E-1_WP,    6.784E-1_WP,    7.778E-1_WP,    &
    & 8.831E-1_WP,    9.946E-1_WP,    1.113_WP,    1.238_WP,    1.370_WP,    1.511_WP,    1.659_WP,    &
    & 1.817_WP,    1.984_WP,    2.160_WP,    2.348_WP,    2.546_WP,    2.756_WP,    2.978_WP,    &
    & 3.214_WP,    3.464_WP,    3.728_WP,    4.008_WP,    4.305_WP,    4.619_WP,    4.951_WP,    &
    & 5.304_WP,    5.677_WP /)

real(WP) :: J(3,nlambdas,nzs)    ! J values bellow are in erg/cm^2/s/Hz/sr

! model = 1
data J(1,    1    ,:)    / 1.278E-27_WP,    1.634E-27_WP,    2.087E-27_WP,    2.662E-27_WP,    3.393E-27_WP,    4.317E-27_WP, &
    & 5.481E-27_WP,    6.944E-27_WP,    8.772E-27_WP,    1.105E-26_WP,    1.385E-26_WP,    1.730E-26_WP,    2.148E-26_WP, &
    & 2.649E-26_WP,    3.235E-26_WP,    3.904E-26_WP,    4.646E-26_WP,    5.439E-26_WP,    6.241E-26_WP,    6.995E-26_WP, &
    & 7.625E-26_WP,    8.044E-26_WP,    8.169E-26_WP,    7.940E-26_WP,    7.346E-26_WP,    6.438E-26_WP,    5.268E-26_WP, &
    & 3.939E-26_WP,    2.649E-26_WP,    1.575E-26_WP,    8.100E-27_WP,    3.513E-27_WP,    1.232E-27_WP,    3.069E-28_WP /
data J(1,    2    ,:)    / 1.418E-27_WP,    1.812E-27_WP,    2.314E-27_WP,    2.953E-27_WP,    3.763E-27_WP,    4.787E-27_WP, &
    & 6.079E-27_WP,    7.701E-27_WP,    9.729E-27_WP,    1.225E-26_WP,    1.536E-26_WP,    1.918E-26_WP,    2.382E-26_WP, &
    & 2.938E-26_WP,    3.587E-26_WP,    4.329E-26_WP,    5.153E-26_WP,    6.031E-26_WP,    6.920E-26_WP,    7.756E-26_WP, &
    & 8.454E-26_WP,    8.919E-26_WP,    9.058E-26_WP,    8.804E-26_WP,    8.146E-26_WP,    7.139E-26_WP,    5.842E-26_WP, &
    & 4.368E-26_WP,    2.938E-26_WP,    1.746E-26_WP,    8.984E-27_WP,    3.896E-27_WP,    1.366E-27_WP,    3.404E-28_WP /
data J(1,    3    ,:)    / 1.573E-27_WP,    2.010E-27_WP,    2.567E-27_WP,    3.275E-27_WP,    4.173E-27_WP,    5.310E-27_WP, &
    & 6.742E-27_WP,    8.541E-27_WP,    1.079E-26_WP,    1.359E-26_WP,    1.704E-26_WP,    2.127E-26_WP,    2.642E-26_WP, &
    & 3.258E-26_WP,    3.978E-26_WP,    4.801E-26_WP,    5.714E-26_WP,    6.687E-26_WP,    7.673E-26_WP,    8.600E-26_WP, &
    & 9.374E-26_WP,    9.889E-26_WP,    1.004E-25_WP,    9.761E-26_WP,    9.032E-26_WP,    7.916E-26_WP,    6.478E-26_WP, &
    & 4.843E-26_WP,    3.258E-26_WP,    1.937E-26_WP,    9.963E-27_WP,    4.321E-27_WP,    1.515E-27_WP,    3.775E-28_WP /
data J(1,    4    ,:)    / 1.744E-27_WP,    2.229E-27_WP,    2.847E-27_WP,    3.632E-27_WP,    4.629E-27_WP,    5.889E-27_WP, &
    & 7.477E-27_WP,    9.472E-27_WP,    1.197E-26_WP,    1.507E-26_WP,    1.890E-26_WP,    2.359E-26_WP,    2.930E-26_WP, &
    & 3.613E-26_WP,    4.411E-26_WP,    5.324E-26_WP,    6.335E-26_WP,    7.415E-26_WP,    8.508E-26_WP,    9.535E-26_WP, &
    & 1.039E-25_WP,    1.096E-25_WP,    1.113E-25_WP,    1.082E-25_WP,    1.001E-25_WP,    8.776E-26_WP,    7.182E-26_WP, &
    & 5.370E-26_WP,    3.613E-26_WP,    2.148E-26_WP,    1.105E-26_WP,    4.793E-27_WP,    1.680E-27_WP,    4.187E-28_WP /
data J(1,    5    ,:)    / 1.934E-27_WP,    2.472E-27_WP,    3.157E-27_WP,    4.029E-27_WP,    5.133E-27_WP,    6.531E-27_WP, &
    & 8.293E-27_WP,    1.051E-26_WP,    1.327E-26_WP,    1.671E-26_WP,    2.096E-26_WP,    2.616E-26_WP,    3.249E-26_WP, &
    & 4.006E-26_WP,    4.892E-26_WP,    5.903E-26_WP,    7.025E-26_WP,    8.221E-26_WP,    9.432E-26_WP,    1.057E-25_WP, &
    & 1.152E-25_WP,    1.215E-25_WP,    1.234E-25_WP,    1.200E-25_WP,    1.110E-25_WP,    9.730E-26_WP,    7.963E-26_WP, &
    & 5.955E-26_WP,    4.006E-26_WP,    2.382E-26_WP,    1.225E-26_WP,    5.315E-27_WP,    1.864E-27_WP,    4.643E-28_WP /
data J(1,    6    ,:)    / 2.145E-27_WP,    2.742E-27_WP,    3.502E-27_WP,    4.468E-27_WP,    5.693E-27_WP,    7.243E-27_WP, &
    & 9.197E-27_WP,    1.165E-26_WP,    1.472E-26_WP,    1.853E-26_WP,    2.324E-26_WP,    2.901E-26_WP,    3.603E-26_WP, &
    & 4.442E-26_WP,    5.424E-26_WP,    6.545E-26_WP,    7.788E-26_WP,    9.114E-26_WP,    1.046E-25_WP,    1.172E-25_WP, &
    & 1.277E-25_WP,    1.347E-25_WP,    1.368E-25_WP,    1.330E-25_WP,    1.230E-25_WP,    1.079E-25_WP,    8.828E-26_WP, &
    & 6.602E-26_WP,    4.442E-26_WP,    2.641E-26_WP,    1.359E-26_WP,    5.895E-27_WP,    2.067E-27_WP,    5.150E-28_WP /
data J(1,    7    ,:)    / 2.379E-27_WP,    3.041E-27_WP,    3.884E-27_WP,    4.955E-27_WP,    6.314E-27_WP,    8.033E-27_WP, &
    & 1.020E-26_WP,    1.292E-26_WP,    1.632E-26_WP,    2.055E-26_WP,    2.577E-26_WP,    3.217E-26_WP,    3.995E-26_WP, &
    & 4.926E-26_WP,    6.014E-26_WP,    7.256E-26_WP,    8.634E-26_WP,    1.010E-25_WP,    1.159E-25_WP,    1.299E-25_WP, &
    & 1.415E-25_WP,    1.493E-25_WP,    1.516E-25_WP,    1.474E-25_WP,    1.364E-25_WP,    1.195E-25_WP,    9.786E-26_WP, &
    & 7.320E-26_WP,    4.925E-26_WP,    2.929E-26_WP,    1.507E-26_WP,    6.538E-27_WP,    2.292E-27_WP,    5.712E-28_WP /
data J(1,    8    ,:)    / 2.639E-27_WP,    3.373E-27_WP,    4.307E-27_WP,    5.496E-27_WP,    7.003E-27_WP,    8.909E-27_WP, &
    & 1.131E-26_WP,    1.433E-26_WP,    1.810E-26_WP,    2.279E-26_WP,    2.858E-26_WP,    3.567E-26_WP,    4.430E-26_WP, &
    & 5.461E-26_WP,    6.667E-26_WP,    8.044E-26_WP,    9.571E-26_WP,    1.120E-25_WP,    1.285E-25_WP,    1.439E-25_WP, &
    & 1.569E-25_WP,    1.654E-25_WP,    1.680E-25_WP,    1.633E-25_WP,    1.511E-25_WP,    1.325E-25_WP,    1.085E-25_WP, &
    & 8.114E-26_WP,    5.461E-26_WP,    3.248E-26_WP,    1.671E-26_WP,    7.250E-27_WP,    2.542E-27_WP,    6.335E-28_WP /
data J(1,    9    ,:)    / 2.927E-27_WP,    3.740E-27_WP,    4.777E-27_WP,    6.095E-27_WP,    7.766E-27_WP,    9.880E-27_WP, &
    & 1.254E-26_WP,    1.589E-26_WP,    2.007E-26_WP,    2.527E-26_WP,    3.168E-26_WP,    3.955E-26_WP,    4.911E-26_WP, &
    & 6.054E-26_WP,    7.391E-26_WP,    8.917E-26_WP,    1.061E-25_WP,    1.241E-25_WP,    1.424E-25_WP,    1.595E-25_WP, &
    & 1.738E-25_WP,    1.833E-25_WP,    1.861E-25_WP,    1.809E-25_WP,    1.674E-25_WP,    1.468E-25_WP,    1.202E-25_WP, &
    & 8.994E-26_WP,    6.054E-26_WP,    3.601E-26_WP,    1.853E-26_WP,    8.040E-27_WP,    2.820E-27_WP,    7.026E-28_WP /
data J(1,    10    ,:)    / 3.246E-27_WP,    4.148E-27_WP,    5.298E-27_WP,    6.759E-27_WP,    8.612E-27_WP,    1.096E-26_WP, &
    & 1.391E-26_WP,    1.762E-26_WP,    2.225E-26_WP,    2.802E-26_WP,    3.513E-26_WP,    4.385E-26_WP,    5.445E-26_WP, &
    & 6.712E-26_WP,    8.192E-26_WP,    9.883E-26_WP,    1.176E-25_WP,    1.375E-25_WP,    1.577E-25_WP,    1.767E-25_WP, &
    & 1.925E-25_WP,    2.030E-25_WP,    2.062E-25_WP,    2.004E-25_WP,    1.855E-25_WP,    1.626E-25_WP,    1.332E-25_WP, &
    & 9.968E-26_WP,    6.711E-26_WP,    3.993E-26_WP,    2.055E-26_WP,    8.917E-27_WP,    3.127E-27_WP,    7.792E-28_WP /
data J(1,    11    ,:)    / 3.600E-27_WP,    4.601E-27_WP,    5.875E-27_WP,    7.496E-27_WP,    9.551E-27_WP,    1.215E-26_WP, &
    & 1.543E-26_WP,    1.954E-26_WP,    2.468E-26_WP,    3.106E-26_WP,    3.895E-26_WP,    4.861E-26_WP,    6.036E-26_WP, &
    & 7.439E-26_WP,    9.080E-26_WP,    1.095E-25_WP,    1.303E-25_WP,    1.524E-25_WP,    1.747E-25_WP,    1.957E-25_WP, &
    & 2.132E-25_WP,    2.248E-25_WP,    2.283E-25_WP,    2.219E-25_WP,    2.054E-25_WP,    1.801E-25_WP,    1.476E-25_WP, &
    & 1.105E-25_WP,    7.439E-26_WP,    4.427E-26_WP,    2.279E-26_WP,    9.888E-27_WP,    3.468E-27_WP,    8.642E-28_WP /
data J(1,    12    ,:)    / 3.992E-27_WP,    5.102E-27_WP,    6.516E-27_WP,    8.312E-27_WP,    1.059E-26_WP,    1.347E-26_WP, &
    & 1.710E-26_WP,    2.166E-26_WP,    2.736E-26_WP,    3.444E-26_WP,    4.318E-26_WP,    5.389E-26_WP,    6.690E-26_WP, &
    & 8.245E-26_WP,    1.006E-25_WP,    1.214E-25_WP,    1.443E-25_WP,    1.688E-25_WP,    1.935E-25_WP,    2.167E-25_WP, &
    & 2.361E-25_WP,    2.489E-25_WP,    2.527E-25_WP,    2.456E-25_WP,    2.274E-25_WP,    1.995E-25_WP,    1.634E-25_WP, &
    & 1.224E-25_WP,    8.245E-26_WP,    4.907E-26_WP,    2.527E-26_WP,    1.096E-26_WP,    3.846E-27_WP,    9.584E-28_WP /
data J(1,    13    ,:)    / 4.428E-27_WP,    5.658E-27_WP,    7.226E-27_WP,    9.218E-27_WP,    1.174E-26_WP,    1.494E-26_WP, &
    & 1.896E-26_WP,    2.402E-26_WP,    3.033E-26_WP,    3.818E-26_WP,    4.786E-26_WP,    5.973E-26_WP,    7.414E-26_WP, &
    & 9.136E-26_WP,    1.115E-25_WP,    1.344E-25_WP,    1.599E-25_WP,    1.869E-25_WP,    2.143E-25_WP,    2.399E-25_WP, &
    & 2.613E-25_WP,    2.754E-25_WP,    2.796E-25_WP,    2.718E-25_WP,    2.516E-25_WP,    2.208E-25_WP,    1.810E-25_WP, &
    & 1.356E-25_WP,    9.136E-26_WP,    5.439E-26_WP,    2.801E-26_WP,    1.216E-26_WP,    4.265E-27_WP,    1.063E-27_WP /
data J(1,    14    ,:)    / 4.910E-27_WP,    6.275E-27_WP,    8.013E-27_WP,    1.022E-26_WP,    1.302E-26_WP,    1.656E-26_WP, &
    & 2.103E-26_WP,    2.663E-26_WP,    3.363E-26_WP,    4.232E-26_WP,    5.305E-26_WP,    6.619E-26_WP,    8.215E-26_WP, &
    & 1.012E-25_WP,    1.235E-25_WP,    1.489E-25_WP,    1.770E-25_WP,    2.069E-25_WP,    2.371E-25_WP,    2.654E-25_WP, &
    & 2.890E-25_WP,    3.047E-25_WP,    3.093E-25_WP,    3.006E-25_WP,    2.784E-25_WP,    2.443E-25_WP,    2.003E-25_WP, &
    & 1.501E-25_WP,    1.012E-25_WP,    6.029E-26_WP,    3.106E-26_WP,    1.348E-26_WP,    4.729E-27_WP,    1.179E-27_WP /
data J(1,    15    ,:)    / 5.445E-27_WP,    6.958E-27_WP,    8.885E-27_WP,    1.133E-26_WP,    1.444E-26_WP,    1.837E-26_WP, &
    & 2.331E-26_WP,    2.952E-26_WP,    3.727E-26_WP,    4.690E-26_WP,    5.879E-26_WP,    7.334E-26_WP,    9.101E-26_WP, &
    & 1.121E-25_WP,    1.368E-25_WP,    1.649E-25_WP,    1.959E-25_WP,    2.290E-25_WP,    2.623E-25_WP,    2.936E-25_WP, &
    & 3.196E-25_WP,    3.368E-25_WP,    3.419E-25_WP,    3.323E-25_WP,    3.078E-25_WP,    2.702E-25_WP,    2.217E-25_WP, &
    & 1.662E-25_WP,    1.121E-25_WP,    6.681E-26_WP,    3.443E-26_WP,    1.495E-26_WP,    5.244E-27_WP,    1.307E-27_WP /
data J(1,    16    ,:)    / 6.038E-27_WP,    7.716E-27_WP,    9.852E-27_WP,    1.257E-26_WP,    1.601E-26_WP,    2.036E-26_WP, &
    & 2.584E-26_WP,    3.272E-26_WP,    4.131E-26_WP,    5.198E-26_WP,    6.514E-26_WP,    8.125E-26_WP,    1.008E-25_WP, &
    & 1.242E-25_WP,    1.514E-25_WP,    1.825E-25_WP,    2.168E-25_WP,    2.533E-25_WP,    2.901E-25_WP,    3.245E-25_WP, &
    & 3.531E-25_WP,    3.721E-25_WP,    3.776E-25_WP,    3.671E-25_WP,    3.401E-25_WP,    2.987E-25_WP,    2.452E-25_WP, &
    & 1.839E-25_WP,    1.242E-25_WP,    7.402E-26_WP,    3.816E-26_WP,    1.657E-26_WP,    5.815E-27_WP,    1.450E-27_WP /
data J(1,    17    ,:)    / 6.696E-27_WP,    8.556E-27_WP,    1.092E-26_WP,    1.393E-26_WP,    1.775E-26_WP,    2.257E-26_WP, &
    & 2.864E-26_WP,    3.626E-26_WP,    4.578E-26_WP,    5.759E-26_WP,    7.216E-26_WP,    8.999E-26_WP,    1.116E-25_WP, &
    & 1.374E-25_WP,    1.676E-25_WP,    2.019E-25_WP,    2.398E-25_WP,    2.800E-25_WP,    3.205E-25_WP,    3.584E-25_WP, &
    & 3.899E-25_WP,    4.108E-25_WP,    4.168E-25_WP,    4.052E-25_WP,    3.755E-25_WP,    3.299E-25_WP,    2.710E-25_WP, &
    & 2.035E-25_WP,    1.374E-25_WP,    8.199E-26_WP,    4.229E-26_WP,    1.837E-26_WP,    6.447E-27_WP,    1.607E-27_WP /
data J(1,    18    ,:)    / 7.424E-27_WP,    9.486E-27_WP,    1.211E-26_WP,    1.545E-26_WP,    1.967E-26_WP,    2.502E-26_WP, &
    & 3.175E-26_WP,    4.019E-26_WP,    5.072E-26_WP,    6.380E-26_WP,    7.992E-26_WP,    9.964E-26_WP,    1.236E-25_WP, &
    & 1.521E-25_WP,    1.854E-25_WP,    2.232E-25_WP,    2.650E-25_WP,    3.093E-25_WP,    3.539E-25_WP,    3.956E-25_WP, &
    & 4.302E-25_WP,    4.530E-25_WP,    4.597E-25_WP,    4.469E-25_WP,    4.142E-25_WP,    3.641E-25_WP,    2.993E-25_WP, &
    & 2.250E-25_WP,    1.521E-25_WP,    9.081E-26_WP,    4.686E-26_WP,    2.036E-26_WP,    7.148E-27_WP,    1.782E-27_WP /
data J(1,    19    ,:)    / 8.232E-27_WP,    1.052E-26_WP,    1.343E-26_WP,    1.712E-26_WP,    2.181E-26_WP,    2.773E-26_WP, &
    & 3.518E-26_WP,    4.452E-26_WP,    5.618E-26_WP,    7.065E-26_WP,    8.848E-26_WP,    1.103E-25_WP,    1.367E-25_WP, &
    & 1.682E-25_WP,    2.049E-25_WP,    2.467E-25_WP,    2.927E-25_WP,    3.415E-25_WP,    3.905E-25_WP,    4.362E-25_WP, &
    & 4.741E-25_WP,    4.991E-25_WP,    5.063E-25_WP,    4.923E-25_WP,    4.564E-25_WP,    4.015E-25_WP,    3.304E-25_WP, &
    & 2.486E-25_WP,    1.682E-25_WP,    1.005E-25_WP,    5.192E-26_WP,    2.257E-26_WP,    7.925E-27_WP,    1.976E-27_WP /
data J(1,    20    ,:)    / 9.127E-27_WP,    1.166E-26_WP,    1.488E-26_WP,    1.898E-26_WP,    2.417E-26_WP,    3.072E-26_WP, &
    & 3.898E-26_WP,    4.932E-26_WP,    6.222E-26_WP,    7.822E-26_WP,    9.794E-26_WP,    1.220E-25_WP,    1.512E-25_WP, &
    & 1.860E-25_WP,    2.264E-25_WP,    2.724E-25_WP,    3.231E-25_WP,    3.766E-25_WP,    4.304E-25_WP,    4.804E-25_WP, &
    & 5.218E-25_WP,    5.491E-25_WP,    5.570E-25_WP,    5.416E-25_WP,    5.023E-25_WP,    4.422E-25_WP,    3.643E-25_WP, &
    & 2.744E-25_WP,    1.860E-25_WP,    1.113E-25_WP,    5.751E-26_WP,    2.501E-26_WP,    8.785E-27_WP,    2.191E-27_WP /
data J(1,    21    ,:)    / 1.012E-26_WP,    1.293E-26_WP,    1.650E-26_WP,    2.104E-26_WP,    2.678E-26_WP,    3.404E-26_WP, &
    & 4.317E-26_WP,    5.462E-26_WP,    6.889E-26_WP,    8.657E-26_WP,    1.084E-25_WP,    1.350E-25_WP,    1.671E-25_WP, &
    & 2.055E-25_WP,    2.500E-25_WP,    3.006E-25_WP,    3.562E-25_WP,    4.149E-25_WP,    4.737E-25_WP,    5.284E-25_WP, &
    & 5.736E-25_WP,    6.033E-25_WP,    6.117E-25_WP,    5.949E-25_WP,    5.519E-25_WP,    4.864E-25_WP,    4.012E-25_WP, &
    & 3.028E-25_WP,    2.055E-25_WP,    1.231E-25_WP,    6.368E-26_WP,    2.772E-26_WP,    9.738E-27_WP,    2.429E-27_WP /
data J(1,    22    ,:)    / 1.122E-26_WP,    1.433E-26_WP,    1.828E-26_WP,    2.331E-26_WP,    2.967E-26_WP,    3.771E-26_WP, &
    & 4.781E-26_WP,    6.047E-26_WP,    7.624E-26_WP,    9.578E-26_WP,    1.198E-25_WP,    1.492E-25_WP,    1.846E-25_WP, &
    & 2.268E-25_WP,    2.758E-25_WP,    3.314E-25_WP,    3.923E-25_WP,    4.566E-25_WP,    5.208E-25_WP,    5.803E-25_WP, &
    & 6.294E-25_WP,    6.615E-25_WP,    6.706E-25_WP,    6.522E-25_WP,    6.055E-25_WP,    5.342E-25_WP,    4.413E-25_WP, &
    & 3.336E-25_WP,    2.269E-25_WP,    1.361E-25_WP,    7.050E-26_WP,    3.070E-26_WP,    1.079E-26_WP,    2.693E-27_WP /
data J(1,    23    ,:)    / 1.243E-26_WP,    1.588E-26_WP,    2.026E-26_WP,    2.582E-26_WP,    3.287E-26_WP,    4.175E-26_WP, &
    & 5.293E-26_WP,    6.692E-26_WP,    8.434E-26_WP,    1.059E-25_WP,    1.324E-25_WP,    1.647E-25_WP,    2.038E-25_WP, &
    & 2.502E-25_WP,    3.040E-25_WP,    3.648E-25_WP,    4.315E-25_WP,    5.016E-25_WP,    5.715E-25_WP,    6.361E-25_WP, &
    & 6.892E-25_WP,    7.239E-25_WP,    7.336E-25_WP,    7.135E-25_WP,    6.629E-25_WP,    5.856E-25_WP,    4.847E-25_WP, &
    & 3.672E-25_WP,    2.503E-25_WP,    1.504E-25_WP,    7.802E-26_WP,    3.401E-26_WP,    1.196E-26_WP,    2.986E-27_WP /
data J(1,    24    ,:)    / 1.378E-26_WP,    1.759E-26_WP,    2.245E-26_WP,    2.860E-26_WP,    3.640E-26_WP,    4.622E-26_WP, &
    & 5.857E-26_WP,    7.403E-26_WP,    9.325E-26_WP,    1.170E-25_WP,    1.462E-25_WP,    1.818E-25_WP,    2.247E-25_WP, &
    & 2.756E-25_WP,    3.346E-25_WP,    4.011E-25_WP,    4.739E-25_WP,    5.502E-25_WP,    6.259E-25_WP,    6.958E-25_WP, &
    & 7.530E-25_WP,    7.903E-25_WP,    8.005E-25_WP,    7.787E-25_WP,    7.239E-25_WP,    6.405E-25_WP,    5.313E-25_WP, &
    & 4.036E-25_WP,    2.758E-25_WP,    1.661E-25_WP,    8.630E-26_WP,    3.766E-26_WP,    1.325E-26_WP,    3.309E-27_WP /
data J(1,    25    ,:)    / 1.527E-26_WP,    1.949E-26_WP,    2.486E-26_WP,    3.168E-26_WP,    4.029E-26_WP,    5.115E-26_WP, &
    & 6.479E-26_WP,    8.184E-26_WP,    1.030E-25_WP,    1.292E-25_WP,    1.614E-25_WP,    2.005E-25_WP,    2.475E-25_WP, &
    & 3.033E-25_WP,    3.677E-25_WP,    4.403E-25_WP,    5.195E-25_WP,    6.021E-25_WP,    6.839E-25_WP,    7.591E-25_WP, &
    & 8.205E-25_WP,    8.603E-25_WP,    8.710E-25_WP,    8.474E-25_WP,    7.885E-25_WP,    6.988E-25_WP,    5.812E-25_WP, &
    & 4.428E-25_WP,    3.035E-25_WP,    1.833E-25_WP,    9.542E-26_WP,    4.169E-26_WP,    1.468E-26_WP,    3.668E-27_WP /
data J(1,    26    ,:)    / 1.691E-26_WP,    2.159E-26_WP,    2.753E-26_WP,    3.507E-26_WP,    4.459E-26_WP,    5.658E-26_WP, &
    & 7.164E-26_WP,    9.044E-26_WP,    1.138E-25_WP,    1.426E-25_WP,    1.779E-25_WP,    2.208E-25_WP,    2.723E-25_WP, &
    & 3.332E-25_WP,    4.035E-25_WP,    4.824E-25_WP,    5.682E-25_WP,    6.575E-25_WP,    7.454E-25_WP,    8.259E-25_WP, &
    & 8.913E-25_WP,    9.335E-25_WP,    9.445E-25_WP,    9.191E-25_WP,    8.560E-25_WP,    7.602E-25_WP,    6.342E-25_WP, &
    & 4.849E-25_WP,    3.336E-25_WP,    2.021E-25_WP,    1.054E-25_WP,    4.613E-26_WP,    1.626E-26_WP,    4.065E-27_WP /
data J(1,    27    ,:)    / 1.874E-26_WP,    2.391E-26_WP,    3.048E-26_WP,    3.880E-26_WP,    4.932E-26_WP,    6.256E-26_WP, &
    & 7.915E-26_WP,    9.985E-26_WP,    1.255E-25_WP,    1.572E-25_WP,    1.959E-25_WP,    2.428E-25_WP,    2.991E-25_WP, &
    & 3.655E-25_WP,    4.418E-25_WP,    5.274E-25_WP,    6.200E-25_WP,    7.159E-25_WP,    8.099E-25_WP,    8.955E-25_WP, &
    & 9.648E-25_WP,    1.009E-24_WP,    1.020E-24_WP,    9.930E-25_WP,    9.260E-25_WP,    8.242E-25_WP,    6.900E-25_WP, &
    & 5.298E-25_WP,    3.660E-25_WP,    2.226E-25_WP,    1.164E-25_WP,    5.103E-26_WP,    1.800E-26_WP,    4.504E-27_WP /
data J(1,    28    ,:)    / 2.075E-26_WP,    2.647E-26_WP,    3.373E-26_WP,    4.292E-26_WP,    5.453E-26_WP,    6.912E-26_WP, &
    & 8.739E-26_WP,    1.102E-25_WP,    1.384E-25_WP,    1.730E-25_WP,    2.154E-25_WP,    2.666E-25_WP,    3.279E-25_WP, &
    & 4.001E-25_WP,    4.828E-25_WP,    5.750E-25_WP,    6.746E-25_WP,    7.771E-25_WP,    8.769E-25_WP,    9.674E-25_WP, &
    & 1.040E-24_WP,    1.086E-24_WP,    1.098E-24_WP,    1.068E-24_WP,    9.975E-25_WP,    8.902E-25_WP,    7.482E-25_WP, &
    & 5.773E-25_WP,    4.008E-25_WP,    2.448E-25_WP,    1.285E-25_WP,    5.643E-26_WP,    1.993E-26_WP,    4.990E-27_WP /
data J(1,    29    ,:)    / 2.297E-26_WP,    2.929E-26_WP,    3.731E-26_WP,    4.745E-26_WP,    6.024E-26_WP,    7.631E-26_WP, &
    & 9.641E-26_WP,    1.214E-25_WP,    1.523E-25_WP,    1.903E-25_WP,    2.365E-25_WP,    2.922E-25_WP,    3.588E-25_WP, &
    & 4.369E-25_WP,    5.261E-25_WP,    6.252E-25_WP,    7.316E-25_WP,    8.405E-25_WP,    9.458E-25_WP,    1.041E-24_WP, &
    & 1.117E-24_WP,    1.164E-24_WP,    1.175E-24_WP,    1.144E-24_WP,    1.070E-24_WP,    9.573E-25_WP,    8.083E-25_WP, &
    & 6.271E-25_WP,    4.378E-25_WP,    2.688E-25_WP,    1.416E-25_WP,    6.236E-26_WP,    2.206E-26_WP,    5.527E-27_WP /
data J(1,    30    ,:)    / 2.542E-26_WP,    3.240E-26_WP,    4.125E-26_WP,    5.243E-26_WP,    6.651E-26_WP,    8.418E-26_WP, &
    & 1.062E-25_WP,    1.336E-25_WP,    1.674E-25_WP,    2.088E-25_WP,    2.591E-25_WP,    3.196E-25_WP,    3.916E-25_WP, &
    & 4.758E-25_WP,    5.715E-25_WP,    6.775E-25_WP,    7.905E-25_WP,    9.054E-25_WP,    1.016E-24_WP,    1.114E-24_WP, &
    & 1.193E-24_WP,    1.241E-24_WP,    1.252E-24_WP,    1.219E-24_WP,    1.141E-24_WP,    1.024E-24_WP,    8.694E-25_WP, &
    & 6.788E-25_WP,    4.771E-25_WP,    2.947E-25_WP,    1.560E-25_WP,    6.887E-26_WP,    2.440E-26_WP,    6.121E-27_WP /
data J(1,    31    ,:)    / 2.811E-26_WP,    3.582E-26_WP,    4.557E-26_WP,    5.788E-26_WP,    7.337E-26_WP,    9.277E-26_WP, &
    & 1.169E-25_WP,    1.469E-25_WP,    1.838E-25_WP,    2.288E-25_WP,    2.833E-25_WP,    3.487E-25_WP,    4.263E-25_WP, &
    & 5.166E-25_WP,    6.188E-25_WP,    7.313E-25_WP,    8.507E-25_WP,    9.711E-25_WP,    1.086E-24_WP,    1.187E-24_WP, &
    & 1.267E-24_WP,    1.316E-24_WP,    1.326E-24_WP,    1.291E-24_WP,    1.210E-24_WP,    1.090E-24_WP,    9.307E-25_WP, &
    & 7.318E-25_WP,    5.183E-25_WP,    3.224E-25_WP,    1.715E-25_WP,    7.600E-26_WP,    2.699E-26_WP,    6.777E-27_WP /
data J(1,    32    ,:)    / 3.108E-26_WP,    3.957E-26_WP,    5.031E-26_WP,    6.385E-26_WP,    8.085E-26_WP,    1.021E-25_WP, &
    & 1.285E-25_WP,    1.612E-25_WP,    2.013E-25_WP,    2.501E-25_WP,    3.090E-25_WP,    3.794E-25_WP,    4.625E-25_WP, &
    & 5.588E-25_WP,    6.674E-25_WP,    7.862E-25_WP,    9.114E-25_WP,    1.036E-24_WP,    1.154E-24_WP,    1.258E-24_WP, &
    & 1.338E-24_WP,    1.387E-24_WP,    1.395E-24_WP,    1.358E-24_WP,    1.276E-24_WP,    1.154E-24_WP,    9.910E-25_WP, &
    & 7.854E-25_WP,    5.612E-25_WP,    3.518E-25_WP,    1.884E-25_WP,    8.380E-26_WP,    2.983E-26_WP,    7.501E-27_WP /
data J(1,    33    ,:)    / 3.434E-26_WP,    4.369E-26_WP,    5.550E-26_WP,    7.036E-26_WP,    8.899E-26_WP,    1.122E-25_WP, &
    & 1.410E-25_WP,    1.766E-25_WP,    2.200E-25_WP,    2.727E-25_WP,    3.361E-25_WP,    4.114E-25_WP,    5.001E-25_WP, &
    & 6.022E-25_WP,    7.167E-25_WP,    8.413E-25_WP,    9.716E-25_WP,    1.100E-24_WP,    1.220E-24_WP,    1.325E-24_WP, &
    & 1.406E-24_WP,    1.453E-24_WP,    1.460E-24_WP,    1.421E-24_WP,    1.336E-24_WP,    1.214E-24_WP,    1.049E-24_WP, &
    & 8.388E-25_WP,    6.052E-25_WP,    3.829E-25_WP,    2.065E-25_WP,    9.232E-26_WP,    3.295E-26_WP,    8.301E-27_WP /
data J(1,    34    ,:)    / 3.792E-26_WP,    4.820E-26_WP,    6.116E-26_WP,    7.744E-26_WP,    9.780E-26_WP,    1.231E-25_WP, &
    & 1.545E-25_WP,    1.930E-25_WP,    2.399E-25_WP,    2.965E-25_WP,    3.643E-25_WP,    4.446E-25_WP,    5.385E-25_WP, &
    & 6.461E-25_WP,    7.661E-25_WP,    8.958E-25_WP,    1.030E-24_WP,    1.162E-24_WP,    1.283E-24_WP,    1.388E-24_WP, &
    & 1.467E-24_WP,    1.513E-24_WP,    1.518E-24_WP,    1.477E-24_WP,    1.391E-24_WP,    1.268E-24_WP,    1.104E-24_WP, &
    & 8.909E-25_WP,    6.498E-25_WP,    4.155E-25_WP,    2.260E-25_WP,    1.016E-25_WP,    3.638E-26_WP,    9.182E-27_WP /
data J(1,    35    ,:)    / 4.183E-26_WP,    5.311E-26_WP,    6.731E-26_WP,    8.511E-26_WP,    1.073E-25_WP,    1.349E-25_WP, &
    & 1.688E-25_WP,    2.104E-25_WP,    2.608E-25_WP,    3.214E-25_WP,    3.935E-25_WP,    4.784E-25_WP,    5.772E-25_WP, &
    & 6.900E-25_WP,    8.148E-25_WP,    9.488E-25_WP,    1.087E-24_WP,    1.220E-24_WP,    1.342E-24_WP,    1.445E-24_WP, &
    & 1.523E-24_WP,    1.566E-24_WP,    1.569E-24_WP,    1.526E-24_WP,    1.438E-24_WP,    1.316E-24_WP,    1.154E-24_WP, &
    & 9.407E-25_WP,    6.942E-25_WP,    4.492E-25_WP,    2.467E-25_WP,    1.116E-25_WP,    4.013E-26_WP,    1.015E-26_WP /
data J(1,    36    ,:)    / 4.609E-26_WP,    5.846E-26_WP,    7.398E-26_WP,    9.339E-26_WP,    1.175E-25_WP,    1.474E-25_WP, &
    & 1.840E-25_WP,    2.286E-25_WP,    2.825E-25_WP,    3.470E-25_WP,    4.232E-25_WP,    5.125E-25_WP,    6.158E-25_WP, &
    & 7.330E-25_WP,    8.620E-25_WP,    9.995E-25_WP,    1.140E-24_WP,    1.274E-24_WP,    1.395E-24_WP,    1.497E-24_WP, &
    & 1.572E-24_WP,    1.613E-24_WP,    1.612E-24_WP,    1.567E-24_WP,    1.478E-24_WP,    1.357E-24_WP,    1.198E-24_WP, &
    & 9.870E-25_WP,    7.378E-25_WP,    4.837E-25_WP,    2.686E-25_WP,    1.225E-25_WP,    4.424E-26_WP,    1.122E-26_WP /
data J(1,    37    ,:)    / 5.074E-26_WP,    6.425E-26_WP,    8.118E-26_WP,    1.023E-25_WP,    1.284E-25_WP,    1.606E-25_WP, &
    & 2.000E-25_WP,    2.477E-25_WP,    3.049E-25_WP,    3.730E-25_WP,    4.531E-25_WP,    5.463E-25_WP,    6.535E-25_WP, &
    & 7.745E-25_WP,    9.068E-25_WP,    1.047E-24_WP,    1.189E-24_WP,    1.324E-24_WP,    1.443E-24_WP,    1.542E-24_WP, &
    & 1.615E-24_WP,    1.652E-24_WP,    1.649E-24_WP,    1.601E-24_WP,    1.510E-24_WP,    1.390E-24_WP,    1.236E-24_WP, &
    & 1.029E-24_WP,    7.795E-25_WP,    5.186E-25_WP,    2.916E-25_WP,    1.341E-25_WP,    4.872E-26_WP,    1.239E-26_WP /
data J(1,    38    ,:)    / 5.577E-26_WP,    7.051E-26_WP,    8.891E-26_WP,    1.118E-25_WP,    1.400E-25_WP,    1.746E-25_WP, &
    & 2.166E-25_WP,    2.673E-25_WP,    3.278E-25_WP,    3.992E-25_WP,    4.827E-25_WP,    5.793E-25_WP,    6.898E-25_WP, &
    & 8.138E-25_WP,    9.488E-25_WP,    1.091E-24_WP,    1.234E-24_WP,    1.368E-24_WP,    1.485E-24_WP,    1.583E-24_WP, &
    & 1.652E-24_WP,    1.686E-24_WP,    1.680E-24_WP,    1.628E-24_WP,    1.536E-24_WP,    1.417E-24_WP,    1.267E-24_WP, &
    & 1.065E-24_WP,    8.185E-25_WP,    5.531E-25_WP,    3.155E-25_WP,    1.466E-25_WP,    5.359E-26_WP,    1.368E-26_WP /
data J(1,    39    ,:)    / 6.120E-26_WP,    7.723E-26_WP,    9.716E-26_WP,    1.218E-25_WP,    1.521E-25_WP,    1.891E-25_WP, &
    & 2.337E-25_WP,    2.872E-25_WP,    3.507E-25_WP,    4.251E-25_WP,    5.116E-25_WP,    6.109E-25_WP,    7.240E-25_WP, &
    & 8.504E-25_WP,    9.874E-25_WP,    1.131E-24_WP,    1.275E-24_WP,    1.408E-24_WP,    1.524E-24_WP,    1.618E-24_WP, &
    & 1.685E-24_WP,    1.717E-24_WP,    1.706E-24_WP,    1.652E-24_WP,    1.556E-24_WP,    1.437E-24_WP,    1.292E-24_WP, &
    & 1.096E-24_WP,    8.540E-25_WP,    5.868E-25_WP,    3.400E-25_WP,    1.599E-25_WP,    5.887E-26_WP,    1.509E-26_WP /
data J(1,    40    ,:)    / 6.704E-26_WP,    8.440E-26_WP,    1.059E-25_WP,    1.324E-25_WP,    1.648E-25_WP,    2.040E-25_WP, &
    & 2.512E-25_WP,    3.073E-25_WP,    3.733E-25_WP,    4.503E-25_WP,    5.391E-25_WP,    6.406E-25_WP,    7.557E-25_WP, &
    & 8.839E-25_WP,    1.022E-24_WP,    1.168E-24_WP,    1.312E-24_WP,    1.445E-24_WP,    1.559E-24_WP,    1.651E-24_WP, &
    & 1.716E-24_WP,    1.745E-24_WP,    1.731E-24_WP,    1.672E-24_WP,    1.573E-24_WP,    1.452E-24_WP,    1.311E-24_WP, &
    & 1.121E-24_WP,    8.853E-25_WP,    6.190E-25_WP,    3.649E-25_WP,    1.739E-25_WP,    6.457E-26_WP,    1.664E-26_WP /
data J(1,    41    ,:)    / 7.328E-26_WP,    9.201E-26_WP,    1.151E-25_WP,    1.434E-25_WP,    1.778E-25_WP,    2.192E-25_WP, &
    & 2.687E-25_WP,    3.271E-25_WP,    3.953E-25_WP,    4.744E-25_WP,    5.649E-25_WP,    6.680E-25_WP,    7.845E-25_WP, &
    & 9.141E-25_WP,    1.054E-24_WP,    1.201E-24_WP,    1.347E-24_WP,    1.479E-24_WP,    1.592E-24_WP,    1.683E-24_WP, &
    & 1.746E-24_WP,    1.772E-24_WP,    1.755E-24_WP,    1.692E-24_WP,    1.588E-24_WP,    1.465E-24_WP,    1.325E-24_WP, &
    & 1.141E-24_WP,    9.119E-25_WP,    6.489E-25_WP,    3.898E-25_WP,    1.885E-25_WP,    7.071E-26_WP,    1.832E-26_WP /
data J(1,    42    ,:)    / 7.989E-26_WP,    1.000E-25_WP,    1.247E-25_WP,    1.548E-25_WP,    1.911E-25_WP,    2.345E-25_WP, &
    & 2.860E-25_WP,    3.463E-25_WP,    4.163E-25_WP,    4.969E-25_WP,    5.887E-25_WP,    6.928E-25_WP,    8.104E-25_WP, &
    & 9.413E-25_WP,    1.083E-24_WP,    1.231E-24_WP,    1.379E-24_WP,    1.512E-24_WP,    1.625E-24_WP,    1.716E-24_WP, &
    & 1.778E-24_WP,    1.802E-24_WP,    1.781E-24_WP,    1.713E-24_WP,    1.603E-24_WP,    1.475E-24_WP,    1.336E-24_WP, &
    & 1.156E-24_WP,    9.337E-25_WP,    6.760E-25_WP,    4.142E-25_WP,    2.037E-25_WP,    7.728E-26_WP,    2.015E-26_WP /
data J(1,    43    ,:)    / 8.686E-26_WP,    1.084E-25_WP,    1.346E-25_WP,    1.664E-25_WP,    2.044E-25_WP,    2.496E-25_WP, &
    & 3.028E-25_WP,    3.646E-25_WP,    4.360E-25_WP,    5.175E-25_WP,    6.101E-25_WP,    7.149E-25_WP,    8.334E-25_WP, &
    & 9.658E-25_WP,    1.109E-24_WP,    1.260E-24_WP,    1.410E-24_WP,    1.546E-24_WP,    1.659E-24_WP,    1.750E-24_WP, &
    & 1.812E-24_WP,    1.835E-24_WP,    1.811E-24_WP,    1.737E-24_WP,    1.620E-24_WP,    1.487E-24_WP,    1.345E-24_WP, &
    & 1.167E-24_WP,    9.510E-25_WP,    6.997E-25_WP,    4.376E-25_WP,    2.193E-25_WP,    8.426E-26_WP,    2.214E-26_WP /
data J(1,    44    ,:)    / 9.414E-26_WP,    1.170E-25_WP,    1.447E-25_WP,    1.780E-25_WP,    2.176E-25_WP,    2.643E-25_WP, &
    & 3.188E-25_WP,    3.818E-25_WP,    4.540E-25_WP,    5.362E-25_WP,    6.292E-25_WP,    7.345E-25_WP,    8.540E-25_WP, &
    & 9.881E-25_WP,    1.134E-24_WP,    1.289E-24_WP,    1.443E-24_WP,    1.581E-24_WP,    1.695E-24_WP,    1.789E-24_WP, &
    & 1.851E-24_WP,    1.872E-24_WP,    1.845E-24_WP,    1.766E-24_WP,    1.641E-24_WP,    1.500E-24_WP,    1.355E-24_WP, &
    & 1.176E-24_WP,    9.641E-25_WP,    7.198E-25_WP,    4.596E-25_WP,    2.350E-25_WP,    9.165E-26_WP,    2.429E-26_WP /
data J(1,    45    ,:)    / 1.017E-25_WP,    1.258E-25_WP,    1.549E-25_WP,    1.895E-25_WP,    2.305E-25_WP,    2.784E-25_WP, &
    & 3.339E-25_WP,    3.976E-25_WP,    4.703E-25_WP,    5.527E-25_WP,    6.461E-25_WP,    7.519E-25_WP,    8.727E-25_WP, &
    & 1.009E-24_WP,    1.159E-24_WP,    1.318E-24_WP,    1.477E-24_WP,    1.619E-24_WP,    1.736E-24_WP,    1.832E-24_WP, &
    & 1.895E-24_WP,    1.916E-24_WP,    1.885E-24_WP,    1.800E-24_WP,    1.666E-24_WP,    1.517E-24_WP,    1.366E-24_WP, &
    & 1.183E-24_WP,    9.740E-25_WP,    7.362E-25_WP,    4.798E-25_WP,    2.506E-25_WP,    9.939E-26_WP,    2.660E-26_WP /
data J(1,    46    ,:)    / 1.094E-25_WP,    1.347E-25_WP,    1.650E-25_WP,    2.008E-25_WP,    2.428E-25_WP,    2.916E-25_WP, &
    & 3.477E-25_WP,    4.118E-25_WP,    4.847E-25_WP,    5.673E-25_WP,    6.610E-25_WP,    7.676E-25_WP,    8.901E-25_WP, &
    & 1.030E-24_WP,    1.183E-24_WP,    1.348E-24_WP,    1.513E-24_WP,    1.660E-24_WP,    1.782E-24_WP,    1.881E-24_WP, &
    & 1.946E-24_WP,    1.966E-24_WP,    1.932E-24_WP,    1.840E-24_WP,    1.697E-24_WP,    1.537E-24_WP,    1.379E-24_WP, &
    & 1.191E-24_WP,    9.816E-25_WP,    7.491E-25_WP,    4.977E-25_WP,    2.659E-25_WP,    1.074E-25_WP,    2.908E-26_WP /
data J(1,    47    ,:)    / 1.171E-25_WP,    1.435E-25_WP,    1.749E-25_WP,    2.116E-25_WP,    2.544E-25_WP,    3.037E-25_WP, &
    & 3.602E-25_WP,    4.245E-25_WP,    4.975E-25_WP,    5.802E-25_WP,    6.743E-25_WP,    7.820E-25_WP,    9.068E-25_WP, &
    & 1.050E-24_WP,    1.209E-24_WP,    1.381E-24_WP,    1.553E-24_WP,    1.707E-24_WP,    1.833E-24_WP,    1.936E-24_WP, &
    & 2.004E-24_WP,    2.024E-24_WP,    1.987E-24_WP,    1.887E-24_WP,    1.734E-24_WP,    1.563E-24_WP,    1.396E-24_WP, &
    & 1.201E-24_WP,    9.880E-25_WP,    7.589E-25_WP,    5.131E-25_WP,    2.806E-25_WP,    1.157E-25_WP,    3.172E-26_WP /
data J(1,    48    ,:)    / 1.249E-25_WP,    1.522E-25_WP,    1.844E-25_WP,    2.218E-25_WP,    2.651E-25_WP,    3.148E-25_WP, &
    & 3.714E-25_WP,    4.357E-25_WP,    5.087E-25_WP,    5.917E-25_WP,    6.865E-25_WP,    7.958E-25_WP,    9.235E-25_WP, &
    & 1.071E-24_WP,    1.237E-24_WP,    1.416E-24_WP,    1.598E-24_WP,    1.759E-24_WP,    1.891E-24_WP,    1.999E-24_WP, &
    & 2.069E-24_WP,    2.090E-24_WP,    2.049E-24_WP,    1.942E-24_WP,    1.777E-24_WP,    1.594E-24_WP,    1.417E-24_WP, &
    & 1.212E-24_WP,    9.941E-25_WP,    7.662E-25_WP,    5.259E-25_WP,    2.944E-25_WP,    1.241E-25_WP,    3.451E-26_WP /
data J(1,    49    ,:)    / 1.325E-25_WP,    1.606E-25_WP,    1.933E-25_WP,    2.313E-25_WP,    2.749E-25_WP,    3.247E-25_WP, &
    & 3.813E-25_WP,    4.457E-25_WP,    5.188E-25_WP,    6.022E-25_WP,    6.980E-25_WP,    8.094E-25_WP,    9.407E-25_WP, &
    & 1.094E-24_WP,    1.267E-24_WP,    1.456E-24_WP,    1.647E-24_WP,    1.817E-24_WP,    1.956E-24_WP,    2.069E-24_WP, &
    & 2.143E-24_WP,    2.164E-24_WP,    2.119E-24_WP,    2.004E-24_WP,    1.827E-24_WP,    1.631E-24_WP,    1.442E-24_WP, &
    & 1.227E-24_WP,    1.001E-24_WP,    7.717E-25_WP,    5.361E-25_WP,    3.069E-25_WP,    1.325E-25_WP,    3.744E-26_WP /
data J(1,    50    ,:)    / 1.398E-25_WP,    1.685E-25_WP,    2.017E-25_WP,    2.400E-25_WP,    2.837E-25_WP,    3.335E-25_WP, &
    & 3.902E-25_WP,    4.545E-25_WP,    5.279E-25_WP,    6.120E-25_WP,    7.093E-25_WP,    8.232E-25_WP,    9.589E-25_WP, &
    & 1.119E-24_WP,    1.300E-24_WP,    1.499E-24_WP,    1.702E-24_WP,    1.881E-24_WP,    2.028E-24_WP,    2.147E-24_WP, &
    & 2.225E-24_WP,    2.246E-24_WP,    2.197E-24_WP,    2.073E-24_WP,    1.884E-24_WP,    1.673E-24_WP,    1.472E-24_WP, &
    & 1.244E-24_WP,    1.009E-24_WP,    7.761E-25_WP,    5.439E-25_WP,    3.181E-25_WP,    1.409E-25_WP,    4.049E-26_WP /
data J(1,    51    ,:)    / 1.469E-25_WP,    1.760E-25_WP,    2.095E-25_WP,    2.479E-25_WP,    2.917E-25_WP,    3.414E-25_WP, &
    & 3.981E-25_WP,    4.626E-25_WP,    5.364E-25_WP,    6.216E-25_WP,    7.207E-25_WP,    8.377E-25_WP,    9.783E-25_WP, &
    & 1.145E-24_WP,    1.336E-24_WP,    1.546E-24_WP,    1.761E-24_WP,    1.952E-24_WP,    2.106E-24_WP,    2.233E-24_WP, &
    & 2.315E-24_WP,    2.337E-24_WP,    2.284E-24_WP,    2.150E-24_WP,    1.947E-24_WP,    1.721E-24_WP,    1.506E-24_WP, &
    & 1.264E-24_WP,    1.018E-24_WP,    7.800E-25_WP,    5.496E-25_WP,    3.276E-25_WP,    1.489E-25_WP,    4.364E-26_WP /
data J(1,    52    ,:)    / 1.535E-25_WP,    1.829E-25_WP,    2.166E-25_WP,    2.550E-25_WP,    2.988E-25_WP,    3.486E-25_WP, &
    & 4.053E-25_WP,    4.701E-25_WP,    5.447E-25_WP,    6.312E-25_WP,    7.326E-25_WP,    8.531E-25_WP,    9.991E-25_WP, &
    & 1.174E-24_WP,    1.375E-24_WP,    1.598E-24_WP,    1.827E-24_WP,    2.029E-24_WP,    2.193E-24_WP,    2.327E-24_WP, &
    & 2.414E-24_WP,    2.436E-24_WP,    2.378E-24_WP,    2.235E-24_WP,    2.017E-24_WP,    1.774E-24_WP,    1.545E-24_WP, &
    & 1.288E-24_WP,    1.029E-24_WP,    7.841E-25_WP,    5.536E-25_WP,    3.355E-25_WP,    1.566E-25_WP,    4.685E-26_WP /
data J(1,    53    ,:)    / 1.597E-25_WP,    1.893E-25_WP,    2.230E-25_WP,    2.615E-25_WP,    3.052E-25_WP,    3.551E-25_WP, &
    & 4.120E-25_WP,    4.774E-25_WP,    5.530E-25_WP,    6.410E-25_WP,    7.450E-25_WP,    8.695E-25_WP,    1.022E-24_WP, &
    & 1.205E-24_WP,    1.418E-24_WP,    1.654E-24_WP,    1.897E-24_WP,    2.112E-24_WP,    2.286E-24_WP,    2.428E-24_WP, &
    & 2.520E-24_WP,    2.543E-24_WP,    2.481E-24_WP,    2.326E-24_WP,    2.093E-24_WP,    1.832E-24_WP,    1.587E-24_WP, &
    & 1.315E-24_WP,    1.042E-24_WP,    7.886E-25_WP,    5.565E-25_WP,    3.418E-25_WP,    1.638E-25_WP,    5.007E-26_WP /
data J(1,    54    ,:)    / 1.655E-25_WP,    1.951E-25_WP,    2.289E-25_WP,    2.674E-25_WP,    3.112E-25_WP,    3.612E-25_WP, &
    & 4.185E-25_WP,    4.846E-25_WP,    5.613E-25_WP,    6.513E-25_WP,    7.582E-25_WP,    8.870E-25_WP,    1.046E-24_WP, &
    & 1.239E-24_WP,    1.463E-24_WP,    1.714E-24_WP,    1.973E-24_WP,    2.202E-24_WP,    2.386E-24_WP,    2.537E-24_WP, &
    & 2.635E-24_WP,    2.658E-24_WP,    2.590E-24_WP,    2.425E-24_WP,    2.174E-24_WP,    1.895E-24_WP,    1.634E-24_WP, &
    & 1.345E-24_WP,    1.058E-24_WP,    7.938E-25_WP,    5.585E-25_WP,    3.466E-25_WP,    1.703E-25_WP,    5.327E-26_WP /
data J(1,    55    ,:)    / 1.708E-25_WP,    2.006E-25_WP,    2.344E-25_WP,    2.729E-25_WP,    3.168E-25_WP,    3.671E-25_WP, &
    & 4.250E-25_WP,    4.919E-25_WP,    5.700E-25_WP,    6.621E-25_WP,    7.722E-25_WP,    9.057E-25_WP,    1.071E-24_WP, &
    & 1.274E-24_WP,    1.511E-24_WP,    1.778E-24_WP,    2.054E-24_WP,    2.297E-24_WP,    2.493E-24_WP,    2.653E-24_WP, &
    & 2.756E-24_WP,    2.780E-24_WP,    2.706E-24_WP,    2.529E-24_WP,    2.262E-24_WP,    1.963E-24_WP,    1.685E-24_WP, &
    & 1.377E-24_WP,    1.075E-24_WP,    7.998E-25_WP,    5.602E-25_WP,    3.500E-25_WP,    1.760E-25_WP,    5.639E-26_WP /
data J(1,    56    ,:)    / 1.758E-25_WP,    2.056E-25_WP,    2.395E-25_WP,    2.781E-25_WP,    3.223E-25_WP,    3.729E-25_WP, &
    & 4.314E-25_WP,    4.994E-25_WP,    5.791E-25_WP,    6.735E-25_WP,    7.870E-25_WP,    9.255E-25_WP,    1.099E-24_WP, &
    & 1.312E-24_WP,    1.563E-24_WP,    1.845E-24_WP,    2.140E-24_WP,    2.399E-24_WP,    2.605E-24_WP,    2.775E-24_WP, &
    & 2.883E-24_WP,    2.908E-24_WP,    2.829E-24_WP,    2.639E-24_WP,    2.353E-24_WP,    2.035E-24_WP,    1.738E-24_WP, &
    & 1.412E-24_WP,    1.093E-24_WP,    8.068E-25_WP,    5.617E-25_WP,    3.524E-25_WP,    1.810E-25_WP,    5.936E-26_WP /
data J(1,    57    ,:)    / 1.805E-25_WP,    2.104E-25_WP,    2.444E-25_WP,    2.832E-25_WP,    3.276E-25_WP,    3.787E-25_WP, &
    & 4.380E-25_WP,    5.071E-25_WP,    5.886E-25_WP,    6.855E-25_WP,    8.026E-25_WP,    9.464E-25_WP,    1.127E-24_WP, &
    & 1.351E-24_WP,    1.616E-24_WP,    1.916E-24_WP,    2.230E-24_WP,    2.505E-24_WP,    2.723E-24_WP,    2.902E-24_WP, &
    & 3.016E-24_WP,    3.041E-24_WP,    2.956E-24_WP,    2.753E-24_WP,    2.449E-24_WP,    2.110E-24_WP,    1.795E-24_WP, &
    & 1.450E-24_WP,    1.113E-24_WP,    8.146E-25_WP,    5.632E-25_WP,    3.539E-25_WP,    1.852E-25_WP,    6.215E-26_WP /
data J(1,    58    ,:)    / 1.850E-25_WP,    2.150E-25_WP,    2.491E-25_WP,    2.881E-25_WP,    3.329E-25_WP,    3.847E-25_WP, &
    & 4.448E-25_WP,    5.152E-25_WP,    5.985E-25_WP,    6.981E-25_WP,    8.190E-25_WP,    9.682E-25_WP,    1.157E-24_WP, &
    & 1.393E-24_WP,    1.672E-24_WP,    1.990E-24_WP,    2.323E-24_WP,    2.615E-24_WP,    2.846E-24_WP,    3.034E-24_WP, &
    & 3.154E-24_WP,    3.179E-24_WP,    3.087E-24_WP,    2.871E-24_WP,    2.549E-24_WP,    2.188E-24_WP,    1.854E-24_WP, &
    & 1.489E-24_WP,    1.135E-24_WP,    8.233E-25_WP,    5.650E-25_WP,    3.549E-25_WP,    1.886E-25_WP,    6.469E-26_WP /
data J(1,    59    ,:)    / 1.894E-25_WP,    2.195E-25_WP,    2.538E-25_WP,    2.931E-25_WP,    3.384E-25_WP,    3.907E-25_WP, &
    & 4.518E-25_WP,    5.236E-25_WP,    6.089E-25_WP,    7.113E-25_WP,    8.361E-25_WP,    9.908E-25_WP,    1.188E-24_WP, &
    & 1.435E-24_WP,    1.730E-24_WP,    2.066E-24_WP,    2.420E-24_WP,    2.729E-24_WP,    2.972E-24_WP,    3.169E-24_WP, &
    & 3.295E-24_WP,    3.319E-24_WP,    3.221E-24_WP,    2.992E-24_WP,    2.650E-24_WP,    2.268E-24_WP,    1.915E-24_WP, &
    & 1.530E-24_WP,    1.158E-24_WP,    8.329E-25_WP,    5.670E-25_WP,    3.554E-25_WP,    1.913E-25_WP,    6.695E-26_WP /
data J(1,    60    ,:)    / 1.937E-25_WP,    2.240E-25_WP,    2.586E-25_WP,    2.982E-25_WP,    3.439E-25_WP,    3.970E-25_WP, &
    & 4.591E-25_WP,    5.324E-25_WP,    6.197E-25_WP,    7.250E-25_WP,    8.538E-25_WP,    1.014E-24_WP,    1.219E-24_WP, &
    & 1.479E-24_WP,    1.789E-24_WP,    2.144E-24_WP,    2.519E-24_WP,    2.846E-24_WP,    3.100E-24_WP,    3.308E-24_WP, &
    & 3.438E-24_WP,    3.463E-24_WP,    3.357E-24_WP,    3.114E-24_WP,    2.754E-24_WP,    2.350E-24_WP,    1.977E-24_WP, &
    & 1.572E-24_WP,    1.182E-24_WP,    8.433E-25_WP,    5.694E-25_WP,    3.558E-25_WP,    1.934E-25_WP,    6.891E-26_WP /
data J(1,    61    ,:)    / 1.981E-25_WP,    2.285E-25_WP,    2.634E-25_WP,    3.034E-25_WP,    3.496E-25_WP,    4.035E-25_WP, &
    & 4.667E-25_WP,    5.415E-25_WP,    6.310E-25_WP,    7.392E-25_WP,    8.720E-25_WP,    1.038E-24_WP,    1.252E-24_WP, &
    & 1.523E-24_WP,    1.849E-24_WP,    2.223E-24_WP,    2.621E-24_WP,    2.965E-24_WP,    3.231E-24_WP,    3.447E-24_WP, &
    & 3.583E-24_WP,    3.607E-24_WP,    3.495E-24_WP,    3.238E-24_WP,    2.859E-24_WP,    2.433E-24_WP,    2.041E-24_WP, &
    & 1.616E-24_WP,    1.207E-24_WP,    8.545E-25_WP,    5.723E-25_WP,    3.561E-25_WP,    1.950E-25_WP,    7.055E-26_WP /
data J(1,    62    ,:)    / 2.025E-25_WP,    2.332E-25_WP,    2.683E-25_WP,    3.087E-25_WP,    3.556E-25_WP,    4.102E-25_WP, &
    & 4.746E-25_WP,    5.510E-25_WP,    6.427E-25_WP,    7.539E-25_WP,    8.907E-25_WP,    1.062E-24_WP,    1.284E-24_WP, &
    & 1.568E-24_WP,    1.909E-24_WP,    2.304E-24_WP,    2.723E-24_WP,    3.086E-24_WP,    3.363E-24_WP,    3.588E-24_WP, &
    & 3.729E-24_WP,    3.752E-24_WP,    3.633E-24_WP,    3.363E-24_WP,    2.965E-24_WP,    2.517E-24_WP,    2.106E-24_WP, &
    & 1.660E-24_WP,    1.234E-24_WP,    8.665E-25_WP,    5.757E-25_WP,    3.564E-25_WP,    1.962E-25_WP,    7.188E-26_WP /
data J(1,    63    ,:)    / 2.069E-25_WP,    2.379E-25_WP,    2.734E-25_WP,    3.143E-25_WP,    3.618E-25_WP,    4.173E-25_WP, &
    & 4.829E-25_WP,    5.609E-25_WP,    6.549E-25_WP,    7.691E-25_WP,    9.099E-25_WP,    1.087E-24_WP,    1.317E-24_WP, &
    & 1.613E-24_WP,    1.970E-24_WP,    2.384E-24_WP,    2.826E-24_WP,    3.208E-24_WP,    3.496E-24_WP,    3.730E-24_WP, &
    & 3.875E-24_WP,    3.898E-24_WP,    3.771E-24_WP,    3.488E-24_WP,    3.071E-24_WP,    2.602E-24_WP,    2.172E-24_WP, &
    & 1.706E-24_WP,    1.261E-24_WP,    8.794E-25_WP,    5.796E-25_WP,    3.569E-25_WP,    1.971E-25_WP,    7.293E-26_WP /
data J(1,    64    ,:)    / 2.116E-25_WP,    2.428E-25_WP,    2.787E-25_WP,    3.201E-25_WP,    3.682E-25_WP,    4.247E-25_WP, &
    & 4.915E-25_WP,    5.713E-25_WP,    6.675E-25_WP,    7.847E-25_WP,    9.296E-25_WP,    1.112E-24_WP,    1.351E-24_WP, &
    & 1.658E-24_WP,    2.031E-24_WP,    2.465E-24_WP,    2.930E-24_WP,    3.330E-24_WP,    3.630E-24_WP,    3.872E-24_WP, &
    & 4.021E-24_WP,    4.043E-24_WP,    3.910E-24_WP,    3.613E-24_WP,    3.177E-24_WP,    2.687E-24_WP,    2.238E-24_WP, &
    & 1.752E-24_WP,    1.289E-24_WP,    8.933E-25_WP,    5.843E-25_WP,    3.577E-25_WP,    1.979E-25_WP,    7.373E-26_WP /
data J(1,    65    ,:)    / 2.164E-25_WP,    2.480E-25_WP,    2.842E-25_WP,    3.262E-25_WP,    3.750E-25_WP,    4.325E-25_WP, &
    & 5.007E-25_WP,    5.822E-25_WP,    6.808E-25_WP,    8.012E-25_WP,    9.502E-25_WP,    1.138E-24_WP,    1.385E-24_WP, &
    & 1.704E-24_WP,    2.093E-24_WP,    2.548E-24_WP,    3.037E-24_WP,    3.455E-24_WP,    3.766E-24_WP,    4.017E-24_WP, &
    & 4.171E-24_WP,    4.192E-24_WP,    4.052E-24_WP,    3.742E-24_WP,    3.288E-24_WP,    2.776E-24_WP,    2.308E-24_WP, &
    & 1.802E-24_WP,    1.320E-24_WP,    9.094E-25_WP,    5.905E-25_WP,    3.593E-25_WP,    1.987E-25_WP,    7.439E-26_WP /
data J(1,    66    ,:)    / 2.215E-25_WP,    2.535E-25_WP,    2.902E-25_WP,    3.328E-25_WP,    3.825E-25_WP,    4.411E-25_WP, &
    & 5.108E-25_WP,    5.945E-25_WP,    6.957E-25_WP,    8.197E-25_WP,    9.733E-25_WP,    1.168E-24_WP,    1.423E-24_WP, &
    & 1.756E-24_WP,    2.163E-24_WP,    2.640E-24_WP,    3.155E-24_WP,    3.595E-24_WP,    3.919E-24_WP,    4.181E-24_WP, &
    & 4.342E-24_WP,    4.364E-24_WP,    4.218E-24_WP,    3.894E-24_WP,    3.420E-24_WP,    2.884E-24_WP,    2.395E-24_WP, &
    & 1.865E-24_WP,    1.362E-24_WP,    9.330E-25_WP,    6.017E-25_WP,    3.638E-25_WP,    2.008E-25_WP,    7.519E-26_WP /
data J(1,    67    ,:)    / 2.273E-25_WP,    2.598E-25_WP,    2.973E-25_WP,    3.409E-25_WP,    3.919E-25_WP,    4.524E-25_WP, &
    & 5.245E-25_WP,    6.114E-25_WP,    7.170E-25_WP,    8.466E-25_WP,    1.008E-24_WP,    1.212E-24_WP,    1.482E-24_WP, &
    & 1.834E-24_WP,    2.267E-24_WP,    2.777E-24_WP,    3.331E-24_WP,    3.804E-24_WP,    4.156E-24_WP,    4.443E-24_WP, &
    & 4.622E-24_WP,    4.652E-24_WP,    4.502E-24_WP,    4.160E-24_WP,    3.656E-24_WP,    3.084E-24_WP,    2.560E-24_WP, &
    & 1.992E-24_WP,    1.451E-24_WP,    9.901E-25_WP,    6.347E-25_WP,    3.809E-25_WP,    2.089E-25_WP,    7.734E-26_WP /
data J(1,    68    ,:)    / 2.352E-25_WP,    2.691E-25_WP,    3.086E-25_WP,    3.548E-25_WP,    4.094E-25_WP,    4.747E-25_WP, &
    & 5.533E-25_WP,    6.489E-25_WP,    7.660E-25_WP,    9.109E-25_WP,    1.092E-24_WP,    1.324E-24_WP,    1.631E-24_WP, &
    & 2.035E-24_WP,    2.534E-24_WP,    3.125E-24_WP,    3.772E-24_WP,    4.331E-24_WP,    4.774E-24_WP,    5.145E-24_WP, &
    & 5.391E-24_WP,    5.461E-24_WP,    5.313E-24_WP,    4.932E-24_WP,    4.352E-24_WP,    3.689E-24_WP,    3.070E-24_WP, &
    & 2.394E-24_WP,    1.745E-24_WP,    1.191E-24_WP,    7.601E-25_WP,    4.516E-25_WP,    2.431E-25_WP,    8.588E-26_WP /
data J(1,    69    ,:)    / 2.518E-25_WP,    2.908E-25_WP,    3.371E-25_WP,    3.927E-25_WP,    4.602E-25_WP,    5.427E-25_WP, &
    & 6.446E-25_WP,    7.712E-25_WP,    9.297E-25_WP,    1.130E-24_WP,    1.384E-24_WP,    1.713E-24_WP,    2.154E-24_WP, &
    & 2.735E-24_WP,    3.459E-24_WP,    4.326E-24_WP,    5.285E-24_WP,    6.140E-24_WP,    6.918E-24_WP,    7.603E-24_WP, &
    & 8.105E-24_WP,    8.334E-24_WP,    8.212E-24_WP,    7.707E-24_WP,    6.873E-24_WP,    5.897E-24_WP,    4.944E-24_WP, &
    & 3.881E-24_WP,    2.848E-24_WP,    1.952E-24_WP,    1.245E-24_WP,    7.304E-25_WP,    3.789E-25_WP,    1.195E-25_WP /
data J(1,    70    ,:)    / 3.022E-25_WP,    3.601E-25_WP,    4.324E-25_WP,    5.234E-25_WP,    6.386E-25_WP,    7.855E-25_WP, &
    & 9.734E-25_WP,    1.215E-24_WP,    1.525E-24_WP,    1.927E-24_WP,    2.449E-24_WP,    3.137E-24_WP,    4.060E-24_WP, &
    & 5.275E-24_WP,    6.802E-24_WP,    8.646E-24_WP,    1.070E-23_WP,    1.261E-23_WP,    1.459E-23_WP,    1.640E-23_WP, &
    & 1.782E-23_WP,    1.863E-23_WP,    1.861E-23_WP,    1.767E-23_WP,    1.594E-23_WP,    1.386E-23_WP,    1.171E-23_WP, &
    & 9.264E-24_WP,    6.847E-24_WP,    4.718E-24_WP,    3.011E-24_WP,    1.748E-24_WP,    8.737E-25_WP,    2.399E-25_WP /
data J(1,    71    ,:)    / 3.022E-25_WP,    3.601E-25_WP,    4.324E-25_WP,    5.234E-25_WP,    6.386E-25_WP,    7.855E-25_WP, &
    & 9.734E-25_WP,    1.215E-24_WP,    1.525E-24_WP,    1.927E-24_WP,    2.449E-24_WP,    3.137E-24_WP,    4.060E-24_WP, &
    & 5.275E-24_WP,    6.802E-24_WP,    8.646E-24_WP,    1.070E-23_WP,    1.261E-23_WP,    1.459E-23_WP,    1.640E-23_WP, &
    & 1.782E-23_WP,    1.863E-23_WP,    1.861E-23_WP,    1.767E-23_WP,    1.594E-23_WP,    1.386E-23_WP,    1.171E-23_WP, &
    & 9.264E-24_WP,    6.847E-24_WP,    4.718E-24_WP,    3.011E-24_WP,    1.748E-24_WP,    8.737E-25_WP,    2.399E-25_WP /
data J(1,    72    ,:)    / 3.551E-25_WP,    4.318E-25_WP,    5.297E-25_WP,    6.554E-25_WP,    8.178E-25_WP,    1.029E-24_WP, &
    & 1.304E-24_WP,    1.663E-24_WP,    2.135E-24_WP,    2.757E-24_WP,    3.584E-24_WP,    4.697E-24_WP,    6.201E-24_WP, &
    & 8.186E-24_WP,    1.070E-23_WP,    1.373E-23_WP,    1.709E-23_WP,    2.101E-23_WP,    2.534E-23_WP,    2.989E-23_WP, &
    & 3.432E-23_WP,    3.818E-23_WP,    4.089E-23_WP,    4.191E-23_WP,    4.098E-23_WP,    3.849E-23_WP,    3.412E-23_WP, &
    & 2.758E-23_WP,    2.000E-23_WP,    1.277E-23_WP,    7.009E-24_WP,    3.212E-24_WP,    1.142E-24_WP,    3.191E-25_WP /
data J(1,    73    ,:)    / 4.207E-25_WP,    5.204E-25_WP,    6.495E-25_WP,    8.174E-25_WP,    1.037E-24_WP,    1.325E-24_WP, &
    & 1.705E-24_WP,    2.208E-24_WP,    2.875E-24_WP,    3.765E-24_WP,    4.964E-24_WP,    6.581E-24_WP,    8.753E-24_WP, &
    & 1.160E-23_WP,    1.515E-23_WP,    1.933E-23_WP,    2.441E-23_WP,    3.034E-23_WP,    3.697E-23_WP,    4.401E-23_WP, &
    & 5.089E-23_WP,    5.684E-23_WP,    6.091E-23_WP,    6.224E-23_WP,    6.053E-23_WP,    5.587E-23_WP,    4.805E-23_WP, &
    & 3.742E-23_WP,    2.596E-23_WP,    1.576E-23_WP,    8.198E-24_WP,    3.545E-24_WP,    1.249E-24_WP,    3.452E-25_WP /
data J(1,    74    ,:)    / 5.009E-25_WP,    6.284E-25_WP,    7.950E-25_WP,    1.014E-24_WP,    1.302E-24_WP,    1.683E-24_WP, &
    & 2.189E-24_WP,    2.863E-24_WP,    3.764E-24_WP,    4.977E-24_WP,    6.612E-24_WP,    8.805E-24_WP,    1.172E-23_WP, &
    & 1.546E-23_WP,    2.001E-23_WP,    2.567E-23_WP,    3.247E-23_WP,    4.035E-23_WP,    4.907E-23_WP,    5.813E-23_WP, &
    & 6.672E-23_WP,    7.373E-23_WP,    7.797E-23_WP,    7.854E-23_WP,    7.494E-23_WP,    6.749E-23_WP,    5.650E-23_WP, &
    & 4.286E-23_WP,    2.904E-23_WP,    1.729E-23_WP,    8.855E-24_WP,    3.816E-24_WP,    1.339E-24_WP,    3.726E-25_WP /
data J(1,    75    ,:)    / 5.978E-25_WP,    7.585E-25_WP,    9.699E-25_WP,    1.249E-24_WP,    1.619E-24_WP,    2.111E-24_WP, &
    & 2.766E-24_WP,    3.643E-24_WP,    4.824E-24_WP,    6.413E-24_WP,    8.542E-24_WP,    1.136E-23_WP,    1.505E-23_WP, &
    & 1.965E-23_WP,    2.547E-23_WP,    3.260E-23_WP,    4.106E-23_WP,    5.071E-23_WP,    6.115E-23_WP,    7.165E-23_WP, &
    & 8.117E-23_WP,    8.839E-23_WP,    9.208E-23_WP,    9.120E-23_WP,    8.544E-23_WP,    7.556E-23_WP,    6.227E-23_WP, &
    & 4.667E-23_WP,    3.137E-23_WP,    1.858E-23_WP,    9.504E-24_WP,    4.089E-24_WP,    1.433E-24_WP,    4.014E-25_WP /
data J(1,    76    ,:)    / 7.138E-25_WP,    9.139E-25_WP,    1.178E-24_WP,    1.529E-24_WP,    1.995E-24_WP,    2.617E-24_WP, &
    & 3.448E-24_WP,    4.567E-24_WP,    6.070E-24_WP,    8.082E-24_WP,    1.075E-23_WP,    1.422E-23_WP,    1.864E-23_WP, &
    & 2.433E-23_WP,    3.139E-23_WP,    3.993E-23_WP,    4.991E-23_WP,    6.103E-23_WP,    7.273E-23_WP,    8.410E-23_WP, &
    & 9.393E-23_WP,    1.009E-22_WP,    1.036E-22_WP,    1.012E-22_WP,    9.365E-23_WP,    8.201E-23_WP,    6.714E-23_WP, &
    & 5.013E-23_WP,    3.362E-23_WP,    1.990E-23_WP,    1.017E-23_WP,    4.373E-24_WP,    1.530E-24_WP,    4.314E-25_WP /
data J(1,    77    ,:)    / 8.265E-25_WP,    1.080E-24_WP,    1.421E-24_WP,    1.882E-24_WP,    2.506E-24_WP,    3.354E-24_WP, &
    & 4.507E-24_WP,    6.084E-24_WP,    8.239E-24_WP,    1.118E-23_WP,    1.516E-23_WP,    2.051E-23_WP,    2.762E-23_WP, &
    & 3.711E-23_WP,    4.941E-23_WP,    6.496E-23_WP,    8.395E-23_WP,    1.058E-22_WP,    1.342E-22_WP,    1.663E-22_WP, &
    & 2.006E-22_WP,    2.346E-22_WP,    2.644E-22_WP,    2.854E-22_WP,    2.936E-22_WP,    2.891E-22_WP,    2.717E-22_WP, &
    & 2.351E-22_WP,    1.841E-22_WP,    1.279E-22_WP,    7.634E-23_WP,    3.737E-23_WP,    1.397E-23_WP,    2.891E-24_WP /
data J(1,    78    ,:)    / 9.910E-25_WP,    1.302E-24_WP,    1.721E-24_WP,    2.287E-24_WP,    3.053E-24_WP,    4.094E-24_WP, &
    & 5.512E-24_WP,    7.443E-24_WP,    1.007E-23_WP,    1.361E-23_WP,    1.835E-23_WP,    2.461E-23_WP,    3.297E-23_WP, &
    & 4.384E-23_WP,    5.756E-23_WP,    7.430E-23_WP,    9.365E-23_WP,    1.182E-22_WP,    1.456E-22_WP,    1.743E-22_WP, &
    & 2.019E-22_WP,    2.247E-22_WP,    2.388E-22_WP,    2.409E-22_WP,    2.307E-22_WP,    2.096E-22_WP,    1.753E-22_WP, &
    & 1.319E-22_WP,    8.747E-23_WP,    4.978E-23_WP,    2.344E-23_WP,    8.719E-24_WP,    2.298E-24_WP,    4.785E-25_WP /
data J(1,    79    ,:)    / 1.187E-24_WP,    1.566E-24_WP,    2.077E-24_WP,    2.766E-24_WP,    3.701E-24_WP,    4.970E-24_WP, &
    & 6.693E-24_WP,    9.025E-24_WP,    1.216E-23_WP,    1.635E-23_WP,    2.187E-23_WP,    2.919E-23_WP,    3.872E-23_WP, &
    & 5.082E-23_WP,    6.561E-23_WP,    8.280E-23_WP,    1.044E-22_WP,    1.284E-22_WP,    1.536E-22_WP,    1.777E-22_WP, &
    & 1.979E-22_WP,    2.106E-22_WP,    2.131E-22_WP,    2.046E-22_WP,    1.861E-22_WP,    1.576E-22_WP,    1.228E-22_WP, &
    & 8.626E-23_WP,    5.391E-23_WP,    2.956E-23_WP,    1.401E-23_WP,    5.609E-24_WP,    1.870E-24_WP,    5.152E-25_WP /
data J(1,    80    ,:)    / 1.417E-24_WP,    1.874E-24_WP,    2.491E-24_WP,    3.324E-24_WP,    4.452E-24_WP,    5.980E-24_WP, &
    & 8.042E-24_WP,    1.081E-23_WP,    1.449E-23_WP,    1.932E-23_WP,    2.571E-23_WP,    3.399E-23_WP,    4.453E-23_WP, &
    & 5.751E-23_WP,    7.271E-23_WP,    9.171E-23_WP,    1.130E-22_WP,    1.355E-22_WP,    1.575E-22_WP,    1.766E-22_WP, &
    & 1.900E-22_WP,    1.956E-22_WP,    1.922E-22_WP,    1.798E-22_WP,    1.583E-22_WP,    1.313E-22_WP,    1.020E-22_WP, &
    & 7.281E-23_WP,    4.716E-23_WP,    2.726E-23_WP,    1.372E-23_WP,    5.818E-24_WP,    1.986E-24_WP,    5.493E-25_WP /
data J(1,    81    ,:)    / 1.683E-24_WP,    2.231E-24_WP,    2.969E-24_WP,    3.967E-24_WP,    5.313E-24_WP,    7.124E-24_WP, &
    & 9.547E-24_WP,    1.276E-23_WP,    1.697E-23_WP,    2.252E-23_WP,    2.969E-23_WP,    3.878E-23_WP,    5.003E-23_WP, &
    & 6.337E-23_WP,    8.002E-23_WP,    9.886E-23_WP,    1.191E-22_WP,    1.394E-22_WP,    1.578E-22_WP,    1.726E-22_WP, &
    & 1.817E-22_WP,    1.841E-22_WP,    1.791E-22_WP,    1.660E-22_WP,    1.465E-22_WP,    1.234E-22_WP,    9.843E-23_WP, &
    & 7.235E-23_WP,    4.815E-23_WP,    2.838E-23_WP,    1.444E-23_WP,    6.135E-24_WP,    2.090E-24_WP,    5.828E-25_WP /
data J(1,    82    ,:)    / 1.990E-24_WP,    2.641E-24_WP,    3.518E-24_WP,    4.698E-24_WP,    6.281E-24_WP,    8.394E-24_WP, &
    & 1.119E-23_WP,    1.484E-23_WP,    1.964E-23_WP,    2.582E-23_WP,    3.364E-23_WP,    4.331E-23_WP,    5.485E-23_WP, &
    & 6.933E-23_WP,    8.590E-23_WP,    1.040E-22_WP,    1.227E-22_WP,    1.406E-22_WP,    1.563E-22_WP,    1.684E-22_WP, &
    & 1.761E-22_WP,    1.783E-22_WP,    1.737E-22_WP,    1.625E-22_WP,    1.456E-22_WP,    1.250E-22_WP,    1.014E-22_WP, &
    & 7.539E-23_WP,    5.047E-23_WP,    2.981E-23_WP,    1.516E-23_WP,    6.433E-24_WP,    2.189E-24_WP,    6.162E-25_WP /
data J(1,    83    ,:)    / 2.341E-24_WP,    3.109E-24_WP,    4.139E-24_WP,    5.518E-24_WP,    7.352E-24_WP,    9.772E-24_WP, &
    & 1.293E-23_WP,    1.706E-23_WP,    2.237E-23_WP,    2.908E-23_WP,    3.736E-23_WP,    4.725E-23_WP,    5.970E-23_WP, &
    & 7.415E-23_WP,    9.020E-23_WP,    1.072E-22_WP,    1.244E-22_WP,    1.406E-22_WP,    1.548E-22_WP,    1.665E-22_WP, &
    & 1.747E-22_WP,    1.778E-22_WP,    1.751E-22_WP,    1.658E-22_WP,    1.503E-22_WP,    1.302E-22_WP,    1.061E-22_WP, &
    & 7.899E-23_WP,    5.288E-23_WP,    3.121E-23_WP,    1.586E-23_WP,    6.728E-24_WP,    2.288E-24_WP,    6.499E-25_WP /
data J(1,    84    ,:)    / 2.740E-24_WP,    3.637E-24_WP,    4.833E-24_WP,    6.421E-24_WP,    8.510E-24_WP,    1.123E-23_WP, &
    & 1.478E-23_WP,    1.933E-23_WP,    2.507E-23_WP,    3.214E-23_WP,    4.060E-23_WP,    5.122E-23_WP,    6.365E-23_WP, &
    & 7.774E-23_WP,    9.307E-23_WP,    1.091E-22_WP,    1.252E-22_WP,    1.408E-22_WP,    1.552E-22_WP,    1.678E-22_WP, &
    & 1.772E-22_WP,    1.821E-22_WP,    1.810E-22_WP,    1.726E-22_WP,    1.571E-22_WP,    1.362E-22_WP,    1.111E-22_WP, &
    & 8.270E-23_WP,    5.534E-23_WP,    3.265E-23_WP,    1.658E-23_WP,    7.028E-24_WP,    2.389E-24_WP,    6.842E-25_WP /
data J(1,    85    ,:)    / 3.189E-24_WP,    4.224E-24_WP,    5.595E-24_WP,    7.394E-24_WP,    9.731E-24_WP,    1.277E-23_WP, &
    & 1.666E-23_WP,    2.156E-23_WP,    2.759E-23_WP,    3.481E-23_WP,    4.385E-23_WP,    5.445E-23_WP,    6.664E-23_WP, &
    & 8.025E-23_WP,    9.496E-23_WP,    1.105E-22_WP,    1.264E-22_WP,    1.425E-22_WP,    1.580E-22_WP,    1.720E-22_WP, &
    & 1.831E-22_WP,    1.895E-22_WP,    1.891E-22_WP,    1.808E-22_WP,    1.646E-22_WP,    1.427E-22_WP,    1.163E-22_WP, &
    & 8.657E-23_WP,    5.791E-23_WP,    3.415E-23_WP,    1.734E-23_WP,    7.344E-24_WP,    2.494E-24_WP,    7.195E-25_WP /
data J(1,    86    ,:)    / 3.687E-24_WP,    4.868E-24_WP,    6.414E-24_WP,    8.418E-24_WP,    1.102E-23_WP,    1.434E-23_WP, &
    & 1.851E-23_WP,    2.364E-23_WP,    2.980E-23_WP,    3.748E-23_WP,    4.652E-23_WP,    5.695E-23_WP,    6.882E-23_WP, &
    & 8.206E-23_WP,    9.653E-23_WP,    1.122E-22_WP,    1.289E-22_WP,    1.462E-22_WP,    1.632E-22_WP,    1.788E-22_WP, &
    & 1.914E-22_WP,    1.986E-22_WP,    1.985E-22_WP,    1.898E-22_WP,    1.727E-22_WP,    1.496E-22_WP,    1.219E-22_WP, &
    & 9.074E-23_WP,    6.068E-23_WP,    3.577E-23_WP,    1.815E-23_WP,    7.684E-24_WP,    2.608E-24_WP,    7.567E-25_WP /
data J(1,    87    ,:)    / 4.247E-24_WP,    5.579E-24_WP,    7.302E-24_WP,    9.534E-24_WP,    1.238E-23_WP,    1.594E-23_WP, &
    & 2.033E-23_WP,    2.560E-23_WP,    3.217E-23_WP,    3.991E-23_WP,    4.890E-23_WP,    5.920E-23_WP,    7.097E-23_WP, &
    & 8.431E-23_WP,    9.927E-23_WP,    1.160E-22_WP,    1.343E-22_WP,    1.534E-22_WP,    1.725E-22_WP,    1.901E-22_WP, &
    & 2.044E-22_WP,    2.128E-22_WP,    2.133E-22_WP,    2.044E-22_WP,    1.864E-22_WP,    1.618E-22_WP,    1.323E-22_WP, &
    & 9.869E-23_WP,    6.615E-23_WP,    3.908E-23_WP,    1.989E-23_WP,    8.460E-24_WP,    2.906E-24_WP,    8.743E-25_WP /
data J(1,    88    ,:)    / 4.867E-24_WP,    6.354E-24_WP,    8.275E-24_WP,    1.072E-23_WP,    1.378E-23_WP,    1.754E-23_WP, &
    & 2.207E-23_WP,    2.771E-23_WP,    3.438E-23_WP,    4.216E-23_WP,    5.116E-23_WP,    6.153E-23_WP,    7.359E-23_WP, &
    & 8.762E-23_WP,    1.038E-22_WP,    1.223E-22_WP,    1.426E-22_WP,    1.642E-22_WP,    1.857E-22_WP,    2.057E-22_WP, &
    & 2.219E-22_WP,    2.318E-22_WP,    2.329E-22_WP,    2.238E-22_WP,    2.046E-22_WP,    1.780E-22_WP,    1.458E-22_WP, &
    & 1.089E-22_WP,    7.310E-23_WP,    4.329E-23_WP,    2.215E-23_WP,    9.564E-24_WP,    3.392E-24_WP,    9.367E-25_WP /
data J(1,    89    ,:)    / 5.528E-24_WP,    7.180E-24_WP,    9.279E-24_WP,    1.190E-23_WP,    1.513E-23_WP,    1.902E-23_WP, &
    & 2.386E-23_WP,    2.960E-23_WP,    3.632E-23_WP,    4.415E-23_WP,    5.326E-23_WP,    6.395E-23_WP,    7.667E-23_WP, &
    & 9.185E-23_WP,    1.096E-22_WP,    1.300E-22_WP,    1.525E-22_WP,    1.763E-22_WP,    2.000E-22_WP,    2.218E-22_WP, &
    & 2.395E-22_WP,    2.503E-22_WP,    2.514E-22_WP,    2.413E-22_WP,    2.202E-22_WP,    1.913E-22_WP,    1.563E-22_WP, &
    & 1.165E-22_WP,    7.807E-23_WP,    4.616E-23_WP,    2.360E-23_WP,    1.018E-23_WP,    3.601E-24_WP,    9.930E-25_WP /
data J(1,    90    ,:)    / 6.233E-24_WP,    8.035E-24_WP,    1.029E-23_WP,    1.306E-23_WP,    1.640E-23_WP,    2.054E-23_WP, &
    & 2.548E-23_WP,    3.129E-23_WP,    3.809E-23_WP,    4.606E-23_WP,    5.550E-23_WP,    6.682E-23_WP,    8.059E-23_WP, &
    & 9.719E-23_WP,    1.167E-22_WP,    1.391E-22_WP,    1.638E-22_WP,    1.897E-22_WP,    2.154E-22_WP,    2.390E-22_WP, &
    & 2.580E-22_WP,    2.694E-22_WP,    2.704E-22_WP,    2.592E-22_WP,    2.362E-22_WP,    2.049E-22_WP,    1.674E-22_WP, &
    & 1.247E-22_WP,    8.356E-23_WP,    4.940E-23_WP,    2.525E-23_WP,    1.089E-23_WP,    3.851E-24_WP,    1.059E-24_WP /
data J(1,    91    ,:)    / 6.964E-24_WP,    8.899E-24_WP,    1.128E-23_WP,    1.414E-23_WP,    1.771E-23_WP,    2.195E-23_WP, &
    & 2.697E-23_WP,    3.288E-23_WP,    3.985E-23_WP,    4.817E-23_WP,    5.821E-23_WP,    7.047E-23_WP,    8.552E-23_WP, &
    & 1.037E-22_WP,    1.251E-22_WP,    1.496E-22_WP,    1.765E-22_WP,    2.047E-22_WP,    2.324E-22_WP,    2.579E-22_WP, &
    & 2.784E-22_WP,    2.906E-22_WP,    2.916E-22_WP,    2.795E-22_WP,    2.547E-22_WP,    2.210E-22_WP,    1.806E-22_WP, &
    & 1.347E-22_WP,    9.023E-23_WP,    5.337E-23_WP,    2.728E-23_WP,    1.177E-23_WP,    4.158E-24_WP,    1.137E-24_WP /
data J(1,    92    ,:)    / 7.706E-24_WP,    9.751E-24_WP,    1.222E-23_WP,    1.528E-23_WP,    1.894E-23_WP,    2.328E-23_WP, &
    & 2.842E-23_WP,    3.452E-23_WP,    4.184E-23_WP,    5.074E-23_WP,    6.165E-23_WP,    7.507E-23_WP,    9.158E-23_WP, &
    & 1.115E-22_WP,    1.350E-22_WP,    1.617E-22_WP,    1.911E-22_WP,    2.217E-22_WP,    2.519E-22_WP,    2.797E-22_WP, &
    & 3.021E-22_WP,    3.155E-22_WP,    3.169E-22_WP,    3.039E-22_WP,    2.772E-22_WP,    2.406E-22_WP,    1.967E-22_WP, &
    & 1.468E-22_WP,    9.843E-23_WP,    5.824E-23_WP,    2.978E-23_WP,    1.285E-23_WP,    4.535E-24_WP,    1.231E-24_WP /
data J(1,    93    ,:)    / 8.445E-24_WP,    1.057E-23_WP,    1.321E-23_WP,    1.637E-23_WP,    2.014E-23_WP,    2.461E-23_WP, &
    & 2.996E-23_WP,    3.641E-23_WP,    4.429E-23_WP,    5.400E-23_WP,    6.598E-23_WP,    8.075E-23_WP,    9.892E-23_WP, &
    & 1.209E-22_WP,    1.466E-22_WP,    1.759E-22_WP,    2.081E-22_WP,    2.417E-22_WP,    2.749E-22_WP,    3.056E-22_WP, &
    & 3.304E-22_WP,    3.456E-22_WP,    3.474E-22_WP,    3.336E-22_WP,    3.045E-22_WP,    2.645E-22_WP,    2.165E-22_WP, &
    & 1.616E-22_WP,    1.084E-22_WP,    6.418E-23_WP,    3.283E-23_WP,    1.416E-23_WP,    4.992E-24_WP,    1.343E-24_WP /
data J(1,    94    ,:)    / 9.166E-24_WP,    1.145E-23_WP,    1.419E-23_WP,    1.746E-23_WP,    2.137E-23_WP,    2.607E-23_WP, &
    & 3.176E-23_WP,    3.875E-23_WP,    4.739E-23_WP,    5.810E-23_WP,    7.133E-23_WP,    8.766E-23_WP,    1.077E-22_WP, &
    & 1.319E-22_WP,    1.603E-22_WP,    1.926E-22_WP,    2.282E-22_WP,    2.654E-22_WP,    3.024E-22_WP,    3.367E-22_WP, &
    & 3.646E-22_WP,    3.818E-22_WP,    3.843E-22_WP,    3.694E-22_WP,    3.375E-22_WP,    2.933E-22_WP,    2.402E-22_WP, &
    & 1.794E-22_WP,    1.204E-22_WP,    7.131E-23_WP,    3.648E-23_WP,    1.573E-23_WP,    5.538E-24_WP,    1.476E-24_WP /
data J(1,    95    ,:)    / 9.942E-24_WP,    1.232E-23_WP,    1.518E-23_WP,    1.860E-23_WP,    2.274E-23_WP,    2.778E-23_WP, &
    & 3.399E-23_WP,    4.169E-23_WP,    5.127E-23_WP,    6.315E-23_WP,    7.784E-23_WP,    9.596E-23_WP,    1.182E-22_WP, &
    & 1.450E-22_WP,    1.765E-22_WP,    2.125E-22_WP,    2.521E-22_WP,    2.937E-22_WP,    3.352E-22_WP,    3.738E-22_WP, &
    & 4.053E-22_WP,    4.250E-22_WP,    4.282E-22_WP,    4.119E-22_WP,    3.765E-22_WP,    3.274E-22_WP,    2.682E-22_WP, &
    & 2.004E-22_WP,    1.346E-22_WP,    7.970E-23_WP,    4.078E-23_WP,    1.757E-23_WP,    6.179E-24_WP,    1.631E-24_WP /
data J(1,    96    ,:)    / 1.073E-23_WP,    1.323E-23_WP,    1.624E-23_WP,    1.989E-23_WP,    2.436E-23_WP,    2.989E-23_WP, &
    & 3.677E-23_WP,    4.535E-23_WP,    5.602E-23_WP,    6.927E-23_WP,    8.565E-23_WP,    1.058E-22_WP,    1.306E-22_WP, &
    & 1.606E-22_WP,    1.958E-22_WP,    2.360E-22_WP,    2.805E-22_WP,    3.272E-22_WP,    3.739E-22_WP,    4.175E-22_WP, &
    & 4.531E-22_WP,    4.755E-22_WP,    4.793E-22_WP,    4.613E-22_WP,    4.219E-22_WP,    3.669E-22_WP,    3.006E-22_WP, &
    & 2.247E-22_WP,    1.509E-22_WP,    8.937E-23_WP,    4.572E-23_WP,    1.969E-23_WP,    6.914E-24_WP,    1.808E-24_WP /
data J(1,    97    ,:)    / 1.083E-23_WP,    1.336E-23_WP,    1.641E-23_WP,    2.010E-23_WP,    2.463E-23_WP,    3.023E-23_WP, &
    & 3.719E-23_WP,    4.587E-23_WP,    5.667E-23_WP,    7.007E-23_WP,    8.662E-23_WP,    1.070E-22_WP,    1.320E-22_WP, &
    & 1.623E-22_WP,    1.978E-22_WP,    2.385E-22_WP,    2.834E-22_WP,    3.306E-22_WP,    3.779E-22_WP,    4.220E-22_WP, &
    & 4.580E-22_WP,    4.806E-22_WP,    4.845E-22_WP,    4.663E-22_WP,    4.265E-22_WP,    3.709E-22_WP,    3.039E-22_WP, &
    & 2.272E-22_WP,    1.525E-22_WP,    9.035E-23_WP,    4.622E-23_WP,    1.991E-23_WP,    6.987E-24_WP,    1.826E-24_WP /
data J(1,    98    ,:)    / 1.191E-23_WP,    1.467E-23_WP,    1.806E-23_WP,    2.224E-23_WP,    2.746E-23_WP,    3.401E-23_WP, &
    & 4.226E-23_WP,    5.264E-23_WP,    6.569E-23_WP,    8.207E-23_WP,    1.026E-22_WP,    1.283E-22_WP,    1.604E-22_WP, &
    & 2.000E-22_WP,    2.476E-22_WP,    3.036E-22_WP,    3.677E-22_WP,    4.391E-22_WP,    5.158E-22_WP,    5.937E-22_WP, &
    & 6.665E-22_WP,    7.258E-22_WP,    7.621E-22_WP,    7.668E-22_WP,    7.365E-22_WP,    6.767E-22_WP,    5.842E-22_WP, &
    & 4.608E-22_WP,    3.267E-22_WP,    2.043E-22_WP,    1.104E-22_WP,    5.027E-23_WP,    1.870E-23_WP,    5.252E-24_WP /
data J(1,    99    ,:)    / 1.308E-23_WP,    1.616E-23_WP,    1.998E-23_WP,    2.478E-23_WP,    3.086E-23_WP,    3.856E-23_WP, &
    & 4.835E-23_WP,    6.077E-23_WP,    7.652E-23_WP,    9.650E-23_WP,    1.218E-22_WP,    1.540E-22_WP,    1.946E-22_WP, &
    & 2.454E-22_WP,    3.075E-22_WP,    3.816E-22_WP,    4.686E-22_WP,    5.681E-22_WP,    6.776E-22_WP,    7.916E-22_WP, &
    & 9.010E-22_WP,    9.936E-22_WP,    1.055E-21_WP,    1.072E-21_WP,    1.039E-21_WP,    9.579E-22_WP,    8.231E-22_WP, &
    & 6.431E-22_WP,    4.496E-22_WP,    2.761E-22_WP,    1.460E-22_WP,    6.490E-23_WP,    2.343E-23_WP,    5.823E-24_WP /
data J(1,    100    ,:)    / 1.440E-23_WP,    1.787E-23_WP,    2.224E-23_WP,    2.780E-23_WP,    3.490E-23_WP,    4.397E-23_WP, &
    & 5.556E-23_WP,    7.039E-23_WP,    8.936E-23_WP,    1.136E-22_WP,    1.447E-22_WP,    1.845E-22_WP,    2.353E-22_WP, &
    & 2.993E-22_WP,    3.779E-22_WP,    4.731E-22_WP,    5.863E-22_WP,    7.167E-22_WP,    8.605E-22_WP,    1.010E-21_WP, &
    & 1.153E-21_WP,    1.272E-21_WP,    1.349E-21_WP,    1.368E-21_WP,    1.319E-21_WP,    1.202E-21_WP,    1.016E-21_WP, &
    & 7.800E-22_WP,    5.351E-22_WP,    3.226E-22_WP,    1.676E-22_WP,    7.328E-23_WP,    2.561E-23_WP,    6.341E-24_WP /
data J(1,    101    ,:)    / 1.589E-23_WP,    1.983E-23_WP,    2.487E-23_WP,    3.132E-23_WP,    3.960E-23_WP,    5.025E-23_WP, &
    & 6.395E-23_WP,    8.158E-23_WP,    1.043E-22_WP,    1.336E-22_WP,    1.715E-22_WP,    2.201E-22_WP,    2.825E-22_WP, &
    & 3.611E-22_WP,    4.586E-22_WP,    5.773E-22_WP,    7.184E-22_WP,    8.801E-22_WP,    1.057E-21_WP,    1.239E-21_WP, &
    & 1.408E-21_WP,    1.545E-21_WP,    1.627E-21_WP,    1.635E-21_WP,    1.557E-21_WP,    1.397E-21_WP,    1.162E-21_WP, &
    & 8.775E-22_WP,    5.934E-22_WP,    3.534E-22_WP,    1.818E-22_WP,    7.845E-23_WP,    2.733E-23_WP,    6.734E-24_WP /
data J(1,    102    ,:)    / 1.759E-23_WP,    2.211E-23_WP,    2.791E-23_WP,    3.538E-23_WP,    4.503E-23_WP,    5.749E-23_WP, &
    & 7.361E-23_WP,    9.449E-23_WP,    1.216E-22_WP,    1.567E-22_WP,    2.023E-22_WP,    2.610E-22_WP,    3.362E-22_WP, &
    & 4.313E-22_WP,    5.495E-22_WP,    6.929E-22_WP,    8.619E-22_WP,    1.054E-21_WP,    1.260E-21_WP,    1.467E-21_WP, &
    & 1.655E-21_WP,    1.800E-21_WP,    1.877E-21_WP,    1.863E-21_WP,    1.750E-21_WP,    1.547E-21_WP,    1.270E-21_WP, &
    & 9.490E-22_WP,    6.362E-22_WP,    3.764E-22_WP,    1.924E-22_WP,    8.289E-23_WP,    2.882E-23_WP,    7.091E-24_WP /
data J(1,    103    ,:)    / 1.860E-23_WP,    2.355E-23_WP,    2.999E-23_WP,    3.836E-23_WP,    4.927E-23_WP,    6.351E-23_WP, &
    & 8.215E-23_WP,    1.066E-22_WP,    1.386E-22_WP,    1.808E-22_WP,    2.361E-22_WP,    3.086E-22_WP,    4.029E-22_WP, &
    & 5.247E-22_WP,    6.788E-22_WP,    8.704E-22_WP,    1.102E-21_WP,    1.373E-21_WP,    1.689E-21_WP,    2.030E-21_WP, &
    & 2.374E-21_WP,    2.690E-21_WP,    2.935E-21_WP,    3.067E-21_WP,    3.053E-21_WP,    2.891E-21_WP,    2.575E-21_WP, &
    & 2.106E-21_WP,    1.559E-21_WP,    1.027E-21_WP,    5.884E-22_WP,    2.859E-22_WP,    1.135E-22_WP,    3.440E-23_WP /
data J(1,    104    ,:)    / 2.079E-23_WP,    2.650E-23_WP,    3.395E-23_WP,    4.368E-23_WP,    5.642E-23_WP,    7.312E-23_WP, &
    & 9.507E-23_WP,    1.240E-22_WP,    1.621E-22_WP,    2.123E-22_WP,    2.782E-22_WP,    3.645E-22_WP,    4.768E-22_WP, &
    & 6.209E-22_WP,    8.019E-22_WP,    1.024E-21_WP,    1.288E-21_WP,    1.600E-21_WP,    1.946E-21_WP,    2.306E-21_WP, &
    & 2.651E-21_WP,    2.940E-21_WP,    3.127E-21_WP,    3.173E-21_WP,    3.061E-21_WP,    2.790E-21_WP,    2.354E-21_WP, &
    & 1.804E-21_WP,    1.235E-21_WP,    7.402E-22_WP,    3.794E-22_WP,    1.614E-22_WP,    5.442E-23_WP,    7.797E-24_WP /
data J(1,    105    ,:)    / 2.335E-23_WP,    2.995E-23_WP,    3.858E-23_WP,    4.990E-23_WP,    6.477E-23_WP,    8.437E-23_WP, &
    & 1.102E-22_WP,    1.444E-22_WP,    1.896E-22_WP,    2.491E-22_WP,    3.272E-22_WP,    4.293E-22_WP,    5.615E-22_WP, &
    & 7.296E-22_WP,    9.383E-22_WP,    1.190E-21_WP,    1.492E-21_WP,    1.835E-21_WP,    2.202E-21_WP,    2.570E-21_WP, &
    & 2.900E-21_WP,    3.149E-21_WP,    3.271E-21_WP,    3.239E-21_WP,    3.039E-21_WP,    2.667E-21_WP,    2.158E-21_WP, &
    & 1.580E-21_WP,    1.031E-21_WP,    5.883E-22_WP,    2.880E-22_WP,    1.179E-22_WP,    3.440E-23_WP,    8.298E-24_WP /
data J(1,    106    ,:)    / 2.631E-23_WP,    3.393E-23_WP,    4.392E-23_WP,    5.708E-23_WP,    7.444E-23_WP,    9.741E-23_WP, &
    & 1.278E-22_WP,    1.681E-22_WP,    2.213E-22_WP,    2.913E-22_WP,    3.830E-22_WP,    5.023E-22_WP,    6.555E-22_WP, &
    & 8.479E-22_WP,    1.083E-21_WP,    1.369E-21_WP,    1.699E-21_WP,    2.062E-21_WP,    2.440E-21_WP,    2.801E-21_WP, &
    & 3.103E-21_WP,    3.302E-21_WP,    3.365E-21_WP,    3.262E-21_WP,    2.979E-21_WP,    2.545E-21_WP,    2.010E-21_WP, &
    & 1.442E-21_WP,    9.284E-22_WP,    5.282E-22_WP,    2.609E-22_WP,    1.055E-22_WP,    3.654E-23_WP,    8.754E-24_WP /
data J(1,    107    ,:)    / 2.972E-23_WP,    3.850E-23_WP,    5.007E-23_WP,    6.535E-23_WP,    8.560E-23_WP,    1.124E-22_WP, &
    & 1.480E-22_WP,    1.952E-22_WP,    2.574E-22_WP,    3.391E-22_WP,    4.457E-22_WP,    5.830E-22_WP,    7.572E-22_WP, &
    & 9.727E-22_WP,    1.238E-21_WP,    1.549E-21_WP,    1.899E-21_WP,    2.274E-21_WP,    2.650E-21_WP,    2.991E-21_WP, &
    & 3.256E-21_WP,    3.410E-21_WP,    3.418E-21_WP,    3.249E-21_WP,    2.915E-21_WP,    2.459E-21_WP,    1.930E-21_WP, &
    & 1.387E-21_WP,    9.007E-22_WP,    5.205E-22_WP,    2.590E-22_WP,    1.114E-22_WP,    3.847E-23_WP,    9.203E-24_WP /
data J(1,    108    ,:)    / 3.363E-23_WP,    4.375E-23_WP,    5.713E-23_WP,    7.487E-23_WP,    9.843E-23_WP,    1.297E-22_WP, &
    & 1.712E-22_WP,    2.260E-22_WP,    2.982E-22_WP,    3.926E-22_WP,    5.147E-22_WP,    6.701E-22_WP,    8.643E-22_WP, &
    & 1.106E-21_WP,    1.393E-21_WP,    1.723E-21_WP,    2.085E-21_WP,    2.461E-21_WP,    2.824E-21_WP,    3.138E-21_WP, &
    & 3.370E-21_WP,    3.484E-21_WP,    3.440E-21_WP,    3.232E-21_WP,    2.881E-21_WP,    2.429E-21_WP,    1.918E-21_WP, &
    & 1.393E-21_WP,    9.164E-22_WP,    5.335E-22_WP,    2.722E-22_WP,    1.169E-22_WP,    4.036E-23_WP,    9.663E-24_WP /
data J(1,    109    ,:)    / 3.810E-23_WP,    4.977E-23_WP,    6.525E-23_WP,    8.581E-23_WP,    1.131E-22_WP,    1.494E-22_WP, &
    & 1.974E-22_WP,    2.608E-22_WP,    3.439E-22_WP,    4.515E-22_WP,    5.891E-22_WP,    7.619E-22_WP,    9.786E-22_WP, &
    & 1.240E-21_WP,    1.544E-21_WP,    1.885E-21_WP,    2.250E-21_WP,    2.618E-21_WP,    2.962E-21_WP,    3.253E-21_WP, &
    & 3.456E-21_WP,    3.531E-21_WP,    3.459E-21_WP,    3.240E-21_WP,    2.894E-21_WP,    2.457E-21_WP,    1.959E-21_WP, &
    & 1.436E-21_WP,    9.499E-22_WP,    5.595E-22_WP,    2.852E-22_WP,    1.225E-22_WP,    4.229E-23_WP,    1.014E-23_WP /
data J(1,    110    ,:)    / 4.324E-23_WP,    5.668E-23_WP,    7.456E-23_WP,    9.834E-23_WP,    1.299E-22_WP,    1.718E-22_WP, &
    & 2.271E-22_WP,    2.997E-22_WP,    3.940E-22_WP,    5.150E-22_WP,    6.675E-22_WP,    8.596E-22_WP,    1.093E-21_WP, &
    & 1.370E-21_WP,    1.685E-21_WP,    2.030E-21_WP,    2.391E-21_WP,    2.746E-21_WP,    3.074E-21_WP,    3.344E-21_WP, &
    & 3.520E-21_WP,    3.577E-21_WP,    3.500E-21_WP,    3.289E-21_WP,    2.957E-21_WP,    2.531E-21_WP,    2.032E-21_WP, &
    & 1.495E-21_WP,    9.951E-22_WP,    5.858E-22_WP,    2.986E-22_WP,    1.282E-22_WP,    4.429E-23_WP,    1.064E-23_WP /
data J(1,    111    ,:)    / 4.912E-23_WP,    6.461E-23_WP,    8.521E-23_WP,    1.126E-22_WP,    1.489E-22_WP,    1.970E-22_WP, &
    & 2.602E-22_WP,    3.424E-22_WP,    4.481E-22_WP,    5.818E-22_WP,    7.510E-22_WP,    9.579E-22_WP,    1.205E-21_WP, &
    & 1.492E-21_WP,    1.812E-21_WP,    2.155E-21_WP,    2.507E-21_WP,    2.852E-21_WP,    3.167E-21_WP,    3.419E-21_WP, &
    & 3.584E-21_WP,    3.642E-21_WP,    3.578E-21_WP,    3.382E-21_WP,    3.060E-21_WP,    2.633E-21_WP,    2.120E-21_WP, &
    & 1.565E-21_WP,    1.042E-21_WP,    6.131E-22_WP,    3.125E-22_WP,    1.342E-22_WP,    4.638E-23_WP,    1.116E-23_WP /
data J(1,    112    ,:)    / 5.585E-23_WP,    7.366E-23_WP,    9.734E-23_WP,    1.288E-22_WP,    1.704E-22_WP,    2.251E-22_WP, &
    & 2.965E-22_WP,    3.885E-22_WP,    5.051E-22_WP,    6.531E-22_WP,    8.350E-22_WP,    1.054E-21_WP,    1.310E-21_WP, &
    & 1.602E-21_WP,    1.922E-21_WP,    2.260E-21_WP,    2.606E-21_WP,    2.943E-21_WP,    3.246E-21_WP,    3.494E-21_WP, &
    & 3.666E-21_WP,    3.740E-21_WP,    3.694E-21_WP,    3.511E-21_WP,    3.190E-21_WP,    2.750E-21_WP,    2.219E-21_WP, &
    & 1.639E-21_WP,    1.090E-21_WP,    6.417E-22_WP,    3.271E-22_WP,    1.405E-22_WP,    4.857E-23_WP,    1.171E-23_WP /
data J(1,    113    ,:)    / 6.355E-23_WP,    8.396E-23_WP,    1.111E-22_WP,    1.470E-22_WP,    1.942E-22_WP,    2.560E-22_WP, &
    & 3.356E-22_WP,    4.369E-22_WP,    5.657E-22_WP,    7.247E-22_WP,    9.168E-22_WP,    1.144E-21_WP,    1.405E-21_WP, &
    & 1.698E-21_WP,    2.015E-21_WP,    2.350E-21_WP,    2.691E-21_WP,    3.022E-21_WP,    3.326E-21_WP,    3.586E-21_WP, &
    & 3.777E-21_WP,    3.872E-21_WP,    3.842E-21_WP,    3.664E-21_WP,    3.334E-21_WP,    2.879E-21_WP,    2.323E-21_WP, &
    & 1.715E-21_WP,    1.141E-21_WP,    6.717E-22_WP,    3.424E-22_WP,    1.471E-22_WP,    5.089E-23_WP,    1.229E-23_WP /
data J(1,    114    ,:)    / 7.228E-23_WP,    9.559E-23_WP,    1.265E-22_WP,    1.672E-22_WP,    2.204E-22_WP,    2.891E-22_WP, &
    & 3.767E-22_WP,    4.884E-22_WP,    6.266E-22_WP,    7.944E-22_WP,    9.939E-22_WP,    1.226E-21_WP,    1.488E-21_WP, &
    & 1.779E-21_WP,    2.094E-21_WP,    2.427E-21_WP,    2.766E-21_WP,    3.101E-21_WP,    3.417E-21_WP,    3.697E-21_WP, &
    & 3.910E-21_WP,    4.024E-21_WP,    4.002E-21_WP,    3.820E-21_WP,    3.478E-21_WP,    3.002E-21_WP,    2.420E-21_WP, &
    & 1.786E-21_WP,    1.187E-21_WP,    6.982E-22_WP,    3.556E-22_WP,    1.526E-22_WP,    5.274E-23_WP,    1.272E-23_WP /
data J(1,    115    ,:)    / 8.209E-23_WP,    1.086E-22_WP,    1.436E-22_WP,    1.893E-22_WP,    2.484E-22_WP,    3.238E-22_WP, &
    & 4.202E-22_WP,    5.398E-22_WP,    6.856E-22_WP,    8.597E-22_WP,    1.063E-21_WP,    1.296E-21_WP,    1.558E-21_WP, &
    & 1.848E-21_WP,    2.162E-21_WP,    2.493E-21_WP,    2.836E-21_WP,    3.183E-21_WP,    3.518E-21_WP,    3.819E-21_WP, &
    & 4.050E-21_WP,    4.174E-21_WP,    4.151E-21_WP,    3.960E-21_WP,    3.601E-21_WP,    3.103E-21_WP,    2.498E-21_WP, &
    & 1.840E-21_WP,    1.221E-21_WP,    7.167E-22_WP,    3.644E-22_WP,    1.561E-22_WP,    5.386E-23_WP,    1.297E-23_WP /
data J(1,    116    ,:)    / 9.307E-23_WP,    1.230E-22_WP,    1.622E-22_WP,    2.129E-22_WP,    2.777E-22_WP,    3.605E-22_WP, &
    & 4.636E-22_WP,    5.896E-22_WP,    7.407E-22_WP,    9.183E-22_WP,    1.123E-21_WP,    1.355E-21_WP,    1.616E-21_WP, &
    & 1.906E-21_WP,    2.219E-21_WP,    2.555E-21_WP,    2.909E-21_WP,    3.272E-21_WP,    3.627E-21_WP,    3.944E-21_WP, &
    & 4.186E-21_WP,    4.310E-21_WP,    4.282E-21_WP,    4.078E-21_WP,    3.701E-21_WP,    3.183E-21_WP,    2.557E-21_WP, &
    & 1.880E-21_WP,    1.245E-21_WP,    7.299E-22_WP,    3.706E-22_WP,    1.586E-22_WP,    5.467E-23_WP,    1.316E-23_WP /
data J(1,    117    ,:)    / 1.052E-22_WP,    1.387E-22_WP,    1.821E-22_WP,    2.376E-22_WP,    3.086E-22_WP,    3.972E-22_WP, &
    & 5.056E-22_WP,    6.362E-22_WP,    7.903E-22_WP,    9.689E-22_WP,    1.173E-21_WP,    1.405E-21_WP,    1.666E-21_WP, &
    & 1.955E-21_WP,    2.272E-21_WP,    2.618E-21_WP,    2.987E-21_WP,    3.368E-21_WP,    3.738E-21_WP,    4.065E-21_WP, &
    & 4.310E-21_WP,    4.431E-21_WP,    4.394E-21_WP,    4.176E-21_WP,    3.783E-21_WP,    3.247E-21_WP,    2.604E-21_WP, &
    & 1.912E-21_WP,    1.265E-21_WP,    7.408E-22_WP,    3.760E-22_WP,    1.608E-22_WP,    5.543E-23_WP,    1.336E-23_WP /
data J(1,    118    ,:)    / 1.184E-22_WP,    1.555E-22_WP,    2.029E-22_WP,    2.637E-22_WP,    3.395E-22_WP,    4.326E-22_WP, &
    & 5.450E-22_WP,    6.781E-22_WP,    8.331E-22_WP,    1.011E-21_WP,    1.215E-21_WP,    1.447E-21_WP,    1.708E-21_WP, &
    & 2.001E-21_WP,    2.327E-21_WP,    2.686E-21_WP,    3.070E-21_WP,    3.465E-21_WP,    3.846E-21_WP,    4.177E-21_WP, &
    & 4.421E-21_WP,    4.537E-21_WP,    4.490E-21_WP,    4.260E-21_WP,    3.853E-21_WP,    3.302E-21_WP,    2.645E-21_WP, &
    & 1.940E-21_WP,    1.283E-21_WP,    7.513E-22_WP,    3.812E-22_WP,    1.631E-22_WP,    5.622E-23_WP,    1.357E-23_WP /
data J(1,    119    ,:)    / 1.326E-22_WP,    1.730E-22_WP,    2.249E-22_WP,    2.897E-22_WP,    3.694E-22_WP,    4.658E-22_WP, &
    & 5.804E-22_WP,    7.143E-22_WP,    8.691E-22_WP,    1.047E-21_WP,    1.252E-21_WP,    1.483E-21_WP,    1.748E-21_WP, &
    & 2.048E-21_WP,    2.386E-21_WP,    2.759E-21_WP,    3.156E-21_WP,    3.561E-21_WP,    3.946E-21_WP,    4.279E-21_WP, &
    & 4.520E-21_WP,    4.630E-21_WP,    4.575E-21_WP,    4.334E-21_WP,    3.915E-21_WP,    3.353E-21_WP,    2.684E-21_WP, &
    & 1.968E-21_WP,    1.301E-21_WP,    7.618E-22_WP,    3.865E-22_WP,    1.654E-22_WP,    5.703E-23_WP,    1.378E-23_WP /
data J(1,    120    ,:)    / 1.474E-22_WP,    1.916E-22_WP,    2.469E-22_WP,    3.150E-22_WP,    3.975E-22_WP,    4.957E-22_WP, &
    & 6.110E-22_WP,    7.448E-22_WP,    8.997E-22_WP,    1.078E-21_WP,    1.283E-21_WP,    1.517E-21_WP,    1.789E-21_WP, &
    & 2.099E-21_WP,    2.449E-21_WP,    2.833E-21_WP,    3.240E-21_WP,    3.649E-21_WP,    4.038E-21_WP,    4.370E-21_WP, &
    & 4.609E-21_WP,    4.714E-21_WP,    4.651E-21_WP,    4.402E-21_WP,    3.973E-21_WP,    3.401E-21_WP,    2.722E-21_WP, &
    & 1.996E-21_WP,    1.320E-21_WP,    7.726E-22_WP,    3.921E-22_WP,    1.678E-22_WP,    5.787E-23_WP,    1.400E-23_WP /
data J(1,    121    ,:)    / 1.630E-22_WP,    2.101E-22_WP,    2.682E-22_WP,    3.386E-22_WP,    4.227E-22_WP,    5.216E-22_WP, &
    & 6.368E-22_WP,    7.707E-22_WP,    9.259E-22_WP,    1.105E-21_WP,    1.312E-21_WP,    1.552E-21_WP,    1.832E-21_WP, &
    & 2.154E-21_WP,    2.514E-21_WP,    2.906E-21_WP,    3.318E-21_WP,    3.732E-21_WP,    4.121E-21_WP,    4.453E-21_WP, &
    & 4.689E-21_WP,    4.791E-21_WP,    4.723E-21_WP,    4.467E-21_WP,    4.031E-21_WP,    3.450E-21_WP,    2.761E-21_WP, &
    & 2.024E-21_WP,    1.339E-21_WP,    7.838E-22_WP,    3.977E-22_WP,    1.702E-22_WP,    5.873E-23_WP,    1.422E-23_WP /
data J(1,    122    ,:)    / 1.787E-22_WP,    2.281E-22_WP,    2.881E-22_WP,    3.599E-22_WP,    4.446E-22_WP,    5.435E-22_WP, &
    & 6.588E-22_WP,    7.930E-22_WP,    9.484E-22_WP,    1.129E-21_WP,    1.342E-21_WP,    1.590E-21_WP,    1.879E-21_WP, &
    & 2.209E-21_WP,    2.577E-21_WP,    2.975E-21_WP,    3.391E-21_WP,    3.807E-21_WP,    4.197E-21_WP,    4.529E-21_WP, &
    & 4.765E-21_WP,    4.864E-21_WP,    4.793E-21_WP,    4.533E-21_WP,    4.090E-21_WP,    3.500E-21_WP,    2.801E-21_WP, &
    & 2.054E-21_WP,    1.358E-21_WP,    7.953E-22_WP,    4.036E-22_WP,    1.727E-22_WP,    5.962E-23_WP,    1.445E-23_WP /
data J(1,    123    ,:)    / 1.938E-22_WP,    2.449E-22_WP,    3.061E-22_WP,    3.784E-22_WP,    4.630E-22_WP,    5.620E-22_WP, &
    & 6.776E-22_WP,    8.121E-22_WP,    9.696E-22_WP,    1.155E-21_WP,    1.374E-21_WP,    1.630E-21_WP,    1.927E-21_WP, &
    & 2.264E-21_WP,    2.637E-21_WP,    3.039E-21_WP,    3.458E-21_WP,    3.876E-21_WP,    4.268E-21_WP,    4.601E-21_WP, &
    & 4.837E-21_WP,    4.936E-21_WP,    4.863E-21_WP,    4.599E-21_WP,    4.149E-21_WP,    3.551E-21_WP,    2.843E-21_WP, &
    & 2.084E-21_WP,    1.378E-21_WP,    8.071E-22_WP,    4.096E-22_WP,    1.753E-22_WP,    6.053E-23_WP,    1.468E-23_WP /
data J(1,    124    ,:)    / 2.080E-22_WP,    2.601E-22_WP,    3.217E-22_WP,    3.940E-22_WP,    4.788E-22_WP,    5.780E-22_WP, &
    & 6.939E-22_WP,    8.302E-22_WP,    9.914E-22_WP,    1.182E-21_WP,    1.408E-21_WP,    1.671E-21_WP,    1.974E-21_WP, &
    & 2.316E-21_WP,    2.693E-21_WP,    3.098E-21_WP,    3.520E-21_WP,    3.941E-21_WP,    4.336E-21_WP,    4.672E-21_WP, &
    & 4.910E-21_WP,    5.010E-21_WP,    4.935E-21_WP,    4.667E-21_WP,    4.211E-21_WP,    3.604E-21_WP,    2.885E-21_WP, &
    & 2.116E-21_WP,    1.399E-21_WP,    8.192E-22_WP,    4.158E-22_WP,    1.780E-22_WP,    6.146E-23_WP,    1.491E-23_WP /
data J(1,    125    ,:)    / 2.208E-22_WP,    2.732E-22_WP,    3.349E-22_WP,    4.073E-22_WP,    4.923E-22_WP,    5.918E-22_WP, &
    & 7.093E-22_WP,    8.487E-22_WP,    1.015E-21_WP,    1.212E-21_WP,    1.443E-21_WP,    1.711E-21_WP,    2.018E-21_WP, &
    & 2.364E-21_WP,    2.744E-21_WP,    3.153E-21_WP,    3.578E-21_WP,    4.003E-21_WP,    4.402E-21_WP,    4.742E-21_WP, &
    & 4.983E-21_WP,    5.084E-21_WP,    5.009E-21_WP,    4.737E-21_WP,    4.274E-21_WP,    3.659E-21_WP,    2.929E-21_WP, &
    & 2.148E-21_WP,    1.420E-21_WP,    8.317E-22_WP,    4.222E-22_WP,    1.807E-22_WP,    6.242E-23_WP,    1.516E-23_WP /
data J(1,    126    ,:)    / 2.319E-22_WP,    2.844E-22_WP,    3.461E-22_WP,    4.188E-22_WP,    5.040E-22_WP,    6.049E-22_WP, &
    & 7.251E-22_WP,    8.687E-22_WP,    1.040E-21_WP,    1.242E-21_WP,    1.478E-21_WP,    1.750E-21_WP,    2.060E-21_WP, &
    & 2.409E-21_WP,    2.793E-21_WP,    3.205E-21_WP,    3.635E-21_WP,    4.065E-21_WP,    4.468E-21_WP,    4.813E-21_WP, &
    & 5.058E-21_WP,    5.160E-21_WP,    5.084E-21_WP,    4.808E-21_WP,    4.339E-21_WP,    3.714E-21_WP,    2.973E-21_WP, &
    & 2.180E-21_WP,    1.442E-21_WP,    8.445E-22_WP,    4.287E-22_WP,    1.835E-22_WP,    6.340E-23_WP,    1.540E-23_WP /
data J(1,    127    ,:)    / 2.413E-22_WP,    2.939E-22_WP,    3.558E-22_WP,    4.287E-22_WP,    5.151E-22_WP,    6.183E-22_WP, &
    & 7.420E-22_WP,    8.898E-22_WP,    1.065E-21_WP,    1.271E-21_WP,    1.510E-21_WP,    1.786E-21_WP,    2.099E-21_WP, &
    & 2.452E-21_WP,    2.839E-21_WP,    3.256E-21_WP,    3.691E-21_WP,    4.126E-21_WP,    4.536E-21_WP,    4.885E-21_WP, &
    & 5.134E-21_WP,    5.239E-21_WP,    5.162E-21_WP,    4.882E-21_WP,    4.406E-21_WP,    3.771E-21_WP,    3.019E-21_WP, &
    & 2.214E-21_WP,    1.464E-21_WP,    8.577E-22_WP,    4.354E-22_WP,    1.864E-22_WP,    6.440E-23_WP,    1.566E-23_WP /
data J(1,    128    ,:)    / 2.493E-22_WP,    3.021E-22_WP,    3.642E-22_WP,    4.381E-22_WP,    5.265E-22_WP,    6.327E-22_WP, &
    & 7.600E-22_WP,    9.115E-22_WP,    1.090E-21_WP,    1.299E-21_WP,    1.541E-21_WP,    1.819E-21_WP,    2.136E-21_WP, &
    & 2.492E-21_WP,    2.884E-21_WP,    3.306E-21_WP,    3.747E-21_WP,    4.188E-21_WP,    4.604E-21_WP,    4.959E-21_WP, &
    & 5.212E-21_WP,    5.319E-21_WP,    5.241E-21_WP,    4.957E-21_WP,    4.474E-21_WP,    3.830E-21_WP,    3.066E-21_WP, &
    & 2.249E-21_WP,    1.487E-21_WP,    8.711E-22_WP,    4.422E-22_WP,    1.894E-22_WP,    6.543E-23_WP,    1.591E-23_WP /
data J(1,    129    ,:)    / 2.563E-22_WP,    3.092E-22_WP,    3.722E-22_WP,    4.478E-22_WP,    5.387E-22_WP,    6.480E-22_WP, &
    & 7.784E-22_WP,    9.329E-22_WP,    1.114E-21_WP,    1.326E-21_WP,    1.570E-21_WP,    1.851E-21_WP,    2.171E-21_WP, &
    & 2.531E-21_WP,    2.928E-21_WP,    3.356E-21_WP,    3.804E-21_WP,    4.252E-21_WP,    4.675E-21_WP,    5.036E-21_WP, &
    & 5.293E-21_WP,    5.402E-21_WP,    5.323E-21_WP,    5.035E-21_WP,    4.544E-21_WP,    3.890E-21_WP,    3.115E-21_WP, &
    & 2.284E-21_WP,    1.511E-21_WP,    8.849E-22_WP,    4.492E-22_WP,    1.924E-22_WP,    6.648E-23_WP,    1.618E-23_WP /
data J(1,    130    ,:)    / 2.623E-22_WP,    3.159E-22_WP,    3.804E-22_WP,    4.581E-22_WP,    5.517E-22_WP,    6.636E-22_WP, &
    & 7.966E-22_WP,    9.533E-22_WP,    1.137E-21_WP,    1.350E-21_WP,    1.598E-21_WP,    1.882E-21_WP,    2.206E-21_WP, &
    & 2.570E-21_WP,    2.973E-21_WP,    3.407E-21_WP,    3.862E-21_WP,    4.317E-21_WP,    4.747E-21_WP,    5.114E-21_WP, &
    & 5.376E-21_WP,    5.486E-21_WP,    5.407E-21_WP,    5.114E-21_WP,    4.616E-21_WP,    3.952E-21_WP,    3.164E-21_WP, &
    & 2.321E-21_WP,    1.535E-21_WP,    8.990E-22_WP,    4.564E-22_WP,    1.955E-22_WP,    6.755E-23_WP,    1.645E-23_WP /
data J(1,    131    ,:)    / 2.680E-22_WP,    3.229E-22_WP,    3.892E-22_WP,    4.691E-22_WP,    5.649E-22_WP,    6.791E-22_WP, &
    & 8.139E-22_WP,    9.725E-22_WP,    1.158E-21_WP,    1.374E-21_WP,    1.624E-21_WP,    1.911E-21_WP,    2.240E-21_WP, &
    & 2.610E-21_WP,    3.018E-21_WP,    3.460E-21_WP,    3.921E-21_WP,    4.384E-21_WP,    4.821E-21_WP,    5.194E-21_WP, &
    & 5.460E-21_WP,    5.573E-21_WP,    5.492E-21_WP,    5.195E-21_WP,    4.689E-21_WP,    4.015E-21_WP,    3.215E-21_WP, &
    & 2.358E-21_WP,    1.560E-21_WP,    9.135E-22_WP,    4.638E-22_WP,    1.986E-22_WP,    6.865E-23_WP,    1.672E-23_WP /
data J(1,    132    ,:)    / 2.739E-22_WP,    3.303E-22_WP,    3.985E-22_WP,    4.803E-22_WP,    5.780E-22_WP,    6.938E-22_WP, &
    & 8.303E-22_WP,    9.907E-22_WP,    1.178E-21_WP,    1.396E-21_WP,    1.649E-21_WP,    1.941E-21_WP,    2.274E-21_WP, &
    & 2.650E-21_WP,    3.065E-21_WP,    3.513E-21_WP,    3.982E-21_WP,    4.453E-21_WP,    4.897E-21_WP,    5.276E-21_WP, &
    & 5.547E-21_WP,    5.662E-21_WP,    5.580E-21_WP,    5.279E-21_WP,    4.764E-21_WP,    4.079E-21_WP,    3.266E-21_WP, &
    & 2.396E-21_WP,    1.585E-21_WP,    9.283E-22_WP,    4.713E-22_WP,    2.018E-22_WP,    6.978E-23_WP,    1.700E-23_WP /
data J(1,    133    ,:)    / 2.802E-22_WP,    3.382E-22_WP,    4.080E-22_WP,    4.914E-22_WP,    5.905E-22_WP,    7.077E-22_WP, &
    & 8.458E-22_WP,    1.008E-21_WP,    1.198E-21_WP,    1.419E-21_WP,    1.675E-21_WP,    1.971E-21_WP,    2.309E-21_WP, &
    & 2.691E-21_WP,    3.112E-21_WP,    3.568E-21_WP,    4.045E-21_WP,    4.523E-21_WP,    4.974E-21_WP,    5.360E-21_WP, &
    & 5.635E-21_WP,    5.752E-21_WP,    5.670E-21_WP,    5.364E-21_WP,    4.841E-21_WP,    4.145E-21_WP,    3.319E-21_WP, &
    & 2.435E-21_WP,    1.610E-21_WP,    9.434E-22_WP,    4.790E-22_WP,    2.051E-22_WP,    7.092E-23_WP,    1.729E-23_WP /
data J(1,    134    ,:)    / 2.868E-22_WP,    3.463E-22_WP,    4.174E-22_WP,    5.020E-22_WP,    6.024E-22_WP,    7.209E-22_WP, &
    & 8.605E-22_WP,    1.024E-21_WP,    1.217E-21_WP,    1.441E-21_WP,    1.701E-21_WP,    2.001E-21_WP,    2.345E-21_WP, &
    & 2.732E-21_WP,    3.161E-21_WP,    3.624E-21_WP,    4.109E-21_WP,    4.595E-21_WP,    5.054E-21_WP,    5.446E-21_WP, &
    & 5.726E-21_WP,    5.845E-21_WP,    5.761E-21_WP,    5.451E-21_WP,    4.920E-21_WP,    4.213E-21_WP,    3.373E-21_WP, &
    & 2.474E-21_WP,    1.637E-21_WP,    9.588E-22_WP,    4.868E-22_WP,    2.085E-22_WP,    7.210E-23_WP,    1.758E-23_WP /
data J(1,    135    ,:)    / 2.936E-22_WP,    3.542E-22_WP,    4.264E-22_WP,    5.121E-22_WP,    6.136E-22_WP,    7.334E-22_WP, &
    & 8.746E-22_WP,    1.041E-21_WP,    1.235E-21_WP,    1.463E-21_WP,    1.727E-21_WP,    2.032E-21_WP,    2.381E-21_WP, &
    & 2.775E-21_WP,    3.211E-21_WP,    3.682E-21_WP,    4.174E-21_WP,    4.669E-21_WP,    5.135E-21_WP,    5.534E-21_WP, &
    & 5.819E-21_WP,    5.940E-21_WP,    5.855E-21_WP,    5.539E-21_WP,    5.000E-21_WP,    4.282E-21_WP,    3.429E-21_WP, &
    & 2.515E-21_WP,    1.664E-21_WP,    9.746E-22_WP,    4.948E-22_WP,    2.120E-22_WP,    7.329E-23_WP,    1.788E-23_WP /
data J(1,    136    ,:)    / 3.004E-22_WP,    3.618E-22_WP,    4.349E-22_WP,    5.216E-22_WP,    6.242E-22_WP,    7.455E-22_WP, &
    & 8.885E-22_WP,    1.057E-21_WP,    1.254E-21_WP,    1.485E-21_WP,    1.754E-21_WP,    2.064E-21_WP,    2.419E-21_WP, &
    & 2.819E-21_WP,    3.262E-21_WP,    3.740E-21_WP,    4.242E-21_WP,    4.744E-21_WP,    5.218E-21_WP,    5.624E-21_WP, &
    & 5.914E-21_WP,    6.037E-21_WP,    5.951E-21_WP,    5.630E-21_WP,    5.082E-21_WP,    4.352E-21_WP,    3.485E-21_WP, &
    & 2.557E-21_WP,    1.691E-21_WP,    9.907E-22_WP,    5.030E-22_WP,    2.155E-22_WP,    7.451E-23_WP,    1.818E-23_WP /
data J(1,    137    ,:)    / 3.068E-22_WP,    3.691E-22_WP,    4.430E-22_WP,    5.306E-22_WP,    6.345E-22_WP,    7.573E-22_WP, &
    & 9.023E-22_WP,    1.073E-21_WP,    1.274E-21_WP,    1.508E-21_WP,    1.781E-21_WP,    2.096E-21_WP,    2.457E-21_WP, &
    & 2.864E-21_WP,    3.314E-21_WP,    3.801E-21_WP,    4.310E-21_WP,    4.821E-21_WP,    5.303E-21_WP,    5.716E-21_WP, &
    & 6.011E-21_WP,    6.137E-21_WP,    6.049E-21_WP,    5.723E-21_WP,    5.166E-21_WP,    4.424E-21_WP,    3.543E-21_WP, &
    & 2.599E-21_WP,    1.719E-21_WP,    1.007E-21_WP,    5.114E-22_WP,    2.191E-22_WP,    7.576E-23_WP,    1.849E-23_WP /
data J(1,    138    ,:)    / 3.129E-22_WP,    3.759E-22_WP,    4.506E-22_WP,    5.394E-22_WP,    6.445E-22_WP,    7.691E-22_WP, &
    & 9.163E-22_WP,    1.090E-21_WP,    1.294E-21_WP,    1.532E-21_WP,    1.809E-21_WP,    2.129E-21_WP,    2.496E-21_WP, &
    & 2.910E-21_WP,    3.368E-21_WP,    3.862E-21_WP,    4.380E-21_WP,    4.900E-21_WP,    5.390E-21_WP,    5.810E-21_WP, &
    & 6.110E-21_WP,    6.238E-21_WP,    6.149E-21_WP,    5.818E-21_WP,    5.252E-21_WP,    4.497E-21_WP,    3.602E-21_WP, &
    & 2.642E-21_WP,    1.748E-21_WP,    1.024E-21_WP,    5.199E-22_WP,    2.227E-22_WP,    7.703E-23_WP,    1.880E-23_WP /
data J(1,    139    ,:)    / 3.187E-22_WP,    3.824E-22_WP,    4.581E-22_WP,    5.479E-22_WP,    6.546E-22_WP,    7.810E-22_WP, &
    & 9.305E-22_WP,    1.107E-21_WP,    1.314E-21_WP,    1.556E-21_WP,    1.838E-21_WP,    2.163E-21_WP,    2.536E-21_WP, &
    & 2.957E-21_WP,    3.422E-21_WP,    3.925E-21_WP,    4.452E-21_WP,    4.980E-21_WP,    5.479E-21_WP,    5.906E-21_WP, &
    & 6.211E-21_WP,    6.341E-21_WP,    6.251E-21_WP,    5.915E-21_WP,    5.339E-21_WP,    4.572E-21_WP,    3.662E-21_WP, &
    & 2.686E-21_WP,    1.777E-21_WP,    1.041E-21_WP,    5.286E-22_WP,    2.265E-22_WP,    7.833E-23_WP,    1.912E-23_WP /
data J(1,    140    ,:)    / 3.243E-22_WP,    3.887E-22_WP,    4.654E-22_WP,    5.565E-22_WP,    6.647E-22_WP,    7.931E-22_WP, &
    & 9.450E-22_WP,    1.124E-21_WP,    1.335E-21_WP,    1.581E-21_WP,    1.867E-21_WP,    2.198E-21_WP,    2.577E-21_WP, &
    & 3.005E-21_WP,    3.478E-21_WP,    3.990E-21_WP,    4.525E-21_WP,    5.063E-21_WP,    5.570E-21_WP,    6.004E-21_WP, &
    & 6.314E-21_WP,    6.447E-21_WP,    6.355E-21_WP,    6.013E-21_WP,    5.429E-21_WP,    4.649E-21_WP,    3.723E-21_WP, &
    & 2.731E-21_WP,    1.807E-21_WP,    1.058E-21_WP,    5.375E-22_WP,    2.303E-22_WP,    7.965E-23_WP,    1.945E-23_WP /
data J(1,    141    ,:)    / 3.296E-22_WP,    3.949E-22_WP,    4.726E-22_WP,    5.651E-22_WP,    6.751E-22_WP,    8.055E-22_WP, &
    & 9.598E-22_WP,    1.142E-21_WP,    1.356E-21_WP,    1.606E-21_WP,    1.897E-21_WP,    2.234E-21_WP,    2.619E-21_WP, &
    & 3.054E-21_WP,    3.535E-21_WP,    4.055E-21_WP,    4.600E-21_WP,    5.147E-21_WP,    5.662E-21_WP,    6.104E-21_WP, &
    & 6.419E-21_WP,    6.554E-21_WP,    6.462E-21_WP,    6.114E-21_WP,    5.520E-21_WP,    4.727E-21_WP,    3.786E-21_WP, &
    & 2.777E-21_WP,    1.837E-21_WP,    1.076E-21_WP,    5.466E-22_WP,    2.342E-22_WP,    8.100E-23_WP,    1.978E-23_WP /
data J(1,    142    ,:)    / 3.349E-22_WP,    4.011E-22_WP,    4.800E-22_WP,    5.739E-22_WP,    6.856E-22_WP,    8.181E-22_WP, &
    & 9.750E-22_WP,    1.160E-21_WP,    1.378E-21_WP,    1.632E-21_WP,    1.928E-21_WP,    2.270E-21_WP,    2.662E-21_WP, &
    & 3.104E-21_WP,    3.594E-21_WP,    4.122E-21_WP,    4.676E-21_WP,    5.232E-21_WP,    5.757E-21_WP,    6.206E-21_WP, &
    & 6.527E-21_WP,    6.664E-21_WP,    6.570E-21_WP,    6.217E-21_WP,    5.612E-21_WP,    4.806E-21_WP,    3.849E-21_WP, &
    & 2.824E-21_WP,    1.868E-21_WP,    1.094E-21_WP,    5.558E-22_WP,    2.381E-22_WP,    8.237E-23_WP,    2.012E-23_WP /
data J(1,    143    ,:)    / 3.401E-22_WP,    4.073E-22_WP,    4.874E-22_WP,    5.829E-22_WP,    6.964E-22_WP,    8.311E-22_WP, &
    & 9.906E-22_WP,    1.179E-21_WP,    1.400E-21_WP,    1.659E-21_WP,    1.960E-21_WP,    2.307E-21_WP,    2.706E-21_WP, &
    & 3.156E-21_WP,    3.653E-21_WP,    4.191E-21_WP,    4.754E-21_WP,    5.320E-21_WP,    5.853E-21_WP,    6.310E-21_WP, &
    & 6.636E-21_WP,    6.776E-21_WP,    6.681E-21_WP,    6.322E-21_WP,    5.707E-21_WP,    4.888E-21_WP,    3.914E-21_WP, &
    & 2.872E-21_WP,    1.900E-21_WP,    1.113E-21_WP,    5.652E-22_WP,    2.422E-22_WP,    8.377E-23_WP,    2.046E-23_WP /
data J(1,    144    ,:)    / 3.454E-22_WP,    4.136E-22_WP,    4.950E-22_WP,    5.921E-22_WP,    7.074E-22_WP,    8.444E-22_WP, &
    & 1.006E-21_WP,    1.198E-21_WP,    1.423E-21_WP,    1.686E-21_WP,    1.992E-21_WP,    2.345E-21_WP,    2.751E-21_WP, &
    & 3.208E-21_WP,    3.714E-21_WP,    4.261E-21_WP,    4.834E-21_WP,    5.409E-21_WP,    5.951E-21_WP,    6.416E-21_WP, &
    & 6.748E-21_WP,    6.891E-21_WP,    6.794E-21_WP,    6.428E-21_WP,    5.804E-21_WP,    4.970E-21_WP,    3.981E-21_WP, &
    & 2.921E-21_WP,    1.932E-21_WP,    1.132E-21_WP,    5.748E-22_WP,    2.463E-22_WP,    8.520E-23_WP,    2.082E-23_WP /
data J(1,    145    ,:)    / 3.508E-22_WP,    4.201E-22_WP,    5.029E-22_WP,    6.015E-22_WP,    7.187E-22_WP,    8.580E-22_WP, &
    & 1.023E-21_WP,    1.217E-21_WP,    1.446E-21_WP,    1.713E-21_WP,    2.025E-21_WP,    2.384E-21_WP,    2.796E-21_WP, &
    & 3.261E-21_WP,    3.776E-21_WP,    4.332E-21_WP,    4.915E-21_WP,    5.500E-21_WP,    6.052E-21_WP,    6.524E-21_WP, &
    & 6.862E-21_WP,    7.007E-21_WP,    6.909E-21_WP,    6.537E-21_WP,    5.902E-21_WP,    5.055E-21_WP,    4.048E-21_WP, &
    & 2.970E-21_WP,    1.965E-21_WP,    1.151E-21_WP,    5.846E-22_WP,    2.505E-22_WP,    8.666E-23_WP,    2.117E-23_WP /
data J(1,    146    ,:)    / 3.563E-22_WP,    4.267E-22_WP,    5.108E-22_WP,    6.111E-22_WP,    7.303E-22_WP,    8.718E-22_WP, &
    & 1.039E-21_WP,    1.237E-21_WP,    1.470E-21_WP,    1.742E-21_WP,    2.058E-21_WP,    2.424E-21_WP,    2.843E-21_WP, &
    & 3.316E-21_WP,    3.839E-21_WP,    4.405E-21_WP,    4.998E-21_WP,    5.593E-21_WP,    6.154E-21_WP,    6.635E-21_WP, &
    & 6.979E-21_WP,    7.126E-21_WP,    7.026E-21_WP,    6.649E-21_WP,    6.002E-21_WP,    5.141E-21_WP,    4.117E-21_WP, &
    & 3.021E-21_WP,    1.998E-21_WP,    1.171E-21_WP,    5.946E-22_WP,    2.547E-22_WP,    8.814E-23_WP,    2.154E-23_WP /
data J(1,    147    ,:)    / 3.619E-22_WP,    4.335E-22_WP,    5.190E-22_WP,    6.209E-22_WP,    7.421E-22_WP,    8.860E-22_WP, &
    & 1.056E-21_WP,    1.257E-21_WP,    1.494E-21_WP,    1.771E-21_WP,    2.092E-21_WP,    2.464E-21_WP,    2.890E-21_WP, &
    & 3.372E-21_WP,    3.904E-21_WP,    4.479E-21_WP,    5.082E-21_WP,    5.687E-21_WP,    6.258E-21_WP,    6.747E-21_WP, &
    & 7.097E-21_WP,    7.247E-21_WP,    7.146E-21_WP,    6.762E-21_WP,    6.105E-21_WP,    5.228E-21_WP,    4.188E-21_WP, &
    & 3.072E-21_WP,    2.033E-21_WP,    1.191E-21_WP,    6.047E-22_WP,    2.591E-22_WP,    8.965E-23_WP,    2.191E-23_WP /
data J(1,    148    ,:)    / 3.676E-22_WP,    4.405E-22_WP,    5.274E-22_WP,    6.310E-22_WP,    7.542E-22_WP,    9.006E-22_WP, &
    & 1.074E-21_WP,    1.278E-21_WP,    1.519E-21_WP,    1.800E-21_WP,    2.127E-21_WP,    2.506E-21_WP,    2.939E-21_WP, &
    & 3.428E-21_WP,    3.970E-21_WP,    4.555E-21_WP,    5.168E-21_WP,    5.784E-21_WP,    6.364E-21_WP,    6.862E-21_WP, &
    & 7.218E-21_WP,    7.371E-21_WP,    7.267E-21_WP,    6.877E-21_WP,    6.209E-21_WP,    5.318E-21_WP,    4.259E-21_WP, &
    & 3.125E-21_WP,    2.067E-21_WP,    1.211E-21_WP,    6.151E-22_WP,    2.635E-22_WP,    9.119E-23_WP,    2.229E-23_WP /
data J(1,    149    ,:)    / 3.735E-22_WP,    4.476E-22_WP,    5.359E-22_WP,    6.413E-22_WP,    7.666E-22_WP,    9.154E-22_WP, &
    & 1.092E-21_WP,    1.299E-21_WP,    1.544E-21_WP,    1.830E-21_WP,    2.163E-21_WP,    2.548E-21_WP,    2.988E-21_WP, &
    & 3.486E-21_WP,    4.037E-21_WP,    4.632E-21_WP,    5.256E-21_WP,    5.882E-21_WP,    6.473E-21_WP,    6.979E-21_WP, &
    & 7.341E-21_WP,    7.497E-21_WP,    7.392E-21_WP,    6.995E-21_WP,    6.315E-21_WP,    5.409E-21_WP,    4.332E-21_WP, &
    & 3.178E-21_WP,    2.103E-21_WP,    1.232E-21_WP,    6.256E-22_WP,    2.681E-22_WP,    9.275E-23_WP,    2.267E-23_WP /
data J(1,    150    ,:)    / 3.796E-22_WP,    4.548E-22_WP,    5.447E-22_WP,    6.518E-22_WP,    7.792E-22_WP,    9.305E-22_WP, &
    & 1.110E-21_WP,    1.321E-21_WP,    1.570E-21_WP,    1.861E-21_WP,    2.199E-21_WP,    2.591E-21_WP,    3.039E-21_WP, &
    & 3.545E-21_WP,    4.105E-21_WP,    4.711E-21_WP,    5.345E-21_WP,    5.982E-21_WP,    6.583E-21_WP,    7.098E-21_WP, &
    & 7.466E-21_WP,    7.625E-21_WP,    7.518E-21_WP,    7.114E-21_WP,    6.423E-21_WP,    5.501E-21_WP,    4.406E-21_WP, &
    & 3.233E-21_WP,    2.139E-21_WP,    1.253E-21_WP,    6.364E-22_WP,    2.727E-22_WP,    9.435E-23_WP,    2.307E-23_WP /
data J(1,    151    ,:)    / 3.857E-22_WP,    4.623E-22_WP,    5.536E-22_WP,    6.625E-22_WP,    7.921E-22_WP,    9.460E-22_WP, &
    & 1.128E-21_WP,    1.343E-21_WP,    1.596E-21_WP,    1.892E-21_WP,    2.236E-21_WP,    2.634E-21_WP,    3.090E-21_WP, &
    & 3.605E-21_WP,    4.175E-21_WP,    4.791E-21_WP,    5.436E-21_WP,    6.084E-21_WP,    6.696E-21_WP,    7.219E-21_WP, &
    & 7.594E-21_WP,    7.756E-21_WP,    7.647E-21_WP,    7.236E-21_WP,    6.533E-21_WP,    5.596E-21_WP,    4.482E-21_WP, &
    & 3.288E-21_WP,    2.176E-21_WP,    1.275E-21_WP,    6.473E-22_WP,    2.774E-22_WP,    9.597E-23_WP,    2.346E-23_WP /
data J(1,    152    ,:)    / 3.920E-22_WP,    4.698E-22_WP,    5.627E-22_WP,    6.735E-22_WP,    8.053E-22_WP,    9.617E-22_WP, &
    & 1.147E-21_WP,    1.366E-21_WP,    1.623E-21_WP,    1.924E-21_WP,    2.274E-21_WP,    2.679E-21_WP,    3.143E-21_WP, &
    & 3.667E-21_WP,    4.246E-21_WP,    4.873E-21_WP,    5.529E-21_WP,    6.188E-21_WP,    6.810E-21_WP,    7.343E-21_WP, &
    & 7.724E-21_WP,    7.889E-21_WP,    7.778E-21_WP,    7.361E-21_WP,    6.646E-21_WP,    5.692E-21_WP,    4.559E-21_WP, &
    & 3.345E-21_WP,    2.213E-21_WP,    1.297E-21_WP,    6.584E-22_WP,    2.821E-22_WP,    9.763E-23_WP,    2.387E-23_WP /
data J(1,    153    ,:)    / 3.985E-22_WP,    4.776E-22_WP,    5.721E-22_WP,    6.847E-22_WP,    8.187E-22_WP,    9.778E-22_WP, &
    & 1.166E-21_WP,    1.389E-21_WP,    1.650E-21_WP,    1.956E-21_WP,    2.313E-21_WP,    2.724E-21_WP,    3.196E-21_WP, &
    & 3.729E-21_WP,    4.319E-21_WP,    4.956E-21_WP,    5.624E-21_WP,    6.294E-21_WP,    6.927E-21_WP,    7.469E-21_WP, &
    & 7.857E-21_WP,    8.024E-21_WP,    7.912E-21_WP,    7.487E-21_WP,    6.760E-21_WP,    5.790E-21_WP,    4.637E-21_WP, &
    & 3.403E-21_WP,    2.251E-21_WP,    1.319E-21_WP,    6.698E-22_WP,    2.870E-22_WP,    9.931E-23_WP,    2.428E-23_WP /
data J(1,    154    ,:)    / 4.050E-22_WP,    4.855E-22_WP,    5.816E-22_WP,    6.961E-22_WP,    8.324E-22_WP,    9.942E-22_WP, &
    & 1.186E-21_WP,    1.412E-21_WP,    1.678E-21_WP,    1.990E-21_WP,    2.352E-21_WP,    2.771E-21_WP,    3.251E-21_WP, &
    & 3.793E-21_WP,    4.393E-21_WP,    5.041E-21_WP,    5.720E-21_WP,    6.402E-21_WP,    7.046E-21_WP,    7.597E-21_WP, &
    & 7.992E-21_WP,    8.162E-21_WP,    8.048E-21_WP,    7.616E-21_WP,    6.876E-21_WP,    5.890E-21_WP,    4.717E-21_WP, &
    & 3.461E-21_WP,    2.290E-21_WP,    1.342E-21_WP,    6.813E-22_WP,    2.919E-22_WP,    1.010E-22_WP,    2.470E-23_WP /
data J(1,    155    ,:)    / 4.117E-22_WP,    4.936E-22_WP,    5.913E-22_WP,    7.078E-22_WP,    8.464E-22_WP,    1.011E-21_WP, &
    & 1.206E-21_WP,    1.436E-21_WP,    1.707E-21_WP,    2.023E-21_WP,    2.392E-21_WP,    2.818E-21_WP,    3.306E-21_WP, &
    & 3.858E-21_WP,    4.468E-21_WP,    5.127E-21_WP,    5.818E-21_WP,    6.512E-21_WP,    7.167E-21_WP,    7.728E-21_WP, &
    & 8.130E-21_WP,    8.303E-21_WP,    8.187E-21_WP,    7.747E-21_WP,    6.995E-21_WP,    5.991E-21_WP,    4.799E-21_WP, &
    & 3.521E-21_WP,    2.329E-21_WP,    1.365E-21_WP,    6.931E-22_WP,    2.970E-22_WP,    1.028E-22_WP,    2.513E-23_WP /
data J(1,    156    ,:)    / 4.186E-22_WP,    5.018E-22_WP,    6.012E-22_WP,    7.197E-22_WP,    8.606E-22_WP,    1.028E-21_WP, &
    & 1.226E-21_WP,    1.460E-21_WP,    1.736E-21_WP,    2.058E-21_WP,    2.433E-21_WP,    2.866E-21_WP,    3.363E-21_WP, &
    & 3.924E-21_WP,    4.544E-21_WP,    5.215E-21_WP,    5.918E-21_WP,    6.624E-21_WP,    7.290E-21_WP,    7.861E-21_WP, &
    & 8.270E-21_WP,    8.446E-21_WP,    8.328E-21_WP,    7.881E-21_WP,    7.116E-21_WP,    6.095E-21_WP,    4.882E-21_WP, &
    & 3.582E-21_WP,    2.370E-21_WP,    1.388E-21_WP,    7.051E-22_WP,    3.021E-22_WP,    1.046E-22_WP,    2.557E-23_WP /
data J(1,    157    ,:)    / 4.256E-22_WP,    5.102E-22_WP,    6.113E-22_WP,    7.318E-22_WP,    8.752E-22_WP,    1.045E-21_WP, &
    & 1.247E-21_WP,    1.485E-21_WP,    1.765E-21_WP,    2.093E-21_WP,    2.474E-21_WP,    2.915E-21_WP,    3.420E-21_WP, &
    & 3.991E-21_WP,    4.623E-21_WP,    5.305E-21_WP,    6.020E-21_WP,    6.738E-21_WP,    7.416E-21_WP,    7.997E-21_WP, &
    & 8.412E-21_WP,    8.591E-21_WP,    8.472E-21_WP,    8.017E-21_WP,    7.239E-21_WP,    6.200E-21_WP,    4.966E-21_WP, &
    & 3.644E-21_WP,    2.411E-21_WP,    1.412E-21_WP,    7.173E-22_WP,    3.073E-22_WP,    1.064E-22_WP,    2.601E-23_WP /
data J(1,    158    ,:)    / 4.327E-22_WP,    5.188E-22_WP,    6.216E-22_WP,    7.442E-22_WP,    8.900E-22_WP,    1.063E-21_WP, &
    & 1.268E-21_WP,    1.511E-21_WP,    1.795E-21_WP,    2.129E-21_WP,    2.517E-21_WP,    2.965E-21_WP,    3.479E-21_WP, &
    & 4.060E-21_WP,    4.702E-21_WP,    5.396E-21_WP,    6.124E-21_WP,    6.854E-21_WP,    7.544E-21_WP,    8.135E-21_WP, &
    & 8.558E-21_WP,    8.740E-21_WP,    8.618E-21_WP,    8.156E-21_WP,    7.364E-21_WP,    6.307E-21_WP,    5.052E-21_WP, &
    & 3.707E-21_WP,    2.452E-21_WP,    1.437E-21_WP,    7.297E-22_WP,    3.127E-22_WP,    1.082E-22_WP,    2.646E-23_WP /
data J(1,    159    ,:)    / 4.400E-22_WP,    5.275E-22_WP,    6.321E-22_WP,    7.568E-22_WP,    9.051E-22_WP,    1.081E-21_WP, &
    & 1.290E-21_WP,    1.536E-21_WP,    1.826E-21_WP,    2.165E-21_WP,    2.560E-21_WP,    3.016E-21_WP,    3.539E-21_WP, &
    & 4.129E-21_WP,    4.783E-21_WP,    5.489E-21_WP,    6.230E-21_WP,    6.973E-21_WP,    7.674E-21_WP,    8.275E-21_WP, &
    & 8.705E-21_WP,    8.891E-21_WP,    8.767E-21_WP,    8.297E-21_WP,    7.491E-21_WP,    6.416E-21_WP,    5.139E-21_WP, &
    & 3.771E-21_WP,    2.495E-21_WP,    1.462E-21_WP,    7.423E-22_WP,    3.181E-22_WP,    1.101E-22_WP,    2.692E-23_WP /
data J(1,    160    ,:)    / 4.474E-22_WP,    5.365E-22_WP,    6.428E-22_WP,    7.696E-22_WP,    9.205E-22_WP,    1.100E-21_WP, &
    & 1.312E-21_WP,    1.563E-21_WP,    1.857E-21_WP,    2.202E-21_WP,    2.604E-21_WP,    3.068E-21_WP,    3.600E-21_WP, &
    & 4.201E-21_WP,    4.865E-21_WP,    5.584E-21_WP,    6.337E-21_WP,    7.093E-21_WP,    7.807E-21_WP,    8.418E-21_WP, &
    & 8.856E-21_WP,    9.045E-21_WP,    8.919E-21_WP,    8.440E-21_WP,    7.621E-21_WP,    6.527E-21_WP,    5.228E-21_WP, &
    & 3.836E-21_WP,    2.538E-21_WP,    1.487E-21_WP,    7.552E-22_WP,    3.236E-22_WP,    1.120E-22_WP,    2.739E-23_WP /
data J(1,    161    ,:)    / 4.550E-22_WP,    5.456E-22_WP,    6.537E-22_WP,    7.827E-22_WP,    9.362E-22_WP,    1.119E-21_WP, &
    & 1.334E-21_WP,    1.589E-21_WP,    1.889E-21_WP,    2.240E-21_WP,    2.649E-21_WP,    3.121E-21_WP,    3.662E-21_WP, &
    & 4.273E-21_WP,    4.949E-21_WP,    5.680E-21_WP,    6.446E-21_WP,    7.216E-21_WP,    7.942E-21_WP,    8.564E-21_WP, &
    & 9.009E-21_WP,    9.201E-21_WP,    9.073E-21_WP,    8.586E-21_WP,    7.753E-21_WP,    6.640E-21_WP,    5.319E-21_WP, &
    & 3.902E-21_WP,    2.582E-21_WP,    1.513E-21_WP,    7.683E-22_WP,    3.292E-22_WP,    1.139E-22_WP,    2.787E-23_WP /
data J(1,    162    ,:)    / 4.627E-22_WP,    5.548E-22_WP,    6.649E-22_WP,    7.961E-22_WP,    9.522E-22_WP,    1.138E-21_WP, &
    & 1.357E-21_WP,    1.617E-21_WP,    1.922E-21_WP,    2.279E-21_WP,    2.694E-21_WP,    3.175E-21_WP,    3.725E-21_WP, &
    & 4.347E-21_WP,    5.035E-21_WP,    5.779E-21_WP,    6.558E-21_WP,    7.340E-21_WP,    8.079E-21_WP,    8.712E-21_WP, &
    & 9.165E-21_WP,    9.361E-21_WP,    9.230E-21_WP,    8.735E-21_WP,    7.887E-21_WP,    6.755E-21_WP,    5.411E-21_WP, &
    & 3.970E-21_WP,    2.627E-21_WP,    1.539E-21_WP,    7.816E-22_WP,    3.349E-22_WP,    1.159E-22_WP,    2.835E-23_WP /
data J(1,    163    ,:)    / 4.706E-22_WP,    5.643E-22_WP,    6.762E-22_WP,    8.097E-22_WP,    9.685E-22_WP,    1.157E-21_WP, &
    & 1.381E-21_WP,    1.644E-21_WP,    1.955E-21_WP,    2.318E-21_WP,    2.741E-21_WP,    3.229E-21_WP,    3.789E-21_WP, &
    & 4.422E-21_WP,    5.122E-21_WP,    5.878E-21_WP,    6.671E-21_WP,    7.467E-21_WP,    8.219E-21_WP,    8.863E-21_WP, &
    & 9.324E-21_WP,    9.523E-21_WP,    9.390E-21_WP,    8.887E-21_WP,    8.024E-21_WP,    6.873E-21_WP,    5.505E-21_WP, &
    & 4.039E-21_WP,    2.672E-21_WP,    1.566E-21_WP,    7.951E-22_WP,    3.407E-22_WP,    1.179E-22_WP,    2.884E-23_WP /
data J(1,    164    ,:)    / 4.786E-22_WP,    5.739E-22_WP,    6.878E-22_WP,    8.235E-22_WP,    9.851E-22_WP,    1.177E-21_WP, &
    & 1.404E-21_WP,    1.673E-21_WP,    1.988E-21_WP,    2.358E-21_WP,    2.788E-21_WP,    3.285E-21_WP,    3.855E-21_WP, &
    & 4.498E-21_WP,    5.210E-21_WP,    5.980E-21_WP,    6.787E-21_WP,    7.597E-21_WP,    8.361E-21_WP,    9.016E-21_WP, &
    & 9.485E-21_WP,    9.688E-21_WP,    9.553E-21_WP,    9.041E-21_WP,    8.163E-21_WP,    6.992E-21_WP,    5.600E-21_WP, &
    & 4.109E-21_WP,    2.719E-21_WP,    1.593E-21_WP,    8.089E-22_WP,    3.466E-22_WP,    1.200E-22_WP,    2.934E-23_WP /
data J(1,    165    ,:)    / 4.867E-22_WP,    5.837E-22_WP,    6.995E-22_WP,    8.376E-22_WP,    1.002E-21_WP,    1.197E-21_WP, &
    & 1.429E-21_WP,    1.702E-21_WP,    2.023E-21_WP,    2.399E-21_WP,    2.836E-21_WP,    3.342E-21_WP,    3.921E-21_WP, &
    & 4.576E-21_WP,    5.301E-21_WP,    6.084E-21_WP,    6.904E-21_WP,    7.728E-21_WP,    8.506E-21_WP,    9.173E-21_WP, &
    & 9.650E-21_WP,    9.856E-21_WP,    9.719E-21_WP,    9.197E-21_WP,    8.305E-21_WP,    7.113E-21_WP,    5.697E-21_WP, &
    & 4.180E-21_WP,    2.766E-21_WP,    1.620E-21_WP,    8.230E-22_WP,    3.527E-22_WP,    1.220E-22_WP,    2.985E-23_WP /
data J(1,    166    ,:)    / 4.951E-22_WP,    5.937E-22_WP,    7.115E-22_WP,    8.520E-22_WP,    1.019E-21_WP,    1.218E-21_WP, &
    & 1.453E-21_WP,    1.731E-21_WP,    2.058E-21_WP,    2.440E-21_WP,    2.885E-21_WP,    3.400E-21_WP,    3.989E-21_WP, &
    & 4.655E-21_WP,    5.392E-21_WP,    6.189E-21_WP,    7.024E-21_WP,    7.862E-21_WP,    8.654E-21_WP,    9.332E-21_WP, &
    & 9.817E-21_WP,    1.003E-20_WP,    9.887E-21_WP,    9.357E-21_WP,    8.449E-21_WP,    7.237E-21_WP,    5.796E-21_WP, &
    & 4.253E-21_WP,    2.814E-21_WP,    1.649E-21_WP,    8.373E-22_WP,    3.588E-22_WP,    1.242E-22_WP,    3.037E-23_WP /
data J(1,    167    ,:)    / 5.035E-22_WP,    6.039E-22_WP,    7.237E-22_WP,    8.667E-22_WP,    1.037E-21_WP,    1.239E-21_WP, &
    & 1.478E-21_WP,    1.761E-21_WP,    2.093E-21_WP,    2.482E-21_WP,    2.935E-21_WP,    3.458E-21_WP,    4.058E-21_WP, &
    & 4.736E-21_WP,    5.486E-21_WP,    6.296E-21_WP,    7.146E-21_WP,    7.999E-21_WP,    8.804E-21_WP,    9.494E-21_WP, &
    & 9.988E-21_WP,    1.020E-20_WP,    1.006E-20_WP,    9.520E-21_WP,    8.595E-21_WP,    7.362E-21_WP,    5.897E-21_WP, &
    & 4.327E-21_WP,    2.863E-21_WP,    1.677E-21_WP,    8.518E-22_WP,    3.650E-22_WP,    1.263E-22_WP,    3.090E-23_WP /
                                                                                                                                                
! model = 2
data J(2,    1    ,:)    / 1.200E-27_WP,    1.509E-27_WP,    1.899E-27_WP,    2.388E-27_WP,    2.999E-27_WP,    3.762E-27_WP, &
    & 4.711E-27_WP,    5.887E-27_WP,    7.336E-27_WP,    9.112E-27_WP,    1.128E-26_WP,    1.389E-26_WP,    1.703E-26_WP, &
    & 2.072E-26_WP,    2.496E-26_WP,    2.971E-26_WP,    3.486E-26_WP,    4.023E-26_WP,    4.552E-26_WP,    5.030E-26_WP, &
    & 5.408E-26_WP,    5.631E-26_WP,    5.648E-26_WP,    5.428E-26_WP,    4.975E-26_WP,    4.328E-26_WP,    3.524E-26_WP, &
    & 2.630E-26_WP,    1.772E-26_WP,    1.060E-26_WP,    5.521E-27_WP,    2.438E-27_WP,    8.759E-28_WP,    2.246E-28_WP /
data J(2,    2    ,:)    / 1.330E-27_WP,    1.674E-27_WP,    2.106E-27_WP,    2.648E-27_WP,    3.326E-27_WP,    4.173E-27_WP, &
    & 5.225E-27_WP,    6.529E-27_WP,    8.136E-27_WP,    1.011E-26_WP,    1.251E-26_WP,    1.541E-26_WP,    1.888E-26_WP, &
    & 2.298E-26_WP,    2.768E-26_WP,    3.295E-26_WP,    3.866E-26_WP,    4.462E-26_WP,    5.048E-26_WP,    5.578E-26_WP, &
    & 5.997E-26_WP,    6.244E-26_WP,    6.263E-26_WP,    6.020E-26_WP,    5.517E-26_WP,    4.799E-26_WP,    3.908E-26_WP, &
    & 2.916E-26_WP,    1.966E-26_WP,    1.176E-26_WP,    6.123E-27_WP,    2.704E-27_WP,    9.714E-28_WP,    2.492E-28_WP /
data J(2,    3    ,:)    / 1.476E-27_WP,    1.857E-27_WP,    2.336E-27_WP,    2.937E-27_WP,    3.689E-27_WP,    4.628E-27_WP, &
    & 5.795E-27_WP,    7.241E-27_WP,    9.023E-27_WP,    1.121E-26_WP,    1.387E-26_WP,    1.709E-26_WP,    2.094E-26_WP, &
    & 2.548E-26_WP,    3.070E-26_WP,    3.654E-26_WP,    4.288E-26_WP,    4.948E-26_WP,    5.597E-26_WP,    6.186E-26_WP, &
    & 6.650E-26_WP,    6.924E-26_WP,    6.945E-26_WP,    6.675E-26_WP,    6.117E-26_WP,    5.322E-26_WP,    4.333E-26_WP, &
    & 3.234E-26_WP,    2.180E-26_WP,    1.304E-26_WP,    6.791E-27_WP,    2.999E-27_WP,    1.077E-27_WP,    2.763E-28_WP /
data J(2,    4    ,:)    / 1.637E-27_WP,    2.059E-27_WP,    2.591E-27_WP,    3.257E-27_WP,    4.092E-27_WP,    5.133E-27_WP, &
    & 6.427E-27_WP,    8.031E-27_WP,    1.001E-26_WP,    1.243E-26_WP,    1.538E-26_WP,    1.895E-26_WP,    2.323E-26_WP, &
    & 2.826E-26_WP,    3.404E-26_WP,    4.052E-26_WP,    4.755E-26_WP,    5.487E-26_WP,    6.207E-26_WP,    6.859E-26_WP, &
    & 7.374E-26_WP,    7.677E-26_WP,    7.700E-26_WP,    7.401E-26_WP,    6.783E-26_WP,    5.901E-26_WP,    4.805E-26_WP, &
    & 3.587E-26_WP,    2.417E-26_WP,    1.447E-26_WP,    7.531E-27_WP,    3.326E-27_WP,    1.195E-27_WP,    3.065E-28_WP /
data J(2,    5    ,:)    / 1.815E-27_WP,    2.284E-27_WP,    2.873E-27_WP,    3.613E-27_WP,    4.538E-27_WP,    5.692E-27_WP, &
    & 7.128E-27_WP,    8.906E-27_WP,    1.110E-26_WP,    1.379E-26_WP,    1.706E-26_WP,    2.102E-26_WP,    2.576E-26_WP, &
    & 3.134E-26_WP,    3.775E-26_WP,    4.493E-26_WP,    5.272E-26_WP,    6.084E-26_WP,    6.882E-26_WP,    7.605E-26_WP, &
    & 8.176E-26_WP,    8.512E-26_WP,    8.537E-26_WP,    8.206E-26_WP,    7.520E-26_WP,    6.543E-26_WP,    5.328E-26_WP, &
    & 3.977E-26_WP,    2.681E-26_WP,    1.604E-26_WP,    8.353E-27_WP,    3.689E-27_WP,    1.325E-27_WP,    3.399E-28_WP /
data J(2,    6    ,:)    / 2.013E-27_WP,    2.533E-27_WP,    3.187E-27_WP,    4.007E-27_WP,    5.033E-27_WP,    6.313E-27_WP, &
    & 7.905E-27_WP,    9.878E-27_WP,    1.231E-26_WP,    1.529E-26_WP,    1.892E-26_WP,    2.331E-26_WP,    2.856E-26_WP, &
    & 3.475E-26_WP,    4.186E-26_WP,    4.982E-26_WP,    5.846E-26_WP,    6.746E-26_WP,    7.630E-26_WP,    8.431E-26_WP, &
    & 9.064E-26_WP,    9.436E-26_WP,    9.465E-26_WP,    9.097E-26_WP,    8.337E-26_WP,    7.254E-26_WP,    5.908E-26_WP, &
    & 4.410E-26_WP,    2.973E-26_WP,    1.779E-26_WP,    9.263E-27_WP,    4.091E-27_WP,    1.470E-27_WP,    3.770E-28_WP /
data J(2,    7    ,:)    / 2.233E-27_WP,    2.810E-27_WP,    3.535E-27_WP,    4.444E-27_WP,    5.582E-27_WP,    7.002E-27_WP, &
    & 8.767E-27_WP,    1.095E-26_WP,    1.365E-26_WP,    1.696E-26_WP,    2.098E-26_WP,    2.585E-26_WP,    3.167E-26_WP, &
    & 3.854E-26_WP,    4.642E-26_WP,    5.524E-26_WP,    6.482E-26_WP,    7.479E-26_WP,    8.459E-26_WP,    9.347E-26_WP, &
    & 1.005E-25_WP,    1.046E-25_WP,    1.049E-25_WP,    1.008E-25_WP,    9.243E-26_WP,    8.042E-26_WP,    6.550E-26_WP, &
    & 4.890E-26_WP,    3.296E-26_WP,    1.973E-26_WP,    1.027E-26_WP,    4.537E-27_WP,    1.630E-27_WP,    4.181E-28_WP /
data J(2,    8    ,:)    / 2.476E-27_WP,    3.116E-27_WP,    3.920E-27_WP,    4.929E-27_WP,    6.190E-27_WP,    7.765E-27_WP, &
    & 9.723E-27_WP,    1.215E-26_WP,    1.514E-26_WP,    1.880E-26_WP,    2.327E-26_WP,    2.866E-26_WP,    3.512E-26_WP, &
    & 4.273E-26_WP,    5.147E-26_WP,    6.125E-26_WP,    7.187E-26_WP,    8.291E-26_WP,    9.378E-26_WP,    1.036E-25_WP, &
    & 1.114E-25_WP,    1.159E-25_WP,    1.163E-25_WP,    1.118E-25_WP,    1.024E-25_WP,    8.915E-26_WP,    7.262E-26_WP, &
    & 5.421E-26_WP,    3.655E-26_WP,    2.188E-26_WP,    1.139E-26_WP,    5.032E-27_WP,    1.808E-27_WP,    4.637E-28_WP /
data J(2,    9    ,:)    / 2.746E-27_WP,    3.456E-27_WP,    4.348E-27_WP,    5.466E-27_WP,    6.866E-27_WP,    8.612E-27_WP, &
    & 1.078E-26_WP,    1.347E-26_WP,    1.679E-26_WP,    2.085E-26_WP,    2.580E-26_WP,    3.178E-26_WP,    3.894E-26_WP, &
    & 4.738E-26_WP,    5.706E-26_WP,    6.791E-26_WP,    7.967E-26_WP,    9.191E-26_WP,    1.040E-25_WP,    1.148E-25_WP, &
    & 1.234E-25_WP,    1.285E-25_WP,    1.289E-25_WP,    1.239E-25_WP,    1.135E-25_WP,    9.881E-26_WP,    8.050E-26_WP, &
    & 6.010E-26_WP,    4.053E-26_WP,    2.426E-26_WP,    1.263E-26_WP,    5.580E-27_WP,    2.005E-27_WP,    5.143E-28_WP /
data J(2,    10    ,:)    / 3.046E-27_WP,    3.833E-27_WP,    4.822E-27_WP,    6.062E-27_WP,    7.614E-27_WP,    9.550E-27_WP, &
    & 1.196E-26_WP,    1.494E-26_WP,    1.862E-26_WP,    2.312E-26_WP,    2.861E-26_WP,    3.524E-26_WP,    4.318E-26_WP, &
    & 5.253E-26_WP,    6.326E-26_WP,    7.528E-26_WP,    8.832E-26_WP,    1.019E-25_WP,    1.152E-25_WP,    1.273E-25_WP, &
    & 1.368E-25_WP,    1.424E-25_WP,    1.428E-25_WP,    1.373E-25_WP,    1.258E-25_WP,    1.095E-25_WP,    8.922E-26_WP, &
    & 6.663E-26_WP,    4.493E-26_WP,    2.690E-26_WP,    1.401E-26_WP,    6.188E-27_WP,    2.224E-27_WP,    5.704E-28_WP /
data J(2,    11    ,:)    / 3.378E-27_WP,    4.251E-27_WP,    5.348E-27_WP,    6.723E-27_WP,    8.444E-27_WP,    1.059E-26_WP, &
    & 1.326E-26_WP,    1.657E-26_WP,    2.064E-26_WP,    2.564E-26_WP,    3.172E-26_WP,    3.907E-26_WP,    4.787E-26_WP, &
    & 5.824E-26_WP,    7.013E-26_WP,    8.345E-26_WP,    9.789E-26_WP,    1.129E-25_WP,    1.277E-25_WP,    1.410E-25_WP, &
    & 1.516E-25_WP,    1.578E-25_WP,    1.582E-25_WP,    1.521E-25_WP,    1.394E-25_WP,    1.213E-25_WP,    9.888E-26_WP, &
    & 7.385E-26_WP,    4.981E-26_WP,    2.983E-26_WP,    1.554E-26_WP,    6.863E-27_WP,    2.466E-27_WP,    6.326E-28_WP /
data J(2,    12    ,:)    / 3.747E-27_WP,    4.714E-27_WP,    5.931E-27_WP,    7.456E-27_WP,    9.364E-27_WP,    1.175E-26_WP, &
    & 1.471E-26_WP,    1.837E-26_WP,    2.289E-26_WP,    2.843E-26_WP,    3.517E-26_WP,    4.332E-26_WP,    5.307E-26_WP, &
    & 6.456E-26_WP,    7.773E-26_WP,    9.249E-26_WP,    1.085E-25_WP,    1.251E-25_WP,    1.415E-25_WP,    1.562E-25_WP, &
    & 1.679E-25_WP,    1.748E-25_WP,    1.753E-25_WP,    1.685E-25_WP,    1.544E-25_WP,    1.344E-25_WP,    1.096E-25_WP, &
    & 8.185E-26_WP,    5.522E-26_WP,    3.307E-26_WP,    1.723E-26_WP,    7.610E-27_WP,    2.735E-27_WP,    7.016E-28_WP /
data J(2,    13    ,:)    / 4.155E-27_WP,    5.228E-27_WP,    6.577E-27_WP,    8.268E-27_WP,    1.038E-26_WP,    1.302E-26_WP, &
    & 1.631E-26_WP,    2.037E-26_WP,    2.538E-26_WP,    3.152E-26_WP,    3.899E-26_WP,    4.802E-26_WP,    5.883E-26_WP, &
    & 7.155E-26_WP,    8.615E-26_WP,    1.025E-25_WP,    1.202E-25_WP,    1.386E-25_WP,    1.567E-25_WP,    1.731E-25_WP, &
    & 1.860E-25_WP,    1.935E-25_WP,    1.941E-25_WP,    1.865E-25_WP,    1.710E-25_WP,    1.489E-25_WP,    1.214E-25_WP, &
    & 9.070E-26_WP,    6.121E-26_WP,    3.666E-26_WP,    1.910E-26_WP,    8.439E-27_WP,    3.033E-27_WP,    7.781E-28_WP /
data J(2,    14    ,:)    / 4.608E-27_WP,    5.798E-27_WP,    7.294E-27_WP,    9.169E-27_WP,    1.152E-26_WP,    1.444E-26_WP, &
    & 1.808E-26_WP,    2.259E-26_WP,    2.814E-26_WP,    3.494E-26_WP,    4.323E-26_WP,    5.323E-26_WP,    6.521E-26_WP, &
    & 7.930E-26_WP,    9.547E-26_WP,    1.136E-25_WP,    1.332E-25_WP,    1.535E-25_WP,    1.735E-25_WP,    1.916E-25_WP, &
    & 2.059E-25_WP,    2.142E-25_WP,    2.148E-25_WP,    2.065E-25_WP,    1.893E-25_WP,    1.649E-25_WP,    1.344E-25_WP, &
    & 1.005E-25_WP,    6.783E-26_WP,    4.064E-26_WP,    2.118E-26_WP,    9.358E-27_WP,    3.363E-27_WP,    8.629E-28_WP /
data J(2,    15    ,:)    / 5.110E-27_WP,    6.430E-27_WP,    8.088E-27_WP,    1.017E-26_WP,    1.277E-26_WP,    1.601E-26_WP, &
    & 2.005E-26_WP,    2.504E-26_WP,    3.120E-26_WP,    3.874E-26_WP,    4.791E-26_WP,    5.900E-26_WP,    7.227E-26_WP, &
    & 8.787E-26_WP,    1.058E-25_WP,    1.258E-25_WP,    1.475E-25_WP,    1.700E-25_WP,    1.921E-25_WP,    2.121E-25_WP, &
    & 2.279E-25_WP,    2.371E-25_WP,    2.377E-25_WP,    2.285E-25_WP,    2.096E-25_WP,    1.825E-25_WP,    1.489E-25_WP, &
    & 1.113E-25_WP,    7.516E-26_WP,    4.504E-26_WP,    2.348E-26_WP,    1.038E-26_WP,    3.729E-27_WP,    9.569E-28_WP /
data J(2,    16    ,:)    / 5.667E-27_WP,    7.131E-27_WP,    8.969E-27_WP,    1.128E-26_WP,    1.416E-26_WP,    1.776E-26_WP, &
    & 2.223E-26_WP,    2.776E-26_WP,    3.458E-26_WP,    4.294E-26_WP,    5.310E-26_WP,    6.538E-26_WP,    8.007E-26_WP, &
    & 9.735E-26_WP,    1.172E-25_WP,    1.393E-25_WP,    1.633E-25_WP,    1.882E-25_WP,    2.126E-25_WP,    2.347E-25_WP, &
    & 2.521E-25_WP,    2.622E-25_WP,    2.629E-25_WP,    2.528E-25_WP,    2.318E-25_WP,    2.020E-25_WP,    1.648E-25_WP, &
    & 1.233E-25_WP,    8.327E-26_WP,    4.992E-26_WP,    2.603E-26_WP,    1.150E-26_WP,    4.135E-27_WP,    1.061E-27_WP /
data J(2,    17    ,:)    / 6.285E-27_WP,    7.907E-27_WP,    9.946E-27_WP,    1.250E-26_WP,    1.570E-26_WP,    1.969E-26_WP, &
    & 2.464E-26_WP,    3.078E-26_WP,    3.833E-26_WP,    4.759E-26_WP,    5.885E-26_WP,    7.244E-26_WP,    8.871E-26_WP, &
    & 1.078E-25_WP,    1.297E-25_WP,    1.542E-25_WP,    1.808E-25_WP,    2.083E-25_WP,    2.352E-25_WP,    2.596E-25_WP, &
    & 2.787E-25_WP,    2.899E-25_WP,    2.907E-25_WP,    2.794E-25_WP,    2.563E-25_WP,    2.234E-25_WP,    1.823E-25_WP, &
    & 1.365E-25_WP,    9.223E-26_WP,    5.532E-26_WP,    2.885E-26_WP,    1.275E-26_WP,    4.585E-27_WP,    1.177E-27_WP /
data J(2,    18    ,:)    / 6.969E-27_WP,    8.768E-27_WP,    1.103E-26_WP,    1.386E-26_WP,    1.741E-26_WP,    2.183E-26_WP, &
    & 2.732E-26_WP,    3.412E-26_WP,    4.249E-26_WP,    5.274E-26_WP,    6.520E-26_WP,    8.025E-26_WP,    9.825E-26_WP, &
    & 1.194E-25_WP,    1.436E-25_WP,    1.707E-25_WP,    2.000E-25_WP,    2.304E-25_WP,    2.601E-25_WP,    2.869E-25_WP, &
    & 3.080E-25_WP,    3.203E-25_WP,    3.211E-25_WP,    3.087E-25_WP,    2.832E-25_WP,    2.469E-25_WP,    2.017E-25_WP, &
    & 1.510E-25_WP,    1.021E-25_WP,    6.128E-26_WP,    3.197E-26_WP,    1.414E-26_WP,    5.084E-27_WP,    1.305E-27_WP /
data J(2,    19    ,:)    / 7.728E-27_WP,    9.723E-27_WP,    1.223E-26_WP,    1.537E-26_WP,    1.930E-26_WP,    2.419E-26_WP, &
    & 3.028E-26_WP,    3.781E-26_WP,    4.708E-26_WP,    5.843E-26_WP,    7.223E-26_WP,    8.888E-26_WP,    1.088E-25_WP, &
    & 1.322E-25_WP,    1.590E-25_WP,    1.889E-25_WP,    2.212E-25_WP,    2.547E-25_WP,    2.874E-25_WP,    3.169E-25_WP, &
    & 3.401E-25_WP,    3.536E-25_WP,    3.544E-25_WP,    3.408E-25_WP,    3.127E-25_WP,    2.727E-25_WP,    2.229E-25_WP, &
    & 1.671E-25_WP,    1.131E-25_WP,    6.788E-26_WP,    3.543E-26_WP,    1.567E-26_WP,    5.637E-27_WP,    1.447E-27_WP /
data J(2,    20    ,:)    / 8.569E-27_WP,    1.078E-26_WP,    1.356E-26_WP,    1.704E-26_WP,    2.139E-26_WP,    2.682E-26_WP, &
    & 3.356E-26_WP,    4.190E-26_WP,    5.216E-26_WP,    6.472E-26_WP,    7.999E-26_WP,    9.840E-26_WP,    1.204E-25_WP, &
    & 1.463E-25_WP,    1.758E-25_WP,    2.088E-25_WP,    2.445E-25_WP,    2.813E-25_WP,    3.174E-25_WP,    3.498E-25_WP, &
    & 3.752E-25_WP,    3.900E-25_WP,    3.909E-25_WP,    3.758E-25_WP,    3.450E-25_WP,    3.010E-25_WP,    2.462E-25_WP, &
    & 1.847E-25_WP,    1.251E-25_WP,    7.517E-26_WP,    3.926E-26_WP,    1.737E-26_WP,    6.249E-27_WP,    1.604E-27_WP /
data J(2,    21    ,:)    / 9.501E-27_WP,    1.195E-26_WP,    1.503E-26_WP,    1.889E-26_WP,    2.371E-26_WP,    2.972E-26_WP, &
    & 3.719E-26_WP,    4.642E-26_WP,    5.778E-26_WP,    7.168E-26_WP,    8.856E-26_WP,    1.089E-25_WP,    1.332E-25_WP, &
    & 1.618E-25_WP,    1.944E-25_WP,    2.308E-25_WP,    2.700E-25_WP,    3.106E-25_WP,    3.502E-25_WP,    3.857E-25_WP, &
    & 4.135E-25_WP,    4.297E-25_WP,    4.306E-25_WP,    4.140E-25_WP,    3.802E-25_WP,    3.320E-25_WP,    2.718E-25_WP, &
    & 2.041E-25_WP,    1.384E-25_WP,    8.321E-26_WP,    4.348E-26_WP,    1.925E-26_WP,    6.927E-27_WP,    1.779E-27_WP /
data J(2,    22    ,:)    / 1.053E-26_WP,    1.325E-26_WP,    1.666E-26_WP,    2.093E-26_WP,    2.628E-26_WP,    3.293E-26_WP, &
    & 4.120E-26_WP,    5.142E-26_WP,    6.398E-26_WP,    7.935E-26_WP,    9.801E-26_WP,    1.205E-25_WP,    1.473E-25_WP, &
    & 1.788E-25_WP,    2.148E-25_WP,    2.548E-25_WP,    2.980E-25_WP,    3.425E-25_WP,    3.859E-25_WP,    4.248E-25_WP, &
    & 4.552E-25_WP,    4.728E-25_WP,    4.737E-25_WP,    4.556E-25_WP,    4.185E-25_WP,    3.657E-25_WP,    2.997E-25_WP, &
    & 2.253E-25_WP,    1.530E-25_WP,    9.208E-26_WP,    4.816E-26_WP,    2.133E-26_WP,    7.677E-27_WP,    1.972E-27_WP /
data J(2,    23    ,:)    / 1.168E-26_WP,    1.469E-26_WP,    1.847E-26_WP,    2.320E-26_WP,    2.912E-26_WP,    3.649E-26_WP, &
    & 4.563E-26_WP,    5.694E-26_WP,    7.083E-26_WP,    8.782E-26_WP,    1.084E-25_WP,    1.332E-25_WP,    1.629E-25_WP, &
    & 1.975E-25_WP,    2.371E-25_WP,    2.812E-25_WP,    3.286E-25_WP,    3.774E-25_WP,    4.248E-25_WP,    4.673E-25_WP, &
    & 5.004E-25_WP,    5.195E-25_WP,    5.204E-25_WP,    5.005E-25_WP,    4.600E-25_WP,    4.023E-25_WP,    3.301E-25_WP, &
    & 2.486E-25_WP,    1.690E-25_WP,    1.019E-25_WP,    5.332E-26_WP,    2.363E-26_WP,    8.509E-27_WP,    2.186E-27_WP /
data J(2,    24    ,:)    / 1.295E-26_WP,    1.628E-26_WP,    2.047E-26_WP,    2.571E-26_WP,    3.226E-26_WP,    4.041E-26_WP, &
    & 5.053E-26_WP,    6.303E-26_WP,    7.839E-26_WP,    9.715E-26_WP,    1.199E-25_WP,    1.473E-25_WP,    1.799E-25_WP, &
    & 2.181E-25_WP,    2.616E-25_WP,    3.099E-25_WP,    3.619E-25_WP,    4.153E-25_WP,    4.670E-25_WP,    5.132E-25_WP, &
    & 5.492E-25_WP,    5.698E-25_WP,    5.706E-25_WP,    5.489E-25_WP,    5.047E-25_WP,    4.419E-25_WP,    3.632E-25_WP, &
    & 2.739E-25_WP,    1.865E-25_WP,    1.126E-25_WP,    5.901E-26_WP,    2.617E-26_WP,    9.428E-27_WP,    2.423E-27_WP /
data J(2,    25    ,:)    / 1.435E-26_WP,    1.804E-26_WP,    2.268E-26_WP,    2.848E-26_WP,    3.573E-26_WP,    4.475E-26_WP, &
    & 5.594E-26_WP,    6.975E-26_WP,    8.671E-26_WP,    1.074E-25_WP,    1.325E-25_WP,    1.626E-25_WP,    1.985E-25_WP, &
    & 2.405E-25_WP,    2.883E-25_WP,    3.412E-25_WP,    3.980E-25_WP,    4.563E-25_WP,    5.125E-25_WP,    5.627E-25_WP, &
    & 6.015E-25_WP,    6.237E-25_WP,    6.244E-25_WP,    6.007E-25_WP,    5.527E-25_WP,    4.845E-25_WP,    3.989E-25_WP, &
    & 3.015E-25_WP,    2.058E-25_WP,    1.244E-25_WP,    6.529E-26_WP,    2.898E-26_WP,    1.045E-26_WP,    2.686E-27_WP /
data J(2,    26    ,:)    / 1.590E-26_WP,    1.999E-26_WP,    2.512E-26_WP,    3.155E-26_WP,    3.956E-26_WP,    4.954E-26_WP, &
    & 6.190E-26_WP,    7.715E-26_WP,    9.587E-26_WP,    1.187E-25_WP,    1.463E-25_WP,    1.795E-25_WP,    2.189E-25_WP, &
    & 2.649E-25_WP,    3.173E-25_WP,    3.752E-25_WP,    4.371E-25_WP,    5.004E-25_WP,    5.614E-25_WP,    6.155E-25_WP, &
    & 6.573E-25_WP,    6.811E-25_WP,    6.816E-25_WP,    6.558E-25_WP,    6.038E-25_WP,    5.301E-25_WP,    4.374E-25_WP, &
    & 3.314E-25_WP,    2.267E-25_WP,    1.374E-25_WP,    7.220E-26_WP,    3.208E-26_WP,    1.157E-26_WP,    2.977E-27_WP /
data J(2,    27    ,:)    / 1.762E-26_WP,    2.215E-26_WP,    2.783E-26_WP,    3.493E-26_WP,    4.380E-26_WP,    5.482E-26_WP, &
    & 6.847E-26_WP,    8.530E-26_WP,    1.059E-25_WP,    1.311E-25_WP,    1.614E-25_WP,    1.979E-25_WP,    2.411E-25_WP, &
    & 2.915E-25_WP,    3.487E-25_WP,    4.118E-25_WP,    4.792E-25_WP,    5.478E-25_WP,    6.136E-25_WP,    6.717E-25_WP, &
    & 7.165E-25_WP,    7.417E-25_WP,    7.419E-25_WP,    7.139E-25_WP,    6.580E-25_WP,    5.786E-25_WP,    4.786E-25_WP, &
    & 3.636E-25_WP,    2.495E-25_WP,    1.516E-25_WP,    7.981E-26_WP,    3.550E-26_WP,    1.282E-26_WP,    3.299E-27_WP /
data J(2,    28    ,:)    / 1.952E-26_WP,    2.453E-26_WP,    3.082E-26_WP,    3.867E-26_WP,    4.847E-26_WP,    6.064E-26_WP, &
    & 7.570E-26_WP,    9.425E-26_WP,    1.170E-25_WP,    1.446E-25_WP,    1.780E-25_WP,    2.179E-25_WP,    2.652E-25_WP, &
    & 3.203E-25_WP,    3.826E-25_WP,    4.512E-25_WP,    5.242E-25_WP,    5.982E-25_WP,    6.688E-25_WP,    7.310E-25_WP, &
    & 7.786E-25_WP,    8.051E-25_WP,    8.050E-25_WP,    7.748E-25_WP,    7.148E-25_WP,    6.298E-25_WP,    5.224E-25_WP, &
    & 3.983E-25_WP,    2.742E-25_WP,    1.671E-25_WP,    8.815E-26_WP,    3.927E-26_WP,    1.419E-26_WP,    3.655E-27_WP /
data J(2,    29    ,:)    / 2.162E-26_WP,    2.717E-26_WP,    3.411E-26_WP,    4.279E-26_WP,    5.361E-26_WP,    6.704E-26_WP, &
    & 8.364E-26_WP,    1.041E-25_WP,    1.290E-25_WP,    1.594E-25_WP,    1.959E-25_WP,    2.396E-25_WP,    2.913E-25_WP, &
    & 3.513E-25_WP,    4.190E-25_WP,    4.933E-25_WP,    5.721E-25_WP,    6.515E-25_WP,    7.269E-25_WP,    7.930E-25_WP, &
    & 8.432E-25_WP,    8.709E-25_WP,    8.703E-25_WP,    8.378E-25_WP,    7.737E-25_WP,    6.832E-25_WP,    5.686E-25_WP, &
    & 4.352E-25_WP,    3.008E-25_WP,    1.839E-25_WP,    9.731E-26_WP,    4.343E-26_WP,    1.571E-26_WP,    4.049E-27_WP /
data J(2,    30    ,:)    / 2.395E-26_WP,    3.007E-26_WP,    3.775E-26_WP,    4.733E-26_WP,    5.926E-26_WP,    7.406E-26_WP, &
    & 9.234E-26_WP,    1.148E-25_WP,    1.422E-25_WP,    1.754E-25_WP,    2.154E-25_WP,    2.631E-25_WP,    3.194E-25_WP, &
    & 3.845E-25_WP,    4.578E-25_WP,    5.380E-25_WP,    6.226E-25_WP,    7.074E-25_WP,    7.874E-25_WP,    8.571E-25_WP, &
    & 9.097E-25_WP,    9.383E-25_WP,    9.370E-25_WP,    9.022E-25_WP,    8.342E-25_WP,    7.385E-25_WP,    6.169E-25_WP, &
    & 4.743E-25_WP,    3.293E-25_WP,    2.022E-25_WP,    1.073E-25_WP,    4.800E-26_WP,    1.738E-26_WP,    4.485E-27_WP /
data J(2,    31    ,:)    / 2.651E-26_WP,    3.328E-26_WP,    4.175E-26_WP,    5.233E-26_WP,    6.548E-26_WP,    8.177E-26_WP, &
    & 1.019E-25_WP,    1.265E-25_WP,    1.565E-25_WP,    1.929E-25_WP,    2.365E-25_WP,    2.884E-25_WP,    3.494E-25_WP, &
    & 4.199E-25_WP,    4.989E-25_WP,    5.851E-25_WP,    6.754E-25_WP,    7.654E-25_WP,    8.498E-25_WP,    9.227E-25_WP, &
    & 9.772E-25_WP,    1.006E-24_WP,    1.004E-24_WP,    9.670E-25_WP,    8.953E-25_WP,    7.949E-25_WP,    6.669E-25_WP, &
    & 5.154E-25_WP,    3.598E-25_WP,    2.220E-25_WP,    1.182E-25_WP,    5.301E-26_WP,    1.923E-26_WP,    4.967E-27_WP /
data J(2,    32    ,:)    / 2.933E-26_WP,    3.681E-26_WP,    4.616E-26_WP,    5.781E-26_WP,    7.229E-26_WP,    9.020E-26_WP, &
    & 1.122E-25_WP,    1.392E-25_WP,    1.721E-25_WP,    2.117E-25_WP,    2.591E-25_WP,    3.154E-25_WP,    3.814E-25_WP, &
    & 4.574E-25_WP,    5.422E-25_WP,    6.342E-25_WP,    7.301E-25_WP,    8.250E-25_WP,    9.132E-25_WP,    9.888E-25_WP, &
    & 1.045E-24_WP,    1.074E-24_WP,    1.071E-24_WP,    1.031E-24_WP,    9.562E-25_WP,    8.515E-25_WP,    7.179E-25_WP, &
    & 5.581E-25_WP,    3.921E-25_WP,    2.433E-25_WP,    1.301E-25_WP,    5.851E-26_WP,    2.127E-26_WP,    5.499E-27_WP /
data J(2,    33    ,:)    / 3.245E-26_WP,    4.070E-26_WP,    5.100E-26_WP,    6.383E-26_WP,    7.974E-26_WP,    9.940E-26_WP, &
    & 1.235E-25_WP,    1.531E-25_WP,    1.888E-25_WP,    2.319E-25_WP,    2.834E-25_WP,    3.441E-25_WP,    4.152E-25_WP, &
    & 4.967E-25_WP,    5.872E-25_WP,    6.849E-25_WP,    7.862E-25_WP,    8.855E-25_WP,    9.769E-25_WP,    1.055E-24_WP, &
    & 1.111E-24_WP,    1.140E-24_WP,    1.136E-24_WP,    1.094E-24_WP,    1.016E-24_WP,    9.075E-25_WP,    7.692E-25_WP, &
    & 6.021E-25_WP,    4.260E-25_WP,    2.660E-25_WP,    1.430E-25_WP,    6.453E-26_WP,    2.351E-26_WP,    6.087E-27_WP /
data J(2,    34    ,:)    / 3.588E-26_WP,    4.497E-26_WP,    5.631E-26_WP,    7.041E-26_WP,    8.788E-26_WP,    1.094E-25_WP, &
    & 1.358E-25_WP,    1.680E-25_WP,    2.069E-25_WP,    2.536E-25_WP,    3.091E-25_WP,    3.745E-25_WP,    4.506E-25_WP, &
    & 5.375E-25_WP,    6.336E-25_WP,    7.368E-25_WP,    8.429E-25_WP,    9.460E-25_WP,    1.040E-24_WP,    1.119E-24_WP, &
    & 1.176E-24_WP,    1.204E-24_WP,    1.197E-24_WP,    1.153E-24_WP,    1.072E-24_WP,    9.617E-25_WP,    8.200E-25_WP, &
    & 6.466E-25_WP,    4.612E-25_WP,    2.902E-25_WP,    1.570E-25_WP,    7.110E-26_WP,    2.597E-26_WP,    6.735E-27_WP /
data J(2,    35    ,:)    / 3.964E-26_WP,    4.965E-26_WP,    6.212E-26_WP,    7.760E-26_WP,    9.672E-26_WP,    1.203E-25_WP, &
    & 1.490E-25_WP,    1.840E-25_WP,    2.261E-25_WP,    2.765E-25_WP,    3.362E-25_WP,    4.062E-25_WP,    4.874E-25_WP, &
    & 5.795E-25_WP,    6.809E-25_WP,    7.891E-25_WP,    8.996E-25_WP,    1.006E-24_WP,    1.101E-24_WP,    1.180E-24_WP, &
    & 1.237E-24_WP,    1.263E-24_WP,    1.255E-24_WP,    1.208E-24_WP,    1.125E-24_WP,    1.013E-24_WP,    8.692E-25_WP, &
    & 6.911E-25_WP,    4.975E-25_WP,    3.157E-25_WP,    1.720E-25_WP,    7.826E-26_WP,    2.867E-26_WP,    7.450E-27_WP /
data J(2,    36    ,:)    / 4.377E-26_WP,    5.478E-26_WP,    6.846E-26_WP,    8.541E-26_WP,    1.063E-25_WP,    1.320E-25_WP, &
    & 1.632E-25_WP,    2.011E-25_WP,    2.466E-25_WP,    3.007E-25_WP,    3.646E-25_WP,    4.391E-25_WP,    5.251E-25_WP, &
    & 6.222E-25_WP,    7.285E-25_WP,    8.411E-25_WP,    9.553E-25_WP,    1.064E-24_WP,    1.160E-24_WP,    1.239E-24_WP, &
    & 1.293E-24_WP,    1.318E-24_WP,    1.307E-24_WP,    1.259E-24_WP,    1.173E-24_WP,    1.060E-24_WP,    9.158E-25_WP, &
    & 7.347E-25_WP,    5.343E-25_WP,    3.424E-25_WP,    1.880E-25_WP,    8.603E-26_WP,    3.163E-26_WP,    8.238E-27_WP /
data J(2,    37    ,:)    / 4.829E-26_WP,    6.037E-26_WP,    7.536E-26_WP,    9.389E-26_WP,    1.167E-25_WP,    1.445E-25_WP, &
    & 1.784E-25_WP,    2.193E-25_WP,    2.681E-25_WP,    3.260E-25_WP,    3.939E-25_WP,    4.727E-25_WP,    5.633E-25_WP, &
    & 6.650E-25_WP,    7.757E-25_WP,    8.922E-25_WP,    1.009E-24_WP,    1.119E-24_WP,    1.215E-24_WP,    1.292E-24_WP, &
    & 1.345E-24_WP,    1.367E-24_WP,    1.354E-24_WP,    1.303E-24_WP,    1.216E-24_WP,    1.102E-24_WP,    9.588E-25_WP, &
    & 7.765E-25_WP,    5.710E-25_WP,    3.701E-25_WP,    2.050E-25_WP,    9.444E-26_WP,    3.487E-26_WP,    9.105E-27_WP /
data J(2,    38    ,:)    / 5.323E-26_WP,    6.646E-26_WP,    8.285E-26_WP,    1.030E-25_WP,    1.278E-25_WP,    1.580E-25_WP, &
    & 1.945E-25_WP,    2.384E-25_WP,    2.906E-25_WP,    3.521E-25_WP,    4.239E-25_WP,    5.068E-25_WP,    6.015E-25_WP, &
    & 7.074E-25_WP,    8.219E-25_WP,    9.415E-25_WP,    1.061E-24_WP,    1.171E-24_WP,    1.266E-24_WP,    1.342E-24_WP, &
    & 1.392E-24_WP,    1.411E-24_WP,    1.395E-24_WP,    1.341E-24_WP,    1.252E-24_WP,    1.139E-24_WP,    9.972E-25_WP, &
    & 8.156E-25_WP,    6.070E-25_WP,    3.983E-25_WP,    2.230E-25_WP,    1.035E-25_WP,    3.840E-26_WP,    1.006E-26_WP /
data J(2,    39    ,:)    / 5.860E-26_WP,    7.307E-26_WP,    9.093E-26_WP,    1.129E-25_WP,    1.397E-25_WP,    1.723E-25_WP, &
    & 2.115E-25_WP,    2.584E-25_WP,    3.138E-25_WP,    3.789E-25_WP,    4.543E-25_WP,    5.409E-25_WP,    6.393E-25_WP, &
    & 7.487E-25_WP,    8.663E-25_WP,    9.885E-25_WP,    1.109E-24_WP,    1.219E-24_WP,    1.313E-24_WP,    1.386E-24_WP, &
    & 1.433E-24_WP,    1.449E-24_WP,    1.429E-24_WP,    1.373E-24_WP,    1.282E-24_WP,    1.169E-24_WP,    1.030E-24_WP, &
    & 8.512E-25_WP,    6.417E-25_WP,    4.268E-25_WP,    2.418E-25_WP,    1.132E-25_WP,    4.224E-26_WP,    1.110E-26_WP /
data J(2,    40    ,:)    / 6.444E-26_WP,    8.021E-26_WP,    9.963E-26_WP,    1.234E-25_WP,    1.523E-25_WP,    1.873E-25_WP, &
    & 2.292E-25_WP,    2.790E-25_WP,    3.376E-25_WP,    4.059E-25_WP,    4.846E-25_WP,    5.744E-25_WP,    6.759E-25_WP, &
    & 7.883E-25_WP,    9.085E-25_WP,    1.033E-24_WP,    1.154E-24_WP,    1.264E-24_WP,    1.355E-24_WP,    1.426E-24_WP, &
    & 1.470E-24_WP,    1.483E-24_WP,    1.459E-24_WP,    1.399E-24_WP,    1.306E-24_WP,    1.193E-24_WP,    1.058E-24_WP, &
    & 8.825E-25_WP,    6.741E-25_WP,    4.550E-25_WP,    2.613E-25_WP,    1.236E-25_WP,    4.641E-26_WP,    1.225E-26_WP /
data J(2,    41    ,:)    / 7.075E-26_WP,    8.789E-26_WP,    1.089E-25_WP,    1.346E-25_WP,    1.657E-25_WP,    2.030E-25_WP, &
    & 2.476E-25_WP,    3.001E-25_WP,    3.616E-25_WP,    4.328E-25_WP,    5.143E-25_WP,    6.068E-25_WP,    7.110E-25_WP, &
    & 8.258E-25_WP,    9.482E-25_WP,    1.074E-24_WP,    1.197E-24_WP,    1.305E-24_WP,    1.395E-24_WP,    1.463E-24_WP, &
    & 1.504E-24_WP,    1.513E-24_WP,    1.486E-24_WP,    1.422E-24_WP,    1.325E-24_WP,    1.212E-24_WP,    1.080E-24_WP, &
    & 9.091E-25_WP,    7.037E-25_WP,    4.825E-25_WP,    2.812E-25_WP,    1.345E-25_WP,    5.093E-26_WP,    1.351E-26_WP /
data J(2,    42    ,:)    / 7.754E-26_WP,    9.612E-26_WP,    1.188E-25_WP,    1.464E-25_WP,    1.796E-25_WP,    2.193E-25_WP, &
    & 2.663E-25_WP,    3.215E-25_WP,    3.856E-25_WP,    4.593E-25_WP,    5.432E-25_WP,    6.379E-25_WP,    7.441E-25_WP, &
    & 8.610E-25_WP,    9.851E-25_WP,    1.112E-24_WP,    1.236E-24_WP,    1.344E-24_WP,    1.431E-24_WP,    1.497E-24_WP, &
    & 1.535E-24_WP,    1.541E-24_WP,    1.510E-24_WP,    1.442E-24_WP,    1.341E-24_WP,    1.227E-24_WP,    1.097E-24_WP, &
    & 9.309E-25_WP,    7.298E-25_WP,    5.087E-25_WP,    3.014E-25_WP,    1.460E-25_WP,    5.578E-26_WP,    1.488E-26_WP /
data J(2,    43    ,:)    / 8.481E-26_WP,    1.049E-25_WP,    1.293E-25_WP,    1.587E-25_WP,    1.940E-25_WP,    2.359E-25_WP, &
    & 2.853E-25_WP,    3.428E-25_WP,    4.091E-25_WP,    4.849E-25_WP,    5.707E-25_WP,    6.671E-25_WP,    7.750E-25_WP, &
    & 8.936E-25_WP,    1.019E-24_WP,    1.148E-24_WP,    1.273E-24_WP,    1.381E-24_WP,    1.466E-24_WP,    1.530E-24_WP, &
    & 1.566E-24_WP,    1.568E-24_WP,    1.533E-24_WP,    1.460E-24_WP,    1.355E-24_WP,    1.238E-24_WP,    1.110E-24_WP, &
    & 9.479E-25_WP,    7.520E-25_WP,    5.329E-25_WP,    3.213E-25_WP,    1.580E-25_WP,    6.099E-26_WP,    1.637E-26_WP /
data J(2,    44    ,:)    / 9.255E-26_WP,    1.141E-25_WP,    1.402E-25_WP,    1.715E-25_WP,    2.088E-25_WP,    2.528E-25_WP, &
    & 3.042E-25_WP,    3.638E-25_WP,    4.319E-25_WP,    5.094E-25_WP,    5.966E-25_WP,    6.943E-25_WP,    8.037E-25_WP, &
    & 9.239E-25_WP,    1.052E-24_WP,    1.182E-24_WP,    1.308E-24_WP,    1.417E-24_WP,    1.501E-24_WP,    1.564E-24_WP, &
    & 1.598E-24_WP,    1.597E-24_WP,    1.558E-24_WP,    1.480E-24_WP,    1.369E-24_WP,    1.248E-24_WP,    1.120E-24_WP, &
    & 9.607E-25_WP,    7.702E-25_WP,    5.548E-25_WP,    3.408E-25_WP,    1.703E-25_WP,    6.654E-26_WP,    1.799E-26_WP /
data J(2,    45    ,:)    / 1.007E-25_WP,    1.238E-25_WP,    1.515E-25_WP,    1.846E-25_WP,    2.238E-25_WP,    2.697E-25_WP, &
    & 3.229E-25_WP,    3.841E-25_WP,    4.537E-25_WP,    5.324E-25_WP,    6.207E-25_WP,    7.195E-25_WP,    8.302E-25_WP, &
    & 9.522E-25_WP,    1.082E-24_WP,    1.215E-24_WP,    1.344E-24_WP,    1.453E-24_WP,    1.538E-24_WP,    1.599E-24_WP, &
    & 1.631E-24_WP,    1.628E-24_WP,    1.585E-24_WP,    1.501E-24_WP,    1.384E-24_WP,    1.257E-24_WP,    1.128E-24_WP, &
    & 9.702E-25_WP,    7.845E-25_WP,    5.739E-25_WP,    3.594E-25_WP,    1.828E-25_WP,    7.242E-26_WP,    1.974E-26_WP /
data J(2,    46    ,:)    / 1.093E-25_WP,    1.339E-25_WP,    1.632E-25_WP,    1.980E-25_WP,    2.388E-25_WP,    2.863E-25_WP, &
    & 3.411E-25_WP,    4.036E-25_WP,    4.743E-25_WP,    5.539E-25_WP,    6.430E-25_WP,    7.427E-25_WP,    8.548E-25_WP, &
    & 9.790E-25_WP,    1.112E-24_WP,    1.248E-24_WP,    1.380E-24_WP,    1.491E-24_WP,    1.576E-24_WP,    1.638E-24_WP, &
    & 1.669E-24_WP,    1.663E-24_WP,    1.615E-24_WP,    1.525E-24_WP,    1.401E-24_WP,    1.268E-24_WP,    1.135E-24_WP, &
    & 9.773E-25_WP,    7.952E-25_WP,    5.899E-25_WP,    3.768E-25_WP,    1.954E-25_WP,    7.859E-26_WP,    2.163E-26_WP /
data J(2,    47    ,:)    / 1.183E-25_WP,    1.443E-25_WP,    1.751E-25_WP,    2.114E-25_WP,    2.537E-25_WP,    3.026E-25_WP, &
    & 3.585E-25_WP,    4.220E-25_WP,    4.935E-25_WP,    5.737E-25_WP,    6.636E-25_WP,    7.643E-25_WP,    8.781E-25_WP, &
    & 1.005E-24_WP,    1.141E-24_WP,    1.282E-24_WP,    1.418E-24_WP,    1.532E-24_WP,    1.618E-24_WP,    1.680E-24_WP, &
    & 1.710E-24_WP,    1.702E-24_WP,    1.650E-24_WP,    1.554E-24_WP,    1.421E-24_WP,    1.281E-24_WP,    1.143E-24_WP, &
    & 9.831E-25_WP,    8.030E-25_WP,    6.027E-25_WP,    3.925E-25_WP,    2.078E-25_WP,    8.503E-26_WP,    2.366E-26_WP /
data J(2,    48    ,:)    / 1.275E-25_WP,    1.548E-25_WP,    1.870E-25_WP,    2.247E-25_WP,    2.682E-25_WP,    3.182E-25_WP, &
    & 3.750E-25_WP,    4.392E-25_WP,    5.114E-25_WP,    5.921E-25_WP,    6.827E-25_WP,    7.846E-25_WP,    9.005E-25_WP, &
    & 1.031E-24_WP,    1.171E-24_WP,    1.317E-24_WP,    1.458E-24_WP,    1.576E-24_WP,    1.664E-24_WP,    1.727E-24_WP, &
    & 1.757E-24_WP,    1.746E-24_WP,    1.690E-24_WP,    1.586E-24_WP,    1.445E-24_WP,    1.296E-24_WP,    1.153E-24_WP, &
    & 9.886E-25_WP,    8.085E-25_WP,    6.126E-25_WP,    4.062E-25_WP,    2.198E-25_WP,    9.168E-26_WP,    2.583E-26_WP /
data J(2,    49    ,:)    / 1.370E-25_WP,    1.655E-25_WP,    1.989E-25_WP,    2.377E-25_WP,    2.823E-25_WP,    3.331E-25_WP, &
    & 3.906E-25_WP,    4.553E-25_WP,    5.279E-25_WP,    6.093E-25_WP,    7.007E-25_WP,    8.042E-25_WP,    9.227E-25_WP, &
    & 1.057E-24_WP,    1.202E-24_WP,    1.354E-24_WP,    1.501E-24_WP,    1.624E-24_WP,    1.715E-24_WP,    1.780E-24_WP, &
    & 1.809E-24_WP,    1.797E-24_WP,    1.735E-24_WP,    1.624E-24_WP,    1.473E-24_WP,    1.315E-24_WP,    1.164E-24_WP, &
    & 9.946E-25_WP,    8.127E-25_WP,    6.198E-25_WP,    4.178E-25_WP,    2.312E-25_WP,    9.847E-26_WP,    2.813E-26_WP /
data J(2,    50    ,:)    / 1.465E-25_WP,    1.762E-25_WP,    2.106E-25_WP,    2.503E-25_WP,    2.957E-25_WP,    3.472E-25_WP, &
    & 4.052E-25_WP,    4.703E-25_WP,    5.434E-25_WP,    6.254E-25_WP,    7.180E-25_WP,    8.233E-25_WP,    9.450E-25_WP, &
    & 1.084E-24_WP,    1.235E-24_WP,    1.394E-24_WP,    1.549E-24_WP,    1.676E-24_WP,    1.771E-24_WP,    1.837E-24_WP, &
    & 1.868E-24_WP,    1.852E-24_WP,    1.786E-24_WP,    1.667E-24_WP,    1.506E-24_WP,    1.337E-24_WP,    1.179E-24_WP, &
    & 1.002E-24_WP,    8.163E-25_WP,    6.248E-25_WP,    4.272E-25_WP,    2.417E-25_WP,    1.053E-25_WP,    3.055E-26_WP /
data J(2,    51    ,:)    / 1.561E-25_WP,    1.867E-25_WP,    2.221E-25_WP,    2.625E-25_WP,    3.085E-25_WP,    3.604E-25_WP, &
    & 4.188E-25_WP,    4.844E-25_WP,    5.580E-25_WP,    6.409E-25_WP,    7.349E-25_WP,    8.426E-25_WP,    9.679E-25_WP, &
    & 1.112E-24_WP,    1.270E-24_WP,    1.437E-24_WP,    1.600E-24_WP,    1.734E-24_WP,    1.832E-24_WP,    1.901E-24_WP, &
    & 1.932E-24_WP,    1.914E-24_WP,    1.843E-24_WP,    1.715E-24_WP,    1.544E-24_WP,    1.363E-24_WP,    1.196E-24_WP, &
    & 1.011E-24_WP,    8.199E-25_WP,    6.282E-25_WP,    4.345E-25_WP,    2.511E-25_WP,    1.121E-25_WP,    3.309E-26_WP /
data J(2,    52    ,:)    / 1.656E-25_WP,    1.971E-25_WP,    2.331E-25_WP,    2.741E-25_WP,    3.206E-25_WP,    3.729E-25_WP, &
    & 4.317E-25_WP,    4.978E-25_WP,    5.721E-25_WP,    6.560E-25_WP,    7.518E-25_WP,    8.622E-25_WP,    9.918E-25_WP, &
    & 1.142E-24_WP,    1.308E-24_WP,    1.484E-24_WP,    1.655E-24_WP,    1.796E-24_WP,    1.898E-24_WP,    1.971E-24_WP, &
    & 2.002E-24_WP,    1.982E-24_WP,    1.905E-24_WP,    1.769E-24_WP,    1.585E-24_WP,    1.393E-24_WP,    1.216E-24_WP, &
    & 1.021E-24_WP,    8.241E-25_WP,    6.305E-25_WP,    4.397E-25_WP,    2.593E-25_WP,    1.187E-25_WP,    3.571E-26_WP /
data J(2,    53    ,:)    / 1.750E-25_WP,    2.071E-25_WP,    2.437E-25_WP,    2.852E-25_WP,    3.321E-25_WP,    3.848E-25_WP, &
    & 4.440E-25_WP,    5.106E-25_WP,    5.858E-25_WP,    6.711E-25_WP,    7.689E-25_WP,    8.824E-25_WP,    1.017E-24_WP, &
    & 1.173E-24_WP,    1.348E-24_WP,    1.533E-24_WP,    1.715E-24_WP,    1.863E-24_WP,    1.970E-24_WP,    2.046E-24_WP, &
    & 2.078E-24_WP,    2.056E-24_WP,    1.973E-24_WP,    1.827E-24_WP,    1.632E-24_WP,    1.426E-24_WP,    1.239E-24_WP, &
    & 1.034E-24_WP,    8.292E-25_WP,    6.322E-25_WP,    4.433E-25_WP,    2.662E-25_WP,    1.251E-25_WP,    3.840E-26_WP /
data J(2,    54    ,:)    / 1.841E-25_WP,    2.168E-25_WP,    2.539E-25_WP,    2.958E-25_WP,    3.430E-25_WP,    3.961E-25_WP, &
    & 4.558E-25_WP,    5.232E-25_WP,    5.994E-25_WP,    6.863E-25_WP,    7.865E-25_WP,    9.035E-25_WP,    1.043E-24_WP, &
    & 1.207E-24_WP,    1.390E-24_WP,    1.586E-24_WP,    1.779E-24_WP,    1.936E-24_WP,    2.048E-24_WP,    2.127E-24_WP, &
    & 2.160E-24_WP,    2.135E-24_WP,    2.046E-24_WP,    1.890E-24_WP,    1.682E-24_WP,    1.463E-24_WP,    1.264E-24_WP, &
    & 1.049E-24_WP,    8.354E-25_WP,    6.338E-25_WP,    4.456E-25_WP,    2.717E-25_WP,    1.311E-25_WP,    4.113E-26_WP /
data J(2,    55    ,:)    / 1.931E-25_WP,    2.262E-25_WP,    2.637E-25_WP,    3.060E-25_WP,    3.536E-25_WP,    4.071E-25_WP, &
    & 4.674E-25_WP,    5.356E-25_WP,    6.131E-25_WP,    7.019E-25_WP,    8.046E-25_WP,    9.255E-25_WP,    1.071E-24_WP, &
    & 1.242E-24_WP,    1.435E-24_WP,    1.643E-24_WP,    1.847E-24_WP,    2.013E-24_WP,    2.130E-24_WP,    2.213E-24_WP, &
    & 2.247E-24_WP,    2.220E-24_WP,    2.124E-24_WP,    1.958E-24_WP,    1.736E-24_WP,    1.503E-24_WP,    1.293E-24_WP, &
    & 1.066E-24_WP,    8.427E-25_WP,    6.356E-25_WP,    4.468E-25_WP,    2.760E-25_WP,    1.366E-25_WP,    4.386E-26_WP /
data J(2,    56    ,:)    / 2.018E-25_WP,    2.353E-25_WP,    2.732E-25_WP,    3.159E-25_WP,    3.639E-25_WP,    4.179E-25_WP, &
    & 4.790E-25_WP,    5.482E-25_WP,    6.271E-25_WP,    7.178E-25_WP,    8.235E-25_WP,    9.484E-25_WP,    1.099E-24_WP, &
    & 1.279E-24_WP,    1.483E-24_WP,    1.702E-24_WP,    1.919E-24_WP,    2.094E-24_WP,    2.217E-24_WP,    2.304E-24_WP, &
    & 2.339E-24_WP,    2.309E-24_WP,    2.206E-24_WP,    2.029E-24_WP,    1.794E-24_WP,    1.546E-24_WP,    1.323E-24_WP, &
    & 1.085E-24_WP,    8.513E-25_WP,    6.377E-25_WP,    4.475E-25_WP,    2.791E-25_WP,    1.416E-25_WP,    4.654E-26_WP /
data J(2,    57    ,:)    / 2.102E-25_WP,    2.442E-25_WP,    2.825E-25_WP,    3.255E-25_WP,    3.740E-25_WP,    4.287E-25_WP, &
    & 4.905E-25_WP,    5.608E-25_WP,    6.413E-25_WP,    7.343E-25_WP,    8.430E-25_WP,    9.722E-25_WP,    1.130E-24_WP, &
    & 1.318E-24_WP,    1.532E-24_WP,    1.764E-24_WP,    1.994E-24_WP,    2.179E-24_WP,    2.309E-24_WP,    2.399E-24_WP, &
    & 2.435E-24_WP,    2.402E-24_WP,    2.292E-24_WP,    2.104E-24_WP,    1.855E-24_WP,    1.592E-24_WP,    1.356E-24_WP, &
    & 1.105E-24_WP,    8.609E-25_WP,    6.404E-25_WP,    4.477E-25_WP,    2.812E-25_WP,    1.459E-25_WP,    4.913E-26_WP /
data J(2,    58    ,:)    / 2.186E-25_WP,    2.529E-25_WP,    2.916E-25_WP,    3.351E-25_WP,    3.841E-25_WP,    4.394E-25_WP, &
    & 5.022E-25_WP,    5.738E-25_WP,    6.560E-25_WP,    7.512E-25_WP,    8.631E-25_WP,    9.968E-25_WP,    1.161E-24_WP, &
    & 1.358E-24_WP,    1.584E-24_WP,    1.829E-24_WP,    2.073E-24_WP,    2.269E-24_WP,    2.404E-24_WP,    2.498E-24_WP, &
    & 2.535E-24_WP,    2.499E-24_WP,    2.382E-24_WP,    2.182E-24_WP,    1.918E-24_WP,    1.639E-24_WP,    1.391E-24_WP, &
    & 1.127E-24_WP,    8.717E-25_WP,    6.436E-25_WP,    4.479E-25_WP,    2.825E-25_WP,    1.496E-25_WP,    5.159E-26_WP /
data J(2,    59    ,:)    / 2.268E-25_WP,    2.615E-25_WP,    3.006E-25_WP,    3.446E-25_WP,    3.942E-25_WP,    4.503E-25_WP, &
    & 5.141E-25_WP,    5.870E-25_WP,    6.710E-25_WP,    7.687E-25_WP,    8.838E-25_WP,    1.022E-24_WP,    1.193E-24_WP, &
    & 1.400E-24_WP,    1.637E-24_WP,    1.896E-24_WP,    2.154E-24_WP,    2.361E-24_WP,    2.502E-24_WP,    2.600E-24_WP, &
    & 2.638E-24_WP,    2.599E-24_WP,    2.473E-24_WP,    2.262E-24_WP,    1.983E-24_WP,    1.689E-24_WP,    1.427E-24_WP, &
    & 1.150E-24_WP,    8.835E-25_WP,    6.474E-25_WP,    4.481E-25_WP,    2.833E-25_WP,    1.527E-25_WP,    5.387E-26_WP /
data J(2,    60    ,:)    / 2.350E-25_WP,    2.702E-25_WP,    3.097E-25_WP,    3.542E-25_WP,    4.044E-25_WP,    4.613E-25_WP, &
    & 5.262E-25_WP,    6.005E-25_WP,    6.864E-25_WP,    7.866E-25_WP,    9.051E-25_WP,    1.048E-24_WP,    1.225E-24_WP, &
    & 1.442E-24_WP,    1.691E-24_WP,    1.964E-24_WP,    2.238E-24_WP,    2.456E-24_WP,    2.603E-24_WP,    2.705E-24_WP, &
    & 2.743E-24_WP,    2.700E-24_WP,    2.567E-24_WP,    2.344E-24_WP,    2.050E-24_WP,    1.740E-24_WP,    1.465E-24_WP, &
    & 1.174E-24_WP,    8.963E-25_WP,    6.519E-25_WP,    4.485E-25_WP,    2.837E-25_WP,    1.552E-25_WP,    5.593E-26_WP /
data J(2,    61    ,:)    / 2.433E-25_WP,    2.788E-25_WP,    3.188E-25_WP,    3.639E-25_WP,    4.148E-25_WP,    4.726E-25_WP, &
    & 5.385E-25_WP,    6.144E-25_WP,    7.022E-25_WP,    8.050E-25_WP,    9.270E-25_WP,    1.075E-24_WP,    1.259E-24_WP, &
    & 1.485E-24_WP,    1.746E-24_WP,    2.034E-24_WP,    2.323E-24_WP,    2.553E-24_WP,    2.707E-24_WP,    2.812E-24_WP, &
    & 2.850E-24_WP,    2.804E-24_WP,    2.663E-24_WP,    2.428E-24_WP,    2.119E-24_WP,    1.793E-24_WP,    1.504E-24_WP, &
    & 1.199E-24_WP,    9.100E-25_WP,    6.571E-25_WP,    4.492E-25_WP,    2.839E-25_WP,    1.572E-25_WP,    5.776E-26_WP /
data J(2,    62    ,:)    / 2.516E-25_WP,    2.876E-25_WP,    3.281E-25_WP,    3.737E-25_WP,    4.254E-25_WP,    4.841E-25_WP, &
    & 5.512E-25_WP,    6.286E-25_WP,    7.184E-25_WP,    8.238E-25_WP,    9.493E-25_WP,    1.102E-24_WP,    1.293E-24_WP, &
    & 1.529E-24_WP,    1.803E-24_WP,    2.105E-24_WP,    2.411E-24_WP,    2.652E-24_WP,    2.811E-24_WP,    2.920E-24_WP, &
    & 2.958E-24_WP,    2.908E-24_WP,    2.759E-24_WP,    2.512E-24_WP,    2.189E-24_WP,    1.846E-24_WP,    1.544E-24_WP, &
    & 1.226E-24_WP,    9.246E-25_WP,    6.629E-25_WP,    4.502E-25_WP,    2.839E-25_WP,    1.587E-25_WP,    5.932E-26_WP /
data J(2,    63    ,:)    / 2.601E-25_WP,    2.966E-25_WP,    3.376E-25_WP,    3.838E-25_WP,    4.362E-25_WP,    4.958E-25_WP, &
    & 5.642E-25_WP,    6.431E-25_WP,    7.350E-25_WP,    8.431E-25_WP,    9.720E-25_WP,    1.129E-24_WP,    1.327E-24_WP, &
    & 1.573E-24_WP,    1.859E-24_WP,    2.177E-24_WP,    2.499E-24_WP,    2.753E-24_WP,    2.917E-24_WP,    3.029E-24_WP, &
    & 3.067E-24_WP,    3.013E-24_WP,    2.856E-24_WP,    2.598E-24_WP,    2.259E-24_WP,    1.901E-24_WP,    1.585E-24_WP, &
    & 1.253E-24_WP,    9.401E-25_WP,    6.694E-25_WP,    4.516E-25_WP,    2.840E-25_WP,    1.599E-25_WP,    6.063E-26_WP /
data J(2,    64    ,:)    / 2.688E-25_WP,    3.057E-25_WP,    3.473E-25_WP,    3.941E-25_WP,    4.473E-25_WP,    5.079E-25_WP, &
    & 5.776E-25_WP,    6.581E-25_WP,    7.521E-25_WP,    8.629E-25_WP,    9.954E-25_WP,    1.157E-24_WP,    1.362E-24_WP, &
    & 1.618E-24_WP,    1.917E-24_WP,    2.250E-24_WP,    2.589E-24_WP,    2.855E-24_WP,    3.025E-24_WP,    3.139E-24_WP, &
    & 3.177E-24_WP,    3.119E-24_WP,    2.955E-24_WP,    2.684E-24_WP,    2.331E-24_WP,    1.956E-24_WP,    1.627E-24_WP, &
    & 1.282E-24_WP,    9.567E-25_WP,    6.769E-25_WP,    4.536E-25_WP,    2.843E-25_WP,    1.609E-25_WP,    6.169E-26_WP /
data J(2,    65    ,:)    / 2.777E-25_WP,    3.152E-25_WP,    3.573E-25_WP,    4.048E-25_WP,    4.588E-25_WP,    5.204E-25_WP, &
    & 5.914E-25_WP,    6.737E-25_WP,    7.699E-25_WP,    8.835E-25_WP,    1.020E-24_WP,    1.186E-24_WP,    1.398E-24_WP, &
    & 1.664E-24_WP,    1.976E-24_WP,    2.325E-24_WP,    2.681E-24_WP,    2.960E-24_WP,    3.135E-24_WP,    3.252E-24_WP, &
    & 3.290E-24_WP,    3.229E-24_WP,    3.056E-24_WP,    2.774E-24_WP,    2.406E-24_WP,    2.015E-24_WP,    1.673E-24_WP, &
    & 1.313E-24_WP,    9.755E-25_WP,    6.861E-25_WP,    4.568E-25_WP,    2.851E-25_WP,    1.619E-25_WP,    6.258E-26_WP /
data J(2,    66    ,:)    / 2.870E-25_WP,    3.250E-25_WP,    3.677E-25_WP,    4.160E-25_WP,    4.709E-25_WP,    5.338E-25_WP, &
    & 6.063E-25_WP,    6.906E-25_WP,    7.893E-25_WP,    9.062E-25_WP,    1.046E-24_WP,    1.218E-24_WP,    1.438E-24_WP, &
    & 1.715E-24_WP,    2.041E-24_WP,    2.408E-24_WP,    2.784E-24_WP,    3.076E-24_WP,    3.259E-24_WP,    3.381E-24_WP, &
    & 3.420E-24_WP,    3.356E-24_WP,    3.175E-24_WP,    2.880E-24_WP,    2.496E-24_WP,    2.088E-24_WP,    1.730E-24_WP, &
    & 1.355E-24_WP,    1.002E-24_WP,    7.010E-25_WP,    4.637E-25_WP,    2.880E-25_WP,    1.637E-25_WP,    6.350E-26_WP /
data J(2,    67    ,:)    / 2.970E-25_WP,    3.358E-25_WP,    3.794E-25_WP,    4.288E-25_WP,    4.852E-25_WP,    5.500E-25_WP, &
    & 6.249E-25_WP,    7.123E-25_WP,    8.151E-25_WP,    9.371E-25_WP,    1.084E-24_WP,    1.264E-24_WP,    1.496E-24_WP, &
    & 1.790E-24_WP,    2.137E-24_WP,    2.530E-24_WP,    2.935E-24_WP,    3.250E-24_WP,    3.450E-24_WP,    3.586E-24_WP, &
    & 3.634E-24_WP,    3.570E-24_WP,    3.382E-24_WP,    3.070E-24_WP,    2.662E-24_WP,    2.227E-24_WP,    1.844E-24_WP, &
    & 1.442E-24_WP,    1.063E-24_WP,    7.405E-25_WP,    4.868E-25_WP,    3.004E-25_WP,    1.702E-25_WP,    6.540E-26_WP /
data J(2,    68    ,:)    / 3.097E-25_WP,    3.501E-25_WP,    3.959E-25_WP,    4.481E-25_WP,    5.082E-25_WP,    5.779E-25_WP, &
    & 6.592E-25_WP,    7.550E-25_WP,    8.687E-25_WP,    1.005E-24_WP,    1.170E-24_WP,    1.374E-24_WP,    1.638E-24_WP, &
    & 1.975E-24_WP,    2.375E-24_WP,    2.830E-24_WP,    3.306E-24_WP,    3.681E-24_WP,    3.943E-24_WP,    4.133E-24_WP, &
    & 4.219E-24_WP,    4.173E-24_WP,    3.975E-24_WP,    3.626E-24_WP,    3.158E-24_WP,    2.654E-24_WP,    2.203E-24_WP, &
    & 1.725E-24_WP,    1.273E-24_WP,    8.847E-25_WP,    5.784E-25_WP,    3.531E-25_WP,    1.972E-25_WP,    7.222E-26_WP /
data J(2,    69    ,:)    / 3.326E-25_WP,    3.786E-25_WP,    4.319E-25_WP,    4.941E-25_WP,    5.674E-25_WP,    6.546E-25_WP, &
    & 7.590E-25_WP,    8.848E-25_WP,    1.038E-24_WP,    1.224E-24_WP,    1.455E-24_WP,    1.745E-24_WP,    2.123E-24_WP, &
    & 2.606E-24_WP,    3.187E-24_WP,    3.857E-24_WP,    4.566E-24_WP,    5.152E-24_WP,    5.647E-24_WP,    6.043E-24_WP, &
    & 6.284E-24_WP,    6.316E-24_WP,    6.100E-24_WP,    5.630E-24_WP,    4.957E-24_WP,    4.220E-24_WP,    3.529E-24_WP, &
    & 2.782E-24_WP,    2.064E-24_WP,    1.437E-24_WP,    9.358E-25_WP,    5.621E-25_WP,    3.045E-25_WP,    9.867E-26_WP /
data J(2,    70    ,:)    / 3.957E-25_WP,    4.625E-25_WP,    5.437E-25_WP,    6.433E-25_WP,    7.661E-25_WP,    9.185E-25_WP, &
    & 1.108E-24_WP,    1.346E-24_WP,    1.643E-24_WP,    2.016E-24_WP,    2.488E-24_WP,    3.092E-24_WP,    3.882E-24_WP, &
    & 4.890E-24_WP,    6.115E-24_WP,    7.541E-24_WP,    9.067E-24_WP,    1.041E-23_WP,    1.174E-23_WP,    1.288E-23_WP, &
    & 1.368E-23_WP,    1.399E-23_WP,    1.372E-23_WP,    1.283E-23_WP,    1.144E-23_WP,    9.867E-24_WP,    8.320E-24_WP, &
    & 6.608E-24_WP,    4.932E-24_WP,    3.448E-24_WP,    2.239E-24_WP,    1.326E-24_WP,    6.956E-25_WP,    1.930E-25_WP /
data J(2,    71    ,:)    / 3.957E-25_WP,    4.625E-25_WP,    5.437E-25_WP,    6.433E-25_WP,    7.661E-25_WP,    9.185E-25_WP, &
    & 1.108E-24_WP,    1.346E-24_WP,    1.643E-24_WP,    2.016E-24_WP,    2.488E-24_WP,    3.092E-24_WP,    3.882E-24_WP, &
    & 4.890E-24_WP,    6.115E-24_WP,    7.541E-24_WP,    9.067E-24_WP,    1.041E-23_WP,    1.174E-23_WP,    1.288E-23_WP, &
    & 1.368E-23_WP,    1.399E-23_WP,    1.372E-23_WP,    1.283E-23_WP,    1.144E-23_WP,    9.867E-24_WP,    8.320E-24_WP, &
    & 6.608E-24_WP,    4.932E-24_WP,    3.448E-24_WP,    2.239E-24_WP,    1.326E-24_WP,    6.956E-25_WP,    1.930E-25_WP /
data J(2,    72    ,:)    / 4.625E-25_WP,    5.502E-25_WP,    6.590E-25_WP,    7.951E-25_WP,    9.662E-25_WP,    1.183E-24_WP, &
    & 1.457E-24_WP,    1.806E-24_WP,    2.251E-24_WP,    2.821E-24_WP,    3.556E-24_WP,    4.519E-24_WP,    5.783E-24_WP, &
    & 7.402E-24_WP,    9.386E-24_WP,    1.170E-23_WP,    1.416E-23_WP,    1.695E-23_WP,    1.992E-23_WP,    2.291E-23_WP, &
    & 2.569E-23_WP,    2.793E-23_WP,    2.930E-23_WP,    2.948E-23_WP,    2.837E-23_WP,    2.631E-23_WP,    2.311E-23_WP, &
    & 1.858E-23_WP,    1.347E-23_WP,    8.635E-24_WP,    4.791E-24_WP,    2.238E-24_WP,    8.099E-25_WP,    2.329E-25_WP /
data J(2,    73    ,:)    / 5.438E-25_WP,    6.564E-25_WP,    7.983E-25_WP,    9.780E-25_WP,    1.207E-24_WP,    1.499E-24_WP, &
    & 1.873E-24_WP,    2.354E-24_WP,    2.974E-24_WP,    3.777E-24_WP,    4.827E-24_WP,    6.204E-24_WP,    8.002E-24_WP, &
    & 1.029E-23_WP,    1.305E-23_WP,    1.620E-23_WP,    1.991E-23_WP,    2.410E-23_WP,    2.863E-23_WP,    3.324E-23_WP, &
    & 3.756E-23_WP,    4.105E-23_WP,    4.312E-23_WP,    4.329E-23_WP,    4.147E-23_WP,    3.781E-23_WP,    3.224E-23_WP, &
    & 2.498E-23_WP,    1.733E-23_WP,    1.057E-23_WP,    5.550E-24_WP,    2.433E-24_WP,    8.784E-25_WP,    2.519E-25_WP /
data J(2,    74    ,:)    / 6.413E-25_WP,    7.837E-25_WP,    9.648E-25_WP,    1.196E-24_WP,    1.492E-24_WP,    1.874E-24_WP, &
    & 2.365E-24_WP,    3.000E-24_WP,    3.826E-24_WP,    4.904E-24_WP,    6.315E-24_WP,    8.154E-24_WP,    1.053E-23_WP, &
    & 1.349E-23_WP,    1.699E-23_WP,    2.122E-23_WP,    2.614E-23_WP,    3.166E-23_WP,    3.755E-23_WP,    4.345E-23_WP, &
    & 4.877E-23_WP,    5.280E-23_WP,    5.480E-23_WP,    5.429E-23_WP,    5.108E-23_WP,    4.549E-23_WP,    3.779E-23_WP, &
    & 2.854E-23_WP,    1.934E-23_WP,    1.156E-23_WP,    5.975E-24_WP,    2.614E-24_WP,    9.412E-25_WP,    2.718E-25_WP /
data J(2,    75    ,:)    / 7.572E-25_WP,    9.346E-25_WP,    1.162E-24_WP,    1.453E-24_WP,    1.828E-24_WP,    2.313E-24_WP, &
    & 2.941E-24_WP,    3.756E-24_WP,    4.822E-24_WP,    6.214E-24_WP,    8.024E-24_WP,    1.036E-23_WP,    1.332E-23_WP, &
    & 1.693E-23_WP,    2.135E-23_WP,    2.662E-23_WP,    3.269E-23_WP,    3.938E-23_WP,    4.637E-23_WP,    5.313E-23_WP, &
    & 5.894E-23_WP,    6.295E-23_WP,    6.444E-23_WP,    6.284E-23_WP,    5.809E-23_WP,    5.085E-23_WP,    4.160E-23_WP, &
    & 3.105E-23_WP,    2.086E-23_WP,    1.241E-23_WP,    6.408E-24_WP,    2.799E-24_WP,    1.006E-24_WP,    2.927E-25_WP /
data J(2,    76    ,:)    / 8.936E-25_WP,    1.112E-24_WP,    1.392E-24_WP,    1.753E-24_WP,    2.220E-24_WP,    2.824E-24_WP, &
    & 3.610E-24_WP,    4.635E-24_WP,    5.971E-24_WP,    7.707E-24_WP,    9.944E-24_WP,    1.278E-23_WP,    1.630E-23_WP, &
    & 2.071E-23_WP,    2.603E-23_WP,    3.228E-23_WP,    3.936E-23_WP,    4.701E-23_WP,    5.478E-23_WP,    6.202E-23_WP, &
    & 6.791E-23_WP,    7.163E-23_WP,    7.236E-23_WP,    6.964E-23_WP,    6.362E-23_WP,    5.516E-23_WP,    4.484E-23_WP, &
    & 3.334E-23_WP,    2.236E-23_WP,    1.329E-23_WP,    6.857E-24_WP,    2.991E-24_WP,    1.073E-24_WP,    3.145E-25_WP /
data J(2,    77    ,:)    / 1.028E-24_WP,    1.304E-24_WP,    1.667E-24_WP,    2.143E-24_WP,    2.771E-24_WP,    3.600E-24_WP, &
    & 4.698E-24_WP,    6.158E-24_WP,    8.097E-24_WP,    1.067E-23_WP,    1.406E-23_WP,    1.851E-23_WP,    2.428E-23_WP, &
    & 3.179E-23_WP,    4.126E-23_WP,    5.292E-23_WP,    6.675E-23_WP,    8.224E-23_WP,    1.020E-22_WP,    1.238E-22_WP, &
    & 1.464E-22_WP,    1.680E-22_WP,    1.860E-22_WP,    1.974E-22_WP,    2.002E-22_WP,    1.947E-22_WP,    1.812E-22_WP, &
    & 1.556E-22_WP,    1.212E-22_WP,    8.389E-23_WP,    4.995E-23_WP,    2.446E-23_WP,    9.242E-24_WP,    1.908E-24_WP /
data J(2,    78    ,:)    / 1.215E-24_WP,    1.549E-24_WP,    1.988E-24_WP,    2.563E-24_WP,    3.322E-24_WP,    4.323E-24_WP, &
    & 5.649E-24_WP,    7.404E-24_WP,    9.722E-24_WP,    1.277E-23_WP,    1.674E-23_WP,    2.186E-23_WP,    2.852E-23_WP, &
    & 3.696E-23_WP,    4.733E-23_WP,    5.964E-23_WP,    7.343E-23_WP,    9.062E-23_WP,    1.092E-22_WP,    1.281E-22_WP, &
    & 1.455E-22_WP,    1.590E-22_WP,    1.661E-22_WP,    1.650E-22_WP,    1.560E-22_WP,    1.402E-22_WP,    1.163E-22_WP, &
    & 8.697E-23_WP,    5.749E-23_WP,    3.269E-23_WP,    1.543E-23_WP,    5.806E-24_WP,    1.568E-24_WP,    3.475E-25_WP /
data J(2,    79    ,:)    / 1.434E-24_WP,    1.836E-24_WP,    2.361E-24_WP,    3.052E-24_WP,    3.962E-24_WP,    5.162E-24_WP, &
    & 6.746E-24_WP,    8.830E-24_WP,    1.156E-23_WP,    1.510E-23_WP,    1.965E-23_WP,    2.554E-23_WP,    3.302E-23_WP, &
    & 4.226E-23_WP,    5.325E-23_WP,    6.566E-23_WP,    8.093E-23_WP,    9.744E-23_WP,    1.141E-22_WP,    1.295E-22_WP, &
    & 1.415E-22_WP,    1.480E-22_WP,    1.474E-22_WP,    1.396E-22_WP,    1.254E-22_WP,    1.052E-22_WP,    8.145E-23_WP, &
    & 5.698E-23_WP,    3.558E-23_WP,    1.957E-23_WP,    9.361E-24_WP,    3.802E-24_WP,    1.302E-24_WP,    3.737E-25_WP /
data J(2,    80    ,:)    / 1.687E-24_WP,    2.164E-24_WP,    2.790E-24_WP,    3.611E-24_WP,    4.691E-24_WP,    6.112E-24_WP, &
    & 7.976E-24_WP,    1.041E-23_WP,    1.356E-23_WP,    1.759E-23_WP,    2.278E-23_WP,    2.935E-23_WP,    3.749E-23_WP, &
    & 4.726E-23_WP,    5.840E-23_WP,    7.202E-23_WP,    8.686E-23_WP,    1.020E-22_WP,    1.162E-22_WP,    1.279E-22_WP, &
    & 1.353E-22_WP,    1.371E-22_WP,    1.328E-22_WP,    1.227E-22_WP,    1.069E-22_WP,    8.793E-23_WP,    6.788E-23_WP, &
    & 4.826E-23_WP,    3.125E-23_WP,    1.813E-23_WP,    9.191E-24_WP,    3.949E-24_WP,    1.381E-24_WP,    3.983E-25_WP /
data J(2,    81    ,:)    / 1.975E-24_WP,    2.538E-24_WP,    3.275E-24_WP,    4.243E-24_WP,    5.511E-24_WP,    7.169E-24_WP, &
    & 9.326E-24_WP,    1.211E-23_WP,    1.567E-23_WP,    2.023E-23_WP,    2.598E-23_WP,    3.309E-23_WP,    4.167E-23_WP, &
    & 5.160E-23_WP,    6.373E-23_WP,    7.708E-23_WP,    9.099E-23_WP,    1.044E-22_WP,    1.162E-22_WP,    1.249E-22_WP, &
    & 1.295E-22_WP,    1.293E-22_WP,    1.241E-22_WP,    1.137E-22_WP,    9.922E-23_WP,    8.288E-23_WP,    6.563E-23_WP, &
    & 4.804E-23_WP,    3.194E-23_WP,    1.888E-23_WP,    9.672E-24_WP,    4.161E-24_WP,    1.452E-24_WP,    4.225E-25_WP /
data J(2,    82    ,:)    / 2.301E-24_WP,    2.960E-24_WP,    3.822E-24_WP,    4.949E-24_WP,    6.418E-24_WP,    8.323E-24_WP, &
    & 1.078E-23_WP,    1.390E-23_WP,    1.790E-23_WP,    2.292E-23_WP,    2.911E-23_WP,    3.658E-23_WP,    4.531E-23_WP, &
    & 5.604E-23_WP,    6.801E-23_WP,    8.073E-23_WP,    9.348E-23_WP,    1.053E-22_WP,    1.151E-22_WP,    1.222E-22_WP, &
    & 1.259E-22_WP,    1.257E-22_WP,    1.208E-22_WP,    1.116E-22_WP,    9.892E-23_WP,    8.409E-23_WP,    6.768E-23_WP, &
    & 5.008E-23_WP,    3.348E-23_WP,    1.982E-23_WP,    1.015E-23_WP,    4.361E-24_WP,    1.521E-24_WP,    4.467E-25_WP /
data J(2,    83    ,:)    / 2.669E-24_WP,    3.434E-24_WP,    4.431E-24_WP,    5.727E-24_WP,    7.404E-24_WP,    9.557E-24_WP, &
    & 1.229E-23_WP,    1.579E-23_WP,    2.016E-23_WP,    2.554E-23_WP,    3.204E-23_WP,    3.963E-23_WP,    4.901E-23_WP, &
    & 5.965E-23_WP,    7.120E-23_WP,    8.315E-23_WP,    9.485E-23_WP,    1.055E-22_WP,    1.145E-22_WP,    1.214E-22_WP, &
    & 1.255E-22_WP,    1.259E-22_WP,    1.222E-22_WP,    1.142E-22_WP,    1.022E-22_WP,    8.759E-23_WP,    7.081E-23_WP, &
    & 5.248E-23_WP,    3.508E-23_WP,    2.076E-23_WP,    1.062E-23_WP,    4.559E-24_WP,    1.589E-24_WP,    4.710E-25_WP /
data J(2,    84    ,:)    / 3.079E-24_WP,    3.959E-24_WP,    5.100E-24_WP,    6.572E-24_WP,    8.457E-24_WP,    1.085E-23_WP, &
    & 1.389E-23_WP,    1.769E-23_WP,    2.237E-23_WP,    2.800E-23_WP,    3.460E-23_WP,    4.272E-23_WP,    5.205E-23_WP, &
    & 6.243E-23_WP,    7.347E-23_WP,    8.478E-23_WP,    9.584E-23_WP,    1.062E-22_WP,    1.153E-22_WP,    1.229E-22_WP, &
    & 1.278E-22_WP,    1.293E-22_WP,    1.266E-22_WP,    1.190E-22_WP,    1.069E-22_WP,    9.173E-23_WP,    7.417E-23_WP, &
    & 5.495E-23_WP,    3.671E-23_WP,    2.171E-23_WP,    1.110E-23_WP,    4.760E-24_WP,    1.658E-24_WP,    4.957E-25_WP /
data J(2,    85    ,:)    / 3.533E-24_WP,    4.535E-24_WP,    5.824E-24_WP,    7.471E-24_WP,    9.556E-24_WP,    1.220E-23_WP, &
    & 1.550E-23_WP,    1.955E-23_WP,    2.443E-23_WP,    3.016E-23_WP,    3.720E-23_WP,    4.530E-23_WP,    5.446E-23_WP, &
    & 6.452E-23_WP,    7.520E-23_WP,    8.621E-23_WP,    9.728E-23_WP,    1.080E-22_WP,    1.180E-22_WP,    1.265E-22_WP, &
    & 1.325E-22_WP,    1.348E-22_WP,    1.325E-22_WP,    1.248E-22_WP,    1.121E-22_WP,    9.611E-23_WP,    7.769E-23_WP, &
    & 5.753E-23_WP,    3.842E-23_WP,    2.270E-23_WP,    1.160E-23_WP,    4.972E-24_WP,    1.730E-24_WP,    5.212E-25_WP /
data J(2,    86    ,:)    / 4.030E-24_WP,    5.158E-24_WP,    6.594E-24_WP,    8.409E-24_WP,    1.071E-23_WP,    1.356E-23_WP, &
    & 1.707E-23_WP,    2.130E-23_WP,    2.627E-23_WP,    3.236E-23_WP,    3.939E-23_WP,    4.739E-23_WP,    5.637E-23_WP, &
    & 6.625E-23_WP,    7.686E-23_WP,    8.809E-23_WP,    9.977E-23_WP,    1.114E-22_WP,    1.224E-22_WP,    1.319E-22_WP, &
    & 1.387E-22_WP,    1.416E-22_WP,    1.392E-22_WP,    1.311E-22_WP,    1.177E-22_WP,    1.008E-22_WP,    8.147E-23_WP, &
    & 6.031E-23_WP,    4.025E-23_WP,    2.378E-23_WP,    1.215E-23_WP,    5.200E-24_WP,    1.808E-24_WP,    5.479E-25_WP /
data J(2,    87    ,:)    / 4.580E-24_WP,    5.837E-24_WP,    7.423E-24_WP,    9.423E-24_WP,    1.191E-23_WP,    1.496E-23_WP, &
    & 1.864E-23_WP,    2.297E-23_WP,    2.827E-23_WP,    3.442E-23_WP,    4.146E-23_WP,    4.942E-23_WP,    5.841E-23_WP, &
    & 6.847E-23_WP,    7.955E-23_WP,    9.164E-23_WP,    1.045E-22_WP,    1.174E-22_WP,    1.297E-22_WP,    1.405E-22_WP, &
    & 1.484E-22_WP,    1.519E-22_WP,    1.497E-22_WP,    1.413E-22_WP,    1.271E-22_WP,    1.091E-22_WP,    8.844E-23_WP, &
    & 6.563E-23_WP,    4.392E-23_WP,    2.601E-23_WP,    1.333E-23_WP,    5.746E-24_WP,    2.029E-24_WP,    6.361E-25_WP /
data J(2,    88    ,:)    / 5.184E-24_WP,    6.573E-24_WP,    8.322E-24_WP,    1.049E-23_WP,    1.316E-23_WP,    1.637E-23_WP, &
    & 2.016E-23_WP,    2.480E-23_WP,    3.021E-23_WP,    3.643E-23_WP,    4.354E-23_WP,    5.165E-23_WP,    6.097E-23_WP, &
    & 7.165E-23_WP,    8.372E-23_WP,    9.711E-23_WP,    1.114E-22_WP,    1.260E-22_WP,    1.400E-22_WP,    1.522E-22_WP, &
    & 1.613E-22_WP,    1.656E-22_WP,    1.636E-22_WP,    1.548E-22_WP,    1.396E-22_WP,    1.201E-22_WP,    9.752E-23_WP, &
    & 7.250E-23_WP,    4.860E-23_WP,    2.889E-23_WP,    1.492E-23_WP,    6.548E-24_WP,    2.380E-24_WP,    6.798E-25_WP /
data J(2,    89    ,:)    / 5.822E-24_WP,    7.352E-24_WP,    9.249E-24_WP,    1.157E-23_WP,    1.438E-23_WP,    1.770E-23_WP, &
    & 2.175E-23_WP,    2.650E-23_WP,    3.200E-23_WP,    3.832E-23_WP,    4.560E-23_WP,    5.405E-23_WP,    6.396E-23_WP, &
    & 7.558E-23_WP,    8.884E-23_WP,    1.036E-22_WP,    1.195E-22_WP,    1.356E-22_WP,    1.510E-22_WP,    1.644E-22_WP, &
    & 1.743E-22_WP,    1.789E-22_WP,    1.767E-22_WP,    1.670E-22_WP,    1.503E-22_WP,    1.291E-22_WP,    1.046E-22_WP, &
    & 7.759E-23_WP,    5.193E-23_WP,    3.082E-23_WP,    1.590E-23_WP,    6.971E-24_WP,    2.526E-24_WP,    7.205E-25_WP /
data J(2,    90    ,:)    / 6.501E-24_WP,    8.158E-24_WP,    1.019E-23_WP,    1.264E-23_WP,    1.554E-23_WP,    1.909E-23_WP, &
    & 2.326E-23_WP,    2.811E-23_WP,    3.372E-23_WP,    4.025E-23_WP,    4.787E-23_WP,    5.688E-23_WP,    6.765E-23_WP, &
    & 8.037E-23_WP,    9.492E-23_WP,    1.112E-22_WP,    1.286E-22_WP,    1.462E-22_WP,    1.628E-22_WP,    1.773E-22_WP, &
    & 1.879E-22_WP,    1.927E-22_WP,    1.901E-22_WP,    1.794E-22_WP,    1.613E-22_WP,    1.383E-22_WP,    1.120E-22_WP, &
    & 8.307E-23_WP,    5.558E-23_WP,    3.298E-23_WP,    1.701E-23_WP,    7.457E-24_WP,    2.701E-24_WP,    7.679E-25_WP /
data J(2,    91    ,:)    / 7.206E-24_WP,    8.979E-24_WP,    1.112E-23_WP,    1.366E-23_WP,    1.677E-23_WP,    2.043E-23_WP, &
    & 2.471E-23_WP,    2.970E-23_WP,    3.554E-23_WP,    4.240E-23_WP,    5.057E-23_WP,    6.036E-23_WP,    7.215E-23_WP, &
    & 8.609E-23_WP,    1.021E-22_WP,    1.198E-22_WP,    1.388E-22_WP,    1.579E-22_WP,    1.759E-22_WP,    1.915E-22_WP, &
    & 2.028E-22_WP,    2.080E-22_WP,    2.052E-22_WP,    1.936E-22_WP,    1.740E-22_WP,    1.492E-22_WP,    1.209E-22_WP, &
    & 8.967E-23_WP,    6.001E-23_WP,    3.562E-23_WP,    1.838E-23_WP,    8.053E-24_WP,    2.915E-24_WP,    8.245E-25_WP /
data J(2,    92    ,:)    / 7.928E-24_WP,    9.804E-24_WP,    1.204E-23_WP,    1.476E-23_WP,    1.798E-23_WP,    2.177E-23_WP, &
    & 2.620E-23_WP,    3.142E-23_WP,    3.760E-23_WP,    4.499E-23_WP,    5.389E-23_WP,    6.462E-23_WP,    7.755E-23_WP, &
    & 9.286E-23_WP,    1.104E-22_WP,    1.298E-22_WP,    1.505E-22_WP,    1.712E-22_WP,    1.908E-22_WP,    2.078E-22_WP, &
    & 2.202E-22_WP,    2.259E-22_WP,    2.230E-22_WP,    2.106E-22_WP,    1.894E-22_WP,    1.625E-22_WP,    1.317E-22_WP, &
    & 9.776E-23_WP,    6.546E-23_WP,    3.886E-23_WP,    2.005E-23_WP,    8.786E-24_WP,    3.178E-24_WP,    8.924E-25_WP /
data J(2,    93    ,:)    / 8.662E-24_WP,    1.063E-23_WP,    1.302E-23_WP,    1.586E-23_WP,    1.921E-23_WP,    2.316E-23_WP, &
    & 2.784E-23_WP,    3.340E-23_WP,    4.009E-23_WP,    4.818E-23_WP,    5.796E-23_WP,    6.977E-23_WP,    8.400E-23_WP, &
    & 1.008E-22_WP,    1.200E-22_WP,    1.413E-22_WP,    1.640E-22_WP,    1.868E-22_WP,    2.084E-22_WP,    2.272E-22_WP, &
    & 2.410E-22_WP,    2.475E-22_WP,    2.446E-22_WP,    2.311E-22_WP,    2.080E-22_WP,    1.786E-22_WP,    1.449E-22_WP, &
    & 1.076E-22_WP,    7.210E-23_WP,    4.282E-23_WP,    2.210E-23_WP,    9.679E-24_WP,    3.497E-24_WP,    9.735E-25_WP /
data J(2,    94    ,:)    / 9.404E-24_WP,    1.152E-23_WP,    1.403E-23_WP,    1.701E-23_WP,    2.053E-23_WP,    2.472E-23_WP, &
    & 2.974E-23_WP,    3.580E-23_WP,    4.316E-23_WP,    5.208E-23_WP,    6.288E-23_WP,    7.593E-23_WP,    9.165E-23_WP, &
    & 1.102E-22_WP,    1.314E-22_WP,    1.549E-22_WP,    1.800E-22_WP,    2.052E-22_WP,    2.293E-22_WP,    2.504E-22_WP, &
    & 2.660E-22_WP,    2.735E-22_WP,    2.706E-22_WP,    2.559E-22_WP,    2.306E-22_WP,    1.981E-22_WP,    1.608E-22_WP, &
    & 1.195E-22_WP,    8.008E-23_WP,    4.757E-23_WP,    2.455E-23_WP,    1.075E-23_WP,    3.879E-24_WP,    1.069E-24_WP /
data J(2,    95    ,:)    / 1.021E-23_WP,    1.244E-23_WP,    1.509E-23_WP,    1.825E-23_WP,    2.202E-23_WP,    2.655E-23_WP, &
    & 3.205E-23_WP,    3.874E-23_WP,    4.690E-23_WP,    5.680E-23_WP,    6.879E-23_WP,    8.327E-23_WP,    1.007E-22_WP, &
    & 1.213E-22_WP,    1.448E-22_WP,    1.710E-22_WP,    1.989E-22_WP,    2.272E-22_WP,    2.542E-22_WP,    2.780E-22_WP, &
    & 2.957E-22_WP,    3.044E-22_WP,    3.014E-22_WP,    2.854E-22_WP,    2.572E-22_WP,    2.211E-22_WP,    1.795E-22_WP, &
    & 1.335E-22_WP,    8.947E-23_WP,    5.316E-23_WP,    2.743E-23_WP,    1.200E-23_WP,    4.326E-24_WP,    1.181E-24_WP /
data J(2,    96    ,:)    / 1.107E-23_WP,    1.343E-23_WP,    1.626E-23_WP,    1.966E-23_WP,    2.376E-23_WP,    2.876E-23_WP, &
    & 3.486E-23_WP,    4.231E-23_WP,    5.139E-23_WP,    6.243E-23_WP,    7.580E-23_WP,    9.194E-23_WP,    1.114E-22_WP, &
    & 1.344E-22_WP,    1.607E-22_WP,    1.900E-22_WP,    2.213E-22_WP,    2.531E-22_WP,    2.836E-22_WP,    3.105E-22_WP, &
    & 3.306E-22_WP,    3.405E-22_WP,    3.374E-22_WP,    3.196E-22_WP,    2.882E-22_WP,    2.477E-22_WP,    2.012E-22_WP, &
    & 1.496E-22_WP,    1.003E-22_WP,    5.960E-23_WP,    3.075E-23_WP,    1.345E-23_WP,    4.838E-24_WP,    1.309E-24_WP /
data J(2,    97    ,:)    / 1.118E-23_WP,    1.357E-23_WP,    1.643E-23_WP,    1.987E-23_WP,    2.403E-23_WP,    2.908E-23_WP, &
    & 3.525E-23_WP,    4.279E-23_WP,    5.198E-23_WP,    6.313E-23_WP,    7.664E-23_WP,    9.293E-23_WP,    1.126E-22_WP, &
    & 1.358E-22_WP,    1.624E-22_WP,    1.919E-22_WP,    2.236E-22_WP,    2.557E-22_WP,    2.866E-22_WP,    3.138E-22_WP, &
    & 3.341E-22_WP,    3.442E-22_WP,    3.410E-22_WP,    3.230E-22_WP,    2.913E-22_WP,    2.504E-22_WP,    2.034E-22_WP, &
    & 1.512E-22_WP,    1.014E-22_WP,    6.025E-23_WP,    3.108E-23_WP,    1.359E-23_WP,    4.890E-24_WP,    1.321E-24_WP /
data J(2,    98    ,:)    / 1.235E-23_WP,    1.500E-23_WP,    1.822E-23_WP,    2.214E-23_WP,    2.695E-23_WP,    3.289E-23_WP, &
    & 4.021E-23_WP,    4.925E-23_WP,    6.037E-23_WP,    7.406E-23_WP,    9.085E-23_WP,    1.115E-22_WP,    1.367E-22_WP, &
    & 1.673E-22_WP,    2.031E-22_WP,    2.442E-22_WP,    2.898E-22_WP,    3.392E-22_WP,    3.905E-22_WP,    4.407E-22_WP, &
    & 4.852E-22_WP,    5.187E-22_WP,    5.352E-22_WP,    5.301E-22_WP,    5.022E-22_WP,    4.562E-22_WP,    3.907E-22_WP, &
    & 3.068E-22_WP,    2.174E-22_WP,    1.367E-22_WP,    7.465E-23_WP,    3.463E-23_WP,    1.323E-23_WP,    3.844E-24_WP /
data J(2,    99    ,:)    / 1.366E-23_WP,    1.664E-23_WP,    2.029E-23_WP,    2.481E-23_WP,    3.043E-23_WP,    3.741E-23_WP, &
    & 4.610E-23_WP,    5.691E-23_WP,    7.035E-23_WP,    8.705E-23_WP,    1.078E-22_WP,    1.336E-22_WP,    1.657E-22_WP, &
    & 2.049E-22_WP,    2.516E-22_WP,    3.059E-22_WP,    3.681E-22_WP,    4.372E-22_WP,    5.109E-22_WP,    5.850E-22_WP, &
    & 6.531E-22_WP,    7.071E-22_WP,    7.378E-22_WP,    7.382E-22_WP,    7.067E-22_WP,    6.444E-22_WP,    5.499E-22_WP, &
    & 4.282E-22_WP,    2.997E-22_WP,    1.852E-22_WP,    9.916E-23_WP,    4.493E-23_WP,    1.666E-23_WP,    4.261E-24_WP /
data J(2,    100    ,:)    / 1.513E-23_WP,    1.850E-23_WP,    2.270E-23_WP,    2.793E-23_WP,    3.448E-23_WP,    4.269E-23_WP, &
    & 5.298E-23_WP,    6.587E-23_WP,    8.203E-23_WP,    1.023E-22_WP,    1.277E-22_WP,    1.597E-22_WP,    1.996E-22_WP, &
    & 2.488E-22_WP,    3.078E-22_WP,    3.775E-22_WP,    4.582E-22_WP,    5.486E-22_WP,    6.453E-22_WP,    7.426E-22_WP, &
    & 8.316E-22_WP,    9.012E-22_WP,    9.397E-22_WP,    9.392E-22_WP,    8.946E-22_WP,    8.073E-22_WP,    6.787E-22_WP, &
    & 5.197E-22_WP,    3.573E-22_WP,    2.169E-22_WP,    1.141E-22_WP,    5.084E-23_WP,    1.821E-23_WP,    4.641E-24_WP /
data J(2,    101    ,:)    / 1.677E-23_WP,    2.062E-23_WP,    2.544E-23_WP,    3.150E-23_WP,    3.913E-23_WP,    4.873E-23_WP, &
    & 6.084E-23_WP,    7.613E-23_WP,    9.543E-23_WP,    1.198E-22_WP,    1.507E-22_WP,    1.895E-22_WP,    2.383E-22_WP, &
    & 2.985E-22_WP,    3.713E-22_WP,    4.577E-22_WP,    5.577E-22_WP,    6.693E-22_WP,    7.878E-22_WP,    9.053E-22_WP, &
    & 1.010E-21_WP,    1.089E-21_WP,    1.129E-21_WP,    1.119E-21_WP,    1.054E-21_WP,    9.377E-22_WP,    7.760E-22_WP, &
    & 5.852E-22_WP,    3.967E-22_WP,    2.379E-22_WP,    1.239E-22_WP,    5.445E-23_WP,    1.944E-23_WP,    4.930E-24_WP /
data J(2,    102    ,:)    / 1.861E-23_WP,    2.301E-23_WP,    2.854E-23_WP,    3.554E-23_WP,    4.438E-23_WP,    5.557E-23_WP, &
    & 6.976E-23_WP,    8.778E-23_WP,    1.107E-22_WP,    1.398E-22_WP,    1.766E-22_WP,    2.232E-22_WP,    2.817E-22_WP, &
    & 3.540E-22_WP,    4.416E-22_WP,    5.453E-22_WP,    6.644E-22_WP,    7.959E-22_WP,    9.333E-22_WP,    1.066E-21_WP, &
    & 1.182E-21_WP,    1.264E-21_WP,    1.299E-21_WP,    1.273E-21_WP,    1.184E-21_WP,    1.039E-21_WP,    8.489E-22_WP, &
    & 6.333E-22_WP,    4.256E-22_WP,    2.536E-22_WP,    1.312E-22_WP,    5.754E-23_WP,    2.050E-23_WP,    5.191E-24_WP /
data J(2,    103    ,:)    / 1.966E-23_WP,    2.449E-23_WP,    3.063E-23_WP,    3.846E-23_WP,    4.847E-23_WP,    6.129E-23_WP, &
    & 7.773E-23_WP,    9.886E-23_WP,    1.261E-22_WP,    1.611E-22_WP,    2.062E-22_WP,    2.640E-22_WP,    3.379E-22_WP, &
    & 4.313E-22_WP,    5.468E-22_WP,    6.870E-22_WP,    8.527E-22_WP,    1.041E-21_WP,    1.257E-21_WP,    1.483E-21_WP, &
    & 1.705E-21_WP,    1.900E-21_WP,    2.042E-21_WP,    2.106E-21_WP,    2.073E-21_WP,    1.945E-21_WP,    1.721E-21_WP, &
    & 1.403E-21_WP,    1.039E-21_WP,    6.870E-22_WP,    3.974E-22_WP,    1.961E-22_WP,    7.972E-23_WP,    2.500E-23_WP /
data J(2,    104    ,:)    / 2.195E-23_WP,    2.748E-23_WP,    3.455E-23_WP,    4.360E-23_WP,    5.520E-23_WP,    7.013E-23_WP, &
    & 8.935E-23_WP,    1.142E-22_WP,    1.462E-22_WP,    1.876E-22_WP,    2.408E-22_WP,    3.091E-22_WP,    3.961E-22_WP, &
    & 5.054E-22_WP,    6.396E-22_WP,    8.002E-22_WP,    9.865E-22_WP,    1.202E-21_WP,    1.435E-21_WP,    1.670E-21_WP, &
    & 1.888E-21_WP,    2.061E-21_WP,    2.161E-21_WP,    2.166E-21_WP,    2.068E-21_WP,    1.871E-21_WP,    1.571E-21_WP, &
    & 1.202E-21_WP,    8.251E-22_WP,    4.982E-22_WP,    2.587E-22_WP,    1.123E-22_WP,    3.902E-23_WP,    5.708E-24_WP /
data J(2,    105    ,:)    / 2.458E-23_WP,    3.092E-23_WP,    3.904E-23_WP,    4.949E-23_WP,    6.294E-23_WP,    8.031E-23_WP, &
    & 1.028E-22_WP,    1.319E-22_WP,    1.695E-22_WP,    2.181E-22_WP,    2.805E-22_WP,    3.604E-22_WP,    4.618E-22_WP, &
    & 5.879E-22_WP,    7.409E-22_WP,    9.210E-22_WP,    1.133E-21_WP,    1.366E-21_WP,    1.610E-21_WP,    1.847E-21_WP, &
    & 2.052E-21_WP,    2.195E-21_WP,    2.250E-21_WP,    2.203E-21_WP,    2.048E-21_WP,    1.786E-21_WP,    1.440E-21_WP, &
    & 1.055E-21_WP,    6.905E-22_WP,    3.975E-22_WP,    1.973E-22_WP,    8.243E-23_WP,    2.447E-23_WP,    6.075E-24_WP /
data J(2,    106    ,:)    / 2.755E-23_WP,    3.481E-23_WP,    4.414E-23_WP,    5.618E-23_WP,    7.175E-23_WP,    9.191E-23_WP, &
    & 1.181E-22_WP,    1.520E-22_WP,    1.958E-22_WP,    2.524E-22_WP,    3.250E-22_WP,    4.174E-22_WP,    5.335E-22_WP, &
    & 6.763E-22_WP,    8.467E-22_WP,    1.050E-21_WP,    1.278E-21_WP,    1.524E-21_WP,    1.772E-21_WP,    2.001E-21_WP, &
    & 2.184E-21_WP,    2.293E-21_WP,    2.308E-21_WP,    2.215E-21_WP,    2.007E-21_WP,    1.705E-21_WP,    1.343E-21_WP, &
    & 9.641E-22_WP,    6.228E-22_WP,    3.571E-22_WP,    1.786E-22_WP,    7.327E-23_WP,    2.599E-23_WP,    6.409E-24_WP /
data J(2,    107    ,:)    / 3.090E-23_WP,    3.920E-23_WP,    4.991E-23_WP,    6.376E-23_WP,    8.173E-23_WP,    1.051E-22_WP, &
    & 1.354E-22_WP,    1.747E-22_WP,    2.254E-22_WP,    2.907E-22_WP,    3.741E-22_WP,    4.793E-22_WP,    6.099E-22_WP, &
    & 7.682E-22_WP,    9.588E-22_WP,    1.177E-21_WP,    1.418E-21_WP,    1.669E-21_WP,    1.914E-21_WP,    2.127E-21_WP, &
    & 2.285E-21_WP,    2.363E-21_WP,    2.343E-21_WP,    2.207E-21_WP,    1.966E-21_WP,    1.650E-21_WP,    1.291E-21_WP, &
    & 9.273E-22_WP,    6.040E-22_WP,    3.514E-22_WP,    1.766E-22_WP,    7.731E-23_WP,    2.736E-23_WP,    6.738E-24_WP /
data J(2,    108    ,:)    / 3.468E-23_WP,    4.415E-23_WP,    5.642E-23_WP,    7.234E-23_WP,    9.303E-23_WP,    1.199E-22_WP, &
    & 1.548E-22_WP,    2.000E-22_WP,    2.583E-22_WP,    3.329E-22_WP,    4.273E-22_WP,    5.451E-22_WP,    6.893E-22_WP, &
    & 8.652E-22_WP,    1.070E-21_WP,    1.299E-21_WP,    1.546E-21_WP,    1.797E-21_WP,    2.031E-21_WP,    2.226E-21_WP, &
    & 2.361E-21_WP,    2.413E-21_WP,    2.359E-21_WP,    2.198E-21_WP,    1.945E-21_WP,    1.631E-21_WP,    1.284E-21_WP, &
    & 9.312E-22_WP,    6.140E-22_WP,    3.595E-22_WP,    1.856E-22_WP,    8.117E-23_WP,    2.871E-23_WP,    7.075E-24_WP /
data J(2,    109    ,:)    / 3.894E-23_WP,    4.974E-23_WP,    6.377E-23_WP,    8.202E-23_WP,    1.058E-22_WP,    1.366E-22_WP, &
    & 1.766E-22_WP,    2.282E-22_WP,    2.944E-22_WP,    3.785E-22_WP,    4.838E-22_WP,    6.134E-22_WP,    7.729E-22_WP, &
    & 9.614E-22_WP,    1.177E-21_WP,    1.413E-21_WP,    1.661E-21_WP,    1.904E-21_WP,    2.126E-21_WP,    2.305E-21_WP, &
    & 2.421E-21_WP,    2.448E-21_WP,    2.375E-21_WP,    2.206E-21_WP,    1.956E-21_WP,    1.651E-21_WP,    1.311E-21_WP, &
    & 9.595E-22_WP,    6.358E-22_WP,    3.770E-22_WP,    1.945E-22_WP,    8.504E-23_WP,    3.009E-23_WP,    7.424E-24_WP /
data J(2,    110    ,:)    / 4.373E-23_WP,    5.604E-23_WP,    7.206E-23_WP,    9.291E-23_WP,    1.200E-22_WP,    1.552E-22_WP, &
    & 2.007E-22_WP,    2.591E-22_WP,    3.335E-22_WP,    4.270E-22_WP,    5.425E-22_WP,    6.853E-22_WP,    8.560E-22_WP, &
    & 1.054E-21_WP,    1.276E-21_WP,    1.514E-21_WP,    1.758E-21_WP,    1.993E-21_WP,    2.204E-21_WP,    2.370E-21_WP, &
    & 2.468E-21_WP,    2.483E-21_WP,    2.407E-21_WP,    2.241E-21_WP,    1.999E-21_WP,    1.700E-21_WP,    1.359E-21_WP, &
    & 9.985E-22_WP,    6.661E-22_WP,    3.948E-22_WP,    2.037E-22_WP,    8.903E-23_WP,    3.151E-23_WP,    7.789E-24_WP /
data J(2,    111    ,:)    / 4.913E-23_WP,    6.314E-23_WP,    8.138E-23_WP,    1.051E-22_WP,    1.359E-22_WP,    1.758E-22_WP, &
    & 2.271E-22_WP,    2.926E-22_WP,    3.750E-22_WP,    4.773E-22_WP,    6.043E-22_WP,    7.569E-22_WP,    9.362E-22_WP, &
    & 1.141E-21_WP,    1.365E-21_WP,    1.602E-21_WP,    1.840E-21_WP,    2.069E-21_WP,    2.271E-21_WP,    2.426E-21_WP, &
    & 2.517E-21_WP,    2.532E-21_WP,    2.462E-21_WP,    2.306E-21_WP,    2.069E-21_WP,    1.768E-21_WP,    1.417E-21_WP, &
    & 1.045E-21_WP,    6.972E-22_WP,    4.132E-22_WP,    2.132E-22_WP,    9.319E-23_WP,    3.299E-23_WP,    8.172E-24_WP /
data J(2,    112    ,:)    / 5.520E-23_WP,    7.110E-23_WP,    9.180E-23_WP,    1.187E-22_WP,    1.535E-22_WP,    1.984E-22_WP, &
    & 2.557E-22_WP,    3.281E-22_WP,    4.181E-22_WP,    5.303E-22_WP,    6.658E-22_WP,    8.261E-22_WP,    1.011E-21_WP, &
    & 1.219E-21_WP,    1.443E-21_WP,    1.677E-21_WP,    1.911E-21_WP,    2.135E-21_WP,    2.331E-21_WP,    2.483E-21_WP, &
    & 2.578E-21_WP,    2.603E-21_WP,    2.544E-21_WP,    2.394E-21_WP,    2.156E-21_WP,    1.847E-21_WP,    1.484E-21_WP, &
    & 1.094E-21_WP,    7.297E-22_WP,    4.324E-22_WP,    2.231E-22_WP,    9.755E-23_WP,    3.456E-23_WP,    8.576E-24_WP /
data J(2,    113    ,:)    / 6.200E-23_WP,    8.000E-23_WP,    1.034E-22_WP,    1.337E-22_WP,    1.728E-22_WP,    2.228E-22_WP, &
    & 2.860E-22_WP,    3.649E-22_WP,    4.635E-22_WP,    5.831E-22_WP,    7.254E-22_WP,    8.912E-22_WP,    1.080E-21_WP, &
    & 1.288E-21_WP,    1.511E-21_WP,    1.743E-21_WP,    1.976E-21_WP,    2.196E-21_WP,    2.393E-21_WP,    2.552E-21_WP, &
    & 2.659E-21_WP,    2.696E-21_WP,    2.646E-21_WP,    2.498E-21_WP,    2.254E-21_WP,    1.933E-21_WP,    1.553E-21_WP, &
    & 1.145E-21_WP,    7.638E-22_WP,    4.527E-22_WP,    2.336E-22_WP,    1.021E-22_WP,    3.620E-23_WP,    9.001E-24_WP /
data J(2,    114    ,:)    / 6.958E-23_WP,    8.986E-23_WP,    1.162E-22_WP,    1.501E-22_WP,    1.935E-22_WP,    2.486E-22_WP, &
    & 3.174E-22_WP,    4.036E-22_WP,    5.085E-22_WP,    6.341E-22_WP,    7.813E-22_WP,    9.502E-22_WP,    1.140E-21_WP, &
    & 1.347E-21_WP,    1.570E-21_WP,    1.802E-21_WP,    2.033E-21_WP,    2.257E-21_WP,    2.462E-21_WP,    2.634E-21_WP, &
    & 2.755E-21_WP,    2.802E-21_WP,    2.756E-21_WP,    2.604E-21_WP,    2.351E-21_WP,    2.016E-21_WP,    1.618E-21_WP, &
    & 1.193E-21_WP,    7.946E-22_WP,    4.705E-22_WP,    2.425E-22_WP,    1.060E-22_WP,    3.752E-23_WP,    9.315E-24_WP /
data J(2,    115    ,:)    / 7.792E-23_WP,    1.007E-22_WP,    1.300E-22_WP,    1.677E-22_WP,    2.154E-22_WP,    2.751E-22_WP, &
    & 3.501E-22_WP,    4.418E-22_WP,    5.518E-22_WP,    6.815E-22_WP,    8.315E-22_WP,    1.001E-21_WP,    1.191E-21_WP, &
    & 1.399E-21_WP,    1.621E-21_WP,    1.853E-21_WP,    2.088E-21_WP,    2.320E-21_WP,    2.536E-21_WP,    2.722E-21_WP, &
    & 2.853E-21_WP,    2.906E-21_WP,    2.858E-21_WP,    2.700E-21_WP,    2.434E-21_WP,    2.084E-21_WP,    1.670E-21_WP, &
    & 1.229E-21_WP,    8.172E-22_WP,    4.831E-22_WP,    2.486E-22_WP,    1.085E-22_WP,    3.833E-23_WP,    9.496E-24_WP /
data J(2,    116    ,:)    / 8.705E-23_WP,    1.124E-22_WP,    1.449E-22_WP,    1.861E-22_WP,    2.378E-22_WP,    3.028E-22_WP, &
    & 3.824E-22_WP,    4.784E-22_WP,    5.920E-22_WP,    7.242E-22_WP,    8.751E-22_WP,    1.045E-21_WP,    1.235E-21_WP, &
    & 1.443E-21_WP,    1.666E-21_WP,    1.901E-21_WP,    2.143E-21_WP,    2.386E-21_WP,    2.615E-21_WP,    2.810E-21_WP, &
    & 2.947E-21_WP,    3.000E-21_WP,    2.948E-21_WP,    2.779E-21_WP,    2.501E-21_WP,    2.137E-21_WP,    1.710E-21_WP, &
    & 1.255E-21_WP,    8.336E-22_WP,    4.920E-22_WP,    2.529E-22_WP,    1.102E-22_WP,    3.890E-23_WP,    9.639E-24_WP /
data J(2,    117    ,:)    / 9.695E-23_WP,    1.249E-22_WP,    1.605E-22_WP,    2.051E-22_WP,    2.612E-22_WP,    3.301E-22_WP, &
    & 4.134E-22_WP,    5.124E-22_WP,    6.282E-22_WP,    7.612E-22_WP,    9.122E-22_WP,    1.083E-21_WP,    1.273E-21_WP, &
    & 1.482E-21_WP,    1.708E-21_WP,    1.949E-21_WP,    2.202E-21_WP,    2.455E-21_WP,    2.694E-21_WP,    2.895E-21_WP, &
    & 3.033E-21_WP,    3.083E-21_WP,    3.024E-21_WP,    2.846E-21_WP,    2.557E-21_WP,    2.180E-21_WP,    1.741E-21_WP, &
    & 1.277E-21_WP,    8.469E-22_WP,    4.994E-22_WP,    2.565E-22_WP,    1.117E-22_WP,    3.944E-23_WP,    9.786E-24_WP /
data J(2,    118    ,:)    / 1.075E-22_WP,    1.381E-22_WP,    1.765E-22_WP,    2.248E-22_WP,    2.843E-22_WP,    3.563E-22_WP, &
    & 4.422E-22_WP,    5.430E-22_WP,    6.596E-22_WP,    7.928E-22_WP,    9.442E-22_WP,    1.115E-21_WP,    1.306E-21_WP, &
    & 1.518E-21_WP,    1.750E-21_WP,    2.001E-21_WP,    2.263E-21_WP,    2.525E-21_WP,    2.770E-21_WP,    2.973E-21_WP, &
    & 3.110E-21_WP,    3.156E-21_WP,    3.090E-21_WP,    2.903E-21_WP,    2.604E-21_WP,    2.217E-21_WP,    1.769E-21_WP, &
    & 1.296E-21_WP,    8.591E-22_WP,    5.065E-22_WP,    2.601E-22_WP,    1.133E-22_WP,    4.001E-23_WP,    9.937E-24_WP /
data J(2,    119    ,:)    / 1.186E-22_WP,    1.516E-22_WP,    1.931E-22_WP,    2.443E-22_WP,    3.064E-22_WP,    3.807E-22_WP, &
    & 4.681E-22_WP,    5.697E-22_WP,    6.864E-22_WP,    8.201E-22_WP,    9.723E-22_WP,    1.144E-21_WP,    1.338E-21_WP, &
    & 1.555E-21_WP,    1.795E-21_WP,    2.054E-21_WP,    2.325E-21_WP,    2.593E-21_WP,    2.840E-21_WP,    3.044E-21_WP, &
    & 3.179E-21_WP,    3.220E-21_WP,    3.148E-21_WP,    2.953E-21_WP,    2.646E-21_WP,    2.251E-21_WP,    1.795E-21_WP, &
    & 1.315E-21_WP,    8.713E-22_WP,    5.136E-22_WP,    2.637E-22_WP,    1.149E-22_WP,    4.058E-23_WP,    1.009E-23_WP /
data J(2,    120    ,:)    / 1.301E-22_WP,    1.657E-22_WP,    2.096E-22_WP,    2.630E-22_WP,    3.270E-22_WP,    4.026E-22_WP, &
    & 4.907E-22_WP,    5.925E-22_WP,    7.097E-22_WP,    8.441E-22_WP,    9.969E-22_WP,    1.171E-21_WP,    1.369E-21_WP, &
    & 1.594E-21_WP,    1.842E-21_WP,    2.108E-21_WP,    2.385E-21_WP,    2.656E-21_WP,    2.905E-21_WP,    3.108E-21_WP, &
    & 3.241E-21_WP,    3.278E-21_WP,    3.200E-21_WP,    2.999E-21_WP,    2.685E-21_WP,    2.284E-21_WP,    1.821E-21_WP, &
    & 1.333E-21_WP,    8.836E-22_WP,    5.209E-22_WP,    2.675E-22_WP,    1.165E-22_WP,    4.118E-23_WP,    1.025E-23_WP /
data J(2,    121    ,:)    / 1.420E-22_WP,    1.796E-22_WP,    2.254E-22_WP,    2.804E-22_WP,    3.456E-22_WP,    4.218E-22_WP, &
    & 5.101E-22_WP,    6.123E-22_WP,    7.302E-22_WP,    8.652E-22_WP,    1.020E-21_WP,    1.199E-21_WP,    1.403E-21_WP, &
    & 1.634E-21_WP,    1.889E-21_WP,    2.161E-21_WP,    2.441E-21_WP,    2.715E-21_WP,    2.964E-21_WP,    3.166E-21_WP, &
    & 3.297E-21_WP,    3.331E-21_WP,    3.249E-21_WP,    3.044E-21_WP,    2.724E-21_WP,    2.317E-21_WP,    1.847E-21_WP, &
    & 1.352E-21_WP,    8.963E-22_WP,    5.284E-22_WP,    2.714E-22_WP,    1.182E-22_WP,    4.179E-23_WP,    1.041E-23_WP /
data J(2,    122    ,:)    / 1.537E-22_WP,    1.929E-22_WP,    2.401E-22_WP,    2.961E-22_WP,    3.618E-22_WP,    4.382E-22_WP, &
    & 5.270E-22_WP,    6.298E-22_WP,    7.482E-22_WP,    8.852E-22_WP,    1.044E-21_WP,    1.227E-21_WP,    1.438E-21_WP, &
    & 1.675E-21_WP,    1.935E-21_WP,    2.211E-21_WP,    2.493E-21_WP,    2.768E-21_WP,    3.018E-21_WP,    3.220E-21_WP, &
    & 3.350E-21_WP,    3.382E-21_WP,    3.298E-21_WP,    3.089E-21_WP,    2.764E-21_WP,    2.350E-21_WP,    1.874E-21_WP, &
    & 1.372E-21_WP,    9.094E-22_WP,    5.361E-22_WP,    2.754E-22_WP,    1.200E-22_WP,    4.242E-23_WP,    1.058E-23_WP /
data J(2,    123    ,:)    / 1.650E-22_WP,    2.054E-22_WP,    2.534E-22_WP,    3.099E-22_WP,    3.758E-22_WP,    4.525E-22_WP, &
    & 5.418E-22_WP,    6.452E-22_WP,    7.654E-22_WP,    9.055E-22_WP,    1.069E-21_WP,    1.258E-21_WP,    1.474E-21_WP, &
    & 1.715E-21_WP,    1.978E-21_WP,    2.257E-21_WP,    2.541E-21_WP,    2.818E-21_WP,    3.069E-21_WP,    3.272E-21_WP, &
    & 3.401E-21_WP,    3.433E-21_WP,    3.346E-21_WP,    3.134E-21_WP,    2.804E-21_WP,    2.385E-21_WP,    1.901E-21_WP, &
    & 1.392E-21_WP,    9.229E-22_WP,    5.441E-22_WP,    2.795E-22_WP,    1.218E-22_WP,    4.307E-23_WP,    1.075E-23_WP /
data J(2,    124    ,:)    / 1.755E-22_WP,    2.166E-22_WP,    2.651E-22_WP,    3.217E-22_WP,    3.879E-22_WP,    4.652E-22_WP, &
    & 5.550E-22_WP,    6.599E-22_WP,    7.828E-22_WP,    9.269E-22_WP,    1.095E-21_WP,    1.288E-21_WP,    1.508E-21_WP, &
    & 1.753E-21_WP,    2.019E-21_WP,    2.300E-21_WP,    2.587E-21_WP,    2.865E-21_WP,    3.118E-21_WP,    3.322E-21_WP, &
    & 3.452E-21_WP,    3.484E-21_WP,    3.396E-21_WP,    3.180E-21_WP,    2.846E-21_WP,    2.420E-21_WP,    1.929E-21_WP, &
    & 1.413E-21_WP,    9.368E-22_WP,    5.523E-22_WP,    2.837E-22_WP,    1.237E-22_WP,    4.373E-23_WP,    1.092E-23_WP /
data J(2,    125    ,:)    / 1.850E-22_WP,    2.265E-22_WP,    2.751E-22_WP,    3.320E-22_WP,    3.987E-22_WP,    4.764E-22_WP, &
    & 5.676E-22_WP,    6.748E-22_WP,    8.012E-22_WP,    9.492E-22_WP,    1.121E-21_WP,    1.318E-21_WP,    1.541E-21_WP, &
    & 1.789E-21_WP,    2.057E-21_WP,    2.341E-21_WP,    2.629E-21_WP,    2.910E-21_WP,    3.165E-21_WP,    3.372E-21_WP, &
    & 3.504E-21_WP,    3.536E-21_WP,    3.447E-21_WP,    3.228E-21_WP,    2.889E-21_WP,    2.457E-21_WP,    1.959E-21_WP, &
    & 1.435E-21_WP,    9.510E-22_WP,    5.607E-22_WP,    2.881E-22_WP,    1.256E-22_WP,    4.441E-23_WP,    1.110E-23_WP /
data J(2,    126    ,:)    / 1.933E-22_WP,    2.349E-22_WP,    2.838E-22_WP,    3.412E-22_WP,    4.083E-22_WP,    4.871E-22_WP, &
    & 5.803E-22_WP,    6.905E-22_WP,    8.202E-22_WP,    9.718E-22_WP,    1.147E-21_WP,    1.347E-21_WP,    1.572E-21_WP, &
    & 1.822E-21_WP,    2.093E-21_WP,    2.379E-21_WP,    2.671E-21_WP,    2.955E-21_WP,    3.213E-21_WP,    3.423E-21_WP, &
    & 3.556E-21_WP,    3.589E-21_WP,    3.499E-21_WP,    3.277E-21_WP,    2.933E-21_WP,    2.494E-21_WP,    1.989E-21_WP, &
    & 1.457E-21_WP,    9.656E-22_WP,    5.694E-22_WP,    2.925E-22_WP,    1.275E-22_WP,    4.511E-23_WP,    1.128E-23_WP /
data J(2,    127    ,:)    / 2.005E-22_WP,    2.424E-22_WP,    2.916E-22_WP,    3.493E-22_WP,    4.174E-22_WP,    4.980E-22_WP, &
    & 5.937E-22_WP,    7.068E-22_WP,    8.396E-22_WP,    9.939E-22_WP,    1.171E-21_WP,    1.374E-21_WP,    1.601E-21_WP, &
    & 1.854E-21_WP,    2.128E-21_WP,    2.417E-21_WP,    2.712E-21_WP,    3.000E-21_WP,    3.262E-21_WP,    3.474E-21_WP, &
    & 3.610E-21_WP,    3.644E-21_WP,    3.552E-21_WP,    3.327E-21_WP,    2.978E-21_WP,    2.533E-21_WP,    2.019E-21_WP, &
    & 1.479E-21_WP,    9.806E-22_WP,    5.782E-22_WP,    2.971E-22_WP,    1.295E-22_WP,    4.583E-23_WP,    1.146E-23_WP /
data J(2,    128    ,:)    / 2.068E-22_WP,    2.490E-22_WP,    2.985E-22_WP,    3.571E-22_WP,    4.267E-22_WP,    5.095E-22_WP, &
    & 6.077E-22_WP,    7.234E-22_WP,    8.586E-22_WP,    1.015E-21_WP,    1.195E-21_WP,    1.399E-21_WP,    1.629E-21_WP, &
    & 1.885E-21_WP,    2.161E-21_WP,    2.454E-21_WP,    2.753E-21_WP,    3.046E-21_WP,    3.312E-21_WP,    3.527E-21_WP, &
    & 3.666E-21_WP,    3.700E-21_WP,    3.607E-21_WP,    3.379E-21_WP,    3.024E-21_WP,    2.572E-21_WP,    2.051E-21_WP, &
    & 1.502E-21_WP,    9.959E-22_WP,    5.873E-22_WP,    3.017E-22_WP,    1.315E-22_WP,    4.656E-23_WP,    1.165E-23_WP /
data J(2,    129    ,:)    / 2.124E-22_WP,    2.549E-22_WP,    3.052E-22_WP,    3.650E-22_WP,    4.364E-22_WP,    5.213E-22_WP, &
    & 6.218E-22_WP,    7.396E-22_WP,    8.767E-22_WP,    1.035E-21_WP,    1.217E-21_WP,    1.423E-21_WP,    1.656E-21_WP, &
    & 1.914E-21_WP,    2.195E-21_WP,    2.492E-21_WP,    2.796E-21_WP,    3.092E-21_WP,    3.363E-21_WP,    3.582E-21_WP, &
    & 3.723E-21_WP,    3.757E-21_WP,    3.663E-21_WP,    3.432E-21_WP,    3.071E-21_WP,    2.613E-21_WP,    2.083E-21_WP, &
    & 1.526E-21_WP,    1.012E-21_WP,    5.966E-22_WP,    3.065E-22_WP,    1.336E-22_WP,    4.731E-23_WP,    1.185E-23_WP /
data J(2,    130    ,:)    / 2.174E-22_WP,    2.605E-22_WP,    3.119E-22_WP,    3.733E-22_WP,    4.465E-22_WP,    5.334E-22_WP, &
    & 6.357E-22_WP,    7.551E-22_WP,    8.939E-22_WP,    1.054E-21_WP,    1.238E-21_WP,    1.446E-21_WP,    1.682E-21_WP, &
    & 1.944E-21_WP,    2.228E-21_WP,    2.530E-21_WP,    2.839E-21_WP,    3.140E-21_WP,    3.415E-21_WP,    3.638E-21_WP, &
    & 3.781E-21_WP,    3.816E-21_WP,    3.721E-21_WP,    3.486E-21_WP,    3.120E-21_WP,    2.654E-21_WP,    2.116E-21_WP, &
    & 1.550E-21_WP,    1.028E-21_WP,    6.061E-22_WP,    3.114E-22_WP,    1.358E-22_WP,    4.807E-23_WP,    1.204E-23_WP /
data J(2,    131    ,:)    / 2.222E-22_WP,    2.662E-22_WP,    3.189E-22_WP,    3.819E-22_WP,    4.568E-22_WP,    5.453E-22_WP, &
    & 6.489E-22_WP,    7.699E-22_WP,    9.103E-22_WP,    1.072E-21_WP,    1.258E-21_WP,    1.469E-21_WP,    1.708E-21_WP, &
    & 1.974E-21_WP,    2.263E-21_WP,    2.569E-21_WP,    2.883E-21_WP,    3.189E-21_WP,    3.468E-21_WP,    3.695E-21_WP, &
    & 3.840E-21_WP,    3.876E-21_WP,    3.780E-21_WP,    3.541E-21_WP,    3.170E-21_WP,    2.696E-21_WP,    2.150E-21_WP, &
    & 1.575E-21_WP,    1.044E-21_WP,    6.158E-22_WP,    3.164E-22_WP,    1.380E-22_WP,    4.885E-23_WP,    1.224E-23_WP /
data J(2,    132    ,:)    / 2.271E-22_WP,    2.722E-22_WP,    3.263E-22_WP,    3.907E-22_WP,    4.669E-22_WP,    5.566E-22_WP, &
    & 6.616E-22_WP,    7.840E-22_WP,    9.259E-22_WP,    1.090E-21_WP,    1.278E-21_WP,    1.492E-21_WP,    1.735E-21_WP, &
    & 2.004E-21_WP,    2.298E-21_WP,    2.609E-21_WP,    2.928E-21_WP,    3.239E-21_WP,    3.523E-21_WP,    3.753E-21_WP, &
    & 3.901E-21_WP,    3.938E-21_WP,    3.840E-21_WP,    3.598E-21_WP,    3.220E-21_WP,    2.740E-21_WP,    2.185E-21_WP, &
    & 1.600E-21_WP,    1.061E-21_WP,    6.258E-22_WP,    3.216E-22_WP,    1.402E-22_WP,    4.965E-23_WP,    1.245E-23_WP /
data J(2,    133    ,:)    / 2.322E-22_WP,    2.785E-22_WP,    3.337E-22_WP,    3.993E-22_WP,    4.766E-22_WP,    5.675E-22_WP, &
    & 6.737E-22_WP,    7.975E-22_WP,    9.411E-22_WP,    1.107E-21_WP,    1.298E-21_WP,    1.515E-21_WP,    1.761E-21_WP, &
    & 2.036E-21_WP,    2.334E-21_WP,    2.650E-21_WP,    2.974E-21_WP,    3.290E-21_WP,    3.579E-21_WP,    3.813E-21_WP, &
    & 3.964E-21_WP,    4.002E-21_WP,    3.902E-21_WP,    3.656E-21_WP,    3.272E-21_WP,    2.784E-21_WP,    2.220E-21_WP, &
    & 1.626E-21_WP,    1.078E-21_WP,    6.360E-22_WP,    3.268E-22_WP,    1.425E-22_WP,    5.047E-23_WP,    1.266E-23_WP /
data J(2,    134    ,:)    / 2.375E-22_WP,    2.848E-22_WP,    3.411E-22_WP,    4.076E-22_WP,    4.859E-22_WP,    5.778E-22_WP, &
    & 6.852E-22_WP,    8.105E-22_WP,    9.561E-22_WP,    1.124E-21_WP,    1.318E-21_WP,    1.539E-21_WP,    1.789E-21_WP, &
    & 2.067E-21_WP,    2.370E-21_WP,    2.692E-21_WP,    3.021E-21_WP,    3.343E-21_WP,    3.636E-21_WP,    3.875E-21_WP, &
    & 4.028E-21_WP,    4.066E-21_WP,    3.966E-21_WP,    3.715E-21_WP,    3.326E-21_WP,    2.829E-21_WP,    2.256E-21_WP, &
    & 1.653E-21_WP,    1.096E-21_WP,    6.464E-22_WP,    3.322E-22_WP,    1.448E-22_WP,    5.130E-23_WP,    1.287E-23_WP /
data J(2,    135    ,:)    / 2.428E-22_WP,    2.910E-22_WP,    3.481E-22_WP,    4.155E-22_WP,    4.947E-22_WP,    5.877E-22_WP, &
    & 6.965E-22_WP,    8.235E-22_WP,    9.710E-22_WP,    1.142E-21_WP,    1.338E-21_WP,    1.563E-21_WP,    1.817E-21_WP, &
    & 2.100E-21_WP,    2.408E-21_WP,    2.735E-21_WP,    3.069E-21_WP,    3.397E-21_WP,    3.695E-21_WP,    3.937E-21_WP, &
    & 4.093E-21_WP,    4.132E-21_WP,    4.030E-21_WP,    3.776E-21_WP,    3.380E-21_WP,    2.876E-21_WP,    2.293E-21_WP, &
    & 1.680E-21_WP,    1.114E-21_WP,    6.570E-22_WP,    3.376E-22_WP,    1.472E-22_WP,    5.215E-23_WP,    1.309E-23_WP /
data J(2,    136    ,:)    / 2.481E-22_WP,    2.970E-22_WP,    3.548E-22_WP,    4.230E-22_WP,    5.032E-22_WP,    5.973E-22_WP, &
    & 7.076E-22_WP,    8.364E-22_WP,    9.862E-22_WP,    1.160E-21_WP,    1.359E-21_WP,    1.587E-21_WP,    1.846E-21_WP, &
    & 2.133E-21_WP,    2.447E-21_WP,    2.779E-21_WP,    3.119E-21_WP,    3.452E-21_WP,    3.755E-21_WP,    4.001E-21_WP, &
    & 4.160E-21_WP,    4.200E-21_WP,    4.096E-21_WP,    3.838E-21_WP,    3.435E-21_WP,    2.923E-21_WP,    2.331E-21_WP, &
    & 1.708E-21_WP,    1.132E-21_WP,    6.679E-22_WP,    3.432E-22_WP,    1.497E-22_WP,    5.302E-23_WP,    1.331E-23_WP /
data J(2,    137    ,:)    / 2.532E-22_WP,    3.028E-22_WP,    3.613E-22_WP,    4.303E-22_WP,    5.115E-22_WP,    6.069E-22_WP, &
    & 7.187E-22_WP,    8.494E-22_WP,    1.002E-21_WP,    1.178E-21_WP,    1.381E-21_WP,    1.613E-21_WP,    1.875E-21_WP, &
    & 2.168E-21_WP,    2.486E-21_WP,    2.824E-21_WP,    3.170E-21_WP,    3.508E-21_WP,    3.816E-21_WP,    4.067E-21_WP, &
    & 4.228E-21_WP,    4.269E-21_WP,    4.164E-21_WP,    3.901E-21_WP,    3.492E-21_WP,    2.971E-21_WP,    2.370E-21_WP, &
    & 1.736E-21_WP,    1.151E-21_WP,    6.789E-22_WP,    3.489E-22_WP,    1.522E-22_WP,    5.391E-23_WP,    1.354E-23_WP /
data J(2,    138    ,:)    / 2.581E-22_WP,    3.083E-22_WP,    3.675E-22_WP,    4.373E-22_WP,    5.196E-22_WP,    6.164E-22_WP, &
    & 7.299E-22_WP,    8.627E-22_WP,    1.017E-21_WP,    1.196E-21_WP,    1.403E-21_WP,    1.638E-21_WP,    1.905E-21_WP, &
    & 2.203E-21_WP,    2.526E-21_WP,    2.869E-21_WP,    3.221E-21_WP,    3.565E-21_WP,    3.879E-21_WP,    4.134E-21_WP, &
    & 4.298E-21_WP,    4.340E-21_WP,    4.233E-21_WP,    3.966E-21_WP,    3.550E-21_WP,    3.021E-21_WP,    2.409E-21_WP, &
    & 1.765E-21_WP,    1.170E-21_WP,    6.903E-22_WP,    3.547E-22_WP,    1.547E-22_WP,    5.481E-23_WP,    1.377E-23_WP /
data J(2,    139    ,:)    / 2.628E-22_WP,    3.135E-22_WP,    3.735E-22_WP,    4.443E-22_WP,    5.278E-22_WP,    6.261E-22_WP, &
    & 7.414E-22_WP,    8.763E-22_WP,    1.033E-21_WP,    1.215E-21_WP,    1.425E-21_WP,    1.665E-21_WP,    1.936E-21_WP, &
    & 2.238E-21_WP,    2.567E-21_WP,    2.916E-21_WP,    3.274E-21_WP,    3.624E-21_WP,    3.943E-21_WP,    4.202E-21_WP, &
    & 4.369E-21_WP,    4.412E-21_WP,    4.303E-21_WP,    4.032E-21_WP,    3.609E-21_WP,    3.071E-21_WP,    2.449E-21_WP, &
    & 1.794E-21_WP,    1.190E-21_WP,    7.018E-22_WP,    3.607E-22_WP,    1.573E-22_WP,    5.574E-23_WP,    1.400E-23_WP /
data J(2,    140    ,:)    / 2.673E-22_WP,    3.187E-22_WP,    3.795E-22_WP,    4.513E-22_WP,    5.361E-22_WP,    6.359E-22_WP, &
    & 7.531E-22_WP,    8.901E-22_WP,    1.050E-21_WP,    1.235E-21_WP,    1.448E-21_WP,    1.692E-21_WP,    1.967E-21_WP, &
    & 2.275E-21_WP,    2.609E-21_WP,    2.964E-21_WP,    3.328E-21_WP,    3.684E-21_WP,    4.008E-21_WP,    4.272E-21_WP, &
    & 4.442E-21_WP,    4.485E-21_WP,    4.375E-21_WP,    4.099E-21_WP,    3.670E-21_WP,    3.122E-21_WP,    2.490E-21_WP, &
    & 1.824E-21_WP,    1.210E-21_WP,    7.136E-22_WP,    3.667E-22_WP,    1.600E-22_WP,    5.668E-23_WP,    1.424E-23_WP /
data J(2,    141    ,:)    / 2.717E-22_WP,    3.238E-22_WP,    3.855E-22_WP,    4.584E-22_WP,    5.445E-22_WP,    6.459E-22_WP, &
    & 7.650E-22_WP,    9.043E-22_WP,    1.067E-21_WP,    1.255E-21_WP,    1.472E-21_WP,    1.719E-21_WP,    2.000E-21_WP, &
    & 2.312E-21_WP,    2.652E-21_WP,    3.013E-21_WP,    3.383E-21_WP,    3.745E-21_WP,    4.075E-21_WP,    4.343E-21_WP, &
    & 4.516E-21_WP,    4.560E-21_WP,    4.448E-21_WP,    4.168E-21_WP,    3.731E-21_WP,    3.175E-21_WP,    2.532E-21_WP, &
    & 1.855E-21_WP,    1.230E-21_WP,    7.256E-22_WP,    3.729E-22_WP,    1.627E-22_WP,    5.764E-23_WP,    1.448E-23_WP /
data J(2,    142    ,:)    / 2.761E-22_WP,    3.289E-22_WP,    3.915E-22_WP,    4.656E-22_WP,    5.531E-22_WP,    6.562E-22_WP, &
    & 7.772E-22_WP,    9.189E-22_WP,    1.084E-21_WP,    1.275E-21_WP,    1.496E-21_WP,    1.747E-21_WP,    2.033E-21_WP, &
    & 2.350E-21_WP,    2.696E-21_WP,    3.063E-21_WP,    3.440E-21_WP,    3.808E-21_WP,    4.143E-21_WP,    4.416E-21_WP, &
    & 4.592E-21_WP,    4.637E-21_WP,    4.523E-21_WP,    4.238E-21_WP,    3.794E-21_WP,    3.228E-21_WP,    2.575E-21_WP, &
    & 1.886E-21_WP,    1.251E-21_WP,    7.378E-22_WP,    3.792E-22_WP,    1.654E-22_WP,    5.862E-23_WP,    1.473E-23_WP /
data J(2,    143    ,:)    / 2.804E-22_WP,    3.341E-22_WP,    3.977E-22_WP,    4.730E-22_WP,    5.619E-22_WP,    6.667E-22_WP, &
    & 7.897E-22_WP,    9.337E-22_WP,    1.102E-21_WP,    1.296E-21_WP,    1.520E-21_WP,    1.776E-21_WP,    2.066E-21_WP, &
    & 2.389E-21_WP,    2.741E-21_WP,    3.114E-21_WP,    3.497E-21_WP,    3.871E-21_WP,    4.213E-21_WP,    4.490E-21_WP, &
    & 4.669E-21_WP,    4.715E-21_WP,    4.599E-21_WP,    4.309E-21_WP,    3.858E-21_WP,    3.283E-21_WP,    2.618E-21_WP, &
    & 1.918E-21_WP,    1.272E-21_WP,    7.503E-22_WP,    3.856E-22_WP,    1.682E-22_WP,    5.961E-23_WP,    1.499E-23_WP /
data J(2,    144    ,:)    / 2.849E-22_WP,    3.394E-22_WP,    4.040E-22_WP,    4.805E-22_WP,    5.709E-22_WP,    6.774E-22_WP, &
    & 8.025E-22_WP,    9.489E-22_WP,    1.120E-21_WP,    1.317E-21_WP,    1.545E-21_WP,    1.805E-21_WP,    2.100E-21_WP, &
    & 2.429E-21_WP,    2.787E-21_WP,    3.166E-21_WP,    3.556E-21_WP,    3.936E-21_WP,    4.283E-21_WP,    4.566E-21_WP, &
    & 4.748E-21_WP,    4.794E-21_WP,    4.677E-21_WP,    4.382E-21_WP,    3.923E-21_WP,    3.338E-21_WP,    2.662E-21_WP, &
    & 1.951E-21_WP,    1.294E-21_WP,    7.631E-22_WP,    3.922E-22_WP,    1.711E-22_WP,    6.063E-23_WP,    1.524E-23_WP /
data J(2,    145    ,:)    / 2.894E-22_WP,    3.448E-22_WP,    4.105E-22_WP,    4.882E-22_WP,    5.801E-22_WP,    6.884E-22_WP, &
    & 8.156E-22_WP,    9.645E-22_WP,    1.138E-21_WP,    1.339E-21_WP,    1.571E-21_WP,    1.835E-21_WP,    2.135E-21_WP, &
    & 2.470E-21_WP,    2.833E-21_WP,    3.220E-21_WP,    3.615E-21_WP,    4.003E-21_WP,    4.356E-21_WP,    4.643E-21_WP, &
    & 4.828E-21_WP,    4.875E-21_WP,    4.756E-21_WP,    4.456E-21_WP,    3.990E-21_WP,    3.395E-21_WP,    2.708E-21_WP, &
    & 1.984E-21_WP,    1.316E-21_WP,    7.761E-22_WP,    3.989E-22_WP,    1.740E-22_WP,    6.166E-23_WP,    1.551E-23_WP /
data J(2,    146    ,:)    / 2.940E-22_WP,    3.503E-22_WP,    4.171E-22_WP,    4.961E-22_WP,    5.896E-22_WP,    6.996E-22_WP, &
    & 8.290E-22_WP,    9.803E-22_WP,    1.157E-21_WP,    1.361E-21_WP,    1.597E-21_WP,    1.866E-21_WP,    2.171E-21_WP, &
    & 2.511E-21_WP,    2.881E-21_WP,    3.274E-21_WP,    3.676E-21_WP,    4.070E-21_WP,    4.429E-21_WP,    4.721E-21_WP, &
    & 4.910E-21_WP,    4.958E-21_WP,    4.837E-21_WP,    4.532E-21_WP,    4.058E-21_WP,    3.453E-21_WP,    2.754E-21_WP, &
    & 2.018E-21_WP,    1.338E-21_WP,    7.893E-22_WP,    4.057E-22_WP,    1.770E-22_WP,    6.272E-23_WP,    1.577E-23_WP /
data J(2,    147    ,:)    / 2.987E-22_WP,    3.559E-22_WP,    4.238E-22_WP,    5.042E-22_WP,    5.992E-22_WP,    7.111E-22_WP, &
    & 8.426E-22_WP,    9.965E-22_WP,    1.176E-21_WP,    1.384E-21_WP,    1.623E-21_WP,    1.897E-21_WP,    2.207E-21_WP, &
    & 2.553E-21_WP,    2.930E-21_WP,    3.329E-21_WP,    3.739E-21_WP,    4.139E-21_WP,    4.504E-21_WP,    4.802E-21_WP, &
    & 4.993E-21_WP,    5.042E-21_WP,    4.919E-21_WP,    4.609E-21_WP,    4.127E-21_WP,    3.512E-21_WP,    2.801E-21_WP, &
    & 2.052E-21_WP,    1.361E-21_WP,    8.028E-22_WP,    4.126E-22_WP,    1.800E-22_WP,    6.379E-23_WP,    1.604E-23_WP /
data J(2,    148    ,:)    / 3.035E-22_WP,    3.617E-22_WP,    4.307E-22_WP,    5.124E-22_WP,    6.090E-22_WP,    7.228E-22_WP, &
    & 8.566E-22_WP,    1.013E-21_WP,    1.196E-21_WP,    1.407E-21_WP,    1.651E-21_WP,    1.929E-21_WP,    2.245E-21_WP, &
    & 2.596E-21_WP,    2.979E-21_WP,    3.385E-21_WP,    3.802E-21_WP,    4.209E-21_WP,    4.581E-21_WP,    4.883E-21_WP, &
    & 5.078E-21_WP,    5.128E-21_WP,    5.003E-21_WP,    4.688E-21_WP,    4.197E-21_WP,    3.571E-21_WP,    2.849E-21_WP, &
    & 2.087E-21_WP,    1.384E-21_WP,    8.165E-22_WP,    4.197E-22_WP,    1.831E-22_WP,    6.489E-23_WP,    1.632E-23_WP /
data J(2,    149    ,:)    / 3.084E-22_WP,    3.675E-22_WP,    4.377E-22_WP,    5.208E-22_WP,    6.191E-22_WP,    7.348E-22_WP, &
    & 8.708E-22_WP,    1.030E-21_WP,    1.215E-21_WP,    1.431E-21_WP,    1.678E-21_WP,    1.962E-21_WP,    2.283E-21_WP, &
    & 2.640E-21_WP,    3.030E-21_WP,    3.443E-21_WP,    3.867E-21_WP,    4.281E-21_WP,    4.659E-21_WP,    4.967E-21_WP, &
    & 5.165E-21_WP,    5.216E-21_WP,    5.088E-21_WP,    4.768E-21_WP,    4.269E-21_WP,    3.633E-21_WP,    2.897E-21_WP, &
    & 2.123E-21_WP,    1.408E-21_WP,    8.305E-22_WP,    4.269E-22_WP,    1.862E-22_WP,    6.600E-23_WP,    1.660E-23_WP /
data J(2,    150    ,:)    / 3.134E-22_WP,    3.736E-22_WP,    4.449E-22_WP,    5.294E-22_WP,    6.293E-22_WP,    7.470E-22_WP, &
    & 8.853E-22_WP,    1.047E-21_WP,    1.236E-21_WP,    1.455E-21_WP,    1.707E-21_WP,    1.995E-21_WP,    2.321E-21_WP, &
    & 2.685E-21_WP,    3.081E-21_WP,    3.501E-21_WP,    3.932E-21_WP,    4.354E-21_WP,    4.739E-21_WP,    5.051E-21_WP, &
    & 5.253E-21_WP,    5.305E-21_WP,    5.175E-21_WP,    4.850E-21_WP,    4.342E-21_WP,    3.695E-21_WP,    2.947E-21_WP, &
    & 2.160E-21_WP,    1.432E-21_WP,    8.447E-22_WP,    4.342E-22_WP,    1.894E-22_WP,    6.714E-23_WP,    1.689E-23_WP /
data J(2,    151    ,:)    / 3.185E-22_WP,    3.797E-22_WP,    4.523E-22_WP,    5.382E-22_WP,    6.398E-22_WP,    7.595E-22_WP, &
    & 9.001E-22_WP,    1.065E-21_WP,    1.257E-21_WP,    1.479E-21_WP,    1.735E-21_WP,    2.029E-21_WP,    2.361E-21_WP, &
    & 2.731E-21_WP,    3.134E-21_WP,    3.561E-21_WP,    4.000E-21_WP,    4.428E-21_WP,    4.820E-21_WP,    5.138E-21_WP, &
    & 5.343E-21_WP,    5.396E-21_WP,    5.264E-21_WP,    4.933E-21_WP,    4.417E-21_WP,    3.758E-21_WP,    2.998E-21_WP, &
    & 2.197E-21_WP,    1.457E-21_WP,    8.593E-22_WP,    4.416E-22_WP,    1.927E-22_WP,    6.829E-23_WP,    1.718E-23_WP /
data J(2,    152    ,:)    / 3.238E-22_WP,    3.860E-22_WP,    4.598E-22_WP,    5.472E-22_WP,    6.505E-22_WP,    7.722E-22_WP, &
    & 9.152E-22_WP,    1.083E-21_WP,    1.278E-21_WP,    1.504E-21_WP,    1.765E-21_WP,    2.063E-21_WP,    2.401E-21_WP, &
    & 2.777E-21_WP,    3.187E-21_WP,    3.622E-21_WP,    4.068E-21_WP,    4.504E-21_WP,    4.902E-21_WP,    5.226E-21_WP, &
    & 5.435E-21_WP,    5.489E-21_WP,    5.354E-21_WP,    5.018E-21_WP,    4.493E-21_WP,    3.823E-21_WP,    3.049E-21_WP, &
    & 2.234E-21_WP,    1.482E-21_WP,    8.740E-22_WP,    4.492E-22_WP,    1.960E-22_WP,    6.947E-23_WP,    1.748E-23_WP /
data J(2,    153    ,:)    / 3.291E-22_WP,    3.924E-22_WP,    4.674E-22_WP,    5.563E-22_WP,    6.614E-22_WP,    7.852E-22_WP, &
    & 9.307E-22_WP,    1.101E-21_WP,    1.299E-21_WP,    1.530E-21_WP,    1.795E-21_WP,    2.098E-21_WP,    2.442E-21_WP, &
    & 2.824E-21_WP,    3.241E-21_WP,    3.684E-21_WP,    4.138E-21_WP,    4.581E-21_WP,    4.986E-21_WP,    5.316E-21_WP, &
    & 5.528E-21_WP,    5.583E-21_WP,    5.446E-21_WP,    5.104E-21_WP,    4.570E-21_WP,    3.889E-21_WP,    3.102E-21_WP, &
    & 2.273E-21_WP,    1.507E-21_WP,    8.891E-22_WP,    4.570E-22_WP,    1.994E-22_WP,    7.067E-23_WP,    1.778E-23_WP /
data J(2,    154    ,:)    / 3.346E-22_WP,    3.989E-22_WP,    4.753E-22_WP,    5.657E-22_WP,    6.725E-22_WP,    7.984E-22_WP, &
    & 9.464E-22_WP,    1.120E-21_WP,    1.322E-21_WP,    1.556E-21_WP,    1.825E-21_WP,    2.134E-21_WP,    2.483E-21_WP, &
    & 2.873E-21_WP,    3.297E-21_WP,    3.747E-21_WP,    4.209E-21_WP,    4.660E-21_WP,    5.072E-21_WP,    5.407E-21_WP, &
    & 5.623E-21_WP,    5.679E-21_WP,    5.540E-21_WP,    5.192E-21_WP,    4.648E-21_WP,    3.956E-21_WP,    3.155E-21_WP, &
    & 2.312E-21_WP,    1.533E-21_WP,    9.044E-22_WP,    4.649E-22_WP,    2.028E-22_WP,    7.189E-23_WP,    1.809E-23_WP /
data J(2,    155    ,:)    / 3.402E-22_WP,    4.056E-22_WP,    4.832E-22_WP,    5.752E-22_WP,    6.838E-22_WP,    8.119E-22_WP, &
    & 9.624E-22_WP,    1.139E-21_WP,    1.344E-21_WP,    1.582E-21_WP,    1.857E-21_WP,    2.170E-21_WP,    2.526E-21_WP, &
    & 2.922E-21_WP,    3.353E-21_WP,    3.811E-21_WP,    4.281E-21_WP,    4.740E-21_WP,    5.159E-21_WP,    5.500E-21_WP, &
    & 5.720E-21_WP,    5.777E-21_WP,    5.636E-21_WP,    5.281E-21_WP,    4.729E-21_WP,    4.024E-21_WP,    3.210E-21_WP, &
    & 2.352E-21_WP,    1.560E-21_WP,    9.200E-22_WP,    4.729E-22_WP,    2.063E-22_WP,    7.313E-23_WP,    1.840E-23_WP /
data J(2,    156    ,:)    / 3.459E-22_WP,    4.124E-22_WP,    4.914E-22_WP,    5.849E-22_WP,    6.954E-22_WP,    8.257E-22_WP, &
    & 9.787E-22_WP,    1.158E-21_WP,    1.367E-21_WP,    1.609E-21_WP,    1.888E-21_WP,    2.207E-21_WP,    2.569E-21_WP, &
    & 2.972E-21_WP,    3.411E-21_WP,    3.877E-21_WP,    4.354E-21_WP,    4.822E-21_WP,    5.248E-21_WP,    5.595E-21_WP, &
    & 5.819E-21_WP,    5.876E-21_WP,    5.733E-21_WP,    5.372E-21_WP,    4.810E-21_WP,    4.093E-21_WP,    3.265E-21_WP, &
    & 2.393E-21_WP,    1.587E-21_WP,    9.359E-22_WP,    4.811E-22_WP,    2.099E-22_WP,    7.440E-23_WP,    1.872E-23_WP /
data J(2,    157    ,:)    / 3.517E-22_WP,    4.193E-22_WP,    4.996E-22_WP,    5.948E-22_WP,    7.072E-22_WP,    8.397E-22_WP, &
    & 9.954E-22_WP,    1.178E-21_WP,    1.390E-21_WP,    1.637E-21_WP,    1.921E-21_WP,    2.245E-21_WP,    2.613E-21_WP, &
    & 3.023E-21_WP,    3.470E-21_WP,    3.943E-21_WP,    4.429E-21_WP,    4.905E-21_WP,    5.338E-21_WP,    5.691E-21_WP, &
    & 5.919E-21_WP,    5.978E-21_WP,    5.832E-21_WP,    5.465E-21_WP,    4.893E-21_WP,    4.164E-21_WP,    3.321E-21_WP, &
    & 2.434E-21_WP,    1.614E-21_WP,    9.521E-22_WP,    4.894E-22_WP,    2.135E-22_WP,    7.569E-23_WP,    1.905E-23_WP /
data J(2,    158    ,:)    / 3.576E-22_WP,    4.264E-22_WP,    5.081E-22_WP,    6.048E-22_WP,    7.192E-22_WP,    8.540E-22_WP, &
    & 1.012E-21_WP,    1.198E-21_WP,    1.414E-21_WP,    1.665E-21_WP,    1.953E-21_WP,    2.284E-21_WP,    2.658E-21_WP, &
    & 3.075E-21_WP,    3.529E-21_WP,    4.011E-21_WP,    4.506E-21_WP,    4.989E-21_WP,    5.430E-21_WP,    5.790E-21_WP, &
    & 6.021E-21_WP,    6.081E-21_WP,    5.933E-21_WP,    5.560E-21_WP,    4.978E-21_WP,    4.236E-21_WP,    3.379E-21_WP, &
    & 2.476E-21_WP,    1.642E-21_WP,    9.686E-22_WP,    4.979E-22_WP,    2.172E-22_WP,    7.700E-23_WP,    1.938E-23_WP /
data J(2,    159    ,:)    / 3.636E-22_WP,    4.336E-22_WP,    5.167E-22_WP,    6.151E-22_WP,    7.314E-22_WP,    8.685E-22_WP, &
    & 1.030E-21_WP,    1.218E-21_WP,    1.438E-21_WP,    1.693E-21_WP,    1.987E-21_WP,    2.323E-21_WP,    2.704E-21_WP, &
    & 3.128E-21_WP,    3.590E-21_WP,    4.081E-21_WP,    4.583E-21_WP,    5.075E-21_WP,    5.524E-21_WP,    5.890E-21_WP, &
    & 6.125E-21_WP,    6.186E-21_WP,    6.035E-21_WP,    5.656E-21_WP,    5.064E-21_WP,    4.309E-21_WP,    3.437E-21_WP, &
    & 2.519E-21_WP,    1.670E-21_WP,    9.854E-22_WP,    5.065E-22_WP,    2.210E-22_WP,    7.833E-23_WP,    1.971E-23_WP /
data J(2,    160    ,:)    / 3.698E-22_WP,    4.410E-22_WP,    5.255E-22_WP,    6.256E-22_WP,    7.439E-22_WP,    8.834E-22_WP, &
    & 1.047E-21_WP,    1.239E-21_WP,    1.463E-21_WP,    1.722E-21_WP,    2.021E-21_WP,    2.363E-21_WP,    2.750E-21_WP, &
    & 3.182E-21_WP,    3.652E-21_WP,    4.151E-21_WP,    4.663E-21_WP,    5.163E-21_WP,    5.620E-21_WP,    5.991E-21_WP, &
    & 6.231E-21_WP,    6.293E-21_WP,    6.140E-21_WP,    5.754E-21_WP,    5.152E-21_WP,    4.384E-21_WP,    3.497E-21_WP, &
    & 2.562E-21_WP,    1.699E-21_WP,    1.002E-21_WP,    5.152E-22_WP,    2.248E-22_WP,    7.969E-23_WP,    2.006E-23_WP /
data J(2,    161    ,:)    / 3.761E-22_WP,    4.485E-22_WP,    5.345E-22_WP,    6.363E-22_WP,    7.566E-22_WP,    8.985E-22_WP, &
    & 1.065E-21_WP,    1.260E-21_WP,    1.488E-21_WP,    1.752E-21_WP,    2.056E-21_WP,    2.404E-21_WP,    2.798E-21_WP, &
    & 3.237E-21_WP,    3.715E-21_WP,    4.223E-21_WP,    4.743E-21_WP,    5.252E-21_WP,    5.717E-21_WP,    6.095E-21_WP, &
    & 6.339E-21_WP,    6.402E-21_WP,    6.246E-21_WP,    5.853E-21_WP,    5.241E-21_WP,    4.460E-21_WP,    3.557E-21_WP, &
    & 2.607E-21_WP,    1.729E-21_WP,    1.020E-21_WP,    5.242E-22_WP,    2.287E-22_WP,    8.107E-23_WP,    2.040E-23_WP /
data J(2,    162    ,:)    / 3.825E-22_WP,    4.562E-22_WP,    5.436E-22_WP,    6.472E-22_WP,    7.696E-22_WP,    9.139E-22_WP, &
    & 1.084E-21_WP,    1.282E-21_WP,    1.514E-21_WP,    1.782E-21_WP,    2.091E-21_WP,    2.445E-21_WP,    2.846E-21_WP, &
    & 3.293E-21_WP,    3.779E-21_WP,    4.296E-21_WP,    4.825E-21_WP,    5.343E-21_WP,    5.816E-21_WP,    6.200E-21_WP, &
    & 6.449E-21_WP,    6.513E-21_WP,    6.354E-21_WP,    5.955E-21_WP,    5.332E-21_WP,    4.537E-21_WP,    3.619E-21_WP, &
    & 2.652E-21_WP,    1.759E-21_WP,    1.037E-21_WP,    5.333E-22_WP,    2.326E-22_WP,    8.248E-23_WP,    2.076E-23_WP /
data J(2,    163    ,:)    / 3.890E-22_WP,    4.639E-22_WP,    5.529E-22_WP,    6.582E-22_WP,    7.828E-22_WP,    9.296E-22_WP, &
    & 1.102E-21_WP,    1.304E-21_WP,    1.540E-21_WP,    1.813E-21_WP,    2.127E-21_WP,    2.487E-21_WP,    2.895E-21_WP, &
    & 3.350E-21_WP,    3.845E-21_WP,    4.370E-21_WP,    4.909E-21_WP,    5.436E-21_WP,    5.917E-21_WP,    6.308E-21_WP, &
    & 6.561E-21_WP,    6.626E-21_WP,    6.464E-21_WP,    6.058E-21_WP,    5.424E-21_WP,    4.616E-21_WP,    3.682E-21_WP, &
    & 2.698E-21_WP,    1.789E-21_WP,    1.055E-21_WP,    5.425E-22_WP,    2.367E-22_WP,    8.391E-23_WP,    2.112E-23_WP /
data J(2,    164    ,:)    / 3.956E-22_WP,    4.719E-22_WP,    5.624E-22_WP,    6.695E-22_WP,    7.962E-22_WP,    9.456E-22_WP, &
    & 1.121E-21_WP,    1.327E-21_WP,    1.566E-21_WP,    1.844E-21_WP,    2.164E-21_WP,    2.530E-21_WP,    2.945E-21_WP, &
    & 3.407E-21_WP,    3.911E-21_WP,    4.446E-21_WP,    4.994E-21_WP,    5.530E-21_WP,    6.019E-21_WP,    6.417E-21_WP, &
    & 6.674E-21_WP,    6.741E-21_WP,    6.576E-21_WP,    6.163E-21_WP,    5.518E-21_WP,    4.696E-21_WP,    3.746E-21_WP, &
    & 2.745E-21_WP,    1.820E-21_WP,    1.074E-21_WP,    5.519E-22_WP,    2.408E-22_WP,    8.537E-23_WP,    2.149E-23_WP /
data J(2,    165    ,:)    / 4.024E-22_WP,    4.800E-22_WP,    5.720E-22_WP,    6.810E-22_WP,    8.099E-22_WP,    9.618E-22_WP, &
    & 1.140E-21_WP,    1.349E-21_WP,    1.593E-21_WP,    1.876E-21_WP,    2.202E-21_WP,    2.574E-21_WP,    2.996E-21_WP, &
    & 3.466E-21_WP,    3.979E-21_WP,    4.523E-21_WP,    5.080E-21_WP,    5.626E-21_WP,    6.123E-21_WP,    6.528E-21_WP, &
    & 6.790E-21_WP,    6.858E-21_WP,    6.690E-21_WP,    6.270E-21_WP,    5.614E-21_WP,    4.777E-21_WP,    3.811E-21_WP, &
    & 2.792E-21_WP,    1.852E-21_WP,    1.092E-21_WP,    5.615E-22_WP,    2.450E-22_WP,    8.685E-23_WP,    2.186E-23_WP /
data J(2,    166    ,:)    / 4.093E-22_WP,    4.882E-22_WP,    5.818E-22_WP,    6.927E-22_WP,    8.238E-22_WP,    9.784E-22_WP, &
    & 1.160E-21_WP,    1.373E-21_WP,    1.621E-21_WP,    1.908E-21_WP,    2.240E-21_WP,    2.619E-21_WP,    3.048E-21_WP, &
    & 3.526E-21_WP,    4.048E-21_WP,    4.601E-21_WP,    5.168E-21_WP,    5.723E-21_WP,    6.230E-21_WP,    6.642E-21_WP, &
    & 6.908E-21_WP,    6.977E-21_WP,    6.806E-21_WP,    6.379E-21_WP,    5.711E-21_WP,    4.860E-21_WP,    3.877E-21_WP, &
    & 2.841E-21_WP,    1.884E-21_WP,    1.111E-21_WP,    5.713E-22_WP,    2.492E-22_WP,    8.836E-23_WP,    2.224E-23_WP /
data J(2,    167    ,:)    / 4.163E-22_WP,    4.966E-22_WP,    5.918E-22_WP,    7.047E-22_WP,    8.380E-22_WP,    9.953E-22_WP, &
    & 1.180E-21_WP,    1.396E-21_WP,    1.649E-21_WP,    1.941E-21_WP,    2.278E-21_WP,    2.664E-21_WP,    3.101E-21_WP, &
    & 3.588E-21_WP,    4.118E-21_WP,    4.681E-21_WP,    5.258E-21_WP,    5.822E-21_WP,    6.338E-21_WP,    6.757E-21_WP, &
    & 7.028E-21_WP,    7.098E-21_WP,    6.925E-21_WP,    6.490E-21_WP,    5.811E-21_WP,    4.945E-21_WP,    3.944E-21_WP, &
    & 2.890E-21_WP,    1.917E-21_WP,    1.131E-21_WP,    5.812E-22_WP,    2.536E-22_WP,    8.989E-23_WP,    2.263E-23_WP /

! model = 3                                                                                                                                                
data J(3,    1    ,:)    / 1.218E-27_WP,    1.535E-27_WP,    1.937E-27_WP,    2.446E-27_WP,    3.091E-27_WP,    3.907E-27_WP, &
    & 4.936E-27_WP,    6.230E-27_WP,    7.853E-27_WP,    9.878E-27_WP,    1.239E-26_WP,    1.547E-26_WP,    1.923E-26_WP, &
    & 2.371E-26_WP,    2.890E-26_WP,    3.470E-26_WP,    4.086E-26_WP,    4.696E-26_WP,    5.241E-26_WP,    5.660E-26_WP, &
    & 5.913E-26_WP,    5.993E-26_WP,    5.927E-26_WP,    5.760E-26_WP,    5.540E-26_WP,    5.299E-26_WP,    5.022E-26_WP, &
    & 4.677E-26_WP,    4.264E-26_WP,    3.783E-26_WP,    3.228E-26_WP,    2.590E-26_WP,    1.854E-26_WP,    1.001E-26_WP /
data J(3,    2    ,:)    / 1.351E-27_WP,    1.702E-27_WP,    2.148E-27_WP,    2.713E-27_WP,    3.428E-27_WP,    4.333E-27_WP, &
    & 5.474E-27_WP,    6.910E-27_WP,    8.710E-27_WP,    1.095E-26_WP,    1.374E-26_WP,    1.716E-26_WP,    2.133E-26_WP, &
    & 2.630E-26_WP,    3.205E-26_WP,    3.848E-26_WP,    4.531E-26_WP,    5.207E-26_WP,    5.812E-26_WP,    6.277E-26_WP, &
    & 6.557E-26_WP,    6.645E-26_WP,    6.572E-26_WP,    6.387E-26_WP,    6.142E-26_WP,    5.876E-26_WP,    5.568E-26_WP, &
    & 5.185E-26_WP,    4.728E-26_WP,    4.195E-26_WP,    3.579E-26_WP,    2.872E-26_WP,    2.056E-26_WP,    1.110E-26_WP /
data J(3,    3    ,:)    / 1.498E-27_WP,    1.888E-27_WP,    2.382E-27_WP,    3.009E-27_WP,    3.802E-27_WP,    4.805E-27_WP, &
    & 6.071E-27_WP,    7.663E-27_WP,    9.659E-27_WP,    1.215E-26_WP,    1.524E-26_WP,    1.903E-26_WP,    2.365E-26_WP, &
    & 2.916E-26_WP,    3.554E-26_WP,    4.268E-26_WP,    5.025E-26_WP,    5.775E-26_WP,    6.444E-26_WP,    6.960E-26_WP, &
    & 7.270E-26_WP,    7.368E-26_WP,    7.287E-26_WP,    7.082E-26_WP,    6.810E-26_WP,    6.515E-26_WP,    6.174E-26_WP, &
    & 5.749E-26_WP,    5.242E-26_WP,    4.651E-26_WP,    3.969E-26_WP,    3.184E-26_WP,    2.280E-26_WP,    1.231E-26_WP /
data J(3,    4    ,:)    / 1.661E-27_WP,    2.094E-27_WP,    2.642E-27_WP,    3.337E-27_WP,    4.217E-27_WP,    5.330E-27_WP, &
    & 6.733E-27_WP,    8.499E-27_WP,    1.071E-26_WP,    1.347E-26_WP,    1.690E-26_WP,    2.111E-26_WP,    2.623E-26_WP, &
    & 3.234E-26_WP,    3.942E-26_WP,    4.732E-26_WP,    5.572E-26_WP,    6.403E-26_WP,    7.146E-26_WP,    7.717E-26_WP, &
    & 8.061E-26_WP,    8.169E-26_WP,    8.079E-26_WP,    7.851E-26_WP,    7.550E-26_WP,    7.222E-26_WP,    6.844E-26_WP, &
    & 6.374E-26_WP,    5.812E-26_WP,    5.156E-26_WP,    4.400E-26_WP,    3.530E-26_WP,    2.528E-26_WP,    1.365E-26_WP /
data J(3,    5    ,:)    / 1.843E-27_WP,    2.322E-27_WP,    2.930E-27_WP,    3.701E-27_WP,    4.677E-27_WP,    5.911E-27_WP, &
    & 7.467E-27_WP,    9.426E-27_WP,    1.188E-26_WP,    1.494E-26_WP,    1.874E-26_WP,    2.341E-26_WP,    2.909E-26_WP, &
    & 3.586E-26_WP,    4.371E-26_WP,    5.247E-26_WP,    6.179E-26_WP,    7.100E-26_WP,    7.923E-26_WP,    8.556E-26_WP, &
    & 8.938E-26_WP,    9.057E-26_WP,    8.956E-26_WP,    8.704E-26_WP,    8.370E-26_WP,    8.006E-26_WP,    7.587E-26_WP, &
    & 7.065E-26_WP,    6.443E-26_WP,    5.716E-26_WP,    4.878E-26_WP,    3.914E-26_WP,    2.803E-26_WP,    1.514E-26_WP /
data J(3,    6    ,:)    / 2.044E-27_WP,    2.576E-27_WP,    3.250E-27_WP,    4.105E-27_WP,    5.187E-27_WP,    6.555E-27_WP, &
    & 8.282E-27_WP,    1.045E-26_WP,    1.318E-26_WP,    1.657E-26_WP,    2.078E-26_WP,    2.596E-26_WP,    3.225E-26_WP, &
    & 3.977E-26_WP,    4.847E-26_WP,    5.818E-26_WP,    6.851E-26_WP,    7.872E-26_WP,    8.784E-26_WP,    9.486E-26_WP, &
    & 9.908E-26_WP,    1.004E-25_WP,    9.928E-26_WP,    9.648E-26_WP,    9.278E-26_WP,    8.874E-26_WP,    8.410E-26_WP, &
    & 7.831E-26_WP,    7.141E-26_WP,    6.336E-26_WP,    5.407E-26_WP,    4.339E-26_WP,    3.108E-26_WP,    1.679E-26_WP /
data J(3,    7    ,:)    / 2.267E-27_WP,    2.857E-27_WP,    3.605E-27_WP,    4.553E-27_WP,    5.753E-27_WP,    7.270E-27_WP, &
    & 9.185E-27_WP,    1.159E-26_WP,    1.461E-26_WP,    1.838E-26_WP,    2.304E-26_WP,    2.878E-26_WP,    3.577E-26_WP, &
    & 4.409E-26_WP,    5.374E-26_WP,    6.451E-26_WP,    7.595E-26_WP,    8.727E-26_WP,    9.738E-26_WP,    1.052E-25_WP, &
    & 1.098E-25_WP,    1.113E-25_WP,    1.100E-25_WP,    1.069E-25_WP,    1.028E-25_WP,    9.835E-26_WP,    9.320E-26_WP, &
    & 8.679E-26_WP,    7.914E-26_WP,    7.023E-26_WP,    5.994E-26_WP,    4.810E-26_WP,    3.446E-26_WP,    1.862E-26_WP /
data J(3,    8    ,:)    / 2.514E-27_WP,    3.168E-27_WP,    3.998E-27_WP,    5.049E-27_WP,    6.380E-27_WP,    8.063E-27_WP, &
    & 1.019E-26_WP,    1.286E-26_WP,    1.620E-26_WP,    2.038E-26_WP,    2.555E-26_WP,    3.192E-26_WP,    3.966E-26_WP, &
    & 4.889E-26_WP,    5.958E-26_WP,    7.152E-26_WP,    8.420E-26_WP,    9.675E-26_WP,    1.079E-25_WP,    1.166E-25_WP, &
    & 1.217E-25_WP,    1.233E-25_WP,    1.220E-25_WP,    1.185E-25_WP,    1.139E-25_WP,    1.090E-25_WP,    1.033E-25_WP, &
    & 9.617E-26_WP,    8.770E-26_WP,    7.782E-26_WP,    6.642E-26_WP,    5.332E-26_WP,    3.820E-26_WP,    2.064E-26_WP /
data J(3,    9    ,:)    / 2.788E-27_WP,    3.514E-27_WP,    4.434E-27_WP,    5.599E-27_WP,    7.076E-27_WP,    8.942E-27_WP, &
    & 1.130E-26_WP,    1.426E-26_WP,    1.797E-26_WP,    2.260E-26_WP,    2.833E-26_WP,    3.539E-26_WP,    4.397E-26_WP, &
    & 5.420E-26_WP,    6.605E-26_WP,    7.929E-26_WP,    9.334E-26_WP,    1.072E-25_WP,    1.196E-25_WP,    1.292E-25_WP, &
    & 1.349E-25_WP,    1.367E-25_WP,    1.351E-25_WP,    1.313E-25_WP,    1.262E-25_WP,    1.207E-25_WP,    1.144E-25_WP, &
    & 1.065E-25_WP,    9.716E-26_WP,    8.622E-26_WP,    7.360E-26_WP,    5.909E-26_WP,    4.235E-26_WP,    2.289E-26_WP /
data J(3,    10    ,:)    / 3.092E-27_WP,    3.897E-27_WP,    4.917E-27_WP,    6.210E-27_WP,    7.847E-27_WP,    9.916E-27_WP, &
    & 1.253E-26_WP,    1.581E-26_WP,    1.992E-26_WP,    2.506E-26_WP,    3.142E-26_WP,    3.924E-26_WP,    4.875E-26_WP, &
    & 6.009E-26_WP,    7.322E-26_WP,    8.789E-26_WP,    1.035E-25_WP,    1.189E-25_WP,    1.326E-25_WP,    1.432E-25_WP, &
    & 1.495E-25_WP,    1.514E-25_WP,    1.497E-25_WP,    1.455E-25_WP,    1.398E-25_WP,    1.337E-25_WP,    1.267E-25_WP, &
    & 1.180E-25_WP,    1.076E-25_WP,    9.551E-26_WP,    8.154E-26_WP,    6.547E-26_WP,    4.694E-26_WP,    2.537E-26_WP /
data J(3,    11    ,:)    / 3.429E-27_WP,    4.322E-27_WP,    5.453E-27_WP,    6.887E-27_WP,    8.702E-27_WP,    1.100E-26_WP, &
    & 1.389E-26_WP,    1.753E-26_WP,    2.209E-26_WP,    2.778E-26_WP,    3.483E-26_WP,    4.350E-26_WP,    5.404E-26_WP, &
    & 6.661E-26_WP,    8.116E-26_WP,    9.741E-26_WP,    1.147E-25_WP,    1.317E-25_WP,    1.469E-25_WP,    1.586E-25_WP, &
    & 1.656E-25_WP,    1.678E-25_WP,    1.658E-25_WP,    1.611E-25_WP,    1.549E-25_WP,    1.481E-25_WP,    1.403E-25_WP, &
    & 1.307E-25_WP,    1.192E-25_WP,    1.058E-25_WP,    9.032E-26_WP,    7.254E-26_WP,    5.201E-26_WP,    2.813E-26_WP /
data J(3,    12    ,:)    / 3.803E-27_WP,    4.793E-27_WP,    6.047E-27_WP,    7.637E-27_WP,    9.650E-27_WP,    1.219E-26_WP, &
    & 1.540E-26_WP,    1.944E-26_WP,    2.450E-26_WP,    3.080E-26_WP,    3.862E-26_WP,    4.822E-26_WP,    5.990E-26_WP, &
    & 7.383E-26_WP,    8.995E-26_WP,    1.079E-25_WP,    1.270E-25_WP,    1.459E-25_WP,    1.628E-25_WP,    1.757E-25_WP, &
    & 1.834E-25_WP,    1.858E-25_WP,    1.836E-25_WP,    1.784E-25_WP,    1.714E-25_WP,    1.639E-25_WP,    1.553E-25_WP, &
    & 1.446E-25_WP,    1.319E-25_WP,    1.171E-25_WP,    1.000E-25_WP,    8.034E-26_WP,    5.763E-26_WP,    3.118E-26_WP /
data J(3,    13    ,:)    / 4.218E-27_WP,    5.315E-27_WP,    6.706E-27_WP,    8.469E-27_WP,    1.070E-26_WP,    1.352E-26_WP, &
    & 1.708E-26_WP,    2.155E-26_WP,    2.716E-26_WP,    3.415E-26_WP,    4.281E-26_WP,    5.345E-26_WP,    6.639E-26_WP, &
    & 8.182E-26_WP,    9.967E-26_WP,    1.196E-25_WP,    1.407E-25_WP,    1.616E-25_WP,    1.803E-25_WP,    1.946E-25_WP, &
    & 2.031E-25_WP,    2.057E-25_WP,    2.033E-25_WP,    1.974E-25_WP,    1.897E-25_WP,    1.814E-25_WP,    1.719E-25_WP, &
    & 1.600E-25_WP,    1.460E-25_WP,    1.296E-25_WP,    1.107E-25_WP,    8.896E-26_WP,    6.384E-26_WP,    3.456E-26_WP /
data J(3,    14    ,:)    / 4.678E-27_WP,    5.894E-27_WP,    7.437E-27_WP,    9.392E-27_WP,    1.187E-26_WP,    1.499E-26_WP, &
    & 1.894E-26_WP,    2.389E-26_WP,    3.011E-26_WP,    3.785E-26_WP,    4.745E-26_WP,    5.924E-26_WP,    7.357E-26_WP, &
    & 9.066E-26_WP,    1.104E-25_WP,    1.325E-25_WP,    1.559E-25_WP,    1.790E-25_WP,    1.996E-25_WP,    2.154E-25_WP, &
    & 2.248E-25_WP,    2.276E-25_WP,    2.249E-25_WP,    2.184E-25_WP,    2.099E-25_WP,    2.006E-25_WP,    1.901E-25_WP, &
    & 1.770E-25_WP,    1.615E-25_WP,    1.434E-25_WP,    1.225E-25_WP,    9.847E-26_WP,    7.071E-26_WP,    3.829E-26_WP /
data J(3,    15    ,:)    / 5.187E-27_WP,    6.536E-27_WP,    8.247E-27_WP,    1.041E-26_WP,    1.316E-26_WP,    1.662E-26_WP, &
    & 2.099E-26_WP,    2.649E-26_WP,    3.337E-26_WP,    4.196E-26_WP,    5.258E-26_WP,    6.564E-26_WP,    8.152E-26_WP, &
    & 1.004E-25_WP,    1.223E-25_WP,    1.467E-25_WP,    1.726E-25_WP,    1.982E-25_WP,    2.209E-25_WP,    2.384E-25_WP, &
    & 2.487E-25_WP,    2.518E-25_WP,    2.488E-25_WP,    2.415E-25_WP,    2.320E-25_WP,    2.218E-25_WP,    2.101E-25_WP, &
    & 1.957E-25_WP,    1.785E-25_WP,    1.585E-25_WP,    1.355E-25_WP,    1.090E-25_WP,    7.828E-26_WP,    4.243E-26_WP /
data J(3,    16    ,:)    / 5.752E-27_WP,    7.248E-27_WP,    9.145E-27_WP,    1.155E-26_WP,    1.459E-26_WP,    1.843E-26_WP, &
    & 2.327E-26_WP,    2.936E-26_WP,    3.699E-26_WP,    4.650E-26_WP,    5.827E-26_WP,    7.273E-26_WP,    9.030E-26_WP, &
    & 1.112E-25_WP,    1.354E-25_WP,    1.624E-25_WP,    1.911E-25_WP,    2.193E-25_WP,    2.444E-25_WP,    2.637E-25_WP, &
    & 2.751E-25_WP,    2.784E-25_WP,    2.750E-25_WP,    2.669E-25_WP,    2.564E-25_WP,    2.450E-25_WP,    2.321E-25_WP, &
    & 2.161E-25_WP,    1.972E-25_WP,    1.751E-25_WP,    1.497E-25_WP,    1.205E-25_WP,    8.664E-26_WP,    4.700E-26_WP /
data J(3,    17    ,:)    / 6.379E-27_WP,    8.037E-27_WP,    1.014E-26_WP,    1.280E-26_WP,    1.617E-26_WP,    2.043E-26_WP, &
    & 2.580E-26_WP,    3.254E-26_WP,    4.099E-26_WP,    5.152E-26_WP,    6.455E-26_WP,    8.056E-26_WP,    1.000E-25_WP, &
    & 1.232E-25_WP,    1.499E-25_WP,    1.798E-25_WP,    2.114E-25_WP,    2.426E-25_WP,    2.703E-25_WP,    2.915E-25_WP, &
    & 3.040E-25_WP,    3.076E-25_WP,    3.037E-25_WP,    2.947E-25_WP,    2.830E-25_WP,    2.704E-25_WP,    2.561E-25_WP, &
    & 2.385E-25_WP,    2.176E-25_WP,    1.934E-25_WP,    1.654E-25_WP,    1.332E-25_WP,    9.586E-26_WP,    5.206E-26_WP /
data J(3,    18    ,:)    / 7.073E-27_WP,    8.912E-27_WP,    1.124E-26_WP,    1.419E-26_WP,    1.793E-26_WP,    2.265E-26_WP, &
    & 2.859E-26_WP,    3.607E-26_WP,    4.542E-26_WP,    5.708E-26_WP,    7.150E-26_WP,    8.921E-26_WP,    1.107E-25_WP, &
    & 1.363E-25_WP,    1.659E-25_WP,    1.989E-25_WP,    2.338E-25_WP,    2.682E-25_WP,    2.988E-25_WP,    3.221E-25_WP, &
    & 3.358E-25_WP,    3.396E-25_WP,    3.352E-25_WP,    3.252E-25_WP,    3.122E-25_WP,    2.982E-25_WP,    2.824E-25_WP, &
    & 2.630E-25_WP,    2.400E-25_WP,    2.133E-25_WP,    1.826E-25_WP,    1.471E-25_WP,    1.060E-25_WP,    5.764E-26_WP /
data J(3,    19    ,:)    / 7.843E-27_WP,    9.881E-27_WP,    1.246E-26_WP,    1.573E-26_WP,    1.987E-26_WP,    2.510E-26_WP, &
    & 3.169E-26_WP,    3.996E-26_WP,    5.032E-26_WP,    6.322E-26_WP,    7.918E-26_WP,    9.876E-26_WP,    1.225E-25_WP, &
    & 1.508E-25_WP,    1.835E-25_WP,    2.199E-25_WP,    2.584E-25_WP,    2.963E-25_WP,    3.300E-25_WP,    3.556E-25_WP, &
    & 3.706E-25_WP,    3.747E-25_WP,    3.697E-25_WP,    3.585E-25_WP,    3.440E-25_WP,    3.285E-25_WP,    3.111E-25_WP, &
    & 2.896E-25_WP,    2.644E-25_WP,    2.351E-25_WP,    2.013E-25_WP,    1.624E-25_WP,    1.171E-25_WP,    6.379E-26_WP /
data J(3,    20    ,:)    / 8.696E-27_WP,    1.095E-26_WP,    1.382E-26_WP,    1.744E-26_WP,    2.203E-26_WP,    2.781E-26_WP, &
    & 3.511E-26_WP,    4.427E-26_WP,    5.573E-26_WP,    7.000E-26_WP,    8.765E-26_WP,    1.093E-25_WP,    1.356E-25_WP, &
    & 1.668E-25_WP,    2.029E-25_WP,    2.430E-25_WP,    2.855E-25_WP,    3.272E-25_WP,    3.641E-25_WP,    3.922E-25_WP, &
    & 4.086E-25_WP,    4.129E-25_WP,    4.072E-25_WP,    3.946E-25_WP,    3.785E-25_WP,    3.614E-25_WP,    3.421E-25_WP, &
    & 3.186E-25_WP,    2.908E-25_WP,    2.587E-25_WP,    2.217E-25_WP,    1.790E-25_WP,    1.294E-25_WP,    7.058E-26_WP /
data J(3,    21    ,:)    / 9.641E-27_WP,    1.214E-26_WP,    1.532E-26_WP,    1.933E-26_WP,    2.441E-26_WP,    3.082E-26_WP, &
    & 3.889E-26_WP,    4.903E-26_WP,    6.171E-26_WP,    7.748E-26_WP,    9.698E-26_WP,    1.209E-25_WP,    1.499E-25_WP, &
    & 1.844E-25_WP,    2.241E-25_WP,    2.683E-25_WP,    3.151E-25_WP,    3.609E-25_WP,    4.014E-25_WP,    4.321E-25_WP, &
    & 4.499E-25_WP,    4.544E-25_WP,    4.479E-25_WP,    4.338E-25_WP,    4.159E-25_WP,    3.969E-25_WP,    3.757E-25_WP, &
    & 3.498E-25_WP,    3.194E-25_WP,    2.843E-25_WP,    2.439E-25_WP,    1.972E-25_WP,    1.427E-25_WP,    7.804E-26_WP /
data J(3,    22    ,:)    / 1.069E-26_WP,    1.346E-26_WP,    1.697E-26_WP,    2.142E-26_WP,    2.704E-26_WP,    3.414E-26_WP, &
    & 4.307E-26_WP,    5.428E-26_WP,    6.830E-26_WP,    8.573E-26_WP,    1.073E-25_WP,    1.337E-25_WP,    1.656E-25_WP, &
    & 2.036E-25_WP,    2.474E-25_WP,    2.961E-25_WP,    3.474E-25_WP,    3.977E-25_WP,    4.421E-25_WP,    4.756E-25_WP, &
    & 4.948E-25_WP,    4.994E-25_WP,    4.919E-25_WP,    4.761E-25_WP,    4.562E-25_WP,    4.351E-25_WP,    4.117E-25_WP, &
    & 3.834E-25_WP,    3.502E-25_WP,    3.118E-25_WP,    2.678E-25_WP,    2.169E-25_WP,    1.573E-25_WP,    8.625E-26_WP /
data J(3,    23    ,:)    / 1.185E-26_WP,    1.492E-26_WP,    1.881E-26_WP,    2.373E-26_WP,    2.996E-26_WP,    3.781E-26_WP, &
    & 4.769E-26_WP,    6.008E-26_WP,    7.557E-26_WP,    9.481E-26_WP,    1.186E-25_WP,    1.477E-25_WP,    1.829E-25_WP, &
    & 2.247E-25_WP,    2.729E-25_WP,    3.263E-25_WP,    3.826E-25_WP,    4.377E-25_WP,    4.862E-25_WP,    5.226E-25_WP, &
    & 5.433E-25_WP,    5.480E-25_WP,    5.393E-25_WP,    5.215E-25_WP,    4.993E-25_WP,    4.759E-25_WP,    4.502E-25_WP, &
    & 4.192E-25_WP,    3.830E-25_WP,    3.414E-25_WP,    2.935E-25_WP,    2.382E-25_WP,    1.732E-25_WP,    9.526E-26_WP /
data J(3,    24    ,:)    / 1.313E-26_WP,    1.653E-26_WP,    2.084E-26_WP,    2.629E-26_WP,    3.318E-26_WP,    4.186E-26_WP, &
    & 5.278E-26_WP,    6.647E-26_WP,    8.357E-26_WP,    1.048E-25_WP,    1.310E-25_WP,    1.630E-25_WP,    2.018E-25_WP, &
    & 2.477E-25_WP,    3.006E-25_WP,    3.592E-25_WP,    4.209E-25_WP,    4.811E-25_WP,    5.338E-25_WP,    5.733E-25_WP, &
    & 5.955E-25_WP,    6.000E-25_WP,    5.899E-25_WP,    5.700E-25_WP,    5.451E-25_WP,    5.193E-25_WP,    4.911E-25_WP, &
    & 4.572E-25_WP,    4.179E-25_WP,    3.728E-25_WP,    3.210E-25_WP,    2.611E-25_WP,    1.905E-25_WP,    1.051E-25_WP /
data J(3,    25    ,:)    / 1.455E-26_WP,    1.832E-26_WP,    2.309E-26_WP,    2.912E-26_WP,    3.673E-26_WP,    4.633E-26_WP, &
    & 5.839E-26_WP,    7.351E-26_WP,    9.236E-26_WP,    1.158E-25_WP,    1.446E-25_WP,    1.798E-25_WP,    2.224E-25_WP, &
    & 2.728E-25_WP,    3.308E-25_WP,    3.949E-25_WP,    4.622E-25_WP,    5.278E-25_WP,    5.851E-25_WP,    6.276E-25_WP, &
    & 6.512E-25_WP,    6.554E-25_WP,    6.437E-25_WP,    6.212E-25_WP,    5.935E-25_WP,    5.649E-25_WP,    5.340E-25_WP, &
    & 4.972E-25_WP,    4.546E-25_WP,    4.059E-25_WP,    3.502E-25_WP,    2.855E-25_WP,    2.091E-25_WP,    1.159E-25_WP /
data J(3,    26    ,:)    / 1.612E-26_WP,    2.029E-26_WP,    2.557E-26_WP,    3.224E-26_WP,    4.065E-26_WP,    5.125E-26_WP, &
    & 6.457E-26_WP,    8.124E-26_WP,    1.020E-25_WP,    1.278E-25_WP,    1.595E-25_WP,    1.981E-25_WP,    2.448E-25_WP, &
    & 3.001E-25_WP,    3.634E-25_WP,    4.334E-25_WP,    5.068E-25_WP,    5.779E-25_WP,    6.398E-25_WP,    6.856E-25_WP, &
    & 7.104E-25_WP,    7.141E-25_WP,    7.003E-25_WP,    6.750E-25_WP,    6.441E-25_WP,    6.124E-25_WP,    5.786E-25_WP, &
    & 5.387E-25_WP,    4.928E-25_WP,    4.405E-25_WP,    3.808E-25_WP,    3.115E-25_WP,    2.291E-25_WP,    1.276E-25_WP /
data J(3,    27    ,:)    / 1.786E-26_WP,    2.247E-26_WP,    2.831E-26_WP,    3.568E-26_WP,    4.498E-26_WP,    5.668E-26_WP, &
    & 7.136E-26_WP,    8.972E-26_WP,    1.126E-25_WP,    1.409E-25_WP,    1.757E-25_WP,    2.181E-25_WP,    2.692E-25_WP, &
    & 3.295E-25_WP,    3.985E-25_WP,    4.747E-25_WP,    5.544E-25_WP,    6.314E-25_WP,    6.980E-25_WP,    7.468E-25_WP, &
    & 7.728E-25_WP,    7.756E-25_WP,    7.595E-25_WP,    7.308E-25_WP,    6.963E-25_WP,    6.613E-25_WP,    6.244E-25_WP, &
    & 5.812E-25_WP,    5.320E-25_WP,    4.763E-25_WP,    4.127E-25_WP,    3.388E-25_WP,    2.504E-25_WP,    1.403E-25_WP /
data J(3,    28    ,:)    / 1.978E-26_WP,    2.488E-26_WP,    3.133E-26_WP,    3.947E-26_WP,    4.973E-26_WP,    6.263E-26_WP, &
    & 7.881E-26_WP,    9.901E-26_WP,    1.241E-25_WP,    1.552E-25_WP,    1.933E-25_WP,    2.396E-25_WP,    2.954E-25_WP, &
    & 3.611E-25_WP,    4.362E-25_WP,    5.189E-25_WP,    6.051E-25_WP,    6.880E-25_WP,    7.594E-25_WP,    8.111E-25_WP, &
    & 8.379E-25_WP,    8.395E-25_WP,    8.205E-25_WP,    7.881E-25_WP,    7.496E-25_WP,    7.109E-25_WP,    6.707E-25_WP, &
    & 6.243E-25_WP,    5.718E-25_WP,    5.126E-25_WP,    4.454E-25_WP,    3.671E-25_WP,    2.730E-25_WP,    1.541E-25_WP /
data J(3,    29    ,:)    / 2.190E-26_WP,    2.754E-26_WP,    3.466E-26_WP,    4.365E-26_WP,    5.496E-26_WP,    6.917E-26_WP, &
    & 8.696E-26_WP,    1.092E-25_WP,    1.367E-25_WP,    1.707E-25_WP,    2.123E-25_WP,    2.629E-25_WP,    3.236E-25_WP, &
    & 3.950E-25_WP,    4.764E-25_WP,    5.657E-25_WP,    6.585E-25_WP,    7.475E-25_WP,    8.235E-25_WP,    8.779E-25_WP, &
    & 9.051E-25_WP,    9.050E-25_WP,    8.827E-25_WP,    8.461E-25_WP,    8.030E-25_WP,    7.604E-25_WP,    7.167E-25_WP, &
    & 6.669E-25_WP,    6.112E-25_WP,    5.490E-25_WP,    4.784E-25_WP,    3.963E-25_WP,    2.967E-25_WP,    1.689E-25_WP /
data J(3,    30    ,:)    / 2.424E-26_WP,    3.047E-26_WP,    3.833E-26_WP,    4.824E-26_WP,    6.070E-26_WP,    7.632E-26_WP, &
    & 9.587E-26_WP,    1.202E-25_WP,    1.504E-25_WP,    1.875E-25_WP,    2.329E-25_WP,    2.878E-25_WP,    3.537E-25_WP, &
    & 4.309E-25_WP,    5.188E-25_WP,    6.149E-25_WP,    7.145E-25_WP,    8.094E-25_WP,    8.897E-25_WP,    9.466E-25_WP, &
    & 9.738E-25_WP,    9.714E-25_WP,    9.452E-25_WP,    9.037E-25_WP,    8.556E-25_WP,    8.087E-25_WP,    7.614E-25_WP, &
    & 7.083E-25_WP,    6.496E-25_WP,    5.845E-25_WP,    5.112E-25_WP,    4.258E-25_WP,    3.214E-25_WP,    1.847E-25_WP /
data J(3,    31    ,:)    / 2.682E-26_WP,    3.369E-26_WP,    4.236E-26_WP,    5.327E-26_WP,    6.697E-26_WP,    8.414E-26_WP, &
    & 1.056E-25_WP,    1.322E-25_WP,    1.651E-25_WP,    2.056E-25_WP,    2.549E-25_WP,    3.144E-25_WP,    3.856E-25_WP, &
    & 4.688E-25_WP,    5.632E-25_WP,    6.662E-25_WP,    7.725E-25_WP,    8.731E-25_WP,    9.575E-25_WP,    1.016E-24_WP, &
    & 1.043E-24_WP,    1.038E-24_WP,    1.007E-24_WP,    9.599E-25_WP,    9.063E-25_WP,    8.547E-25_WP,    8.037E-25_WP, &
    & 7.473E-25_WP,    6.858E-25_WP,    6.184E-25_WP,    5.429E-25_WP,    4.550E-25_WP,    3.466E-25_WP,    2.015E-25_WP /
data J(3,    32    ,:)    / 2.966E-26_WP,    3.724E-26_WP,    4.678E-26_WP,    5.878E-26_WP,    7.383E-26_WP,    9.265E-26_WP, &
    & 1.161E-25_WP,    1.452E-25_WP,    1.810E-25_WP,    2.249E-25_WP,    2.783E-25_WP,    3.426E-25_WP,    4.191E-25_WP, &
    & 5.084E-25_WP,    6.094E-25_WP,    7.192E-25_WP,    8.319E-25_WP,    9.379E-25_WP,    1.026E-24_WP,    1.086E-24_WP, &
    & 1.111E-24_WP,    1.103E-24_WP,    1.067E-24_WP,    1.014E-24_WP,    9.540E-25_WP,    8.973E-25_WP,    8.424E-25_WP, &
    & 7.828E-25_WP,    7.188E-25_WP,    6.495E-25_WP,    5.727E-25_WP,    4.834E-25_WP,    3.721E-25_WP,    2.192E-25_WP /
data J(3,    33    ,:)    / 3.278E-26_WP,    4.113E-26_WP,    5.162E-26_WP,    6.480E-26_WP,    8.130E-26_WP,    1.019E-25_WP, &
    & 1.275E-25_WP,    1.591E-25_WP,    1.980E-25_WP,    2.455E-25_WP,    3.030E-25_WP,    3.721E-25_WP,    4.541E-25_WP, &
    & 5.494E-25_WP,    6.568E-25_WP,    7.731E-25_WP,    8.921E-25_WP,    1.003E-24_WP,    1.094E-24_WP,    1.155E-24_WP, &
    & 1.178E-24_WP,    1.165E-24_WP,    1.123E-24_WP,    1.064E-24_WP,    9.973E-25_WP,    9.353E-25_WP,    8.764E-25_WP, &
    & 8.137E-25_WP,    7.475E-25_WP,    6.770E-25_WP,    5.996E-25_WP,    5.100E-25_WP,    3.973E-25_WP,    2.376E-25_WP /
data J(3,    34    ,:)    / 3.620E-26_WP,    4.539E-26_WP,    5.691E-26_WP,    7.136E-26_WP,    8.940E-26_WP,    1.119E-25_WP, &
    & 1.397E-25_WP,    1.740E-25_WP,    2.160E-25_WP,    2.672E-25_WP,    3.289E-25_WP,    4.027E-25_WP,    4.901E-25_WP, &
    & 5.912E-25_WP,    7.048E-25_WP,    8.274E-25_WP,    9.522E-25_WP,    1.068E-24_WP,    1.161E-24_WP,    1.222E-24_WP, &
    & 1.243E-24_WP,    1.225E-24_WP,    1.176E-24_WP,    1.109E-24_WP,    1.035E-24_WP,    9.677E-25_WP,    9.048E-25_WP, &
    & 8.391E-25_WP,    7.710E-25_WP,    6.998E-25_WP,    6.227E-25_WP,    5.342E-25_WP,    4.217E-25_WP,    2.565E-25_WP /
data J(3,    35    ,:)    / 3.996E-26_WP,    5.004E-26_WP,    6.268E-26_WP,    7.847E-26_WP,    9.815E-26_WP,    1.226E-25_WP, &
    & 1.528E-25_WP,    1.898E-25_WP,    2.351E-25_WP,    2.899E-25_WP,    3.558E-25_WP,    4.343E-25_WP,    5.267E-25_WP, &
    & 6.334E-25_WP,    7.529E-25_WP,    8.813E-25_WP,    1.011E-24_WP,    1.131E-24_WP,    1.226E-24_WP,    1.286E-24_WP, &
    & 1.304E-24_WP,    1.280E-24_WP,    1.224E-24_WP,    1.149E-24_WP,    1.068E-24_WP,    9.939E-25_WP,    9.269E-25_WP, &
    & 8.583E-25_WP,    7.886E-25_WP,    7.171E-25_WP,    6.412E-25_WP,    5.550E-25_WP,    4.446E-25_WP,    2.758E-25_WP /
data J(3,    36    ,:)    / 4.406E-26_WP,    5.511E-26_WP,    6.893E-26_WP,    8.616E-26_WP,    1.076E-25_WP,    1.341E-25_WP, &
    & 1.667E-25_WP,    2.065E-25_WP,    2.550E-25_WP,    3.134E-25_WP,    3.833E-25_WP,    4.662E-25_WP,    5.635E-25_WP, &
    & 6.754E-25_WP,    8.002E-25_WP,    9.339E-25_WP,    1.069E-24_WP,    1.191E-24_WP,    1.288E-24_WP,    1.347E-24_WP, &
    & 1.360E-24_WP,    1.330E-24_WP,    1.266E-24_WP,    1.183E-24_WP,    1.094E-24_WP,    1.014E-24_WP,    9.424E-25_WP, &
    & 8.709E-25_WP,    7.998E-25_WP,    7.285E-25_WP,    6.544E-25_WP,    5.716E-25_WP,    4.652E-25_WP,    2.950E-25_WP /
data J(3,    37    ,:)    / 4.853E-26_WP,    6.062E-26_WP,    7.569E-26_WP,    9.444E-26_WP,    1.176E-25_WP,    1.462E-25_WP, &
    & 1.813E-25_WP,    2.240E-25_WP,    2.756E-25_WP,    3.375E-25_WP,    4.112E-25_WP,    4.981E-25_WP,    5.998E-25_WP, &
    & 7.164E-25_WP,    8.460E-25_WP,    9.846E-25_WP,    1.124E-24_WP,    1.249E-24_WP,    1.346E-24_WP,    1.403E-24_WP, &
    & 1.413E-24_WP,    1.376E-24_WP,    1.303E-24_WP,    1.211E-24_WP,    1.114E-24_WP,    1.027E-24_WP,    9.516E-25_WP, &
    & 8.772E-25_WP,    8.048E-25_WP,    7.337E-25_WP,    6.618E-25_WP,    5.834E-25_WP,    4.829E-25_WP,    3.137E-25_WP /
data J(3,    38    ,:)    / 5.338E-26_WP,    6.657E-26_WP,    8.297E-26_WP,    1.033E-25_WP,    1.284E-25_WP,    1.591E-25_WP, &
    & 1.966E-25_WP,    2.420E-25_WP,    2.966E-25_WP,    3.618E-25_WP,    4.389E-25_WP,    5.295E-25_WP,    6.350E-25_WP, &
    & 7.557E-25_WP,    8.897E-25_WP,    1.033E-24_WP,    1.176E-24_WP,    1.303E-24_WP,    1.401E-24_WP,    1.456E-24_WP, &
    & 1.460E-24_WP,    1.416E-24_WP,    1.335E-24_WP,    1.234E-24_WP,    1.128E-24_WP,    1.035E-24_WP,    9.550E-25_WP, &
    & 8.778E-25_WP,    8.040E-25_WP,    7.332E-25_WP,    6.636E-25_WP,    5.900E-25_WP,    4.970E-25_WP,    3.315E-25_WP /
data J(3,    39    ,:)    / 5.864E-26_WP,    7.299E-26_WP,    9.077E-26_WP,    1.127E-25_WP,    1.396E-25_WP,    1.725E-25_WP, &
    & 2.124E-25_WP,    2.604E-25_WP,    3.179E-25_WP,    3.860E-25_WP,    4.661E-25_WP,    5.598E-25_WP,    6.686E-25_WP, &
    & 7.929E-25_WP,    9.306E-25_WP,    1.078E-24_WP,    1.224E-24_WP,    1.355E-24_WP,    1.452E-24_WP,    1.505E-24_WP, &
    & 1.505E-24_WP,    1.453E-24_WP,    1.364E-24_WP,    1.252E-24_WP,    1.138E-24_WP,    1.038E-24_WP,    9.538E-25_WP, &
    & 8.736E-25_WP,    7.983E-25_WP,    7.276E-25_WP,    6.601E-25_WP,    5.914E-25_WP,    5.070E-25_WP,    3.478E-25_WP /
data J(3,    40    ,:)    / 6.430E-26_WP,    7.986E-26_WP,    9.906E-26_WP,    1.226E-25_WP,    1.514E-25_WP,    1.864E-25_WP, &
    & 2.286E-25_WP,    2.790E-25_WP,    3.389E-25_WP,    4.095E-25_WP,    4.922E-25_WP,    5.884E-25_WP,    6.999E-25_WP, &
    & 8.273E-25_WP,    9.685E-25_WP,    1.119E-24_WP,    1.270E-24_WP,    1.403E-24_WP,    1.500E-24_WP,    1.552E-24_WP, &
    & 1.547E-24_WP,    1.488E-24_WP,    1.389E-24_WP,    1.268E-24_WP,    1.144E-24_WP,    1.037E-24_WP,    9.492E-25_WP, &
    & 8.659E-25_WP,    7.889E-25_WP,    7.179E-25_WP,    6.521E-25_WP,    5.880E-25_WP,    5.126E-25_WP,    3.621E-25_WP /
data J(3,    41    ,:)    / 7.037E-26_WP,    8.717E-26_WP,    1.078E-25_WP,    1.330E-25_WP,    1.637E-25_WP,    2.006E-25_WP, &
    & 2.449E-25_WP,    2.974E-25_WP,    3.595E-25_WP,    4.321E-25_WP,    5.167E-25_WP,    6.149E-25_WP,    7.287E-25_WP, &
    & 8.587E-25_WP,    1.003E-24_WP,    1.158E-24_WP,    1.313E-24_WP,    1.448E-24_WP,    1.547E-24_WP,    1.597E-24_WP, &
    & 1.587E-24_WP,    1.521E-24_WP,    1.413E-24_WP,    1.282E-24_WP,    1.149E-24_WP,    1.035E-24_WP,    9.426E-25_WP, &
    & 8.561E-25_WP,    7.771E-25_WP,    7.055E-25_WP,    6.406E-25_WP,    5.803E-25_WP,    5.138E-25_WP,    3.741E-25_WP /
data J(3,    42    ,:)    / 7.684E-26_WP,    9.491E-26_WP,    1.170E-25_WP,    1.438E-25_WP,    1.762E-25_WP,    2.149E-25_WP, &
    & 2.610E-25_WP,    3.154E-25_WP,    3.791E-25_WP,    4.533E-25_WP,    5.394E-25_WP,    6.390E-25_WP,    7.546E-25_WP, &
    & 8.871E-25_WP,    1.035E-24_WP,    1.194E-24_WP,    1.354E-24_WP,    1.493E-24_WP,    1.593E-24_WP,    1.642E-24_WP, &
    & 1.628E-24_WP,    1.554E-24_WP,    1.437E-24_WP,    1.296E-24_WP,    1.154E-24_WP,    1.033E-24_WP,    9.353E-25_WP, &
    & 8.455E-25_WP,    7.642E-25_WP,    6.915E-25_WP,    6.270E-25_WP,    5.694E-25_WP,    5.109E-25_WP,    3.831E-25_WP /
data J(3,    43    ,:)    / 8.369E-26_WP,    1.030E-25_WP,    1.265E-25_WP,    1.548E-25_WP,    1.888E-25_WP,    2.291E-25_WP, &
    & 2.768E-25_WP,    3.326E-25_WP,    3.975E-25_WP,    4.727E-25_WP,    5.597E-25_WP,    6.604E-25_WP,    7.776E-25_WP, &
    & 9.127E-25_WP,    1.064E-24_WP,    1.228E-24_WP,    1.394E-24_WP,    1.538E-24_WP,    1.640E-24_WP,    1.689E-24_WP, &
    & 1.671E-24_WP,    1.590E-24_WP,    1.463E-24_WP,    1.312E-24_WP,    1.160E-24_WP,    1.031E-24_WP,    9.284E-25_WP, &
    & 8.350E-25_WP,    7.512E-25_WP,    6.771E-25_WP,    6.123E-25_WP,    5.564E-25_WP,    5.045E-25_WP,    3.890E-25_WP /
data J(3,    44    ,:)    / 9.089E-26_WP,    1.114E-25_WP,    1.363E-25_WP,    1.660E-25_WP,    2.013E-25_WP,    2.430E-25_WP, &
    & 2.919E-25_WP,    3.487E-25_WP,    4.144E-25_WP,    4.902E-25_WP,    5.777E-25_WP,    6.792E-25_WP,    7.979E-25_WP, &
    & 9.358E-25_WP,    1.092E-24_WP,    1.262E-24_WP,    1.434E-24_WP,    1.584E-24_WP,    1.690E-24_WP,    1.739E-24_WP, &
    & 1.717E-24_WP,    1.628E-24_WP,    1.491E-24_WP,    1.329E-24_WP,    1.167E-24_WP,    1.030E-24_WP,    9.227E-25_WP, &
    & 8.254E-25_WP,    7.390E-25_WP,    6.631E-25_WP,    5.974E-25_WP,    5.422E-25_WP,    4.954E-25_WP,    3.917E-25_WP /
data J(3,    45    ,:)    / 9.837E-26_WP,    1.201E-25_WP,    1.462E-25_WP,    1.771E-25_WP,    2.136E-25_WP,    2.564E-25_WP, &
    & 3.061E-25_WP,    3.634E-25_WP,    4.295E-25_WP,    5.055E-25_WP,    5.933E-25_WP,    6.955E-25_WP,    8.160E-25_WP, &
    & 9.572E-25_WP,    1.118E-24_WP,    1.295E-24_WP,    1.476E-24_WP,    1.632E-24_WP,    1.743E-24_WP,    1.793E-24_WP, &
    & 1.767E-24_WP,    1.671E-24_WP,    1.523E-24_WP,    1.350E-24_WP,    1.177E-24_WP,    1.031E-24_WP,    9.188E-25_WP, &
    & 8.175E-25_WP,    7.280E-25_WP,    6.501E-25_WP,    5.832E-25_WP,    5.279E-25_WP,    4.845E-25_WP,    3.913E-25_WP /
data J(3,    46    ,:)    / 1.061E-25_WP,    1.289E-25_WP,    1.561E-25_WP,    1.881E-25_WP,    2.255E-25_WP,    2.689E-25_WP, &
    & 3.191E-25_WP,    3.767E-25_WP,    4.428E-25_WP,    5.187E-25_WP,    6.067E-25_WP,    7.097E-25_WP,    8.323E-25_WP, &
    & 9.775E-25_WP,    1.145E-24_WP,    1.330E-24_WP,    1.520E-24_WP,    1.685E-24_WP,    1.801E-24_WP,    1.852E-24_WP, &
    & 1.824E-24_WP,    1.719E-24_WP,    1.560E-24_WP,    1.375E-24_WP,    1.190E-24_WP,    1.036E-24_WP,    9.172E-25_WP, &
    & 8.115E-25_WP,    7.187E-25_WP,    6.385E-25_WP,    5.701E-25_WP,    5.141E-25_WP,    4.726E-25_WP,    3.882E-25_WP /
data J(3,    47    ,:)    / 1.140E-25_WP,    1.378E-25_WP,    1.658E-25_WP,    1.986E-25_WP,    2.367E-25_WP,    2.805E-25_WP, &
    & 3.308E-25_WP,    3.883E-25_WP,    4.542E-25_WP,    5.300E-25_WP,    6.182E-25_WP,    7.223E-25_WP,    8.475E-25_WP, &
    & 9.975E-25_WP,    1.172E-24_WP,    1.367E-24_WP,    1.568E-24_WP,    1.743E-24_WP,    1.865E-24_WP,    1.919E-24_WP, &
    & 1.887E-24_WP,    1.773E-24_WP,    1.602E-24_WP,    1.404E-24_WP,    1.207E-24_WP,    1.043E-24_WP,    9.182E-25_WP, &
    & 8.076E-25_WP,    7.113E-25_WP,    6.285E-25_WP,    5.584E-25_WP,    5.013E-25_WP,    4.605E-25_WP,    3.828E-25_WP /
data J(3,    48    ,:)    / 1.219E-25_WP,    1.465E-25_WP,    1.753E-25_WP,    2.087E-25_WP,    2.471E-25_WP,    2.911E-25_WP, &
    & 3.412E-25_WP,    3.985E-25_WP,    4.640E-25_WP,    5.397E-25_WP,    6.283E-25_WP,    7.339E-25_WP,    8.624E-25_WP, &
    & 1.018E-24_WP,    1.201E-24_WP,    1.407E-24_WP,    1.621E-24_WP,    1.807E-24_WP,    1.936E-24_WP,    1.992E-24_WP, &
    & 1.957E-24_WP,    1.834E-24_WP,    1.650E-24_WP,    1.437E-24_WP,    1.228E-24_WP,    1.053E-24_WP,    9.219E-25_WP, &
    & 8.061E-25_WP,    7.059E-25_WP,    6.203E-25_WP,    5.482E-25_WP,    4.898E-25_WP,    4.487E-25_WP,    3.758E-25_WP /
data J(3,    49    ,:)    / 1.298E-25_WP,    1.551E-25_WP,    1.844E-25_WP,    2.181E-25_WP,    2.566E-25_WP,    3.005E-25_WP, &
    & 3.503E-25_WP,    4.071E-25_WP,    4.724E-25_WP,    5.481E-25_WP,    6.374E-25_WP,    7.450E-25_WP,    8.774E-25_WP, &
    & 1.040E-24_WP,    1.232E-24_WP,    1.451E-24_WP,    1.678E-24_WP,    1.877E-24_WP,    2.015E-24_WP,    2.074E-24_WP, &
    & 2.035E-24_WP,    1.902E-24_WP,    1.704E-24_WP,    1.476E-24_WP,    1.253E-24_WP,    1.067E-24_WP,    9.285E-25_WP, &
    & 8.070E-25_WP,    7.026E-25_WP,    6.139E-25_WP,    5.396E-25_WP,    4.796E-25_WP,    4.377E-25_WP,    3.678E-25_WP /
data J(3,    50    ,:)    / 1.375E-25_WP,    1.633E-25_WP,    1.930E-25_WP,    2.268E-25_WP,    2.652E-25_WP,    3.088E-25_WP, &
    & 3.582E-25_WP,    4.146E-25_WP,    4.796E-25_WP,    5.556E-25_WP,    6.460E-25_WP,    7.561E-25_WP,    8.932E-25_WP, &
    & 1.063E-24_WP,    1.266E-24_WP,    1.499E-24_WP,    1.742E-24_WP,    1.954E-24_WP,    2.101E-24_WP,    2.164E-24_WP, &
    & 2.121E-24_WP,    1.977E-24_WP,    1.764E-24_WP,    1.520E-24_WP,    1.282E-24_WP,    1.084E-24_WP,    9.379E-25_WP, &
    & 8.104E-25_WP,    7.014E-25_WP,    6.093E-25_WP,    5.325E-25_WP,    4.707E-25_WP,    4.276E-25_WP,    3.595E-25_WP /
data J(3,    51    ,:)    / 1.451E-25_WP,    1.712E-25_WP,    2.010E-25_WP,    2.347E-25_WP,    2.729E-25_WP,    3.161E-25_WP, &
    & 3.651E-25_WP,    4.212E-25_WP,    4.862E-25_WP,    5.627E-25_WP,    6.546E-25_WP,    7.677E-25_WP,    9.101E-25_WP, &
    & 1.088E-24_WP,    1.304E-24_WP,    1.551E-24_WP,    1.812E-24_WP,    2.039E-24_WP,    2.196E-24_WP,    2.263E-24_WP, &
    & 2.216E-24_WP,    2.060E-24_WP,    1.831E-24_WP,    1.570E-24_WP,    1.315E-24_WP,    1.104E-24_WP,    9.502E-25_WP, &
    & 8.162E-25_WP,    7.022E-25_WP,    6.064E-25_WP,    5.271E-25_WP,    4.633E-25_WP,    4.186E-25_WP,    3.512E-25_WP /
data J(3,    52    ,:)    / 1.523E-25_WP,    1.786E-25_WP,    2.084E-25_WP,    2.420E-25_WP,    2.798E-25_WP,    3.226E-25_WP, &
    & 3.713E-25_WP,    4.272E-25_WP,    4.924E-25_WP,    5.697E-25_WP,    6.635E-25_WP,    7.800E-25_WP,    9.284E-25_WP, &
    & 1.116E-24_WP,    1.344E-24_WP,    1.609E-24_WP,    1.888E-24_WP,    2.131E-24_WP,    2.299E-24_WP,    2.370E-24_WP, &
    & 2.319E-24_WP,    2.151E-24_WP,    1.904E-24_WP,    1.624E-24_WP,    1.353E-24_WP,    1.128E-24_WP,    9.654E-25_WP, &
    & 8.244E-25_WP,    7.051E-25_WP,    6.054E-25_WP,    5.231E-25_WP,    4.573E-25_WP,    4.107E-25_WP,    3.435E-25_WP /
data J(3,    53    ,:)    / 1.592E-25_WP,    1.856E-25_WP,    2.152E-25_WP,    2.486E-25_WP,    2.861E-25_WP,    3.286E-25_WP, &
    & 3.770E-25_WP,    4.329E-25_WP,    4.985E-25_WP,    5.769E-25_WP,    6.729E-25_WP,    7.934E-25_WP,    9.484E-25_WP, &
    & 1.146E-24_WP,    1.388E-24_WP,    1.671E-24_WP,    1.970E-24_WP,    2.231E-24_WP,    2.411E-24_WP,    2.487E-24_WP, &
    & 2.430E-24_WP,    2.249E-24_WP,    1.984E-24_WP,    1.684E-24_WP,    1.394E-24_WP,    1.156E-24_WP,    9.832E-25_WP, &
    & 8.349E-25_WP,    7.099E-25_WP,    6.060E-25_WP,    5.206E-25_WP,    4.526E-25_WP,    4.040E-25_WP,    3.365E-25_WP /
data J(3,    54    ,:)    / 1.658E-25_WP,    1.921E-25_WP,    2.216E-25_WP,    2.547E-25_WP,    2.919E-25_WP,    3.341E-25_WP, &
    & 3.824E-25_WP,    4.385E-25_WP,    5.048E-25_WP,    5.846E-25_WP,    6.832E-25_WP,    8.080E-25_WP,    9.701E-25_WP, &
    & 1.179E-24_WP,    1.436E-24_WP,    1.737E-24_WP,    2.058E-24_WP,    2.338E-24_WP,    2.530E-24_WP,    2.611E-24_WP, &
    & 2.549E-24_WP,    2.354E-24_WP,    2.069E-24_WP,    1.749E-24_WP,    1.440E-24_WP,    1.186E-24_WP,    1.004E-24_WP, &
    & 8.476E-25_WP,    7.167E-25_WP,    6.083E-25_WP,    5.196E-25_WP,    4.491E-25_WP,    3.984E-25_WP,    3.304E-25_WP /
data J(3,    55    ,:)    / 1.721E-25_WP,    1.982E-25_WP,    2.275E-25_WP,    2.604E-25_WP,    2.974E-25_WP,    3.395E-25_WP, &
    & 3.878E-25_WP,    4.443E-25_WP,    5.114E-25_WP,    5.929E-25_WP,    6.943E-25_WP,    8.237E-25_WP,    9.934E-25_WP, &
    & 1.213E-24_WP,    1.486E-24_WP,    1.808E-24_WP,    2.152E-24_WP,    2.452E-24_WP,    2.657E-24_WP,    2.743E-24_WP, &
    & 2.676E-24_WP,    2.466E-24_WP,    2.160E-24_WP,    1.818E-24_WP,    1.489E-24_WP,    1.219E-24_WP,    1.027E-24_WP, &
    & 8.624E-25_WP,    7.252E-25_WP,    6.121E-25_WP,    5.200E-25_WP,    4.469E-25_WP,    3.940E-25_WP,    3.252E-25_WP /
data J(3,    56    ,:)    / 1.781E-25_WP,    2.041E-25_WP,    2.333E-25_WP,    2.659E-25_WP,    3.028E-25_WP,    3.448E-25_WP, &
    & 3.934E-25_WP,    4.504E-25_WP,    5.185E-25_WP,    6.019E-25_WP,    7.063E-25_WP,    8.407E-25_WP,    1.018E-24_WP, &
    & 1.250E-24_WP,    1.540E-24_WP,    1.883E-24_WP,    2.251E-24_WP,    2.573E-24_WP,    2.791E-24_WP,    2.882E-24_WP, &
    & 2.809E-24_WP,    2.584E-24_WP,    2.256E-24_WP,    1.891E-24_WP,    1.541E-24_WP,    1.255E-24_WP,    1.052E-24_WP, &
    & 8.790E-25_WP,    7.354E-25_WP,    6.174E-25_WP,    5.217E-25_WP,    4.459E-25_WP,    3.906E-25_WP,    3.210E-25_WP /
data J(3,    57    ,:)    / 1.839E-25_WP,    2.099E-25_WP,    2.389E-25_WP,    2.714E-25_WP,    3.083E-25_WP,    3.504E-25_WP, &
    & 3.992E-25_WP,    4.569E-25_WP,    5.262E-25_WP,    6.116E-25_WP,    7.192E-25_WP,    8.586E-25_WP,    1.044E-24_WP, &
    & 1.289E-24_WP,    1.596E-24_WP,    1.961E-24_WP,    2.355E-24_WP,    2.699E-24_WP,    2.931E-24_WP,    3.027E-24_WP, &
    & 2.949E-24_WP,    2.707E-24_WP,    2.357E-24_WP,    1.967E-24_WP,    1.596E-24_WP,    1.293E-24_WP,    1.079E-24_WP, &
    & 8.974E-25_WP,    7.471E-25_WP,    6.241E-25_WP,    5.246E-25_WP,    4.460E-25_WP,    3.882E-25_WP,    3.177E-25_WP /
data J(3,    58    ,:)    / 1.897E-25_WP,    2.156E-25_WP,    2.445E-25_WP,    2.770E-25_WP,    3.139E-25_WP,    3.562E-25_WP, &
    & 4.055E-25_WP,    4.639E-25_WP,    5.346E-25_WP,    6.220E-25_WP,    7.329E-25_WP,    8.776E-25_WP,    1.072E-24_WP, &
    & 1.329E-24_WP,    1.653E-24_WP,    2.042E-24_WP,    2.462E-24_WP,    2.829E-24_WP,    3.076E-24_WP,    3.178E-24_WP, &
    & 3.093E-24_WP,    2.834E-24_WP,    2.461E-24_WP,    2.046E-24_WP,    1.653E-24_WP,    1.333E-24_WP,    1.108E-24_WP, &
    & 9.174E-25_WP,    7.602E-25_WP,    6.320E-25_WP,    5.286E-25_WP,    4.472E-25_WP,    3.869E-25_WP,    3.153E-25_WP /
data J(3,    59    ,:)    / 1.955E-25_WP,    2.214E-25_WP,    2.503E-25_WP,    2.828E-25_WP,    3.198E-25_WP,    3.624E-25_WP, &
    & 4.121E-25_WP,    4.714E-25_WP,    5.435E-25_WP,    6.331E-25_WP,    7.473E-25_WP,    8.972E-25_WP,    1.100E-24_WP, &
    & 1.370E-24_WP,    1.713E-24_WP,    2.125E-24_WP,    2.573E-24_WP,    2.963E-24_WP,    3.225E-24_WP,    3.332E-24_WP, &
    & 3.241E-24_WP,    2.965E-24_WP,    2.567E-24_WP,    2.128E-24_WP,    1.713E-24_WP,    1.375E-24_WP,    1.138E-24_WP, &
    & 9.387E-25_WP,    7.746E-25_WP,    6.411E-25_WP,    5.337E-25_WP,    4.494E-25_WP,    3.865E-25_WP,    3.137E-25_WP /
data J(3,    60    ,:)    / 2.014E-25_WP,    2.273E-25_WP,    2.563E-25_WP,    2.889E-25_WP,    3.261E-25_WP,    3.690E-25_WP, &
    & 4.193E-25_WP,    4.795E-25_WP,    5.530E-25_WP,    6.448E-25_WP,    7.623E-25_WP,    9.175E-25_WP,    1.128E-24_WP, &
    & 1.411E-24_WP,    1.773E-24_WP,    2.209E-24_WP,    2.685E-24_WP,    3.100E-24_WP,    3.377E-24_WP,    3.490E-24_WP, &
    & 3.392E-24_WP,    3.098E-24_WP,    2.676E-24_WP,    2.212E-24_WP,    1.773E-24_WP,    1.418E-24_WP,    1.170E-24_WP, &
    & 9.613E-25_WP,    7.901E-25_WP,    6.512E-25_WP,    5.399E-25_WP,    4.525E-25_WP,    3.871E-25_WP,    3.130E-25_WP /
data J(3,    61    ,:)    / 2.075E-25_WP,    2.335E-25_WP,    2.626E-25_WP,    2.954E-25_WP,    3.328E-25_WP,    3.760E-25_WP, &
    & 4.270E-25_WP,    4.881E-25_WP,    5.630E-25_WP,    6.570E-25_WP,    7.779E-25_WP,    9.381E-25_WP,    1.157E-24_WP, &
    & 1.453E-24_WP,    1.833E-24_WP,    2.294E-24_WP,    2.799E-24_WP,    3.240E-24_WP,    3.531E-24_WP,    3.650E-24_WP, &
    & 3.545E-24_WP,    3.233E-24_WP,    2.787E-24_WP,    2.296E-24_WP,    1.835E-24_WP,    1.462E-24_WP,    1.202E-24_WP, &
    & 9.849E-25_WP,    8.066E-25_WP,    6.623E-25_WP,    5.469E-25_WP,    4.564E-25_WP,    3.884E-25_WP,    3.130E-25_WP /
data J(3,    62    ,:)    / 2.140E-25_WP,    2.401E-25_WP,    2.693E-25_WP,    3.022E-25_WP,    3.399E-25_WP,    3.836E-25_WP, &
    & 4.351E-25_WP,    4.972E-25_WP,    5.736E-25_WP,    6.698E-25_WP,    7.939E-25_WP,    9.590E-25_WP,    1.186E-24_WP, &
    & 1.495E-24_WP,    1.893E-24_WP,    2.379E-24_WP,    2.914E-24_WP,    3.380E-24_WP,    3.687E-24_WP,    3.811E-24_WP, &
    & 3.699E-24_WP,    3.369E-24_WP,    2.898E-24_WP,    2.382E-24_WP,    1.898E-24_WP,    1.507E-24_WP,    1.236E-24_WP, &
    & 1.009E-24_WP,    8.240E-25_WP,    6.743E-25_WP,    5.547E-25_WP,    4.612E-25_WP,    3.906E-25_WP,    3.138E-25_WP /
data J(3,    63    ,:)    / 2.208E-25_WP,    2.470E-25_WP,    2.763E-25_WP,    3.095E-25_WP,    3.475E-25_WP,    3.916E-25_WP, &
    & 4.438E-25_WP,    5.069E-25_WP,    5.848E-25_WP,    6.830E-25_WP,    8.102E-25_WP,    9.802E-25_WP,    1.215E-24_WP, &
    & 1.537E-24_WP,    1.953E-24_WP,    2.464E-24_WP,    3.029E-24_WP,    3.521E-24_WP,    3.843E-24_WP,    3.973E-24_WP, &
    & 3.854E-24_WP,    3.505E-24_WP,    3.010E-24_WP,    2.468E-24_WP,    1.962E-24_WP,    1.553E-24_WP,    1.270E-24_WP, &
    & 1.035E-24_WP,    8.422E-25_WP,    6.871E-25_WP,    5.634E-25_WP,    4.668E-25_WP,    3.936E-25_WP,    3.153E-25_WP /
data J(3,    64    ,:)    / 2.279E-25_WP,    2.543E-25_WP,    2.839E-25_WP,    3.173E-25_WP,    3.555E-25_WP,    4.002E-25_WP, &
    & 4.531E-25_WP,    5.172E-25_WP,    5.965E-25_WP,    6.967E-25_WP,    8.270E-25_WP,    1.002E-24_WP,    1.244E-24_WP, &
    & 1.578E-24_WP,    2.012E-24_WP,    2.548E-24_WP,    3.144E-24_WP,    3.662E-24_WP,    4.000E-24_WP,    4.135E-24_WP, &
    & 4.010E-24_WP,    3.643E-24_WP,    3.122E-24_WP,    2.555E-24_WP,    2.027E-24_WP,    1.600E-24_WP,    1.306E-24_WP, &
    & 1.061E-24_WP,    8.614E-25_WP,    7.008E-25_WP,    5.730E-25_WP,    4.732E-25_WP,    3.974E-25_WP,    3.175E-25_WP /
data J(3,    65    ,:)    / 2.356E-25_WP,    2.622E-25_WP,    2.919E-25_WP,    3.256E-25_WP,    3.642E-25_WP,    4.093E-25_WP, &
    & 4.630E-25_WP,    5.281E-25_WP,    6.089E-25_WP,    7.112E-25_WP,    8.445E-25_WP,    1.024E-24_WP,    1.274E-24_WP, &
    & 1.620E-24_WP,    2.072E-24_WP,    2.633E-24_WP,    3.260E-24_WP,    3.806E-24_WP,    4.159E-24_WP,    4.301E-24_WP, &
    & 4.169E-24_WP,    3.783E-24_WP,    3.238E-24_WP,    2.645E-24_WP,    2.094E-24_WP,    1.649E-24_WP,    1.344E-24_WP, &
    & 1.089E-24_WP,    8.825E-25_WP,    7.162E-25_WP,    5.840E-25_WP,    4.810E-25_WP,    4.025E-25_WP,    3.208E-25_WP /
data J(3,    66    ,:)    / 2.438E-25_WP,    2.706E-25_WP,    3.006E-25_WP,    3.346E-25_WP,    3.737E-25_WP,    4.195E-25_WP, &
    & 4.740E-25_WP,    5.404E-25_WP,    6.228E-25_WP,    7.276E-25_WP,    8.642E-25_WP,    1.048E-24_WP,    1.307E-24_WP, &
    & 1.667E-24_WP,    2.138E-24_WP,    2.726E-24_WP,    3.388E-24_WP,    3.963E-24_WP,    4.335E-24_WP,    4.485E-24_WP, &
    & 4.347E-24_WP,    3.943E-24_WP,    3.372E-24_WP,    2.751E-24_WP,    2.175E-24_WP,    1.710E-24_WP,    1.391E-24_WP, &
    & 1.126E-24_WP,    9.107E-25_WP,    7.377E-25_WP,    6.002E-25_WP,    4.931E-25_WP,    4.115E-25_WP,    3.270E-25_WP /
data J(3,    67    ,:)    / 2.529E-25_WP,    2.801E-25_WP,    3.106E-25_WP,    3.453E-25_WP,    3.853E-25_WP,    4.323E-25_WP, &
    & 4.884E-25_WP,    5.570E-25_WP,    6.426E-25_WP,    7.516E-25_WP,    8.942E-25_WP,    1.087E-24_WP,    1.359E-24_WP, &
    & 1.739E-24_WP,    2.239E-24_WP,    2.867E-24_WP,    3.576E-24_WP,    4.195E-24_WP,    4.600E-24_WP,    4.768E-24_WP, &
    & 4.627E-24_WP,    4.201E-24_WP,    3.595E-24_WP,    2.934E-24_WP,    2.320E-24_WP,    1.825E-24_WP,    1.484E-24_WP, &
    & 1.201E-24_WP,    9.709E-25_WP,    7.857E-25_WP,    6.387E-25_WP,    5.241E-25_WP,    4.365E-25_WP,    3.449E-25_WP /
data J(3,    68    ,:)    / 2.646E-25_WP,    2.930E-25_WP,    3.252E-25_WP,    3.620E-25_WP,    4.048E-25_WP,    4.556E-25_WP, &
    & 5.170E-25_WP,    5.926E-25_WP,    6.876E-25_WP,    8.096E-25_WP,    9.702E-25_WP,    1.188E-24_WP,    1.497E-24_WP, &
    & 1.930E-24_WP,    2.503E-24_WP,    3.225E-24_WP,    4.046E-24_WP,    4.767E-24_WP,    5.265E-24_WP,    5.493E-24_WP, &
    & 5.364E-24_WP,    4.899E-24_WP,    4.216E-24_WP,    3.460E-24_WP,    2.751E-24_WP,    2.177E-24_WP,    1.777E-24_WP, &
    & 1.443E-24_WP,    1.169E-24_WP,    9.481E-25_WP,    7.720E-25_WP,    6.347E-25_WP,    5.293E-25_WP,    4.121E-25_WP /
data J(3,    69    ,:)    / 2.861E-25_WP,    3.192E-25_WP,    3.575E-25_WP,    4.026E-25_WP,    4.566E-25_WP,    5.223E-25_WP, &
    & 6.036E-25_WP,    7.059E-25_WP,    8.369E-25_WP,    1.008E-24_WP,    1.236E-24_WP,    1.547E-24_WP,    1.990E-24_WP, &
    & 2.609E-24_WP,    3.431E-24_WP,    4.469E-24_WP,    5.655E-24_WP,    6.713E-24_WP,    7.541E-24_WP,    7.992E-24_WP, &
    & 7.923E-24_WP,    7.346E-24_WP,    6.416E-24_WP,    5.343E-24_WP,    4.308E-24_WP,    3.464E-24_WP,    2.857E-24_WP, &
    & 2.340E-24_WP,    1.910E-24_WP,    1.559E-24_WP,    1.277E-24_WP,    1.057E-24_WP,    8.866E-25_WP,    6.699E-25_WP /
data J(3,    70    ,:)    / 3.447E-25_WP,    3.958E-25_WP,    4.582E-25_WP,    5.354E-25_WP,    6.320E-25_WP,    7.543E-25_WP, &
    & 9.109E-25_WP,    1.114E-24_WP,    1.379E-24_WP,    1.732E-24_WP,    2.209E-24_WP,    2.867E-24_WP,    3.797E-24_WP, &
    & 5.088E-24_WP,    6.800E-24_WP,    8.957E-24_WP,    1.141E-23_WP,    1.366E-23_WP,    1.566E-23_WP,    1.690E-23_WP, &
    & 1.705E-23_WP,    1.609E-23_WP,    1.429E-23_WP,    1.210E-23_WP,    9.913E-24_WP,    8.104E-24_WP,    6.757E-24_WP, &
    & 5.585E-24_WP,    4.595E-24_WP,    3.777E-24_WP,    3.114E-24_WP,    2.593E-24_WP,    2.187E-24_WP,    1.599E-24_WP /
data J(3,    71    ,:)    / 3.447E-25_WP,    3.958E-25_WP,    4.582E-25_WP,    5.354E-25_WP,    6.320E-25_WP,    7.543E-25_WP, &
    & 9.109E-25_WP,    1.114E-24_WP,    1.379E-24_WP,    1.732E-24_WP,    2.209E-24_WP,    2.867E-24_WP,    3.797E-24_WP, &
    & 5.088E-24_WP,    6.800E-24_WP,    8.957E-24_WP,    1.141E-23_WP,    1.366E-23_WP,    1.566E-23_WP,    1.690E-23_WP, &
    & 1.705E-23_WP,    1.609E-23_WP,    1.429E-23_WP,    1.210E-23_WP,    9.913E-24_WP,    8.104E-24_WP,    6.757E-24_WP, &
    & 5.585E-24_WP,    4.595E-24_WP,    3.777E-24_WP,    3.114E-24_WP,    2.593E-24_WP,    2.187E-24_WP,    1.599E-24_WP /
data J(3,    72    ,:)    / 4.053E-25_WP,    4.739E-25_WP,    5.593E-25_WP,    6.671E-25_WP,    8.047E-25_WP,    9.823E-25_WP, &
    & 1.214E-24_WP,    1.520E-24_WP,    1.928E-24_WP,    2.483E-24_WP,    3.248E-24_WP,    4.325E-24_WP,    5.850E-24_WP, &
    & 7.958E-24_WP,    1.074E-23_WP,    1.418E-23_WP,    1.796E-23_WP,    2.206E-23_WP,    2.589E-23_WP,    2.882E-23_WP, &
    & 3.034E-23_WP,    3.037E-23_WP,    2.915E-23_WP,    2.718E-23_WP,    2.498E-23_WP,    2.321E-23_WP,    2.176E-23_WP, &
    & 2.022E-23_WP,    1.861E-23_WP,    1.697E-23_WP,    1.529E-23_WP,    1.360E-23_WP,    1.177E-23_WP,    9.638E-24_WP /
data J(3,    73    ,:)    / 4.783E-25_WP,    5.678E-25_WP,    6.809E-25_WP,    8.256E-25_WP,    1.013E-24_WP,    1.257E-24_WP, &
    & 1.580E-24_WP,    2.012E-24_WP,    2.597E-24_WP,    3.404E-24_WP,    4.532E-24_WP,    6.121E-24_WP,    8.348E-24_WP, &
    & 1.138E-23_WP,    1.527E-23_WP,    1.988E-23_WP,    2.529E-23_WP,    3.100E-23_WP,    3.633E-23_WP,    4.045E-23_WP, &
    & 4.278E-23_WP,    4.316E-23_WP,    4.190E-23_WP,    3.958E-23_WP,    3.698E-23_WP,    3.467E-23_WP,    3.249E-23_WP, &
    & 3.002E-23_WP,    2.734E-23_WP,    2.452E-23_WP,    2.162E-23_WP,    1.866E-23_WP,    1.561E-23_WP,    1.035E-23_WP /
data J(3,    74    ,:)    / 5.653E-25_WP,    6.796E-25_WP,    8.257E-25_WP,    1.014E-24_WP,    1.261E-24_WP,    1.586E-24_WP, &
    & 2.020E-24_WP,    2.606E-24_WP,    3.410E-24_WP,    4.529E-24_WP,    6.096E-24_WP,    8.283E-24_WP,    1.129E-23_WP, &
    & 1.528E-23_WP,    2.019E-23_WP,    2.621E-23_WP,    3.304E-23_WP,    4.011E-23_WP,    4.655E-23_WP,    5.143E-23_WP, &
    & 5.409E-23_WP,    5.436E-23_WP,    5.266E-23_WP,    4.974E-23_WP,    4.637E-23_WP,    4.315E-23_WP,    3.998E-23_WP, &
    & 3.643E-23_WP,    3.265E-23_WP,    2.876E-23_WP,    2.485E-23_WP,    2.100E-23_WP,    1.652E-23_WP,    1.110E-23_WP /
data J(3,    75    ,:)    / 6.681E-25_WP,    8.118E-25_WP,    9.970E-25_WP,    1.238E-24_WP,    1.555E-24_WP,    1.977E-24_WP, &
    & 2.546E-24_WP,    3.321E-24_WP,    4.394E-24_WP,    5.887E-24_WP,    7.961E-24_WP,    1.081E-23_WP,    1.463E-23_WP, &
    & 1.949E-23_WP,    2.561E-23_WP,    3.287E-23_WP,    4.093E-23_WP,    4.905E-23_WP,    5.628E-23_WP,    6.158E-23_WP, &
    & 6.423E-23_WP,    6.412E-23_WP,    6.180E-23_WP,    5.805E-23_WP,    5.370E-23_WP,    4.947E-23_WP,    4.529E-23_WP, &
    & 4.075E-23_WP,    3.605E-23_WP,    3.134E-23_WP,    2.674E-23_WP,    2.209E-23_WP,    1.743E-23_WP,    1.186E-23_WP /
data J(3,    76    ,:)    / 7.887E-25_WP,    9.670E-25_WP,    1.198E-24_WP,    1.502E-24_WP,    1.904E-24_WP,    2.442E-24_WP, &
    & 3.173E-24_WP,    4.179E-24_WP,    5.570E-24_WP,    7.492E-24_WP,    1.012E-23_WP,    1.365E-23_WP,    1.822E-23_WP, &
    & 2.410E-23_WP,    3.129E-23_WP,    3.963E-23_WP,    4.868E-23_WP,    5.762E-23_WP,    6.538E-23_WP,    7.087E-23_WP, &
    & 7.334E-23_WP,    7.275E-23_WP,    6.967E-23_WP,    6.496E-23_WP,    5.958E-23_WP,    5.438E-23_WP,    4.932E-23_WP, &
    & 4.397E-23_WP,    3.855E-23_WP,    3.325E-23_WP,    2.808E-23_WP,    2.318E-23_WP,    1.833E-23_WP,    1.263E-23_WP /
data J(3,    77    ,:)    / 9.157E-25_WP,    1.147E-24_WP,    1.452E-24_WP,    1.859E-24_WP,    2.406E-24_WP,    3.149E-24_WP, &
    & 4.168E-24_WP,    5.582E-24_WP,    7.556E-24_WP,    1.032E-23_WP,    1.417E-23_WP,    1.949E-23_WP,    2.670E-23_WP, &
    & 3.644E-23_WP,    4.910E-23_WP,    6.491E-23_WP,    8.353E-23_WP,    1.035E-22_WP,    1.269E-22_WP,    1.493E-22_WP, &
    & 1.678E-22_WP,    1.802E-22_WP,    1.860E-22_WP,    1.860E-22_WP,    1.820E-22_WP,    1.775E-22_WP,    1.746E-22_WP, &
    & 1.697E-22_WP,    1.630E-22_WP,    1.550E-22_WP,    1.460E-22_WP,    1.364E-22_WP,    1.259E-22_WP,    1.031E-22_WP /
data J(3,    78    ,:)    / 1.081E-24_WP,    1.362E-24_WP,    1.734E-24_WP,    2.232E-24_WP,    2.904E-24_WP,    3.822E-24_WP, &
    & 5.087E-24_WP,    6.843E-24_WP,    9.284E-24_WP,    1.267E-23_WP,    1.732E-23_WP,    2.360E-23_WP,    3.210E-23_WP, &
    & 4.321E-23_WP,    5.716E-23_WP,    7.377E-23_WP,    9.194E-23_WP,    1.130E-22_WP,    1.331E-22_WP,    1.494E-22_WP, &
    & 1.595E-22_WP,    1.625E-22_WP,    1.591E-22_WP,    1.512E-22_WP,    1.418E-22_WP,    1.330E-22_WP,    1.229E-22_WP, &
    & 1.112E-22_WP,    9.856E-23_WP,    8.544E-23_WP,    7.235E-23_WP,    5.952E-23_WP,    4.406E-23_WP,    1.419E-23_WP /
data J(3,    79    ,:)    / 1.274E-24_WP,    1.614E-24_WP,    2.067E-24_WP,    2.674E-24_WP,    3.498E-24_WP,    4.627E-24_WP, &
    & 6.185E-24_WP,    8.339E-24_WP,    1.131E-23_WP,    1.537E-23_WP,    2.083E-23_WP,    2.819E-23_WP,    3.784E-23_WP, &
    & 5.007E-23_WP,    6.479E-23_WP,    8.125E-23_WP,    1.005E-22_WP,    1.195E-22_WP,    1.354E-22_WP,    1.460E-22_WP, &
    & 1.498E-22_WP,    1.470E-22_WP,    1.390E-22_WP,    1.282E-22_WP,    1.169E-22_WP,    1.050E-22_WP,    9.289E-23_WP, &
    & 8.028E-23_WP,    6.781E-23_WP,    5.601E-23_WP,    4.519E-23_WP,    3.466E-23_WP,    2.210E-23_WP,    1.518E-23_WP /
data J(3,    80    ,:)    / 1.499E-24_WP,    1.907E-24_WP,    2.454E-24_WP,    3.190E-24_WP,    4.194E-24_WP,    5.571E-24_WP, &
    & 7.462E-24_WP,    1.006E-23_WP,    1.358E-23_WP,    1.831E-23_WP,    2.466E-23_WP,    3.295E-23_WP,    4.352E-23_WP, &
    & 5.639E-23_WP,    7.107E-23_WP,    8.851E-23_WP,    1.063E-22_WP,    1.224E-22_WP,    1.346E-22_WP,    1.411E-22_WP, &
    & 1.413E-22_WP,    1.357E-22_WP,    1.263E-22_WP,    1.148E-22_WP,    1.021E-22_WP,    8.983E-23_WP,    7.816E-23_WP, &
    & 6.677E-23_WP,    5.612E-23_WP,    4.648E-23_WP,    3.770E-23_WP,    2.929E-23_WP,    2.316E-23_WP,    1.601E-23_WP /
data J(3,    81    ,:)    / 1.757E-24_WP,    2.246E-24_WP,    2.902E-24_WP,    3.791E-24_WP,    5.002E-24_WP,    6.657E-24_WP, &
    & 8.914E-24_WP,    1.197E-23_WP,    1.604E-23_WP,    2.149E-23_WP,    2.859E-23_WP,    3.761E-23_WP,    4.869E-23_WP, &
    & 6.156E-23_WP,    7.706E-23_WP,    9.348E-23_WP,    1.094E-22_WP,    1.229E-22_WP,    1.325E-22_WP,    1.370E-22_WP, &
    & 1.358E-22_WP,    1.296E-22_WP,    1.200E-22_WP,    1.079E-22_WP,    9.526E-23_WP,    8.354E-23_WP,    7.293E-23_WP, &
    & 6.279E-23_WP,    5.340E-23_WP,    4.480E-23_WP,    3.691E-23_WP,    3.029E-23_WP,    2.395E-23_WP,    1.676E-23_WP /
data J(3,    82    ,:)    / 2.053E-24_WP,    2.636E-24_WP,    3.420E-24_WP,    4.483E-24_WP,    5.926E-24_WP,    7.884E-24_WP, &
    & 1.052E-23_WP,    1.402E-23_WP,    1.868E-23_WP,    2.473E-23_WP,    3.241E-23_WP,    4.183E-23_WP,    5.292E-23_WP, &
    & 6.646E-23_WP,    8.124E-23_WP,    9.632E-23_WP,    1.106E-22_WP,    1.227E-22_WP,    1.312E-22_WP,    1.352E-22_WP, &
    & 1.341E-22_WP,    1.281E-22_WP,    1.183E-22_WP,    1.062E-22_WP,    9.383E-23_WP,    8.264E-23_WP,    7.266E-23_WP, &
    & 6.303E-23_WP,    5.396E-23_WP,    4.554E-23_WP,    3.800E-23_WP,    3.110E-23_WP,    2.460E-23_WP,    1.745E-23_WP /
data J(3,    83    ,:)    / 2.393E-24_WP,    3.084E-24_WP,    4.015E-24_WP,    5.271E-24_WP,    6.966E-24_WP,    9.236E-24_WP, &
    & 1.224E-23_WP,    1.621E-23_WP,    2.136E-23_WP,    2.787E-23_WP,    3.586E-23_WP,    4.528E-23_WP,    5.690E-23_WP, &
    & 6.993E-23_WP,    8.377E-23_WP,    9.782E-23_WP,    1.113E-22_WP,    1.231E-22_WP,    1.318E-22_WP,    1.364E-22_WP, &
    & 1.357E-22_WP,    1.298E-22_WP,    1.200E-22_WP,    1.080E-22_WP,    9.561E-23_WP,    8.446E-23_WP,    7.449E-23_WP, &
    & 6.474E-23_WP,    5.545E-23_WP,    4.686E-23_WP,    3.898E-23_WP,    3.183E-23_WP,    2.519E-23_WP,    1.812E-23_WP /
data J(3,    84    ,:)    / 2.782E-24_WP,    3.596E-24_WP,    4.688E-24_WP,    6.153E-24_WP,    8.106E-24_WP,    1.068E-23_WP, &
    & 1.406E-23_WP,    1.843E-23_WP,    2.394E-23_WP,    3.070E-23_WP,    3.868E-23_WP,    4.853E-23_WP,    5.975E-23_WP, &
    & 7.212E-23_WP,    8.531E-23_WP,    9.901E-23_WP,    1.126E-22_WP,    1.250E-22_WP,    1.347E-22_WP,    1.401E-22_WP, &
    & 1.398E-22_WP,    1.341E-22_WP,    1.241E-22_WP,    1.118E-22_WP,    9.901E-23_WP,    8.744E-23_WP,    7.706E-23_WP, &
    & 6.687E-23_WP,    5.718E-23_WP,    4.817E-23_WP,    3.996E-23_WP,    3.256E-23_WP,    2.577E-23_WP,    1.876E-23_WP /
data J(3,    85    ,:)    / 3.224E-24_WP,    4.173E-24_WP,    5.439E-24_WP,    7.117E-24_WP,    9.316E-24_WP,    1.220E-23_WP, &
    & 1.590E-23_WP,    2.056E-23_WP,    2.627E-23_WP,    3.302E-23_WP,    4.133E-23_WP,    5.088E-23_WP,    6.163E-23_WP, &
    & 7.358E-23_WP,    8.668E-23_WP,    1.008E-22_WP,    1.152E-22_WP,    1.288E-22_WP,    1.396E-22_WP,    1.457E-22_WP, &
    & 1.458E-22_WP,    1.400E-22_WP,    1.297E-22_WP,    1.167E-22_WP,    1.032E-22_WP,    9.092E-23_WP,    7.995E-23_WP, &
    & 6.920E-23_WP,    5.900E-23_WP,    4.956E-23_WP,    4.100E-23_WP,    3.334E-23_WP,    2.637E-23_WP,    1.940E-23_WP /
data J(3,    86    ,:)    / 3.721E-24_WP,    4.815E-24_WP,    6.258E-24_WP,    8.139E-24_WP,    1.059E-23_WP,    1.374E-23_WP, &
    & 1.767E-23_WP,    2.249E-23_WP,    2.819E-23_WP,    3.520E-23_WP,    4.328E-23_WP,    5.247E-23_WP,    6.297E-23_WP, &
    & 7.497E-23_WP,    8.856E-23_WP,    1.036E-22_WP,    1.193E-22_WP,    1.342E-22_WP,    1.460E-22_WP,    1.528E-22_WP, &
    & 1.532E-22_WP,    1.472E-22_WP,    1.362E-22_WP,    1.224E-22_WP,    1.079E-22_WP,    9.477E-23_WP,    8.312E-23_WP, &
    & 7.175E-23_WP,    6.102E-23_WP,    5.112E-23_WP,    4.219E-23_WP,    3.423E-23_WP,    2.706E-23_WP,    2.007E-23_WP /
data J(3,    87    ,:)    / 4.289E-24_WP,    5.536E-24_WP,    7.154E-24_WP,    9.256E-24_WP,    1.194E-23_WP,    1.528E-23_WP, &
    & 1.937E-23_WP,    2.421E-23_WP,    3.015E-23_WP,    3.703E-23_WP,    4.491E-23_WP,    5.401E-23_WP,    6.472E-23_WP, &
    & 7.736E-23_WP,    9.203E-23_WP,    1.085E-22_WP,    1.259E-22_WP,    1.424E-22_WP,    1.556E-22_WP,    1.635E-22_WP, &
    & 1.643E-22_WP,    1.581E-22_WP,    1.465E-22_WP,    1.317E-22_WP,    1.161E-22_WP,    1.021E-22_WP,    8.970E-23_WP, &
    & 7.757E-23_WP,    6.609E-23_WP,    5.549E-23_WP,    4.590E-23_WP,    3.733E-23_WP,    2.962E-23_WP,    2.218E-23_WP /
data J(3,    88    ,:)    / 4.928E-24_WP,    6.329E-24_WP,    8.140E-24_WP,    1.044E-23_WP,    1.330E-23_WP,    1.679E-23_WP, &
    & 2.092E-23_WP,    2.598E-23_WP,    3.186E-23_WP,    3.865E-23_WP,    4.655E-23_WP,    5.596E-23_WP,    6.735E-23_WP, &
    & 8.112E-23_WP,    9.733E-23_WP,    1.156E-22_WP,    1.350E-22_WP,    1.535E-22_WP,    1.685E-22_WP,    1.776E-22_WP, &
    & 1.789E-22_WP,    1.725E-22_WP,    1.601E-22_WP,    1.441E-22_WP,    1.273E-22_WP,    1.121E-22_WP,    9.866E-23_WP, &
    & 8.548E-23_WP,    7.295E-23_WP,    6.136E-23_WP,    5.084E-23_WP,    4.146E-23_WP,    3.315E-23_WP,    2.476E-23_WP /
data J(3,    89    ,:)    / 5.615E-24_WP,    7.177E-24_WP,    9.152E-24_WP,    1.160E-23_WP,    1.458E-23_WP,    1.811E-23_WP, &
    & 2.243E-23_WP,    2.746E-23_WP,    3.329E-23_WP,    4.012E-23_WP,    4.832E-23_WP,    5.833E-23_WP,    7.071E-23_WP, &
    & 8.583E-23_WP,    1.037E-22_WP,    1.238E-22_WP,    1.452E-22_WP,    1.656E-22_WP,    1.821E-22_WP,    1.921E-22_WP, &
    & 1.936E-22_WP,    1.865E-22_WP,    1.729E-22_WP,    1.553E-22_WP,    1.369E-22_WP,    1.203E-22_WP,    1.055E-22_WP, &
    & 9.115E-23_WP,    7.754E-23_WP,    6.499E-23_WP,    5.367E-23_WP,    4.364E-23_WP,    3.478E-23_WP,    2.591E-23_WP /
data J(3,    90    ,:)    / 6.349E-24_WP,    8.050E-24_WP,    1.016E-23_WP,    1.271E-23_WP,    1.573E-23_WP,    1.942E-23_WP, &
    & 2.373E-23_WP,    2.873E-23_WP,    3.464E-23_WP,    4.175E-23_WP,    5.049E-23_WP,    6.135E-23_WP,    7.489E-23_WP, &
    & 9.147E-23_WP,    1.110E-22_WP,    1.331E-22_WP,    1.565E-22_WP,    1.787E-22_WP,    1.967E-22_WP,    2.075E-22_WP, &
    & 2.089E-22_WP,    2.011E-22_WP,    1.861E-22_WP,    1.669E-22_WP,    1.468E-22_WP,    1.286E-22_WP,    1.126E-22_WP, &
    & 9.708E-23_WP,    8.244E-23_WP,    6.900E-23_WP,    5.691E-23_WP,    4.623E-23_WP,    3.679E-23_WP,    2.733E-23_WP /
data J(3,    91    ,:)    / 7.108E-24_WP,    8.922E-24_WP,    1.112E-23_WP,    1.371E-23_WP,    1.688E-23_WP,    2.057E-23_WP, &
    & 2.488E-23_WP,    2.998E-23_WP,    3.615E-23_WP,    4.376E-23_WP,    5.324E-23_WP,    6.512E-23_WP,    7.994E-23_WP, &
    & 9.805E-23_WP,    1.194E-22_WP,    1.436E-22_WP,    1.691E-22_WP,    1.933E-22_WP,    2.128E-22_WP,    2.244E-22_WP, &
    & 2.259E-22_WP,    2.174E-22_WP,    2.009E-22_WP,    1.800E-22_WP,    1.581E-22_WP,    1.384E-22_WP,    1.211E-22_WP, &
    & 1.043E-22_WP,    8.856E-23_WP,    7.410E-23_WP,    6.110E-23_WP,    4.961E-23_WP,    3.945E-23_WP,    2.911E-23_WP /
data J(3,    92    ,:)    / 7.870E-24_WP,    9.765E-24_WP,    1.200E-23_WP,    1.473E-23_WP,    1.791E-23_WP,    2.164E-23_WP, &
    & 2.606E-23_WP,    3.142E-23_WP,    3.804E-23_WP,    4.631E-23_WP,    5.669E-23_WP,    6.969E-23_WP,    8.591E-23_WP, &
    & 1.057E-22_WP,    1.291E-22_WP,    1.555E-22_WP,    1.834E-22_WP,    2.098E-22_WP,    2.312E-22_WP,    2.440E-22_WP, &
    & 2.457E-22_WP,    2.364E-22_WP,    2.185E-22_WP,    1.958E-22_WP,    1.719E-22_WP,    1.504E-22_WP,    1.316E-22_WP, &
    & 1.134E-22_WP,    9.631E-23_WP,    8.059E-23_WP,    6.646E-23_WP,    5.397E-23_WP,    4.287E-23_WP,    3.135E-23_WP /
data J(3,    93    ,:)    / 8.617E-24_WP,    1.056E-23_WP,    1.291E-23_WP,    1.567E-23_WP,    1.891E-23_WP,    2.276E-23_WP, &
    & 2.743E-23_WP,    3.321E-23_WP,    4.043E-23_WP,    4.951E-23_WP,    6.090E-23_WP,    7.516E-23_WP,    9.296E-23_WP, &
    & 1.147E-22_WP,    1.404E-22_WP,    1.694E-22_WP,    2.000E-22_WP,    2.291E-22_WP,    2.527E-22_WP,    2.670E-22_WP, &
    & 2.691E-22_WP,    2.591E-22_WP,    2.397E-22_WP,    2.148E-22_WP,    1.887E-22_WP,    1.652E-22_WP,    1.447E-22_WP, &
    & 1.247E-22_WP,    1.059E-22_WP,    8.868E-23_WP,    7.316E-23_WP,    5.942E-23_WP,    4.717E-23_WP,    3.410E-23_WP /
data J(3,    94    ,:)    / 9.333E-24_WP,    1.138E-23_WP,    1.379E-23_WP,    1.661E-23_WP,    1.998E-23_WP,    2.408E-23_WP, &
    & 2.914E-23_WP,    3.547E-23_WP,    4.343E-23_WP,    5.343E-23_WP,    6.597E-23_WP,    8.169E-23_WP,    1.013E-22_WP, &
    & 1.253E-22_WP,    1.536E-22_WP,    1.856E-22_WP,    2.196E-22_WP,    2.519E-22_WP,    2.783E-22_WP,    2.944E-22_WP, &
    & 2.971E-22_WP,    2.863E-22_WP,    2.651E-22_WP,    2.378E-22_WP,    2.090E-22_WP,    1.831E-22_WP,    1.604E-22_WP, &
    & 1.384E-22_WP,    1.176E-22_WP,    9.853E-23_WP,    8.132E-23_WP,    6.606E-23_WP,    5.241E-23_WP,    3.744E-23_WP /
data J(3,    95    ,:)    / 1.009E-23_WP,    1.219E-23_WP,    1.468E-23_WP,    1.764E-23_WP,    2.125E-23_WP,    2.572E-23_WP, &
    & 3.130E-23_WP,    3.830E-23_WP,    4.710E-23_WP,    5.816E-23_WP,    7.204E-23_WP,    8.944E-23_WP,    1.112E-22_WP, &
    & 1.378E-22_WP,    1.692E-22_WP,    2.049E-22_WP,    2.427E-22_WP,    2.789E-22_WP,    3.085E-22_WP,    3.268E-22_WP, &
    & 3.302E-22_WP,    3.185E-22_WP,    2.952E-22_WP,    2.650E-22_WP,    2.331E-22_WP,    2.043E-22_WP,    1.792E-22_WP, &
    & 1.547E-22_WP,    1.315E-22_WP,    1.102E-22_WP,    9.098E-23_WP,    7.393E-23_WP,    5.861E-23_WP,    4.139E-23_WP /
data J(3,    96    ,:)    / 1.084E-23_WP,    1.304E-23_WP,    1.566E-23_WP,    1.886E-23_WP,    2.282E-23_WP,    2.776E-23_WP, &
    & 3.396E-23_WP,    4.175E-23_WP,    5.153E-23_WP,    6.382E-23_WP,    7.926E-23_WP,    9.863E-23_WP,    1.228E-22_WP, &
    & 1.525E-22_WP,    1.877E-22_WP,    2.276E-22_WP,    2.700E-22_WP,    3.106E-22_WP,    3.440E-22_WP,    3.647E-22_WP, &
    & 3.688E-22_WP,    3.561E-22_WP,    3.302E-22_WP,    2.967E-22_WP,    2.612E-22_WP,    2.290E-22_WP,    2.009E-22_WP, &
    & 1.734E-22_WP,    1.475E-22_WP,    1.236E-22_WP,    1.021E-22_WP,    8.300E-23_WP,    6.576E-23_WP,    4.594E-23_WP /
data J(3,    97    ,:)    / 1.095E-23_WP,    1.317E-23_WP,    1.583E-23_WP,    1.907E-23_WP,    2.308E-23_WP,    2.809E-23_WP, &
    & 3.436E-23_WP,    4.224E-23_WP,    5.212E-23_WP,    6.454E-23_WP,    8.013E-23_WP,    9.968E-23_WP,    1.241E-22_WP, &
    & 1.541E-22_WP,    1.896E-22_WP,    2.299E-22_WP,    2.728E-22_WP,    3.138E-22_WP,    3.476E-22_WP,    3.686E-22_WP, &
    & 3.727E-22_WP,    3.599E-22_WP,    3.337E-22_WP,    2.999E-22_WP,    2.640E-22_WP,    2.315E-22_WP,    2.031E-22_WP, &
    & 1.753E-22_WP,    1.491E-22_WP,    1.250E-22_WP,    1.032E-22_WP,    8.391E-23_WP,    6.648E-23_WP,    4.640E-23_WP /
data J(3,    98    ,:)    / 1.199E-23_WP,    1.444E-23_WP,    1.743E-23_WP,    2.115E-23_WP,    2.581E-23_WP,    3.169E-23_WP, &
    & 3.912E-23_WP,    4.853E-23_WP,    6.048E-23_WP,    7.567E-23_WP,    9.501E-23_WP,    1.197E-22_WP,    1.510E-22_WP, &
    & 1.903E-22_WP,    2.378E-22_WP,    2.932E-22_WP,    3.541E-22_WP,    4.163E-22_WP,    4.730E-22_WP,    5.157E-22_WP, &
    & 5.380E-22_WP,    5.384E-22_WP,    5.203E-22_WP,    4.904E-22_WP,    4.565E-22_WP,    4.276E-22_WP,    4.017E-22_WP, &
    & 3.741E-22_WP,    3.457E-22_WP,    3.171E-22_WP,    2.886E-22_WP,    2.600E-22_WP,    2.278E-22_WP,    1.716E-22_WP /
data J(3,    99    ,:)    / 1.316E-23_WP,    1.591E-23_WP,    1.933E-23_WP,    2.364E-23_WP,    2.909E-23_WP,    3.600E-23_WP, &
    & 4.481E-23_WP,    5.607E-23_WP,    7.050E-23_WP,    8.907E-23_WP,    1.130E-22_WP,    1.439E-22_WP,    1.838E-22_WP, &
    & 2.341E-22_WP,    2.958E-22_WP,    3.685E-22_WP,    4.501E-22_WP,    5.352E-22_WP,    6.143E-22_WP,    6.765E-22_WP, &
    & 7.136E-22_WP,    7.231E-22_WP,    7.095E-22_WP,    6.813E-22_WP,    6.497E-22_WP,    6.223E-22_WP,    5.962E-22_WP, &
    & 5.662E-22_WP,    5.331E-22_WP,    4.977E-22_WP,    4.602E-22_WP,    4.177E-22_WP,    3.535E-22_WP,    1.902E-22_WP /
data J(3,    100    ,:)    / 1.449E-23_WP,    1.762E-23_WP,    2.157E-23_WP,    2.657E-23_WP,    3.294E-23_WP,    4.107E-23_WP, &
    & 5.152E-23_WP,    6.498E-23_WP,    8.241E-23_WP,    1.051E-22_WP,    1.346E-22_WP,    1.730E-22_WP,    2.228E-22_WP, &
    & 2.861E-22_WP,    3.637E-22_WP,    4.558E-22_WP,    5.593E-22_WP,    6.667E-22_WP,    7.663E-22_WP,    8.449E-22_WP, &
    & 8.930E-22_WP,    9.085E-22_WP,    8.973E-22_WP,    8.710E-22_WP,    8.401E-22_WP,    8.120E-22_WP,    7.836E-22_WP, &
    & 7.491E-22_WP,    7.095E-22_WP,    6.654E-22_WP,    6.149E-22_WP,    5.441E-22_WP,    3.882E-22_WP,    2.077E-22_WP /
data J(3,    101    ,:)    / 1.601E-23_WP,    1.959E-23_WP,    2.414E-23_WP,    2.994E-23_WP,    3.737E-23_WP,    4.693E-23_WP, &
    & 5.930E-23_WP,    7.537E-23_WP,    9.636E-23_WP,    1.239E-22_WP,    1.599E-22_WP,    2.071E-22_WP,    2.682E-22_WP, &
    & 3.456E-22_WP,    4.406E-22_WP,    5.527E-22_WP,    6.771E-22_WP,    8.044E-22_WP,    9.207E-22_WP,    1.012E-21_WP, &
    & 1.067E-21_WP,    1.087E-21_WP,    1.078E-21_WP,    1.052E-21_WP,    1.020E-21_WP,    9.896E-22_WP,    9.572E-22_WP, &
    & 9.167E-22_WP,    8.692E-22_WP,    8.134E-22_WP,    7.381E-22_WP,    5.890E-22_WP,    4.175E-22_WP,    2.211E-22_WP /
data J(3,    102    ,:)    / 1.773E-23_WP,    2.184E-23_WP,    2.708E-23_WP,    3.380E-23_WP,    4.245E-23_WP,    5.367E-23_WP, &
    & 6.829E-23_WP,    8.743E-23_WP,    1.126E-22_WP,    1.458E-22_WP,    1.894E-22_WP,    2.463E-22_WP,    3.198E-22_WP, &
    & 4.124E-22_WP,    5.252E-22_WP,    6.561E-22_WP,    7.990E-22_WP,    9.427E-22_WP,    1.072E-21_WP,    1.172E-21_WP, &
    & 1.233E-21_WP,    1.256E-21_WP,    1.248E-21_WP,    1.222E-21_WP,    1.187E-21_WP,    1.153E-21_WP,    1.115E-21_WP, &
    & 1.068E-21_WP,    1.010E-21_WP,    9.326E-22_WP,    7.906E-22_WP,    6.274E-22_WP,    4.414E-22_WP,    2.328E-22_WP /
data J(3,    103    ,:)    / 1.875E-23_WP,    2.327E-23_WP,    2.910E-23_WP,    3.665E-23_WP,    4.648E-23_WP,    5.933E-23_WP, &
    & 7.625E-23_WP,    9.865E-23_WP,    1.284E-22_WP,    1.681E-22_WP,    2.210E-22_WP,    2.911E-22_WP,    3.831E-22_WP, &
    & 5.017E-22_WP,    6.499E-22_WP,    8.281E-22_WP,    1.031E-21_WP,    1.247E-21_WP,    1.468E-21_WP,    1.665E-21_WP, &
    & 1.821E-21_WP,    1.927E-21_WP,    1.985E-21_WP,    2.007E-21_WP,    2.008E-21_WP,    2.009E-21_WP,    2.008E-21_WP, &
    & 1.984E-21_WP,    1.938E-21_WP,    1.859E-21_WP,    1.702E-21_WP,    1.510E-21_WP,    1.275E-21_WP,    9.433E-22_WP /
data J(3,    104    ,:)    / 2.091E-23_WP,    2.612E-23_WP,    3.286E-23_WP,    4.163E-23_WP,    5.312E-23_WP,    6.825E-23_WP, &
    & 8.829E-23_WP,    1.150E-22_WP,    1.506E-22_WP,    1.982E-22_WP,    2.614E-22_WP,    3.449E-22_WP,    4.539E-22_WP, &
    & 5.925E-22_WP,    7.625E-22_WP,    9.619E-22_WP,    1.182E-21_WP,    1.418E-21_WP,    1.641E-21_WP,    1.830E-21_WP, &
    & 1.968E-21_WP,    2.050E-21_WP,    2.084E-21_WP,    2.086E-21_WP,    2.078E-21_WP,    2.070E-21_WP,    2.047E-21_WP, &
    & 2.001E-21_WP,    1.923E-21_WP,    1.776E-21_WP,    1.597E-21_WP,    1.379E-21_WP,    1.077E-21_WP,    2.572E-22_WP /
data J(3,    105    ,:)    / 2.340E-23_WP,    2.941E-23_WP,    3.722E-23_WP,    4.744E-23_WP,    6.090E-23_WP,    7.874E-23_WP, &
    & 1.025E-22_WP,    1.343E-22_WP,    1.768E-22_WP,    2.335E-22_WP,    3.085E-22_WP,    4.070E-22_WP,    5.337E-22_WP, &
    & 6.921E-22_WP,    8.820E-22_WP,    1.099E-21_WP,    1.339E-21_WP,    1.581E-21_WP,    1.803E-21_WP,    1.981E-21_WP, &
    & 2.104E-21_WP,    2.169E-21_WP,    2.188E-21_WP,    2.184E-21_WP,    2.169E-21_WP,    2.144E-21_WP,    2.102E-21_WP, &
    & 2.029E-21_WP,    1.893E-21_WP,    1.727E-21_WP,    1.524E-21_WP,    1.247E-21_WP,    5.313E-22_WP,    2.740E-22_WP /
data J(3,    106    ,:)    / 2.625E-23_WP,    3.318E-23_WP,    4.223E-23_WP,    5.415E-23_WP,    6.995E-23_WP,    9.099E-23_WP, &
    & 1.191E-22_WP,    1.568E-22_WP,    2.071E-22_WP,    2.739E-22_WP,    3.619E-22_WP,    4.758E-22_WP,    6.198E-22_WP, &
    & 7.959E-22_WP,    1.002E-21_WP,    1.237E-21_WP,    1.486E-21_WP,    1.730E-21_WP,    1.946E-21_WP,    2.116E-21_WP, &
    & 2.227E-21_WP,    2.283E-21_WP,    2.298E-21_WP,    2.288E-21_WP,    2.258E-21_WP,    2.216E-21_WP,    2.152E-21_WP, &
    & 2.031E-21_WP,    1.878E-21_WP,    1.690E-21_WP,    1.437E-21_WP,    8.110E-22_WP,    5.662E-22_WP,    2.891E-22_WP /
data J(3,    107    ,:)    / 2.950E-23_WP,    3.749E-23_WP,    4.800E-23_WP,    6.192E-23_WP,    8.045E-23_WP,    1.052E-22_WP, &
    & 1.384E-22_WP,    1.828E-22_WP,    2.417E-22_WP,    3.196E-22_WP,    4.208E-22_WP,    5.496E-22_WP,    7.091E-22_WP, &
    & 8.994E-22_WP,    1.122E-21_WP,    1.367E-21_WP,    1.619E-21_WP,    1.862E-21_WP,    2.074E-21_WP,    2.238E-21_WP, &
    & 2.345E-21_WP,    2.398E-21_WP,    2.410E-21_WP,    2.387E-21_WP,    2.343E-21_WP,    2.283E-21_WP,    2.178E-21_WP, &
    & 2.041E-21_WP,    1.870E-21_WP,    1.639E-21_WP,    1.091E-21_WP,    8.595E-22_WP,    5.966E-22_WP,    3.039E-22_WP /
data J(3,    108    ,:)    / 3.322E-23_WP,    4.245E-23_WP,    5.466E-23_WP,    7.091E-23_WP,    9.262E-23_WP,    1.217E-22_WP, &
    & 1.606E-22_WP,    2.123E-22_WP,    2.807E-22_WP,    3.699E-22_WP,    4.839E-22_WP,    6.261E-22_WP,    7.982E-22_WP, &
    & 1.004E-21_WP,    1.235E-21_WP,    1.484E-21_WP,    1.738E-21_WP,    1.981E-21_WP,    2.192E-21_WP,    2.355E-21_WP, &
    & 2.463E-21_WP,    2.516E-21_WP,    2.518E-21_WP,    2.483E-21_WP,    2.422E-21_WP,    2.327E-21_WP,    2.209E-21_WP, &
    & 2.056E-21_WP,    1.848E-21_WP,    1.368E-21_WP,    1.151E-21_WP,    9.033E-22_WP,    6.258E-22_WP,    3.189E-22_WP /
data J(3,    109    ,:)    / 3.748E-23_WP,    4.816E-23_WP,    6.235E-23_WP,    8.131E-23_WP,    1.067E-22_WP,    1.406E-22_WP, &
    & 1.858E-22_WP,    2.455E-22_WP,    3.236E-22_WP,    4.238E-22_WP,    5.493E-22_WP,    7.024E-22_WP,    8.879E-22_WP, &
    & 1.102E-21_WP,    1.338E-21_WP,    1.590E-21_WP,    1.846E-21_WP,    2.091E-21_WP,    2.305E-21_WP,    2.473E-21_WP, &
    & 2.583E-21_WP,    2.629E-21_WP,    2.622E-21_WP,    2.574E-21_WP,    2.483E-21_WP,    2.376E-21_WP,    2.244E-21_WP, &
    & 2.060E-21_WP,    1.641E-21_WP,    1.439E-21_WP,    1.207E-21_WP,    9.461E-22_WP,    6.553E-22_WP,    3.345E-22_WP /
data J(3,    110    ,:)    / 4.237E-23_WP,    5.474E-23_WP,    7.124E-23_WP,    9.330E-23_WP,    1.228E-22_WP,    1.621E-22_WP, &
    & 2.141E-22_WP,    2.821E-22_WP,    3.695E-22_WP,    4.796E-22_WP,    6.146E-22_WP,    7.793E-22_WP,    9.722E-22_WP, &
    & 1.191E-21_WP,    1.432E-21_WP,    1.687E-21_WP,    1.947E-21_WP,    2.197E-21_WP,    2.419E-21_WP,    2.593E-21_WP, &
    & 2.701E-21_WP,    2.742E-21_WP,    2.724E-21_WP,    2.650E-21_WP,    2.549E-21_WP,    2.430E-21_WP,    2.271E-21_WP, &
    & 1.909E-21_WP,    1.724E-21_WP,    1.508E-21_WP,    1.263E-21_WP,    9.896E-22_WP,    6.858E-22_WP,    3.508E-22_WP /
data J(3,    111    ,:)    / 4.800E-23_WP,    6.233E-23_WP,    8.147E-23_WP,    1.070E-22_WP,    1.411E-22_WP,    1.861E-22_WP, &
    & 2.452E-22_WP,    3.212E-22_WP,    4.171E-22_WP,    5.353E-22_WP,    6.804E-22_WP,    8.518E-22_WP,    1.050E-21_WP, &
    & 1.273E-21_WP,    1.518E-21_WP,    1.779E-21_WP,    2.046E-21_WP,    2.306E-21_WP,    2.536E-21_WP,    2.712E-21_WP, &
    & 2.818E-21_WP,    2.852E-21_WP,    2.814E-21_WP,    2.732E-21_WP,    2.620E-21_WP,    2.478E-21_WP,    2.169E-21_WP, &
    & 2.002E-21_WP,    1.804E-21_WP,    1.577E-21_WP,    1.321E-21_WP,    1.035E-21_WP,    7.175E-22_WP,    3.679E-22_WP /
data J(3,    112    ,:)    / 5.448E-23_WP,    7.105E-23_WP,    9.317E-23_WP,    1.226E-22_WP,    1.616E-22_WP,    2.126E-22_WP, &
    & 2.784E-22_WP,    3.617E-22_WP,    4.646E-22_WP,    5.915E-22_WP,    7.425E-22_WP,    9.184E-22_WP,    1.120E-21_WP, &
    & 1.348E-21_WP,    1.599E-21_WP,    1.868E-21_WP,    2.146E-21_WP,    2.417E-21_WP,    2.653E-21_WP,    2.832E-21_WP, &
    & 2.936E-21_WP,    2.955E-21_WP,    2.910E-21_WP,    2.819E-21_WP,    2.687E-21_WP,    2.417E-21_WP,    2.273E-21_WP, &
    & 2.095E-21_WP,    1.887E-21_WP,    1.648E-21_WP,    1.380E-21_WP,    1.082E-21_WP,    7.508E-22_WP,    3.859E-22_WP /
data J(3,    113    ,:)    / 6.192E-23_WP,    8.103E-23_WP,    1.064E-22_WP,    1.400E-22_WP,    1.841E-22_WP,    2.409E-22_WP, &
    & 3.129E-22_WP,    4.021E-22_WP,    5.126E-22_WP,    6.446E-22_WP,    7.998E-22_WP,    9.796E-22_WP,    1.186E-21_WP, &
    & 1.420E-21_WP,    1.679E-21_WP,    1.959E-21_WP,    2.250E-21_WP,    2.529E-21_WP,    2.772E-21_WP,    2.952E-21_WP, &
    & 3.048E-21_WP,    3.064E-21_WP,    3.013E-21_WP,    2.904E-21_WP,    2.664E-21_WP,    2.533E-21_WP,    2.378E-21_WP, &
    & 2.191E-21_WP,    1.972E-21_WP,    1.723E-21_WP,    1.443E-21_WP,    1.131E-21_WP,    7.859E-22_WP,    4.049E-22_WP /
data J(3,    114    ,:)    / 7.040E-23_WP,    9.228E-23_WP,    1.212E-22_WP,    1.591E-22_WP,    2.081E-22_WP,    2.701E-22_WP, &
    & 3.473E-22_WP,    4.429E-22_WP,    5.578E-22_WP,    6.935E-22_WP,    8.522E-22_WP,    1.036E-21_WP,    1.248E-21_WP, &
    & 1.490E-21_WP,    1.760E-21_WP,    2.052E-21_WP,    2.352E-21_WP,    2.639E-21_WP,    2.887E-21_WP,    3.064E-21_WP, &
    & 3.159E-21_WP,    3.172E-21_WP,    3.107E-21_WP,    2.908E-21_WP,    2.783E-21_WP,    2.643E-21_WP,    2.480E-21_WP, &
    & 2.284E-21_WP,    2.055E-21_WP,    1.795E-21_WP,    1.503E-21_WP,    1.178E-21_WP,    8.173E-22_WP,    4.199E-22_WP /
data J(3,    115    ,:)    / 7.991E-23_WP,    1.048E-22_WP,    1.374E-22_WP,    1.794E-22_WP,    2.328E-22_WP,    2.992E-22_WP, &
    & 3.817E-22_WP,    4.812E-22_WP,    5.993E-22_WP,    7.381E-22_WP,    9.003E-22_WP,    1.089E-21_WP,    1.308E-21_WP, &
    & 1.559E-21_WP,    1.839E-21_WP,    2.141E-21_WP,    2.449E-21_WP,    2.743E-21_WP,    2.989E-21_WP,    3.168E-21_WP, &
    & 3.262E-21_WP,    3.264E-21_WP,    3.123E-21_WP,    3.020E-21_WP,    2.888E-21_WP,    2.740E-21_WP,    2.570E-21_WP, &
    & 2.365E-21_WP,    2.126E-21_WP,    1.855E-21_WP,    1.551E-21_WP,    1.213E-21_WP,    8.383E-22_WP,    4.283E-22_WP /
data J(3,    116    ,:)    / 9.047E-23_WP,    1.184E-22_WP,    1.545E-22_WP,    2.003E-22_WP,    2.573E-22_WP,    3.283E-22_WP, &
    & 4.141E-22_WP,    5.162E-22_WP,    6.369E-22_WP,    7.788E-22_WP,    9.453E-22_WP,    1.140E-21_WP,    1.367E-21_WP, &
    & 1.627E-21_WP,    1.916E-21_WP,    2.225E-21_WP,    2.539E-21_WP,    2.833E-21_WP,    3.082E-21_WP,    3.261E-21_WP, &
    & 3.347E-21_WP,    3.286E-21_WP,    3.227E-21_WP,    3.118E-21_WP,    2.979E-21_WP,    2.825E-21_WP,    2.647E-21_WP, &
    & 2.433E-21_WP,    2.185E-21_WP,    1.904E-21_WP,    1.588E-21_WP,    1.237E-21_WP,    8.516E-22_WP,    4.347E-22_WP /
data J(3,    117    ,:)    / 1.020E-22_WP,    1.328E-22_WP,    1.721E-22_WP,    2.210E-22_WP,    2.819E-22_WP,    3.556E-22_WP, &
    & 4.437E-22_WP,    5.481E-22_WP,    6.714E-22_WP,    8.170E-22_WP,    9.885E-22_WP,    1.190E-21_WP,    1.425E-21_WP, &
    & 1.693E-21_WP,    1.988E-21_WP,    2.303E-21_WP,    2.618E-21_WP,    2.916E-21_WP,    3.167E-21_WP,    3.340E-21_WP, &
    & 3.375E-21_WP,    3.383E-21_WP,    3.319E-21_WP,    3.204E-21_WP,    3.059E-21_WP,    2.898E-21_WP,    2.713E-21_WP, &
    & 2.491E-21_WP,    2.234E-21_WP,    1.942E-21_WP,    1.616E-21_WP,    1.255E-21_WP,    8.631E-22_WP,    4.412E-22_WP /
data J(3,    118    ,:)    / 1.142E-22_WP,    1.477E-22_WP,    1.896E-22_WP,    2.417E-22_WP,    3.050E-22_WP,    3.807E-22_WP, &
    & 4.707E-22_WP,    5.773E-22_WP,    7.038E-22_WP,    8.536E-22_WP,    1.031E-21_WP,    1.240E-21_WP,    1.481E-21_WP, &
    & 1.754E-21_WP,    2.055E-21_WP,    2.372E-21_WP,    2.691E-21_WP,    2.991E-21_WP,    3.240E-21_WP,    3.371E-21_WP, &
    & 3.465E-21_WP,    3.470E-21_WP,    3.401E-21_WP,    3.281E-21_WP,    3.130E-21_WP,    2.962E-21_WP,    2.770E-21_WP, &
    & 2.540E-21_WP,    2.274E-21_WP,    1.973E-21_WP,    1.638E-21_WP,    1.272E-21_WP,    8.749E-22_WP,    4.479E-22_WP /
data J(3,    119    ,:)    / 1.267E-22_WP,    1.625E-22_WP,    2.071E-22_WP,    2.612E-22_WP,    3.261E-22_WP,    4.035E-22_WP, &
    & 4.954E-22_WP,    6.047E-22_WP,    7.349E-22_WP,    8.899E-22_WP,    1.073E-21_WP,    1.287E-21_WP,    1.534E-21_WP, &
    & 1.812E-21_WP,    2.115E-21_WP,    2.435E-21_WP,    2.758E-21_WP,    3.057E-21_WP,    3.272E-21_WP,    3.455E-21_WP, &
    & 3.547E-21_WP,    3.549E-21_WP,    3.476E-21_WP,    3.351E-21_WP,    3.194E-21_WP,    3.020E-21_WP,    2.821E-21_WP, &
    & 2.583E-21_WP,    2.309E-21_WP,    2.000E-21_WP,    1.659E-21_WP,    1.289E-21_WP,    8.872E-22_WP,    4.547E-22_WP /
data J(3,    120    ,:)    / 1.392E-22_WP,    1.773E-22_WP,    2.235E-22_WP,    2.791E-22_WP,    3.454E-22_WP,    4.244E-22_WP, &
    & 5.186E-22_WP,    6.311E-22_WP,    7.656E-22_WP,    9.256E-22_WP,    1.114E-21_WP,    1.332E-21_WP,    1.583E-21_WP, &
    & 1.863E-21_WP,    2.170E-21_WP,    2.494E-21_WP,    2.817E-21_WP,    3.089E-21_WP,    3.348E-21_WP,    3.532E-21_WP, &
    & 3.623E-21_WP,    3.622E-21_WP,    3.546E-21_WP,    3.415E-21_WP,    3.253E-21_WP,    3.073E-21_WP,    2.867E-21_WP, &
    & 2.622E-21_WP,    2.341E-21_WP,    2.027E-21_WP,    1.682E-21_WP,    1.306E-21_WP,    8.998E-22_WP,    4.617E-22_WP /
data J(3,    121    ,:)    / 1.516E-22_WP,    1.912E-22_WP,    2.386E-22_WP,    2.954E-22_WP,    3.631E-22_WP,    4.440E-22_WP, &
    & 5.409E-22_WP,    6.571E-22_WP,    7.959E-22_WP,    9.598E-22_WP,    1.152E-21_WP,    1.374E-21_WP,    1.627E-21_WP, &
    & 1.911E-21_WP,    2.221E-21_WP,    2.547E-21_WP,    2.848E-21_WP,    3.159E-21_WP,    3.420E-21_WP,    3.604E-21_WP, &
    & 3.694E-21_WP,    3.691E-21_WP,    3.610E-21_WP,    3.475E-21_WP,    3.307E-21_WP,    3.121E-21_WP,    2.909E-21_WP, &
    & 2.658E-21_WP,    2.372E-21_WP,    2.054E-21_WP,    1.705E-21_WP,    1.325E-21_WP,    9.128E-22_WP,    4.690E-22_WP /
data J(3,    122    ,:)    / 1.634E-22_WP,    2.039E-22_WP,    2.524E-22_WP,    3.103E-22_WP,    3.797E-22_WP,    4.629E-22_WP, &
    & 5.629E-22_WP,    6.828E-22_WP,    8.249E-22_WP,    9.924E-22_WP,    1.188E-21_WP,    1.412E-21_WP,    1.668E-21_WP, &
    & 1.956E-21_WP,    2.267E-21_WP,    2.575E-21_WP,    2.910E-21_WP,    3.224E-21_WP,    3.487E-21_WP,    3.672E-21_WP, &
    & 3.762E-21_WP,    3.756E-21_WP,    3.672E-21_WP,    3.532E-21_WP,    3.359E-21_WP,    3.167E-21_WP,    2.950E-21_WP, &
    & 2.694E-21_WP,    2.405E-21_WP,    2.082E-21_WP,    1.729E-21_WP,    1.343E-21_WP,    9.262E-22_WP,    4.763E-22_WP /
data J(3,    123    ,:)    / 1.741E-22_WP,    2.155E-22_WP,    2.650E-22_WP,    3.244E-22_WP,    3.957E-22_WP,    4.816E-22_WP, &
    & 5.847E-22_WP,    7.074E-22_WP,    8.526E-22_WP,    1.023E-21_WP,    1.220E-21_WP,    1.447E-21_WP,    1.707E-21_WP, &
    & 1.995E-21_WP,    2.293E-21_WP,    2.630E-21_WP,    2.969E-21_WP,    3.286E-21_WP,    3.552E-21_WP,    3.738E-21_WP, &
    & 3.827E-21_WP,    3.819E-21_WP,    3.731E-21_WP,    3.587E-21_WP,    3.408E-21_WP,    3.212E-21_WP,    2.991E-21_WP, &
    & 2.732E-21_WP,    2.438E-21_WP,    2.112E-21_WP,    1.753E-21_WP,    1.363E-21_WP,    9.400E-22_WP,    4.839E-22_WP /
data J(3,    124    ,:)    / 1.840E-22_WP,    2.262E-22_WP,    2.769E-22_WP,    3.379E-22_WP,    4.115E-22_WP,    5.000E-22_WP, &
    & 6.056E-22_WP,    7.309E-22_WP,    8.786E-22_WP,    1.050E-21_WP,    1.250E-21_WP,    1.480E-21_WP,    1.741E-21_WP, &
    & 2.018E-21_WP,    2.341E-21_WP,    2.682E-21_WP,    3.025E-21_WP,    3.345E-21_WP,    3.614E-21_WP,    3.801E-21_WP, &
    & 3.890E-21_WP,    3.880E-21_WP,    3.789E-21_WP,    3.640E-21_WP,    3.457E-21_WP,    3.257E-21_WP,    3.033E-21_WP, &
    & 2.771E-21_WP,    2.473E-21_WP,    2.142E-21_WP,    1.779E-21_WP,    1.383E-21_WP,    9.542E-22_WP,    4.916E-22_WP /
data J(3,    125    ,:)    / 1.930E-22_WP,    2.363E-22_WP,    2.884E-22_WP,    3.513E-22_WP,    4.271E-22_WP,    5.177E-22_WP, &
    & 6.255E-22_WP,    7.530E-22_WP,    9.019E-22_WP,    1.076E-21_WP,    1.278E-21_WP,    1.510E-21_WP,    1.761E-21_WP, &
    & 2.060E-21_WP,    2.387E-21_WP,    2.732E-21_WP,    3.079E-21_WP,    3.403E-21_WP,    3.675E-21_WP,    3.863E-21_WP, &
    & 3.952E-21_WP,    3.940E-21_WP,    3.846E-21_WP,    3.693E-21_WP,    3.507E-21_WP,    3.304E-21_WP,    3.077E-21_WP, &
    & 2.811E-21_WP,    2.509E-21_WP,    2.173E-21_WP,    1.805E-21_WP,    1.403E-21_WP,    9.688E-22_WP,    4.995E-22_WP /
data J(3,    126    ,:)    / 2.015E-22_WP,    2.460E-22_WP,    2.997E-22_WP,    3.645E-22_WP,    4.421E-22_WP,    5.346E-22_WP, &
    & 6.443E-22_WP,    7.728E-22_WP,    9.238E-22_WP,    1.100E-21_WP,    1.303E-21_WP,    1.527E-21_WP,    1.797E-21_WP, &
    & 2.100E-21_WP,    2.431E-21_WP,    2.780E-21_WP,    3.132E-21_WP,    3.460E-21_WP,    3.734E-21_WP,    3.925E-21_WP, &
    & 4.013E-21_WP,    4.000E-21_WP,    3.902E-21_WP,    3.747E-21_WP,    3.558E-21_WP,    3.352E-21_WP,    3.122E-21_WP, &
    & 2.852E-21_WP,    2.546E-21_WP,    2.205E-21_WP,    1.832E-21_WP,    1.425E-21_WP,    9.837E-22_WP,    5.076E-22_WP /
data J(3,    127    ,:)    / 2.097E-22_WP,    2.556E-22_WP,    3.109E-22_WP,    3.772E-22_WP,    4.565E-22_WP,    5.505E-22_WP, &
    & 6.611E-22_WP,    7.915E-22_WP,    9.443E-22_WP,    1.122E-21_WP,    1.319E-21_WP,    1.558E-21_WP,    1.832E-21_WP, &
    & 2.139E-21_WP,    2.474E-21_WP,    2.828E-21_WP,    3.184E-21_WP,    3.516E-21_WP,    3.794E-21_WP,    3.986E-21_WP, &
    & 4.074E-21_WP,    4.060E-21_WP,    3.961E-21_WP,    3.803E-21_WP,    3.611E-21_WP,    3.402E-21_WP,    3.168E-21_WP, &
    & 2.894E-21_WP,    2.584E-21_WP,    2.238E-21_WP,    1.859E-21_WP,    1.446E-21_WP,    9.990E-22_WP,    5.159E-22_WP /
data J(3,    128    ,:)    / 2.178E-22_WP,    2.650E-22_WP,    3.217E-22_WP,    3.894E-22_WP,    4.699E-22_WP,    5.649E-22_WP, &
    & 6.770E-22_WP,    8.089E-22_WP,    9.629E-22_WP,    1.135E-21_WP,    1.346E-21_WP,    1.588E-21_WP,    1.865E-21_WP, &
    & 2.176E-21_WP,    2.516E-21_WP,    2.875E-21_WP,    3.235E-21_WP,    3.572E-21_WP,    3.853E-21_WP,    4.047E-21_WP, &
    & 4.136E-21_WP,    4.121E-21_WP,    4.020E-21_WP,    3.860E-21_WP,    3.665E-21_WP,    3.453E-21_WP,    3.216E-21_WP, &
    & 2.938E-21_WP,    2.623E-21_WP,    2.272E-21_WP,    1.888E-21_WP,    1.469E-21_WP,    1.015E-21_WP,    5.243E-22_WP /
data J(3,    129    ,:)    / 2.259E-22_WP,    2.742E-22_WP,    3.320E-22_WP,    4.008E-22_WP,    4.821E-22_WP,    5.783E-22_WP, &
    & 6.919E-22_WP,    8.248E-22_WP,    9.744E-22_WP,    1.158E-21_WP,    1.371E-21_WP,    1.617E-21_WP,    1.898E-21_WP, &
    & 2.213E-21_WP,    2.557E-21_WP,    2.921E-21_WP,    3.287E-21_WP,    3.628E-21_WP,    3.913E-21_WP,    4.109E-21_WP, &
    & 4.199E-21_WP,    4.184E-21_WP,    4.082E-21_WP,    3.919E-21_WP,    3.721E-21_WP,    3.506E-21_WP,    3.264E-21_WP, &
    & 2.982E-21_WP,    2.663E-21_WP,    2.307E-21_WP,    1.917E-21_WP,    1.492E-21_WP,    1.031E-21_WP,    5.330E-22_WP /
data J(3,    130    ,:)    / 2.336E-22_WP,    2.829E-22_WP,    3.417E-22_WP,    4.111E-22_WP,    4.936E-22_WP,    5.910E-22_WP, &
    & 7.054E-22_WP,    8.347E-22_WP,    9.940E-22_WP,    1.180E-21_WP,    1.396E-21_WP,    1.645E-21_WP,    1.930E-21_WP, &
    & 2.249E-21_WP,    2.599E-21_WP,    2.968E-21_WP,    3.339E-21_WP,    3.685E-21_WP,    3.973E-21_WP,    4.173E-21_WP, &
    & 4.264E-21_WP,    4.248E-21_WP,    4.144E-21_WP,    3.979E-21_WP,    3.778E-21_WP,    3.560E-21_WP,    3.315E-21_WP, &
    & 3.028E-21_WP,    2.704E-21_WP,    2.343E-21_WP,    1.947E-21_WP,    1.515E-21_WP,    1.047E-21_WP,    5.418E-22_WP /
data J(3,    131    ,:)    / 2.410E-22_WP,    2.911E-22_WP,    3.504E-22_WP,    4.209E-22_WP,    5.043E-22_WP,    6.025E-22_WP, &
    & 7.139E-22_WP,    8.515E-22_WP,    1.013E-21_WP,    1.201E-21_WP,    1.420E-21_WP,    1.673E-21_WP,    1.962E-21_WP, &
    & 2.286E-21_WP,    2.640E-21_WP,    3.014E-21_WP,    3.391E-21_WP,    3.742E-21_WP,    4.035E-21_WP,    4.238E-21_WP, &
    & 4.331E-21_WP,    4.315E-21_WP,    4.209E-21_WP,    4.041E-21_WP,    3.837E-21_WP,    3.615E-21_WP,    3.366E-21_WP, &
    & 3.076E-21_WP,    2.746E-21_WP,    2.380E-21_WP,    1.977E-21_WP,    1.539E-21_WP,    1.064E-21_WP,    5.508E-22_WP /
data J(3,    132    ,:)    / 2.479E-22_WP,    2.985E-22_WP,    3.587E-22_WP,    4.300E-22_WP,    5.142E-22_WP,    6.099E-22_WP, &
    & 7.282E-22_WP,    8.676E-22_WP,    1.031E-21_WP,    1.222E-21_WP,    1.444E-21_WP,    1.701E-21_WP,    1.993E-21_WP, &
    & 2.322E-21_WP,    2.682E-21_WP,    3.062E-21_WP,    3.443E-21_WP,    3.800E-21_WP,    4.098E-21_WP,    4.304E-21_WP, &
    & 4.399E-21_WP,    4.382E-21_WP,    4.275E-21_WP,    4.104E-21_WP,    3.897E-21_WP,    3.672E-21_WP,    3.419E-21_WP, &
    & 3.124E-21_WP,    2.789E-21_WP,    2.417E-21_WP,    2.009E-21_WP,    1.564E-21_WP,    1.081E-21_WP,    5.599E-22_WP /
data J(3,    133    ,:)    / 2.542E-22_WP,    3.055E-22_WP,    3.665E-22_WP,    4.384E-22_WP,    5.204E-22_WP,    6.220E-22_WP, &
    & 7.420E-22_WP,    8.832E-22_WP,    1.049E-21_WP,    1.243E-21_WP,    1.468E-21_WP,    1.728E-21_WP,    2.025E-21_WP, &
    & 2.359E-21_WP,    2.724E-21_WP,    3.109E-21_WP,    3.497E-21_WP,    3.860E-21_WP,    4.163E-21_WP,    4.372E-21_WP, &
    & 4.468E-21_WP,    4.452E-21_WP,    4.343E-21_WP,    4.169E-21_WP,    3.958E-21_WP,    3.730E-21_WP,    3.473E-21_WP, &
    & 3.173E-21_WP,    2.834E-21_WP,    2.456E-21_WP,    2.041E-21_WP,    1.589E-21_WP,    1.099E-21_WP,    5.693E-22_WP /
data J(3,    134    ,:)    / 2.602E-22_WP,    3.121E-22_WP,    3.736E-22_WP,    4.438E-22_WP,    5.308E-22_WP,    6.337E-22_WP, &
    & 7.553E-22_WP,    8.985E-22_WP,    1.067E-21_WP,    1.263E-21_WP,    1.492E-21_WP,    1.756E-21_WP,    2.057E-21_WP, &
    & 2.396E-21_WP,    2.766E-21_WP,    3.158E-21_WP,    3.553E-21_WP,    3.921E-21_WP,    4.229E-21_WP,    4.442E-21_WP, &
    & 4.540E-21_WP,    4.523E-21_WP,    4.412E-21_WP,    4.236E-21_WP,    4.022E-21_WP,    3.789E-21_WP,    3.528E-21_WP, &
    & 3.224E-21_WP,    2.879E-21_WP,    2.495E-21_WP,    2.074E-21_WP,    1.614E-21_WP,    1.117E-21_WP,    5.789E-22_WP /
data J(3,    135    ,:)    / 2.658E-22_WP,    3.182E-22_WP,    3.782E-22_WP,    4.526E-22_WP,    5.408E-22_WP,    6.451E-22_WP, &
    & 7.684E-22_WP,    9.136E-22_WP,    1.084E-21_WP,    1.284E-21_WP,    1.516E-21_WP,    1.783E-21_WP,    2.090E-21_WP, &
    & 2.434E-21_WP,    2.810E-21_WP,    3.208E-21_WP,    3.609E-21_WP,    3.984E-21_WP,    4.297E-21_WP,    4.514E-21_WP, &
    & 4.613E-21_WP,    4.596E-21_WP,    4.483E-21_WP,    4.304E-21_WP,    4.086E-21_WP,    3.850E-21_WP,    3.585E-21_WP, &
    & 3.276E-21_WP,    2.925E-21_WP,    2.535E-21_WP,    2.107E-21_WP,    1.641E-21_WP,    1.135E-21_WP,    5.886E-22_WP /
data J(3,    136    ,:)    / 2.709E-22_WP,    3.221E-22_WP,    3.857E-22_WP,    4.611E-22_WP,    5.505E-22_WP,    6.563E-22_WP, &
    & 7.813E-22_WP,    9.286E-22_WP,    1.102E-21_WP,    1.304E-21_WP,    1.540E-21_WP,    1.812E-21_WP,    2.122E-21_WP, &
    & 2.472E-21_WP,    2.855E-21_WP,    3.260E-21_WP,    3.667E-21_WP,    4.048E-21_WP,    4.366E-21_WP,    4.587E-21_WP, &
    & 4.688E-21_WP,    4.670E-21_WP,    4.556E-21_WP,    4.373E-21_WP,    4.152E-21_WP,    3.912E-21_WP,    3.643E-21_WP, &
    & 3.329E-21_WP,    2.973E-21_WP,    2.576E-21_WP,    2.141E-21_WP,    1.668E-21_WP,    1.154E-21_WP,    5.985E-22_WP /
data J(3,    137    ,:)    / 2.743E-22_WP,    3.285E-22_WP,    3.929E-22_WP,    4.693E-22_WP,    5.600E-22_WP,    6.673E-22_WP, &
    & 7.942E-22_WP,    9.437E-22_WP,    1.119E-21_WP,    1.325E-21_WP,    1.564E-21_WP,    1.840E-21_WP,    2.156E-21_WP, &
    & 2.511E-21_WP,    2.900E-21_WP,    3.312E-21_WP,    3.727E-21_WP,    4.114E-21_WP,    4.437E-21_WP,    4.661E-21_WP, &
    & 4.764E-21_WP,    4.746E-21_WP,    4.630E-21_WP,    4.444E-21_WP,    4.220E-21_WP,    3.976E-21_WP,    3.702E-21_WP, &
    & 3.383E-21_WP,    3.021E-21_WP,    2.619E-21_WP,    2.177E-21_WP,    1.695E-21_WP,    1.173E-21_WP,    6.087E-22_WP /
data J(3,    138    ,:)    / 2.797E-22_WP,    3.347E-22_WP,    3.999E-22_WP,    4.775E-22_WP,    5.694E-22_WP,    6.783E-22_WP, &
    & 8.070E-22_WP,    9.588E-22_WP,    1.137E-21_WP,    1.346E-21_WP,    1.589E-21_WP,    1.869E-21_WP,    2.190E-21_WP, &
    & 2.552E-21_WP,    2.947E-21_WP,    3.366E-21_WP,    3.787E-21_WP,    4.181E-21_WP,    4.510E-21_WP,    4.738E-21_WP, &
    & 4.842E-21_WP,    4.824E-21_WP,    4.706E-21_WP,    4.517E-21_WP,    4.289E-21_WP,    4.041E-21_WP,    3.763E-21_WP, &
    & 3.438E-21_WP,    3.071E-21_WP,    2.662E-21_WP,    2.212E-21_WP,    1.723E-21_WP,    1.193E-21_WP,    6.190E-22_WP /
data J(3,    139    ,:)    / 2.850E-22_WP,    3.407E-22_WP,    4.069E-22_WP,    4.855E-22_WP,    5.788E-22_WP,    6.893E-22_WP, &
    & 8.200E-22_WP,    9.740E-22_WP,    1.155E-21_WP,    1.367E-21_WP,    1.614E-21_WP,    1.899E-21_WP,    2.226E-21_WP, &
    & 2.593E-21_WP,    2.995E-21_WP,    3.421E-21_WP,    3.849E-21_WP,    4.250E-21_WP,    4.584E-21_WP,    4.815E-21_WP, &
    & 4.922E-21_WP,    4.903E-21_WP,    4.783E-21_WP,    4.591E-21_WP,    4.359E-21_WP,    4.107E-21_WP,    3.825E-21_WP, &
    & 3.495E-21_WP,    3.121E-21_WP,    2.705E-21_WP,    2.249E-21_WP,    1.752E-21_WP,    1.213E-21_WP,    6.295E-22_WP /
data J(3,    140    ,:)    / 2.901E-22_WP,    3.465E-22_WP,    4.137E-22_WP,    4.935E-22_WP,    5.882E-22_WP,    7.004E-22_WP, &
    & 8.330E-22_WP,    9.894E-22_WP,    1.173E-21_WP,    1.389E-21_WP,    1.640E-21_WP,    1.930E-21_WP,    2.262E-21_WP, &
    & 2.635E-21_WP,    3.044E-21_WP,    3.477E-21_WP,    3.912E-21_WP,    4.320E-21_WP,    4.660E-21_WP,    4.895E-21_WP, &
    & 5.003E-21_WP,    4.985E-21_WP,    4.862E-21_WP,    4.667E-21_WP,    4.431E-21_WP,    4.175E-21_WP,    3.888E-21_WP, &
    & 3.553E-21_WP,    3.173E-21_WP,    2.750E-21_WP,    2.286E-21_WP,    1.781E-21_WP,    1.233E-21_WP,    6.402E-22_WP /
data J(3,    141    ,:)    / 2.951E-22_WP,    3.524E-22_WP,    4.205E-22_WP,    5.015E-22_WP,    5.976E-22_WP,    7.115E-22_WP, &
    & 8.462E-22_WP,    1.005E-21_WP,    1.192E-21_WP,    1.411E-21_WP,    1.666E-21_WP,    1.961E-21_WP,    2.298E-21_WP, &
    & 2.678E-21_WP,    3.094E-21_WP,    3.534E-21_WP,    3.977E-21_WP,    4.391E-21_WP,    4.737E-21_WP,    4.976E-21_WP, &
    & 5.086E-21_WP,    5.067E-21_WP,    4.943E-21_WP,    4.745E-21_WP,    4.505E-21_WP,    4.244E-21_WP,    3.952E-21_WP, &
    & 3.612E-21_WP,    3.226E-21_WP,    2.796E-21_WP,    2.325E-21_WP,    1.811E-21_WP,    1.254E-21_WP,    6.512E-22_WP /
data J(3,    142    ,:)    / 3.001E-22_WP,    3.582E-22_WP,    4.274E-22_WP,    5.096E-22_WP,    6.072E-22_WP,    7.228E-22_WP, &
    & 8.596E-22_WP,    1.021E-21_WP,    1.211E-21_WP,    1.434E-21_WP,    1.693E-21_WP,    1.993E-21_WP,    2.336E-21_WP, &
    & 2.722E-21_WP,    3.145E-21_WP,    3.592E-21_WP,    4.043E-21_WP,    4.464E-21_WP,    4.816E-21_WP,    5.059E-21_WP, &
    & 5.171E-21_WP,    5.152E-21_WP,    5.025E-21_WP,    4.824E-21_WP,    4.580E-21_WP,    4.315E-21_WP,    4.018E-21_WP, &
    & 3.672E-21_WP,    3.279E-21_WP,    2.843E-21_WP,    2.363E-21_WP,    1.841E-21_WP,    1.275E-21_WP,    6.623E-22_WP /
data J(3,    143    ,:)    / 3.050E-22_WP,    3.640E-22_WP,    4.342E-22_WP,    5.177E-22_WP,    6.168E-22_WP,    7.343E-22_WP, &
    & 8.733E-22_WP,    1.038E-21_WP,    1.231E-21_WP,    1.457E-21_WP,    1.721E-21_WP,    2.025E-21_WP,    2.374E-21_WP, &
    & 2.767E-21_WP,    3.197E-21_WP,    3.652E-21_WP,    4.110E-21_WP,    4.539E-21_WP,    4.896E-21_WP,    5.144E-21_WP, &
    & 5.257E-21_WP,    5.238E-21_WP,    5.109E-21_WP,    4.905E-21_WP,    4.656E-21_WP,    4.387E-21_WP,    4.086E-21_WP, &
    & 3.733E-21_WP,    3.334E-21_WP,    2.891E-21_WP,    2.403E-21_WP,    1.872E-21_WP,    1.297E-21_WP,    6.736E-22_WP /
data J(3,    144    ,:)    / 3.100E-22_WP,    3.699E-22_WP,    4.412E-22_WP,    5.259E-22_WP,    6.266E-22_WP,    7.460E-22_WP, &
    & 8.874E-22_WP,    1.054E-21_WP,    1.251E-21_WP,    1.481E-21_WP,    1.749E-21_WP,    2.059E-21_WP,    2.414E-21_WP, &
    & 2.813E-21_WP,    3.250E-21_WP,    3.713E-21_WP,    4.179E-21_WP,    4.615E-21_WP,    4.978E-21_WP,    5.230E-21_WP, &
    & 5.346E-21_WP,    5.326E-21_WP,    5.195E-21_WP,    4.987E-21_WP,    4.734E-21_WP,    4.461E-21_WP,    4.154E-21_WP, &
    & 3.796E-21_WP,    3.390E-21_WP,    2.939E-21_WP,    2.444E-21_WP,    1.904E-21_WP,    1.319E-21_WP,    6.852E-22_WP /
data J(3,    145    ,:)    / 3.150E-22_WP,    3.758E-22_WP,    4.482E-22_WP,    5.343E-22_WP,    6.366E-22_WP,    7.580E-22_WP, &
    & 9.017E-22_WP,    1.071E-21_WP,    1.271E-21_WP,    1.505E-21_WP,    1.778E-21_WP,    2.093E-21_WP,    2.454E-21_WP, &
    & 2.860E-21_WP,    3.304E-21_WP,    3.775E-21_WP,    4.249E-21_WP,    4.692E-21_WP,    5.062E-21_WP,    5.318E-21_WP, &
    & 5.436E-21_WP,    5.416E-21_WP,    5.283E-21_WP,    5.071E-21_WP,    4.814E-21_WP,    4.536E-21_WP,    4.224E-21_WP, &
    & 3.860E-21_WP,    3.448E-21_WP,    2.989E-21_WP,    2.485E-21_WP,    1.936E-21_WP,    1.341E-21_WP,    6.969E-22_WP /
data J(3,    146    ,:)    / 3.200E-22_WP,    3.818E-22_WP,    4.553E-22_WP,    5.428E-22_WP,    6.469E-22_WP,    7.703E-22_WP, &
    & 9.164E-22_WP,    1.089E-21_WP,    1.292E-21_WP,    1.530E-21_WP,    1.807E-21_WP,    2.128E-21_WP,    2.495E-21_WP, &
    & 2.907E-21_WP,    3.360E-21_WP,    3.839E-21_WP,    4.321E-21_WP,    4.771E-21_WP,    5.148E-21_WP,    5.408E-21_WP, &
    & 5.528E-21_WP,    5.507E-21_WP,    5.372E-21_WP,    5.156E-21_WP,    4.895E-21_WP,    4.612E-21_WP,    4.296E-21_WP, &
    & 3.925E-21_WP,    3.506E-21_WP,    3.039E-21_WP,    2.527E-21_WP,    1.969E-21_WP,    1.364E-21_WP,    7.089E-22_WP /
data J(3,    147    ,:)    / 3.251E-22_WP,    3.879E-22_WP,    4.626E-22_WP,    5.516E-22_WP,    6.573E-22_WP,    7.828E-22_WP, &
    & 9.314E-22_WP,    1.107E-21_WP,    1.313E-21_WP,    1.555E-21_WP,    1.837E-21_WP,    2.163E-21_WP,    2.536E-21_WP, &
    & 2.956E-21_WP,    3.416E-21_WP,    3.903E-21_WP,    4.394E-21_WP,    4.852E-21_WP,    5.235E-21_WP,    5.500E-21_WP, &
    & 5.621E-21_WP,    5.601E-21_WP,    5.463E-21_WP,    5.244E-21_WP,    4.978E-21_WP,    4.691E-21_WP,    4.368E-21_WP, &
    & 3.992E-21_WP,    3.565E-21_WP,    3.091E-21_WP,    2.570E-21_WP,    2.003E-21_WP,    1.387E-21_WP,    7.211E-22_WP /
data J(3,    148    ,:)    / 3.303E-22_WP,    3.941E-22_WP,    4.701E-22_WP,    5.605E-22_WP,    6.680E-22_WP,    7.956E-22_WP, &
    & 9.467E-22_WP,    1.125E-21_WP,    1.335E-21_WP,    1.581E-21_WP,    1.868E-21_WP,    2.199E-21_WP,    2.579E-21_WP, &
    & 3.006E-21_WP,    3.474E-21_WP,    3.969E-21_WP,    4.468E-21_WP,    4.934E-21_WP,    5.324E-21_WP,    5.593E-21_WP, &
    & 5.717E-21_WP,    5.696E-21_WP,    5.556E-21_WP,    5.333E-21_WP,    5.063E-21_WP,    4.770E-21_WP,    4.442E-21_WP, &
    & 4.060E-21_WP,    3.626E-21_WP,    3.144E-21_WP,    2.614E-21_WP,    2.037E-21_WP,    1.411E-21_WP,    7.336E-22_WP /
data J(3,    149    ,:)    / 3.356E-22_WP,    4.004E-22_WP,    4.777E-22_WP,    5.697E-22_WP,    6.790E-22_WP,    8.087E-22_WP, &
    & 9.623E-22_WP,    1.144E-21_WP,    1.357E-21_WP,    1.607E-21_WP,    1.899E-21_WP,    2.236E-21_WP,    2.622E-21_WP, &
    & 3.057E-21_WP,    3.533E-21_WP,    4.036E-21_WP,    4.544E-21_WP,    5.018E-21_WP,    5.414E-21_WP,    5.688E-21_WP, &
    & 5.814E-21_WP,    5.793E-21_WP,    5.650E-21_WP,    5.424E-21_WP,    5.149E-21_WP,    4.851E-21_WP,    4.518E-21_WP, &
    & 4.129E-21_WP,    3.688E-21_WP,    3.197E-21_WP,    2.658E-21_WP,    2.072E-21_WP,    1.435E-21_WP,    7.462E-22_WP /
data J(3,    150    ,:)    / 3.410E-22_WP,    4.069E-22_WP,    4.855E-22_WP,    5.790E-22_WP,    6.902E-22_WP,    8.221E-22_WP, &
    & 9.783E-22_WP,    1.163E-21_WP,    1.380E-21_WP,    1.634E-21_WP,    1.931E-21_WP,    2.274E-21_WP,    2.667E-21_WP, &
    & 3.108E-21_WP,    3.593E-21_WP,    4.105E-21_WP,    4.621E-21_WP,    5.103E-21_WP,    5.506E-21_WP,    5.785E-21_WP, &
    & 5.913E-21_WP,    5.891E-21_WP,    5.746E-21_WP,    5.516E-21_WP,    5.237E-21_WP,    4.934E-21_WP,    4.595E-21_WP, &
    & 4.199E-21_WP,    3.751E-21_WP,    3.252E-21_WP,    2.704E-21_WP,    2.107E-21_WP,    1.460E-21_WP,    7.591E-22_WP /
data J(3,    151    ,:)    / 3.466E-22_WP,    4.136E-22_WP,    4.935E-22_WP,    5.886E-22_WP,    7.016E-22_WP,    8.357E-22_WP, &
    & 9.946E-22_WP,    1.182E-21_WP,    1.403E-21_WP,    1.662E-21_WP,    1.963E-21_WP,    2.312E-21_WP,    2.712E-21_WP, &
    & 3.161E-21_WP,    3.654E-21_WP,    4.175E-21_WP,    4.700E-21_WP,    5.191E-21_WP,    5.600E-21_WP,    5.884E-21_WP, &
    & 6.014E-21_WP,    5.992E-21_WP,    5.845E-21_WP,    5.610E-21_WP,    5.326E-21_WP,    5.018E-21_WP,    4.674E-21_WP, &
    & 4.271E-21_WP,    3.815E-21_WP,    3.307E-21_WP,    2.750E-21_WP,    2.143E-21_WP,    1.485E-21_WP,    7.722E-22_WP /
data J(3,    152    ,:)    / 3.522E-22_WP,    4.204E-22_WP,    5.016E-22_WP,    5.983E-22_WP,    7.132E-22_WP,    8.497E-22_WP, &
    & 1.011E-21_WP,    1.202E-21_WP,    1.427E-21_WP,    1.690E-21_WP,    1.997E-21_WP,    2.351E-21_WP,    2.758E-21_WP, &
    & 3.215E-21_WP,    3.716E-21_WP,    4.246E-21_WP,    4.780E-21_WP,    5.279E-21_WP,    5.696E-21_WP,    5.985E-21_WP, &
    & 6.117E-21_WP,    6.095E-21_WP,    5.945E-21_WP,    5.706E-21_WP,    5.417E-21_WP,    5.104E-21_WP,    4.754E-21_WP, &
    & 4.344E-21_WP,    3.880E-21_WP,    3.364E-21_WP,    2.797E-21_WP,    2.180E-21_WP,    1.511E-21_WP,    7.856E-22_WP /
data J(3,    153    ,:)    / 3.580E-22_WP,    4.273E-22_WP,    5.099E-22_WP,    6.082E-22_WP,    7.251E-22_WP,    8.639E-22_WP, &
    & 1.028E-21_WP,    1.222E-21_WP,    1.451E-21_WP,    1.718E-21_WP,    2.030E-21_WP,    2.391E-21_WP,    2.805E-21_WP, &
    & 3.270E-21_WP,    3.779E-21_WP,    4.319E-21_WP,    4.862E-21_WP,    5.370E-21_WP,    5.794E-21_WP,    6.087E-21_WP, &
    & 6.222E-21_WP,    6.199E-21_WP,    6.046E-21_WP,    5.804E-21_WP,    5.510E-21_WP,    5.192E-21_WP,    4.835E-21_WP, &
    & 4.419E-21_WP,    3.947E-21_WP,    3.422E-21_WP,    2.845E-21_WP,    2.218E-21_WP,    1.537E-21_WP,    7.992E-22_WP /
data J(3,    154    ,:)    / 3.639E-22_WP,    4.344E-22_WP,    5.184E-22_WP,    6.184E-22_WP,    7.373E-22_WP,    8.784E-22_WP, &
    & 1.045E-21_WP,    1.243E-21_WP,    1.475E-21_WP,    1.748E-21_WP,    2.065E-21_WP,    2.432E-21_WP,    2.852E-21_WP, &
    & 3.325E-21_WP,    3.844E-21_WP,    4.392E-21_WP,    4.945E-21_WP,    5.462E-21_WP,    5.893E-21_WP,    6.192E-21_WP, &
    & 6.329E-21_WP,    6.306E-21_WP,    6.150E-21_WP,    5.904E-21_WP,    5.605E-21_WP,    5.281E-21_WP,    4.918E-21_WP, &
    & 4.494E-21_WP,    4.014E-21_WP,    3.481E-21_WP,    2.894E-21_WP,    2.256E-21_WP,    1.563E-21_WP,    8.130E-22_WP /
data J(3,    155    ,:)    / 3.699E-22_WP,    4.416E-22_WP,    5.270E-22_WP,    6.287E-22_WP,    7.497E-22_WP,    8.932E-22_WP, &
    & 1.063E-21_WP,    1.264E-21_WP,    1.500E-21_WP,    1.777E-21_WP,    2.100E-21_WP,    2.474E-21_WP,    2.901E-21_WP, &
    & 3.382E-21_WP,    3.910E-21_WP,    4.468E-21_WP,    5.030E-21_WP,    5.555E-21_WP,    5.994E-21_WP,    6.298E-21_WP, &
    & 6.438E-21_WP,    6.414E-21_WP,    6.256E-21_WP,    6.005E-21_WP,    5.701E-21_WP,    5.372E-21_WP,    5.003E-21_WP, &
    & 4.572E-21_WP,    4.083E-21_WP,    3.541E-21_WP,    2.944E-21_WP,    2.295E-21_WP,    1.590E-21_WP,    8.271E-22_WP /
data J(3,    156    ,:)    / 3.761E-22_WP,    4.490E-22_WP,    5.359E-22_WP,    6.393E-22_WP,    7.623E-22_WP,    9.083E-22_WP, &
    & 1.081E-21_WP,    1.285E-21_WP,    1.526E-21_WP,    1.808E-21_WP,    2.136E-21_WP,    2.516E-21_WP,    2.951E-21_WP, &
    & 3.440E-21_WP,    3.977E-21_WP,    4.544E-21_WP,    5.116E-21_WP,    5.651E-21_WP,    6.097E-21_WP,    6.406E-21_WP, &
    & 6.548E-21_WP,    6.524E-21_WP,    6.364E-21_WP,    6.108E-21_WP,    5.799E-21_WP,    5.464E-21_WP,    5.089E-21_WP, &
    & 4.650E-21_WP,    4.154E-21_WP,    3.602E-21_WP,    2.995E-21_WP,    2.334E-21_WP,    1.618E-21_WP,    8.414E-22_WP /
data J(3,    157    ,:)    / 3.824E-22_WP,    4.565E-22_WP,    5.449E-22_WP,    6.501E-22_WP,    7.752E-22_WP,    9.236E-22_WP, &
    & 1.099E-21_WP,    1.307E-21_WP,    1.552E-21_WP,    1.838E-21_WP,    2.172E-21_WP,    2.559E-21_WP,    3.001E-21_WP, &
    & 3.499E-21_WP,    4.045E-21_WP,    4.623E-21_WP,    5.204E-21_WP,    5.748E-21_WP,    6.202E-21_WP,    6.517E-21_WP, &
    & 6.661E-21_WP,    6.637E-21_WP,    6.473E-21_WP,    6.214E-21_WP,    5.899E-21_WP,    5.558E-21_WP,    5.176E-21_WP, &
    & 4.731E-21_WP,    4.225E-21_WP,    3.664E-21_WP,    3.047E-21_WP,    2.374E-21_WP,    1.645E-21_WP,    8.560E-22_WP /
data J(3,    158    ,:)    / 3.888E-22_WP,    4.642E-22_WP,    5.541E-22_WP,    6.611E-22_WP,    7.883E-22_WP,    9.393E-22_WP, &
    & 1.118E-21_WP,    1.329E-21_WP,    1.578E-21_WP,    1.870E-21_WP,    2.210E-21_WP,    2.603E-21_WP,    3.053E-21_WP, &
    & 3.559E-21_WP,    4.114E-21_WP,    4.702E-21_WP,    5.294E-21_WP,    5.847E-21_WP,    6.309E-21_WP,    6.629E-21_WP, &
    & 6.776E-21_WP,    6.751E-21_WP,    6.585E-21_WP,    6.321E-21_WP,    6.001E-21_WP,    5.654E-21_WP,    5.266E-21_WP, &
    & 4.812E-21_WP,    4.298E-21_WP,    3.727E-21_WP,    3.099E-21_WP,    2.416E-21_WP,    1.674E-21_WP,    8.708E-22_WP /
data J(3,    159    ,:)    / 3.953E-22_WP,    4.720E-22_WP,    5.634E-22_WP,    6.723E-22_WP,    8.017E-22_WP,    9.553E-22_WP, &
    & 1.137E-21_WP,    1.352E-21_WP,    1.605E-21_WP,    1.902E-21_WP,    2.247E-21_WP,    2.647E-21_WP,    3.105E-21_WP, &
    & 3.621E-21_WP,    4.185E-21_WP,    4.783E-21_WP,    5.385E-21_WP,    5.948E-21_WP,    6.418E-21_WP,    6.744E-21_WP, &
    & 6.893E-21_WP,    6.868E-21_WP,    6.699E-21_WP,    6.430E-21_WP,    6.104E-21_WP,    5.752E-21_WP,    5.357E-21_WP, &
    & 4.895E-21_WP,    4.373E-21_WP,    3.791E-21_WP,    3.153E-21_WP,    2.457E-21_WP,    1.703E-21_WP,    8.859E-22_WP /
data J(3,    160    ,:)    / 4.020E-22_WP,    4.800E-22_WP,    5.730E-22_WP,    6.837E-22_WP,    8.153E-22_WP,    9.716E-22_WP, &
    & 1.157E-21_WP,    1.375E-21_WP,    1.633E-21_WP,    1.934E-21_WP,    2.286E-21_WP,    2.693E-21_WP,    3.159E-21_WP, &
    & 3.683E-21_WP,    4.257E-21_WP,    4.866E-21_WP,    5.478E-21_WP,    6.051E-21_WP,    6.529E-21_WP,    6.860E-21_WP, &
    & 7.012E-21_WP,    6.986E-21_WP,    6.814E-21_WP,    6.541E-21_WP,    6.210E-21_WP,    5.851E-21_WP,    5.449E-21_WP, &
    & 4.980E-21_WP,    4.448E-21_WP,    3.857E-21_WP,    3.207E-21_WP,    2.500E-21_WP,    1.732E-21_WP,    9.013E-22_WP /
data J(3,    161    ,:)    / 4.088E-22_WP,    4.881E-22_WP,    5.827E-22_WP,    6.953E-22_WP,    8.292E-22_WP,    9.882E-22_WP, &
    & 1.176E-21_WP,    1.399E-21_WP,    1.661E-21_WP,    1.968E-21_WP,    2.325E-21_WP,    2.739E-21_WP,    3.213E-21_WP, &
    & 3.746E-21_WP,    4.331E-21_WP,    4.950E-21_WP,    5.573E-21_WP,    6.155E-21_WP,    6.642E-21_WP,    6.979E-21_WP, &
    & 7.133E-21_WP,    7.107E-21_WP,    6.932E-21_WP,    6.654E-21_WP,    6.317E-21_WP,    5.952E-21_WP,    5.543E-21_WP, &
    & 5.066E-21_WP,    4.525E-21_WP,    3.924E-21_WP,    3.263E-21_WP,    2.543E-21_WP,    1.762E-21_WP,    9.170E-22_WP /
data J(3,    162    ,:)    / 4.157E-22_WP,    4.964E-22_WP,    5.926E-22_WP,    7.072E-22_WP,    8.434E-22_WP,    1.005E-21_WP, &
    & 1.197E-21_WP,    1.423E-21_WP,    1.689E-21_WP,    2.001E-21_WP,    2.365E-21_WP,    2.786E-21_WP,    3.268E-21_WP, &
    & 3.811E-21_WP,    4.406E-21_WP,    5.035E-21_WP,    5.669E-21_WP,    6.262E-21_WP,    6.757E-21_WP,    7.099E-21_WP, &
    & 7.257E-21_WP,    7.230E-21_WP,    7.052E-21_WP,    6.769E-21_WP,    6.426E-21_WP,    6.055E-21_WP,    5.639E-21_WP, &
    & 5.154E-21_WP,    4.603E-21_WP,    3.991E-21_WP,    3.319E-21_WP,    2.587E-21_WP,    1.793E-21_WP,    9.329E-22_WP /
data J(3,    163    ,:)    / 4.228E-22_WP,    5.049E-22_WP,    6.027E-22_WP,    7.193E-22_WP,    8.578E-22_WP,    1.022E-21_WP, &
    & 1.217E-21_WP,    1.447E-21_WP,    1.718E-21_WP,    2.036E-21_WP,    2.406E-21_WP,    2.834E-21_WP,    3.325E-21_WP, &
    & 3.877E-21_WP,    4.482E-21_WP,    5.122E-21_WP,    5.767E-21_WP,    6.370E-21_WP,    6.874E-21_WP,    7.222E-21_WP, &
    & 7.382E-21_WP,    7.355E-21_WP,    7.174E-21_WP,    6.886E-21_WP,    6.538E-21_WP,    6.160E-21_WP,    5.737E-21_WP, &
    & 5.243E-21_WP,    4.683E-21_WP,    4.061E-21_WP,    3.377E-21_WP,    2.632E-21_WP,    1.824E-21_WP,    9.491E-22_WP /
data J(3,    164    ,:)    / 4.300E-22_WP,    5.135E-22_WP,    6.130E-22_WP,    7.316E-22_WP,    8.725E-22_WP,    1.040E-21_WP, &
    & 1.238E-21_WP,    1.472E-21_WP,    1.748E-21_WP,    2.071E-21_WP,    2.448E-21_WP,    2.883E-21_WP,    3.382E-21_WP, &
    & 3.944E-21_WP,    4.559E-21_WP,    5.211E-21_WP,    5.867E-21_WP,    6.480E-21_WP,    6.993E-21_WP,    7.347E-21_WP, &
    & 7.510E-21_WP,    7.483E-21_WP,    7.298E-21_WP,    7.006E-21_WP,    6.651E-21_WP,    6.267E-21_WP,    5.836E-21_WP, &
    & 5.334E-21_WP,    4.764E-21_WP,    4.131E-21_WP,    3.435E-21_WP,    2.678E-21_WP,    1.856E-21_WP,    9.656E-22_WP /
data J(3,    165    ,:)    / 4.374E-22_WP,    5.223E-22_WP,    6.235E-22_WP,    7.441E-22_WP,    8.875E-22_WP,    1.058E-21_WP, &
    & 1.259E-21_WP,    1.498E-21_WP,    1.778E-21_WP,    2.107E-21_WP,    2.490E-21_WP,    2.933E-21_WP,    3.441E-21_WP, &
    & 4.012E-21_WP,    4.638E-21_WP,    5.301E-21_WP,    5.968E-21_WP,    6.593E-21_WP,    7.114E-21_WP,    7.475E-21_WP, &
    & 7.640E-21_WP,    7.612E-21_WP,    7.425E-21_WP,    7.127E-21_WP,    6.766E-21_WP,    6.375E-21_WP,    5.938E-21_WP, &
    & 5.426E-21_WP,    4.847E-21_WP,    4.203E-21_WP,    3.495E-21_WP,    2.724E-21_WP,    1.888E-21_WP,    9.824E-22_WP /
data J(3,    166    ,:)    / 4.448E-22_WP,    5.312E-22_WP,    6.342E-22_WP,    7.569E-22_WP,    9.027E-22_WP,    1.076E-21_WP, &
    & 1.281E-21_WP,    1.523E-21_WP,    1.809E-21_WP,    2.143E-21_WP,    2.533E-21_WP,    2.984E-21_WP,    3.500E-21_WP, &
    & 4.081E-21_WP,    4.718E-21_WP,    5.393E-21_WP,    6.072E-21_WP,    6.707E-21_WP,    7.237E-21_WP,    7.604E-21_WP, &
    & 7.773E-21_WP,    7.744E-21_WP,    7.554E-21_WP,    7.251E-21_WP,    6.884E-21_WP,    6.486E-21_WP,    6.040E-21_WP, &
    & 5.520E-21_WP,    4.931E-21_WP,    4.276E-21_WP,    3.556E-21_WP,    2.771E-21_WP,    1.921E-21_WP,    9.995E-22_WP /
data J(3,    167    ,:)    / 4.524E-22_WP,    5.403E-22_WP,    6.451E-22_WP,    7.699E-22_WP,    9.183E-22_WP,    1.094E-21_WP, &
    & 1.303E-21_WP,    1.550E-21_WP,    1.840E-21_WP,    2.180E-21_WP,    2.577E-21_WP,    3.035E-21_WP,    3.561E-21_WP, &
    & 4.152E-21_WP,    4.800E-21_WP,    5.486E-21_WP,    6.177E-21_WP,    6.823E-21_WP,    7.363E-21_WP,    7.736E-21_WP, &
    & 7.908E-21_WP,    7.879E-21_WP,    7.685E-21_WP,    7.377E-21_WP,    7.003E-21_WP,    6.598E-21_WP,    6.145E-21_WP, &
    & 5.616E-21_WP,    5.017E-21_WP,    4.350E-21_WP,    3.617E-21_WP,    2.820E-21_WP,    1.954E-21_WP,    1.017E-21_WP /

integer :: i    ! counter
real(WP) :: logJ_at_z(nlambdas)

Fuvb = 0.0_WP
if (lambda>=lambdas(1) .and. lambda<=lambdas(nlambdas) .and. z>=zs(1) .and. z<=zs(nzs)) then
    ! first logJ is interpolated at z at all values of lambdas, and stored in logJ_at_z:
    do i=1, nlambdas
        logJ_at_z(i) = LInterpol(zs,log10(J(model,i,:)),z)
    end do
    ! then the value of logJ_at_z is interpolated at lambda:
    Fuvb = 4.0_WP*pi/c*10.0_WP**LInterpol(log10(lambdas),logJ_at_z,log10(lambda))
end if

end function Fuvb

!---------------------------------------------------------------------------
!*------- returns the energy density of the hot halo radiation field -------
!* Source:
!* Viegas, S.M. & Friaa, A.C.S., 1995: MNRAS 272, L35
!---------------------------------------------------------------------------
function Fhothalo(lambda,r,t)
real(WP) :: Fhothalo    ! in erg/cm^3/Hz
real(WP), intent(in) :: lambda    ! in 
! The next two variables are parameters of the hot halo model:
integer, intent(in) :: r    ! distance from the center of the galaxy
                            ! r=1 corresponds to r=10 kpc
                            ! r=2 corresponds to r=30 kpc
                            ! r=3 corresponds to r=100 kpc
integer, intent(in) :: t    ! age of the galaxy
                            ! t=1 corresponds to t=0.206 Gyr
                            ! t=2 corresponds to t=0.3644 Gyr

integer, parameter :: nenergies = 284    ! number of points in the interpolation tables bellow.
                                        
                                        ! Energies, in eV:
real(WP) :: Energies(nenergies) = (/ 7.200E+0_WP,    7.550E+0_WP,    7.900E+0_WP, &
    & 8.250E+0_WP,    8.600E+0_WP,    9.050E+0_WP,    9.500E+0_WP,    9.930E+0_WP,    1.036E+1_WP,    1.081E+1_WP, &
    & 1.126E+1_WP,    1.143E+1_WP,    1.160E+1_WP,    1.228E+1_WP,    1.297E+1_WP,    1.328E+1_WP,    1.360E+1_WP, &
    & 1.365E+1_WP,    1.370E+1_WP,    1.375E+1_WP,    1.380E+1_WP,    1.390E+1_WP,    1.400E+1_WP,    1.413E+1_WP, &
    & 1.426E+1_WP,    1.440E+1_WP,    1.453E+1_WP,    1.465E+1_WP,    1.478E+1_WP,    1.491E+1_WP,    1.504E+1_WP, &
    & 1.540E+1_WP,    1.576E+1_WP,    1.598E+1_WP,    1.620E+1_WP,    1.750E+1_WP,    1.880E+1_WP,    2.018E+1_WP, &
    & 2.156E+1_WP,    2.244E+1_WP,    2.333E+1_WP,    2.357E+1_WP,    2.381E+1_WP,    2.409E+1_WP,    2.438E+1_WP, &
    & 2.449E+1_WP,    2.459E+1_WP,    2.469E+1_WP,    2.480E+1_WP,    2.490E+1_WP,    2.500E+1_WP,    2.515E+1_WP, &
    & 2.530E+1_WP,    2.545E+1_WP,    2.560E+1_WP,    2.580E+1_WP,    2.600E+1_WP,    2.625E+1_WP,    2.650E+1_WP, &
    & 2.675E+1_WP,    2.700E+1_WP,    2.731E+1_WP,    2.763E+1_WP,    2.861E+1_WP,    2.960E+1_WP,    3.070E+1_WP, &
    & 3.180E+1_WP,    3.295E+1_WP,    3.410E+1_WP,    3.461E+1_WP,    3.512E+1_WP,    3.646E+1_WP,    3.780E+1_WP, &
    & 3.925E+1_WP,    4.070E+1_WP,    4.080E+1_WP,    4.090E+1_WP,    4.195E+1_WP,    4.300E+1_WP,    4.407E+1_WP, &
    & 4.514E+1_WP,    4.627E+1_WP,    4.740E+1_WP,    4.765E+1_WP,    4.789E+1_WP,    4.990E+1_WP,    5.192E+1_WP, &
    & 5.269E+1_WP,    5.346E+1_WP,    5.394E+1_WP,    5.442E+1_WP,    5.453E+1_WP,    5.465E+1_WP,    5.479E+1_WP, &
    & 5.493E+1_WP,    5.528E+1_WP,    5.563E+1_WP,    5.599E+1_WP,    5.635E+1_WP,    5.717E+1_WP,    5.800E+1_WP, &
    & 5.890E+1_WP,    5.981E+1_WP,    6.090E+1_WP,    6.200E+1_WP,    6.300E+1_WP,    6.400E+1_WP,    6.535E+1_WP, &
    & 6.670E+1_WP,    6.850E+1_WP,    7.030E+1_WP,    7.215E+1_WP,    7.400E+1_WP,    7.571E+1_WP,    7.741E+1_WP, &
    & 8.046E+1_WP,    8.350E+1_WP,    8.650E+1_WP,    8.950E+1_WP,    9.350E+1_WP,    9.750E+1_WP,    1.018E+2_WP, &
    & 1.060E+2_WP,    1.099E+2_WP,    1.139E+2_WP,    1.162E+2_WP,    1.186E+2_WP,    1.224E+2_WP,    1.262E+2_WP, &
    & 1.322E+2_WP,    1.381E+2_WP,    1.403E+2_WP,    1.424E+2_WP,    1.445E+2_WP,    1.466E+2_WP,    1.488E+2_WP, &
    & 1.511E+2_WP,    1.545E+2_WP,    1.579E+2_WP,    1.630E+2_WP,    1.680E+2_WP,    1.768E+2_WP,    1.855E+2_WP, &
    & 1.925E+2_WP,    1.994E+2_WP,    2.034E+2_WP,    2.073E+2_WP,    2.118E+2_WP,    2.163E+2_WP,    2.199E+2_WP, &
    & 2.235E+2_WP,    2.302E+2_WP,    2.370E+2_WP,    2.417E+2_WP,    2.465E+2_WP,    2.510E+2_WP,    2.555E+2_WP, &
    & 2.597E+2_WP,    2.640E+2_WP,    2.709E+2_WP,    2.778E+2_WP,    2.829E+2_WP,    2.880E+2_WP,    2.979E+2_WP, &
    & 3.078E+2_WP,    3.138E+2_WP,    3.198E+2_WP,    3.237E+2_WP,    3.276E+2_WP,    3.382E+2_WP,    3.489E+2_WP, &
    & 3.576E+2_WP,    3.663E+2_WP,    3.792E+2_WP,    3.920E+2_WP,    3.990E+2_WP,    4.060E+2_WP,    4.164E+2_WP, &
    & 4.268E+2_WP,    4.388E+2_WP,    4.507E+2_WP,    4.627E+2_WP,    4.747E+2_WP,    4.839E+2_WP,    4.930E+2_WP, &
    & 5.155E+2_WP,    5.380E+2_WP,    5.532E+2_WP,    5.683E+2_WP,    5.805E+2_WP,    5.928E+2_WP,    6.068E+2_WP, &
    & 6.209E+2_WP,    6.350E+2_WP,    6.490E+2_WP,    6.639E+2_WP,    6.788E+2_WP,    6.944E+2_WP,    7.100E+2_WP, &
    & 7.246E+2_WP,    7.391E+2_WP,    7.576E+2_WP,    7.760E+2_WP,    7.940E+2_WP,    8.120E+2_WP,    8.302E+2_WP, &
    & 8.485E+2_WP,    8.598E+2_WP,    8.711E+2_WP,    8.865E+2_WP,    9.018E+2_WP,    9.182E+2_WP,    9.346E+2_WP, &
    & 9.544E+2_WP,    9.743E+2_WP,    9.902E+2_WP,    1.006E+3_WP,    1.024E+3_WP,    1.043E+3_WP,    1.061E+3_WP, &
    & 1.079E+3_WP,    1.098E+3_WP,    1.118E+3_WP,    1.157E+3_WP,    1.196E+3_WP,    1.215E+3_WP,    1.235E+3_WP, &
    & 1.268E+3_WP,    1.301E+3_WP,    1.332E+3_WP,    1.362E+3_WP,    1.416E+3_WP,    1.470E+3_WP,    1.512E+3_WP, &
    & 1.555E+3_WP,    1.599E+3_WP,    1.643E+3_WP,    1.703E+3_WP,    1.763E+3_WP,    1.802E+3_WP,    1.840E+3_WP, &
    & 1.884E+3_WP,    1.929E+3_WP,    1.953E+3_WP,    1.977E+3_WP,    2.008E+3_WP,    2.040E+3_WP,    2.096E+3_WP, &
    & 2.153E+3_WP,    2.224E+3_WP,    2.295E+3_WP,    2.377E+3_WP,    2.460E+3_WP,    2.493E+3_WP,    2.526E+3_WP, &
    & 2.593E+3_WP,    2.660E+3_WP,    2.734E+3_WP,    2.807E+3_WP,    2.852E+3_WP,    2.896E+3_WP,    2.974E+3_WP, &
    & 3.053E+3_WP,    3.116E+3_WP,    3.180E+3_WP,    3.230E+3_WP,    3.280E+3_WP,    3.387E+3_WP,    3.494E+3_WP, &
    & 3.567E+3_WP,    3.640E+3_WP,    3.735E+3_WP,    3.830E+3_WP,    3.930E+3_WP,    4.030E+3_WP,    4.228E+3_WP, &
    & 4.426E+3_WP,    4.713E+3_WP,    5.000E+3_WP,    5.300E+3_WP,    5.600E+3_WP,    5.950E+3_WP,    6.300E+3_WP, &
    & 6.662E+3_WP /)

real(WP) :: PHhothalo(2,3,nenergies)    ! photon flux, in photons/cm^2/s/eV

! t = 0.206 Gyr:
                        !   r=10kpc        r=30kpc        r=100kpc
data PHhothalo(1,:,    1    ) / 2.526E+8_WP,    2.807E+7_WP,    2.526E+6_WP /
data PHhothalo(1,:,    2    ) / 3.748E+8_WP,    4.164E+7_WP,    3.748E+6_WP /
data PHhothalo(1,:,    3    ) / 1.864E+9_WP,    2.071E+8_WP,    1.864E+7_WP /
data PHhothalo(1,:,    4    ) / 6.925E+8_WP,    7.695E+7_WP,    6.925E+6_WP /
data PHhothalo(1,:,    5    ) / 2.083E+6_WP,    2.314E+5_WP,    2.083E+4_WP /
data PHhothalo(1,:,    6    ) / 2.532E+9_WP,    2.814E+8_WP,    2.532E+7_WP /
data PHhothalo(1,:,    7    ) / 5.833E+6_WP,    6.482E+5_WP,    5.834E+4_WP /
data PHhothalo(1,:,    8    ) / 5.597E+8_WP,    6.219E+7_WP,    5.597E+6_WP /
data PHhothalo(1,:,    9    ) / 1.131E+9_WP,    1.257E+8_WP,    1.131E+7_WP /
data PHhothalo(1,:,    10    ) / 1.758E+7_WP,    1.953E+6_WP,    1.758E+5_WP /
data PHhothalo(1,:,    11    ) / 2.842E+7_WP,    3.158E+6_WP,    2.842E+5_WP /
data PHhothalo(1,:,    12    ) / 2.842E+7_WP,    3.158E+6_WP,    2.842E+5_WP /
data PHhothalo(1,:,    13    ) / 2.842E+7_WP,    3.158E+6_WP,    2.842E+5_WP /
data PHhothalo(1,:,    14    ) / 3.984E+8_WP,    4.426E+7_WP,    3.984E+6_WP /
data PHhothalo(1,:,    15    ) / 9.419E+7_WP,    1.047E+7_WP,    9.419E+5_WP /
data PHhothalo(1,:,    16    ) / 9.419E+7_WP,    1.047E+7_WP,    9.419E+5_WP /
data PHhothalo(1,:,    17    ) / 4.334E+7_WP,    4.816E+6_WP,    4.334E+5_WP /
data PHhothalo(1,:,    18    ) / 4.334E+7_WP,    4.816E+6_WP,    4.334E+5_WP /
data PHhothalo(1,:,    19    ) / 4.334E+7_WP,    4.816E+6_WP,    4.334E+5_WP /
data PHhothalo(1,:,    20    ) / 4.334E+7_WP,    4.816E+6_WP,    4.334E+5_WP /
data PHhothalo(1,:,    21    ) / 4.334E+7_WP,    4.816E+6_WP,    4.334E+5_WP /
data PHhothalo(1,:,    22    ) / 4.334E+7_WP,    4.816E+6_WP,    4.334E+5_WP /
data PHhothalo(1,:,    23    ) / 5.404E+7_WP,    6.005E+6_WP,    5.404E+5_WP /
data PHhothalo(1,:,    24    ) / 5.404E+7_WP,    6.005E+6_WP,    5.404E+5_WP /
data PHhothalo(1,:,    25    ) / 5.404E+7_WP,    6.005E+6_WP,    5.404E+5_WP /
data PHhothalo(1,:,    26    ) / 5.404E+7_WP,    6.005E+6_WP,    5.404E+5_WP /
data PHhothalo(1,:,    27    ) / 5.404E+7_WP,    6.005E+6_WP,    5.404E+5_WP /
data PHhothalo(1,:,    28    ) / 9.017E+8_WP,    1.002E+8_WP,    9.017E+6_WP /
data PHhothalo(1,:,    29    ) / 9.017E+8_WP,    1.002E+8_WP,    9.017E+6_WP /
data PHhothalo(1,:,    30    ) / 9.017E+8_WP,    1.002E+8_WP,    9.017E+6_WP /
data PHhothalo(1,:,    31    ) / 9.017E+8_WP,    1.002E+8_WP,    9.017E+6_WP /
data PHhothalo(1,:,    32    ) / 2.151E+9_WP,    2.390E+8_WP,    2.151E+7_WP /
data PHhothalo(1,:,    33    ) / 2.151E+9_WP,    2.390E+8_WP,    2.151E+7_WP /
data PHhothalo(1,:,    34    ) / 2.151E+9_WP,    2.390E+8_WP,    2.151E+7_WP /
data PHhothalo(1,:,    35    ) / 9.575E+8_WP,    1.064E+8_WP,    9.575E+6_WP /
data PHhothalo(1,:,    36    ) / 1.437E+7_WP,    1.596E+6_WP,    1.437E+5_WP /
data PHhothalo(1,:,    37    ) / 2.305E+7_WP,    2.561E+6_WP,    2.305E+5_WP /
data PHhothalo(1,:,    38    ) / 1.461E+9_WP,    1.624E+8_WP,    1.461E+7_WP /
data PHhothalo(1,:,    39    ) / 3.724E+8_WP,    4.137E+7_WP,    3.724E+6_WP /
data PHhothalo(1,:,    40    ) / 1.371E+9_WP,    1.523E+8_WP,    1.371E+7_WP /
data PHhothalo(1,:,    41    ) / 6.399E+6_WP,    7.111E+5_WP,    6.400E+4_WP /
data PHhothalo(1,:,    42    ) / 6.399E+6_WP,    7.111E+5_WP,    6.400E+4_WP /
data PHhothalo(1,:,    43    ) / 6.399E+6_WP,    7.111E+5_WP,    6.400E+4_WP /
data PHhothalo(1,:,    44    ) / 6.399E+6_WP,    7.111E+5_WP,    6.400E+4_WP /
data PHhothalo(1,:,    45    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    46    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    47    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    48    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    49    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    50    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    51    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    52    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    53    ) / 1.552E+8_WP,    1.725E+7_WP,    1.552E+6_WP /
data PHhothalo(1,:,    54    ) / 1.121E+8_WP,    1.245E+7_WP,    1.121E+6_WP /
data PHhothalo(1,:,    55    ) / 1.121E+8_WP,    1.245E+7_WP,    1.121E+6_WP /
data PHhothalo(1,:,    56    ) / 1.121E+8_WP,    1.245E+7_WP,    1.121E+6_WP /
data PHhothalo(1,:,    57    ) / 1.121E+8_WP,    1.245E+7_WP,    1.121E+6_WP /
data PHhothalo(1,:,    58    ) / 1.121E+8_WP,    1.245E+7_WP,    1.121E+6_WP /
data PHhothalo(1,:,    59    ) / 1.121E+8_WP,    1.245E+7_WP,    1.121E+6_WP /
data PHhothalo(1,:,    60    ) / 2.480E+8_WP,    2.755E+7_WP,    2.480E+6_WP /
data PHhothalo(1,:,    61    ) / 2.480E+8_WP,    2.755E+7_WP,    2.480E+6_WP /
data PHhothalo(1,:,    62    ) / 2.480E+8_WP,    2.755E+7_WP,    2.480E+6_WP /
data PHhothalo(1,:,    63    ) / 2.480E+8_WP,    2.755E+7_WP,    2.480E+6_WP /
data PHhothalo(1,:,    64    ) / 1.196E+8_WP,    1.329E+7_WP,    1.196E+6_WP /
data PHhothalo(1,:,    65    ) / 3.549E+6_WP,    3.944E+5_WP,    3.550E+4_WP /
data PHhothalo(1,:,    66    ) / 2.722E+8_WP,    3.025E+7_WP,    2.722E+6_WP /
data PHhothalo(1,:,    67    ) / 2.722E+8_WP,    3.025E+7_WP,    2.722E+6_WP /
data PHhothalo(1,:,    68    ) / 1.058E+7_WP,    1.176E+6_WP,    1.058E+5_WP /
data PHhothalo(1,:,    69    ) / 1.321E+8_WP,    1.467E+7_WP,    1.321E+6_WP /
data PHhothalo(1,:,    70    ) / 1.321E+8_WP,    1.467E+7_WP,    1.321E+6_WP /
data PHhothalo(1,:,    71    ) / 3.372E+7_WP,    3.746E+6_WP,    3.372E+5_WP /
data PHhothalo(1,:,    72    ) / 3.372E+7_WP,    3.746E+6_WP,    3.372E+5_WP /
data PHhothalo(1,:,    73    ) / 1.929E+6_WP,    2.144E+5_WP,    1.929E+4_WP /
data PHhothalo(1,:,    74    ) / 4.272E+7_WP,    4.747E+6_WP,    4.272E+5_WP /
data PHhothalo(1,:,    75    ) / 5.478E+7_WP,    6.086E+6_WP,    5.478E+5_WP /
data PHhothalo(1,:,    76    ) / 5.478E+7_WP,    6.086E+6_WP,    5.478E+5_WP /
data PHhothalo(1,:,    77    ) / 5.478E+7_WP,    6.086E+6_WP,    5.478E+5_WP /
data PHhothalo(1,:,    78    ) / 5.478E+7_WP,    6.086E+6_WP,    5.478E+5_WP /
data PHhothalo(1,:,    79    ) / 3.663E+6_WP,    4.070E+5_WP,    3.663E+4_WP /
data PHhothalo(1,:,    80    ) / 3.663E+6_WP,    4.070E+5_WP,    3.663E+4_WP /
data PHhothalo(1,:,    81    ) / 9.478E+7_WP,    1.053E+7_WP,    9.478E+5_WP /
data PHhothalo(1,:,    82    ) / 1.399E+7_WP,    1.555E+6_WP,    1.399E+5_WP /
data PHhothalo(1,:,    83    ) / 1.399E+7_WP,    1.555E+6_WP,    1.399E+5_WP /
data PHhothalo(1,:,    84    ) / 1.399E+7_WP,    1.555E+6_WP,    1.399E+5_WP /
data PHhothalo(1,:,    85    ) / 1.399E+7_WP,    1.555E+6_WP,    1.399E+5_WP /
data PHhothalo(1,:,    86    ) / 1.816E+7_WP,    2.017E+6_WP,    1.816E+5_WP /
data PHhothalo(1,:,    87    ) / 1.938E+7_WP,    2.153E+6_WP,    1.938E+5_WP /
data PHhothalo(1,:,    88    ) / 1.938E+7_WP,    2.153E+6_WP,    1.938E+5_WP /
data PHhothalo(1,:,    89    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    90    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    91    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    92    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    93    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    94    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    95    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    96    ) / 2.482E+6_WP,    2.758E+5_WP,    2.482E+4_WP /
data PHhothalo(1,:,    97    ) / 6.652E+6_WP,    7.391E+5_WP,    6.652E+4_WP /
data PHhothalo(1,:,    98    ) / 6.652E+6_WP,    7.391E+5_WP,    6.652E+4_WP /
data PHhothalo(1,:,    99    ) / 6.652E+6_WP,    7.391E+5_WP,    6.652E+4_WP /
data PHhothalo(1,:,    100    ) / 6.652E+6_WP,    7.391E+5_WP,    6.652E+4_WP /
data PHhothalo(1,:,    101    ) / 6.652E+6_WP,    7.391E+5_WP,    6.652E+4_WP /
data PHhothalo(1,:,    102    ) / 9.252E+5_WP,    1.028E+5_WP,    9.252E+3_WP /
data PHhothalo(1,:,    103    ) / 9.252E+5_WP,    1.028E+5_WP,    9.252E+3_WP /
data PHhothalo(1,:,    104    ) / 9.252E+5_WP,    1.028E+5_WP,    9.252E+3_WP /
data PHhothalo(1,:,    105    ) / 8.394E+6_WP,    9.327E+5_WP,    8.394E+4_WP /
data PHhothalo(1,:,    106    ) / 8.394E+6_WP,    9.327E+5_WP,    8.394E+4_WP /
data PHhothalo(1,:,    107    ) / 2.976E+6_WP,    3.307E+5_WP,    2.976E+4_WP /
data PHhothalo(1,:,    108    ) / 2.976E+6_WP,    3.307E+5_WP,    2.976E+4_WP /
data PHhothalo(1,:,    109    ) / 2.976E+6_WP,    3.307E+5_WP,    2.976E+4_WP /
data PHhothalo(1,:,    110    ) / 2.016E+7_WP,    2.239E+6_WP,    2.016E+5_WP /
data PHhothalo(1,:,    111    ) / 3.838E+7_WP,    4.265E+6_WP,    3.838E+5_WP /
data PHhothalo(1,:,    112    ) / 3.838E+7_WP,    4.265E+6_WP,    3.838E+5_WP /
data PHhothalo(1,:,    113    ) / 1.251E+7_WP,    1.389E+6_WP,    1.251E+5_WP /
data PHhothalo(1,:,    114    ) / 1.251E+7_WP,    1.389E+6_WP,    1.251E+5_WP /
data PHhothalo(1,:,    115    ) / 2.896E+5_WP,    3.218E+4_WP,    2.897E+3_WP /
data PHhothalo(1,:,    116    ) / 1.256E+7_WP,    1.395E+6_WP,    1.256E+5_WP /
data PHhothalo(1,:,    117    ) / 1.256E+7_WP,    1.395E+6_WP,    1.256E+5_WP /
data PHhothalo(1,:,    118    ) / 8.444E+6_WP,    9.382E+5_WP,    8.444E+4_WP /
data PHhothalo(1,:,    119    ) / 4.480E+6_WP,    4.978E+5_WP,    4.480E+4_WP /
data PHhothalo(1,:,    120    ) / 5.598E+6_WP,    6.220E+5_WP,    5.598E+4_WP /
data PHhothalo(1,:,    121    ) / 5.465E+4_WP,    6.073E+3_WP,    5.468E+2_WP /
data PHhothalo(1,:,    122    ) / 1.317E+7_WP,    1.463E+6_WP,    1.317E+5_WP /
data PHhothalo(1,:,    123    ) / 3.090E+6_WP,    3.433E+5_WP,    3.090E+4_WP /
data PHhothalo(1,:,    124    ) / 3.090E+6_WP,    3.433E+5_WP,    3.090E+4_WP /
data PHhothalo(1,:,    125    ) / 3.289E+6_WP,    3.654E+5_WP,    3.289E+4_WP /
data PHhothalo(1,:,    126    ) / 3.289E+6_WP,    3.654E+5_WP,    3.289E+4_WP /
data PHhothalo(1,:,    127    ) / 9.188E+5_WP,    1.021E+5_WP,    9.189E+3_WP /
data PHhothalo(1,:,    128    ) / 2.613E+6_WP,    2.903E+5_WP,    2.613E+4_WP /
data PHhothalo(1,:,    129    ) / 2.613E+6_WP,    2.903E+5_WP,    2.613E+4_WP /
data PHhothalo(1,:,    130    ) / 5.543E+6_WP,    6.159E+5_WP,    5.543E+4_WP /
data PHhothalo(1,:,    131    ) / 5.247E+4_WP,    5.830E+3_WP,    5.248E+2_WP /
data PHhothalo(1,:,    132    ) / 9.110E+5_WP,    1.012E+5_WP,    9.110E+3_WP /
data PHhothalo(1,:,    133    ) / 9.110E+5_WP,    1.012E+5_WP,    9.110E+3_WP /
data PHhothalo(1,:,    134    ) / 9.110E+5_WP,    1.012E+5_WP,    9.110E+3_WP /
data PHhothalo(1,:,    135    ) / 1.565E+6_WP,    1.738E+5_WP,    1.565E+4_WP /
data PHhothalo(1,:,    136    ) / 1.565E+6_WP,    1.738E+5_WP,    1.565E+4_WP /
data PHhothalo(1,:,    137    ) / 1.565E+6_WP,    1.738E+5_WP,    1.565E+4_WP /
data PHhothalo(1,:,    138    ) / 8.793E+5_WP,    9.770E+4_WP,    8.793E+3_WP /
data PHhothalo(1,:,    139    ) / 8.793E+5_WP,    9.770E+4_WP,    8.793E+3_WP /
data PHhothalo(1,:,    140    ) / 5.826E+5_WP,    6.474E+4_WP,    5.827E+3_WP /
data PHhothalo(1,:,    141    ) / 1.365E+6_WP,    1.516E+5_WP,    1.365E+4_WP /
data PHhothalo(1,:,    142    ) / 6.160E+5_WP,    6.844E+4_WP,    6.160E+3_WP /
data PHhothalo(1,:,    143    ) / 7.474E+4_WP,    8.305E+3_WP,    7.475E+2_WP /
data PHhothalo(1,:,    144    ) / 7.474E+4_WP,    8.305E+3_WP,    7.475E+2_WP /
data PHhothalo(1,:,    145    ) / 7.248E+5_WP,    8.053E+4_WP,    7.248E+3_WP /
data PHhothalo(1,:,    146    ) / 1.344E+6_WP,    1.493E+5_WP,    1.344E+4_WP /
data PHhothalo(1,:,    147    ) / 1.344E+6_WP,    1.493E+5_WP,    1.344E+4_WP /
data PHhothalo(1,:,    148    ) / 1.510E+5_WP,    1.678E+4_WP,    1.510E+3_WP /
data PHhothalo(1,:,    149    ) / 1.510E+5_WP,    1.678E+4_WP,    1.510E+3_WP /
data PHhothalo(1,:,    150    ) / 1.510E+5_WP,    1.678E+4_WP,    1.510E+3_WP /
data PHhothalo(1,:,    151    ) / 2.098E+5_WP,    2.331E+4_WP,    2.098E+3_WP /
data PHhothalo(1,:,    152    ) / 2.098E+5_WP,    2.331E+4_WP,    2.098E+3_WP /
data PHhothalo(1,:,    153    ) / 1.982E+5_WP,    2.203E+4_WP,    1.982E+3_WP /
data PHhothalo(1,:,    154    ) / 1.982E+5_WP,    2.203E+4_WP,    1.982E+3_WP /
data PHhothalo(1,:,    155    ) / 1.131E+5_WP,    1.257E+4_WP,    1.131E+3_WP /
data PHhothalo(1,:,    156    ) / 1.131E+5_WP,    1.257E+4_WP,    1.131E+3_WP /
data PHhothalo(1,:,    157    ) / 5.774E+4_WP,    6.416E+3_WP,    5.775E+2_WP /
data PHhothalo(1,:,    158    ) / 5.774E+4_WP,    6.416E+3_WP,    5.775E+2_WP /
data PHhothalo(1,:,    159    ) / 5.774E+4_WP,    6.416E+3_WP,    5.775E+2_WP /
data PHhothalo(1,:,    160    ) / 4.766E+4_WP,    5.296E+3_WP,    4.767E+2_WP /
data PHhothalo(1,:,    161    ) / 4.766E+4_WP,    5.296E+3_WP,    4.767E+2_WP /
data PHhothalo(1,:,    162    ) / 6.873E+4_WP,    7.637E+3_WP,    6.874E+2_WP /
data PHhothalo(1,:,    163    ) / 6.873E+4_WP,    7.637E+3_WP,    6.874E+2_WP /
data PHhothalo(1,:,    164    ) / 1.663E+5_WP,    1.848E+4_WP,    1.663E+3_WP /
data PHhothalo(1,:,    165    ) / 1.180E+5_WP,    1.311E+4_WP,    1.180E+3_WP /
data PHhothalo(1,:,    166    ) / 1.180E+5_WP,    1.311E+4_WP,    1.180E+3_WP /
data PHhothalo(1,:,    167    ) / 1.180E+5_WP,    1.311E+4_WP,    1.180E+3_WP /
data PHhothalo(1,:,    168    ) / 1.660E+4_WP,    1.845E+3_WP,    1.661E+2_WP /
data PHhothalo(1,:,    169    ) / 1.660E+4_WP,    1.845E+3_WP,    1.661E+2_WP /
data PHhothalo(1,:,    170    ) / 8.628E+3_WP,    9.589E+2_WP,    8.634E+1_WP /
data PHhothalo(1,:,    171    ) / 8.628E+3_WP,    9.589E+2_WP,    8.634E+1_WP /
data PHhothalo(1,:,    172    ) / 2.693E+4_WP,    2.992E+3_WP,    2.693E+2_WP /
data PHhothalo(1,:,    173    ) / 2.693E+4_WP,    2.992E+3_WP,    2.693E+2_WP /
data PHhothalo(1,:,    174    ) / 5.820E+4_WP,    6.467E+3_WP,    5.822E+2_WP /
data PHhothalo(1,:,    175    ) / 4.921E+3_WP,    5.470E+2_WP,    4.926E+1_WP /
data PHhothalo(1,:,    176    ) / 4.921E+3_WP,    5.470E+2_WP,    4.926E+1_WP /
data PHhothalo(1,:,    177    ) / 1.119E+4_WP,    1.243E+3_WP,    1.120E+2_WP /
data PHhothalo(1,:,    178    ) / 1.119E+4_WP,    1.243E+3_WP,    1.120E+2_WP /
data PHhothalo(1,:,    179    ) / 1.732E+4_WP,    1.925E+3_WP,    1.732E+2_WP /
data PHhothalo(1,:,    180    ) / 1.732E+4_WP,    1.925E+3_WP,    1.732E+2_WP /
data PHhothalo(1,:,    181    ) / 3.869E+3_WP,    4.300E+2_WP,    3.873E+1_WP /
data PHhothalo(1,:,    182    ) / 1.746E+3_WP,    1.941E+2_WP,    1.749E+1_WP /
data PHhothalo(1,:,    183    ) / 1.746E+3_WP,    1.941E+2_WP,    1.749E+1_WP /
data PHhothalo(1,:,    184    ) / 1.746E+3_WP,    1.941E+2_WP,    1.749E+1_WP /
data PHhothalo(1,:,    185    ) / 8.511E+3_WP,    9.459E+2_WP,    8.516E+1_WP /
data PHhothalo(1,:,    186    ) / 1.150E+3_WP,    1.279E+2_WP,    1.154E+1_WP /
data PHhothalo(1,:,    187    ) / 1.329E+3_WP,    1.478E+2_WP,    1.333E+1_WP /
data PHhothalo(1,:,    188    ) / 1.329E+3_WP,    1.478E+2_WP,    1.333E+1_WP /
data PHhothalo(1,:,    189    ) / 6.229E+4_WP,    6.921E+3_WP,    6.229E+2_WP /
data PHhothalo(1,:,    190    ) / 6.229E+4_WP,    6.921E+3_WP,    6.229E+2_WP /
data PHhothalo(1,:,    191    ) / 6.306E+2_WP,    7.013E+1_WP,    6.323E+0_WP /
data PHhothalo(1,:,    192    ) / 6.306E+2_WP,    7.013E+1_WP,    6.323E+0_WP /
data PHhothalo(1,:,    193    ) / 2.193E+2_WP,    2.441E+1_WP,    2.206E+0_WP /
data PHhothalo(1,:,    194    ) / 2.193E+2_WP,    2.441E+1_WP,    2.206E+0_WP /
data PHhothalo(1,:,    195    ) / 1.109E+4_WP,    1.232E+3_WP,    1.110E+2_WP /
data PHhothalo(1,:,    196    ) / 1.109E+4_WP,    1.232E+3_WP,    1.110E+2_WP /
data PHhothalo(1,:,    197    ) / 1.255E+3_WP,    1.395E+2_WP,    1.256E+1_WP /
data PHhothalo(1,:,    198    ) / 1.255E+3_WP,    1.395E+2_WP,    1.256E+1_WP /
data PHhothalo(1,:,    199    ) / 1.724E+3_WP,    1.925E+2_WP,    1.747E+1_WP /
data PHhothalo(1,:,    200    ) / 1.724E+3_WP,    1.925E+2_WP,    1.747E+1_WP /
data PHhothalo(1,:,    201    ) / 2.103E+2_WP,    2.367E+1_WP,    2.182E+0_WP /
data PHhothalo(1,:,    202    ) / 2.103E+2_WP,    2.367E+1_WP,    2.182E+0_WP /
data PHhothalo(1,:,    203    ) / 4.985E+2_WP,    5.553E+1_WP,    5.023E+0_WP /
data PHhothalo(1,:,    204    ) / 4.985E+2_WP,    5.553E+1_WP,    5.023E+0_WP /
data PHhothalo(1,:,    205    ) / 3.480E+2_WP,    3.980E+1_WP,    3.761E+0_WP /
data PHhothalo(1,:,    206    ) / 3.480E+2_WP,    3.980E+1_WP,    3.761E+0_WP /
data PHhothalo(1,:,    207    ) / 1.132E+2_WP,    1.269E+1_WP,    1.165E+0_WP /
data PHhothalo(1,:,    208    ) / 1.132E+2_WP,    1.269E+1_WP,    1.165E+0_WP /
data PHhothalo(1,:,    209    ) / 1.132E+2_WP,    1.269E+1_WP,    1.165E+0_WP /
data PHhothalo(1,:,    210    ) / 7.141E+2_WP,    7.954E+1_WP,    7.191E+0_WP /
data PHhothalo(1,:,    211    ) / 7.141E+2_WP,    7.954E+1_WP,    7.191E+0_WP /
data PHhothalo(1,:,    212    ) / 7.141E+2_WP,    7.954E+1_WP,    7.191E+0_WP /
data PHhothalo(1,:,    213    ) / 8.412E+1_WP,    9.374E+0_WP,    8.496E-1_WP /
data PHhothalo(1,:,    214    ) / 8.412E+1_WP,    9.374E+0_WP,    8.496E-1_WP /
data PHhothalo(1,:,    215    ) / 6.282E+1_WP,    7.033E+0_WP,    6.428E-1_WP /
data PHhothalo(1,:,    216    ) / 6.282E+1_WP,    7.033E+0_WP,    6.428E-1_WP /
data PHhothalo(1,:,    217    ) / 6.282E+1_WP,    7.033E+0_WP,    6.428E-1_WP /
data PHhothalo(1,:,    218    ) / 5.078E+1_WP,    5.921E+0_WP,    5.785E-1_WP /
data PHhothalo(1,:,    219    ) / 5.078E+1_WP,    5.921E+0_WP,    5.785E-1_WP /
data PHhothalo(1,:,    220    ) / 6.587E+1_WP,    7.375E+0_WP,    6.752E-1_WP /
data PHhothalo(1,:,    221    ) / 6.587E+1_WP,    7.375E+0_WP,    6.752E-1_WP /
data PHhothalo(1,:,    222    ) / 6.587E+1_WP,    7.375E+0_WP,    6.752E-1_WP /
data PHhothalo(1,:,    223    ) / 3.989E+1_WP,    4.511E+0_WP,    4.196E-1_WP /
data PHhothalo(1,:,    224    ) / 3.989E+1_WP,    4.511E+0_WP,    4.196E-1_WP /
data PHhothalo(1,:,    225    ) / 1.800E+1_WP,    2.019E+0_WP,    1.853E-1_WP /
data PHhothalo(1,:,    226    ) / 1.800E+1_WP,    2.019E+0_WP,    1.853E-1_WP /
data PHhothalo(1,:,    227    ) / 1.285E+1_WP,    1.459E+0_WP,    1.371E-1_WP /
data PHhothalo(1,:,    228    ) / 1.285E+1_WP,    1.459E+0_WP,    1.371E-1_WP /
data PHhothalo(1,:,    229    ) / 1.062E+1_WP,    1.201E+0_WP,    1.120E-1_WP /
data PHhothalo(1,:,    230    ) / 1.062E+1_WP,    1.201E+0_WP,    1.120E-1_WP /
data PHhothalo(1,:,    231    ) / 1.099E+1_WP,    1.261E+0_WP,    1.197E-1_WP /
data PHhothalo(1,:,    232    ) / 4.364E+0_WP,    4.902E-1_WP,    4.522E-2_WP /
data PHhothalo(1,:,    233    ) / 3.025E+0_WP,    3.509E-1_WP,    3.464E-2_WP /
data PHhothalo(1,:,    234    ) / 3.025E+0_WP,    3.509E-1_WP,    3.464E-2_WP /
data PHhothalo(1,:,    235    ) / 2.279E+0_WP,    2.618E-1_WP,    2.515E-2_WP /
data PHhothalo(1,:,    236    ) / 2.279E+0_WP,    2.618E-1_WP,    2.515E-2_WP /
data PHhothalo(1,:,    237    ) / 1.361E+0_WP,    1.543E-1_WP,    1.456E-2_WP /
data PHhothalo(1,:,    238    ) / 9.616E-1_WP,    1.119E-1_WP,    1.107E-2_WP /
data PHhothalo(1,:,    239    ) / 6.184E-1_WP,    7.594E-2_WP,    8.266E-3_WP /
data PHhothalo(1,:,    240    ) / 6.184E-1_WP,    7.594E-2_WP,    8.266E-3_WP /
data PHhothalo(1,:,    241    ) / 6.184E-1_WP,    7.594E-2_WP,    8.266E-3_WP /
data PHhothalo(1,:,    242    ) / 4.601E-1_WP,    6.623E-2_WP,    8.902E-3_WP /
data PHhothalo(1,:,    243    ) / 2.445E-1_WP,    2.924E-2_WP,    3.153E-3_WP /
data PHhothalo(1,:,    244    ) / 2.445E-1_WP,    2.924E-2_WP,    3.153E-3_WP /
data PHhothalo(1,:,    245    ) / 2.445E-1_WP,    2.924E-2_WP,    3.153E-3_WP /
data PHhothalo(1,:,    246    ) / 2.445E-1_WP,    2.924E-2_WP,    3.153E-3_WP /
data PHhothalo(1,:,    247    ) / 1.532E-1_WP,    1.820E-2_WP,    1.908E-3_WP /
data PHhothalo(1,:,    248    ) / 1.532E-1_WP,    1.820E-2_WP,    1.908E-3_WP /
data PHhothalo(1,:,    249    ) / 1.015E-1_WP,    1.355E-2_WP,    1.707E-3_WP /
data PHhothalo(1,:,    250    ) / 5.992E-2_WP,    7.776E-3_WP,    9.511E-4_WP /
data PHhothalo(1,:,    251    ) / 5.992E-2_WP,    7.776E-3_WP,    9.511E-4_WP /
data PHhothalo(1,:,    252    ) / 3.735E-2_WP,    5.260E-3_WP,    7.302E-4_WP /
data PHhothalo(1,:,    253    ) / 2.807E-2_WP,    5.818E-3_WP,    1.139E-3_WP /
data PHhothalo(1,:,    254    ) / 2.807E-2_WP,    5.818E-3_WP,    1.139E-3_WP /
data PHhothalo(1,:,    255    ) / 2.807E-2_WP,    5.818E-3_WP,    1.139E-3_WP /
data PHhothalo(1,:,    256    ) / 1.257E-2_WP,    1.800E-3_WP,    2.624E-4_WP /
data PHhothalo(1,:,    257    ) / 1.257E-2_WP,    1.800E-3_WP,    2.624E-4_WP /
data PHhothalo(1,:,    258    ) / 7.374E-3_WP,    1.128E-3_WP,    1.801E-4_WP /
data PHhothalo(1,:,    259    ) / 4.669E-3_WP,    8.925E-4_WP,    1.755E-4_WP /
data PHhothalo(1,:,    260    ) / 4.669E-3_WP,    8.925E-4_WP,    1.755E-4_WP /
data PHhothalo(1,:,    261    ) / 4.669E-3_WP,    8.925E-4_WP,    1.755E-4_WP /
data PHhothalo(1,:,    262    ) / 2.510E-3_WP,    4.583E-4_WP,    8.936E-5_WP /
data PHhothalo(1,:,    263    ) / 2.510E-3_WP,    4.583E-4_WP,    8.936E-5_WP /
data PHhothalo(1,:,    264    ) / 1.656E-3_WP,    4.296E-4_WP,    1.069E-4_WP /
data PHhothalo(1,:,    265    ) / 1.656E-3_WP,    4.296E-4_WP,    1.069E-4_WP /
data PHhothalo(1,:,    266    ) / 8.728E-4_WP,    1.987E-4_WP,    4.679E-5_WP /
data PHhothalo(1,:,    267    ) / 8.728E-4_WP,    1.987E-4_WP,    4.679E-5_WP /
data PHhothalo(1,:,    268    ) / 5.205E-4_WP,    1.323E-4_WP,    3.378E-5_WP /
data PHhothalo(1,:,    269    ) / 5.205E-4_WP,    1.323E-4_WP,    3.378E-5_WP /
data PHhothalo(1,:,    270    ) / 3.147E-4_WP,    8.883E-5_WP,    2.439E-5_WP /
data PHhothalo(1,:,    271    ) / 3.147E-4_WP,    8.883E-5_WP,    2.439E-5_WP /
data PHhothalo(1,:,    272    ) / 2.011E-4_WP,    6.586E-5_WP,    1.962E-5_WP /
data PHhothalo(1,:,    273    ) / 2.011E-4_WP,    6.586E-5_WP,    1.962E-5_WP /
data PHhothalo(1,:,    274    ) / 1.213E-4_WP,    4.163E-5_WP,    1.326E-5_WP /
data PHhothalo(1,:,    275    ) / 7.552E-5_WP,    2.755E-5_WP,    9.034E-6_WP /
data PHhothalo(1,:,    276    ) / 4.773E-5_WP,    1.860E-5_WP,    6.416E-6_WP /
data PHhothalo(1,:,    277    ) / 3.020E-5_WP,    1.249E-5_WP,    4.522E-6_WP /
data PHhothalo(1,:,    278    ) / 1.906E-5_WP,    8.332E-6_WP,    3.161E-6_WP /
data PHhothalo(1,:,    279    ) / 1.195E-5_WP,    5.512E-6_WP,    2.190E-6_WP /
data PHhothalo(1,:,    280    ) / 7.429E-6_WP,    3.615E-6_WP,    1.503E-6_WP /
data PHhothalo(1,:,    281    ) / 2.781E-6_WP,    1.513E-6_WP,    6.867E-7_WP /
data PHhothalo(1,:,    282    ) / 1.672E-6_WP,    9.651E-7_WP,    4.567E-7_WP /
data PHhothalo(1,:,    283    ) / 9.943E-7_WP,    6.101E-7_WP,    3.002E-7_WP /
data PHhothalo(1,:,    284    ) / 5.845E-7_WP,    3.820E-7_WP,    1.950E-7_WP /

! t = 0.3644 Gyr:
                        !   r=10kpc        r=30kpc        r=100kpc
data PHhothalo(2,:,    1    ) / 5.543E+5_WP,    6.159E+4_WP,    5.544E+3_WP /
data PHhothalo(2,:,    2    ) / 1.121E+6_WP,    1.246E+5_WP,    1.122E+4_WP /
data PHhothalo(2,:,    3    ) / 1.120E+7_WP,    1.244E+6_WP,    1.120E+5_WP /
data PHhothalo(2,:,    4    ) / 2.338E+6_WP,    2.598E+5_WP,    2.338E+4_WP /
data PHhothalo(2,:,    5    ) / 3.182E+4_WP,    3.536E+3_WP,    3.186E+2_WP /
data PHhothalo(2,:,    6    ) / 2.681E+7_WP,    2.979E+6_WP,    2.681E+5_WP /
data PHhothalo(2,:,    7    ) / 1.126E+5_WP,    1.252E+4_WP,    1.127E+3_WP /
data PHhothalo(2,:,    8    ) / 8.015E+6_WP,    8.906E+5_WP,    8.016E+4_WP /
data PHhothalo(2,:,    9    ) / 2.783E+7_WP,    3.092E+6_WP,    2.783E+5_WP /
data PHhothalo(2,:,    10    ) / 5.023E+5_WP,    5.581E+4_WP,    5.023E+3_WP /
data PHhothalo(2,:,    11    ) / 5.765E+4_WP,    6.406E+3_WP,    5.768E+2_WP /
data PHhothalo(2,:,    12    ) / 5.765E+4_WP,    6.406E+3_WP,    5.768E+2_WP /
data PHhothalo(2,:,    13    ) / 5.765E+4_WP,    6.406E+3_WP,    5.768E+2_WP /
data PHhothalo(2,:,    14    ) / 2.583E+6_WP,    2.870E+5_WP,    2.583E+4_WP /
data PHhothalo(2,:,    15    ) / 2.265E+6_WP,    2.517E+5_WP,    2.265E+4_WP /
data PHhothalo(2,:,    16    ) / 2.265E+6_WP,    2.517E+5_WP,    2.265E+4_WP /
data PHhothalo(2,:,    17    ) / 1.064E+6_WP,    1.182E+5_WP,    1.064E+4_WP /
data PHhothalo(2,:,    18    ) / 1.064E+6_WP,    1.182E+5_WP,    1.064E+4_WP /
data PHhothalo(2,:,    19    ) / 1.064E+6_WP,    1.182E+5_WP,    1.064E+4_WP /
data PHhothalo(2,:,    20    ) / 1.064E+6_WP,    1.182E+5_WP,    1.064E+4_WP /
data PHhothalo(2,:,    21    ) / 1.064E+6_WP,    1.182E+5_WP,    1.064E+4_WP /
data PHhothalo(2,:,    22    ) / 1.064E+6_WP,    1.182E+5_WP,    1.064E+4_WP /
data PHhothalo(2,:,    23    ) / 6.707E+5_WP,    7.452E+4_WP,    6.707E+3_WP /
data PHhothalo(2,:,    24    ) / 6.707E+5_WP,    7.452E+4_WP,    6.707E+3_WP /
data PHhothalo(2,:,    25    ) / 6.707E+5_WP,    7.452E+4_WP,    6.707E+3_WP /
data PHhothalo(2,:,    26    ) / 6.707E+5_WP,    7.452E+4_WP,    6.707E+3_WP /
data PHhothalo(2,:,    27    ) / 6.707E+5_WP,    7.452E+4_WP,    6.707E+3_WP /
data PHhothalo(2,:,    28    ) / 3.315E+6_WP,    3.683E+5_WP,    3.315E+4_WP /
data PHhothalo(2,:,    29    ) / 3.315E+6_WP,    3.683E+5_WP,    3.315E+4_WP /
data PHhothalo(2,:,    30    ) / 3.315E+6_WP,    3.683E+5_WP,    3.315E+4_WP /
data PHhothalo(2,:,    31    ) / 3.315E+6_WP,    3.683E+5_WP,    3.315E+4_WP /
data PHhothalo(2,:,    32    ) / 2.032E+7_WP,    2.257E+6_WP,    2.032E+5_WP /
data PHhothalo(2,:,    33    ) / 2.032E+7_WP,    2.257E+6_WP,    2.032E+5_WP /
data PHhothalo(2,:,    34    ) / 2.032E+7_WP,    2.257E+6_WP,    2.032E+5_WP /
data PHhothalo(2,:,    35    ) / 1.271E+7_WP,    1.412E+6_WP,    1.271E+5_WP /
data PHhothalo(2,:,    36    ) / 1.639E+5_WP,    1.821E+4_WP,    1.639E+3_WP /
data PHhothalo(2,:,    37    ) / 8.287E+4_WP,    9.208E+3_WP,    8.290E+2_WP /
data PHhothalo(2,:,    38    ) / 4.324E+7_WP,    4.804E+6_WP,    4.324E+5_WP /
data PHhothalo(2,:,    39    ) / 1.188E+7_WP,    1.320E+6_WP,    1.188E+5_WP /
data PHhothalo(2,:,    40    ) / 1.635E+7_WP,    1.817E+6_WP,    1.635E+5_WP /
data PHhothalo(2,:,    41    ) / 9.045E+4_WP,    1.005E+4_WP,    9.047E+2_WP /
data PHhothalo(2,:,    42    ) / 9.045E+4_WP,    1.005E+4_WP,    9.047E+2_WP /
data PHhothalo(2,:,    43    ) / 9.045E+4_WP,    1.005E+4_WP,    9.047E+2_WP /
data PHhothalo(2,:,    44    ) / 9.045E+4_WP,    1.005E+4_WP,    9.047E+2_WP /
data PHhothalo(2,:,    45    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    46    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    47    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    48    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    49    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    50    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    51    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    52    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    53    ) / 7.040E+5_WP,    7.823E+4_WP,    7.040E+3_WP /
data PHhothalo(2,:,    54    ) / 3.764E+6_WP,    4.182E+5_WP,    3.764E+4_WP /
data PHhothalo(2,:,    55    ) / 3.764E+6_WP,    4.182E+5_WP,    3.764E+4_WP /
data PHhothalo(2,:,    56    ) / 3.764E+6_WP,    4.182E+5_WP,    3.764E+4_WP /
data PHhothalo(2,:,    57    ) / 3.764E+6_WP,    4.182E+5_WP,    3.764E+4_WP /
data PHhothalo(2,:,    58    ) / 3.764E+6_WP,    4.182E+5_WP,    3.764E+4_WP /
data PHhothalo(2,:,    59    ) / 3.764E+6_WP,    4.182E+5_WP,    3.764E+4_WP /
data PHhothalo(2,:,    60    ) / 6.663E+6_WP,    7.404E+5_WP,    6.663E+4_WP /
data PHhothalo(2,:,    61    ) / 6.663E+6_WP,    7.404E+5_WP,    6.663E+4_WP /
data PHhothalo(2,:,    62    ) / 6.663E+6_WP,    7.404E+5_WP,    6.663E+4_WP /
data PHhothalo(2,:,    63    ) / 6.663E+6_WP,    7.404E+5_WP,    6.663E+4_WP /
data PHhothalo(2,:,    64    ) / 3.102E+6_WP,    3.447E+5_WP,    3.102E+4_WP /
data PHhothalo(2,:,    65    ) / 3.980E+4_WP,    4.422E+3_WP,    3.981E+2_WP /
data PHhothalo(2,:,    66    ) / 7.651E+6_WP,    8.501E+5_WP,    7.651E+4_WP /
data PHhothalo(2,:,    67    ) / 7.651E+6_WP,    8.501E+5_WP,    7.651E+4_WP /
data PHhothalo(2,:,    68    ) / 7.485E+4_WP,    8.317E+3_WP,    7.486E+2_WP /
data PHhothalo(2,:,    69    ) / 3.532E+6_WP,    3.925E+5_WP,    3.532E+4_WP /
data PHhothalo(2,:,    70    ) / 3.532E+6_WP,    3.925E+5_WP,    3.532E+4_WP /
data PHhothalo(2,:,    71    ) / 8.064E+5_WP,    8.960E+4_WP,    8.064E+3_WP /
data PHhothalo(2,:,    72    ) / 8.064E+5_WP,    8.960E+4_WP,    8.064E+3_WP /
data PHhothalo(2,:,    73    ) / 1.332E+4_WP,    1.480E+3_WP,    1.332E+2_WP /
data PHhothalo(2,:,    74    ) / 8.283E+5_WP,    9.203E+4_WP,    8.283E+3_WP /
data PHhothalo(2,:,    75    ) / 3.701E+5_WP,    4.113E+4_WP,    3.701E+3_WP /
data PHhothalo(2,:,    76    ) / 3.701E+5_WP,    4.113E+4_WP,    3.701E+3_WP /
data PHhothalo(2,:,    77    ) / 3.701E+5_WP,    4.113E+4_WP,    3.701E+3_WP /
data PHhothalo(2,:,    78    ) / 3.701E+5_WP,    4.113E+4_WP,    3.701E+3_WP /
data PHhothalo(2,:,    79    ) / 1.037E+4_WP,    1.152E+3_WP,    1.038E+2_WP /
data PHhothalo(2,:,    80    ) / 1.037E+4_WP,    1.152E+3_WP,    1.038E+2_WP /
data PHhothalo(2,:,    81    ) / 2.469E+6_WP,    2.744E+5_WP,    2.469E+4_WP /
data PHhothalo(2,:,    82    ) / 8.460E+4_WP,    9.400E+3_WP,    8.462E+2_WP /
data PHhothalo(2,:,    83    ) / 8.460E+4_WP,    9.400E+3_WP,    8.462E+2_WP /
data PHhothalo(2,:,    84    ) / 8.460E+4_WP,    9.400E+3_WP,    8.462E+2_WP /
data PHhothalo(2,:,    85    ) / 8.460E+4_WP,    9.400E+3_WP,    8.462E+2_WP /
data PHhothalo(2,:,    86    ) / 4.896E+5_WP,    5.440E+4_WP,    4.896E+3_WP /
data PHhothalo(2,:,    87    ) / 2.988E+5_WP,    3.320E+4_WP,    2.988E+3_WP /
data PHhothalo(2,:,    88    ) / 2.988E+5_WP,    3.320E+4_WP,    2.988E+3_WP /
data PHhothalo(2,:,    89    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    90    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    91    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    92    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    93    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    94    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    95    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    96    ) / 7.846E+3_WP,    8.718E+2_WP,    7.851E+1_WP /
data PHhothalo(2,:,    97    ) / 1.673E+5_WP,    1.859E+4_WP,    1.674E+3_WP /
data PHhothalo(2,:,    98    ) / 1.673E+5_WP,    1.859E+4_WP,    1.674E+3_WP /
data PHhothalo(2,:,    99    ) / 1.673E+5_WP,    1.859E+4_WP,    1.674E+3_WP /
data PHhothalo(2,:,    100    ) / 1.673E+5_WP,    1.859E+4_WP,    1.674E+3_WP /
data PHhothalo(2,:,    101    ) / 1.673E+5_WP,    1.859E+4_WP,    1.674E+3_WP /
data PHhothalo(2,:,    102    ) / 8.298E+3_WP,    9.221E+2_WP,    8.304E+1_WP /
data PHhothalo(2,:,    103    ) / 8.298E+3_WP,    9.221E+2_WP,    8.304E+1_WP /
data PHhothalo(2,:,    104    ) / 8.298E+3_WP,    9.221E+2_WP,    8.304E+1_WP /
data PHhothalo(2,:,    105    ) / 1.017E+5_WP,    1.131E+4_WP,    1.018E+3_WP /
data PHhothalo(2,:,    106    ) / 1.017E+5_WP,    1.131E+4_WP,    1.018E+3_WP /
data PHhothalo(2,:,    107    ) / 2.389E+4_WP,    2.655E+3_WP,    2.391E+2_WP /
data PHhothalo(2,:,    108    ) / 2.389E+4_WP,    2.655E+3_WP,    2.391E+2_WP /
data PHhothalo(2,:,    109    ) / 2.389E+4_WP,    2.655E+3_WP,    2.391E+2_WP /
data PHhothalo(2,:,    110    ) / 4.616E+5_WP,    5.129E+4_WP,    4.616E+3_WP /
data PHhothalo(2,:,    111    ) / 8.018E+5_WP,    8.909E+4_WP,    8.018E+3_WP /
data PHhothalo(2,:,    112    ) / 8.018E+5_WP,    8.909E+4_WP,    8.018E+3_WP /
data PHhothalo(2,:,    113    ) / 3.524E+5_WP,    3.915E+4_WP,    3.524E+3_WP /
data PHhothalo(2,:,    114    ) / 3.524E+5_WP,    3.915E+4_WP,    3.524E+3_WP /
data PHhothalo(2,:,    115    ) / 3.415E+3_WP,    3.795E+2_WP,    3.418E+1_WP /
data PHhothalo(2,:,    116    ) / 3.284E+5_WP,    3.649E+4_WP,    3.284E+3_WP /
data PHhothalo(2,:,    117    ) / 3.284E+5_WP,    3.649E+4_WP,    3.284E+3_WP /
data PHhothalo(2,:,    118    ) / 2.181E+5_WP,    2.423E+4_WP,    2.181E+3_WP /
data PHhothalo(2,:,    119    ) / 1.172E+5_WP,    1.302E+4_WP,    1.172E+3_WP /
data PHhothalo(2,:,    120    ) / 1.457E+5_WP,    1.619E+4_WP,    1.457E+3_WP /
data PHhothalo(2,:,    121    ) / 7.738E+2_WP,    8.600E+1_WP,    7.757E+0_WP /
data PHhothalo(2,:,    122    ) / 3.591E+5_WP,    3.990E+4_WP,    3.591E+3_WP /
data PHhothalo(2,:,    123    ) / 7.979E+4_WP,    8.866E+3_WP,    7.979E+2_WP /
data PHhothalo(2,:,    124    ) / 7.979E+4_WP,    8.866E+3_WP,    7.979E+2_WP /
data PHhothalo(2,:,    125    ) / 8.833E+4_WP,    9.814E+3_WP,    8.833E+2_WP /
data PHhothalo(2,:,    126    ) / 8.833E+4_WP,    9.814E+3_WP,    8.833E+2_WP /
data PHhothalo(2,:,    127    ) / 2.049E+4_WP,    2.277E+3_WP,    2.050E+2_WP /
data PHhothalo(2,:,    128    ) / 6.587E+4_WP,    7.319E+3_WP,    6.588E+2_WP /
data PHhothalo(2,:,    129    ) / 6.587E+4_WP,    7.319E+3_WP,    6.588E+2_WP /
data PHhothalo(2,:,    130    ) / 1.441E+5_WP,    1.601E+4_WP,    1.441E+3_WP /
data PHhothalo(2,:,    131    ) / 5.018E+2_WP,    5.577E+1_WP,    5.029E+0_WP /
data PHhothalo(2,:,    132    ) / 1.640E+4_WP,    1.822E+3_WP,    1.640E+2_WP /
data PHhothalo(2,:,    133    ) / 1.640E+4_WP,    1.822E+3_WP,    1.640E+2_WP /
data PHhothalo(2,:,    134    ) / 1.640E+4_WP,    1.822E+3_WP,    1.640E+2_WP /
data PHhothalo(2,:,    135    ) / 3.394E+4_WP,    3.771E+3_WP,    3.394E+2_WP /
data PHhothalo(2,:,    136    ) / 3.394E+4_WP,    3.771E+3_WP,    3.394E+2_WP /
data PHhothalo(2,:,    137    ) / 3.394E+4_WP,    3.771E+3_WP,    3.394E+2_WP /
data PHhothalo(2,:,    138    ) / 1.531E+4_WP,    1.702E+3_WP,    1.531E+2_WP /
data PHhothalo(2,:,    139    ) / 1.531E+4_WP,    1.702E+3_WP,    1.531E+2_WP /
data PHhothalo(2,:,    140    ) / 5.690E+3_WP,    6.323E+2_WP,    5.691E+1_WP /
data PHhothalo(2,:,    141    ) / 2.273E+4_WP,    2.526E+3_WP,    2.274E+2_WP /
data PHhothalo(2,:,    142    ) / 5.190E+3_WP,    5.766E+2_WP,    5.191E+1_WP /
data PHhothalo(2,:,    143    ) / 4.396E+2_WP,    4.885E+1_WP,    4.403E+0_WP /
data PHhothalo(2,:,    144    ) / 4.396E+2_WP,    4.885E+1_WP,    4.403E+0_WP /
data PHhothalo(2,:,    145    ) / 6.439E+3_WP,    7.155E+2_WP,    6.440E+1_WP /
data PHhothalo(2,:,    146    ) / 1.140E+4_WP,    1.266E+3_WP,    1.140E+2_WP /
data PHhothalo(2,:,    147    ) / 1.140E+4_WP,    1.266E+3_WP,    1.140E+2_WP /
data PHhothalo(2,:,    148    ) / 4.865E+2_WP,    5.406E+1_WP,    4.873E+0_WP /
data PHhothalo(2,:,    149    ) / 4.865E+2_WP,    5.406E+1_WP,    4.873E+0_WP /
data PHhothalo(2,:,    150    ) / 4.865E+2_WP,    5.406E+1_WP,    4.873E+0_WP /
data PHhothalo(2,:,    151    ) / 8.669E+2_WP,    9.633E+1_WP,    8.679E+0_WP /
data PHhothalo(2,:,    152    ) / 8.669E+2_WP,    9.633E+1_WP,    8.679E+0_WP /
data PHhothalo(2,:,    153    ) / 2.181E+3_WP,    2.424E+2_WP,    2.182E+1_WP /
data PHhothalo(2,:,    154    ) / 2.181E+3_WP,    2.424E+2_WP,    2.182E+1_WP /
data PHhothalo(2,:,    155    ) / 4.659E+2_WP,    5.178E+1_WP,    4.671E+0_WP /
data PHhothalo(2,:,    156    ) / 4.659E+2_WP,    5.178E+1_WP,    4.671E+0_WP /
data PHhothalo(2,:,    157    ) / 4.091E+2_WP,    4.546E+1_WP,    4.096E+0_WP /
data PHhothalo(2,:,    158    ) / 4.091E+2_WP,    4.546E+1_WP,    4.096E+0_WP /
data PHhothalo(2,:,    159    ) / 4.091E+2_WP,    4.546E+1_WP,    4.096E+0_WP /
data PHhothalo(2,:,    160    ) / 1.429E+2_WP,    1.589E+1_WP,    1.435E+0_WP /
data PHhothalo(2,:,    161    ) / 1.429E+2_WP,    1.589E+1_WP,    1.435E+0_WP /
data PHhothalo(2,:,    162    ) / 3.373E+1_WP,    3.760E+0_WP,    3.478E-1_WP /
data PHhothalo(2,:,    163    ) / 3.373E+1_WP,    3.760E+0_WP,    3.478E-1_WP /
data PHhothalo(2,:,    164    ) / 1.574E+3_WP,    1.749E+2_WP,    1.575E+1_WP /
data PHhothalo(2,:,    165    ) / 1.693E+3_WP,    1.881E+2_WP,    1.693E+1_WP /
data PHhothalo(2,:,    166    ) / 1.693E+3_WP,    1.881E+2_WP,    1.693E+1_WP /
data PHhothalo(2,:,    167    ) / 1.693E+3_WP,    1.881E+2_WP,    1.693E+1_WP /
data PHhothalo(2,:,    168    ) / 1.152E+1_WP,    1.284E+0_WP,    1.185E-1_WP /
data PHhothalo(2,:,    169    ) / 1.152E+1_WP,    1.284E+0_WP,    1.185E-1_WP /
data PHhothalo(2,:,    170    ) / 8.509E+0_WP,    9.507E-1_WP,    8.941E-2_WP /
data PHhothalo(2,:,    171    ) / 8.509E+0_WP,    9.507E-1_WP,    8.941E-2_WP /
data PHhothalo(2,:,    172    ) / 2.278E+2_WP,    2.531E+1_WP,    2.281E+0_WP /
data PHhothalo(2,:,    173    ) / 2.278E+2_WP,    2.531E+1_WP,    2.281E+0_WP /
data PHhothalo(2,:,    174    ) / 2.794E+2_WP,    3.105E+1_WP,    2.801E+0_WP /
data PHhothalo(2,:,    175    ) / 8.267E+0_WP,    9.223E-1_WP,    8.590E-2_WP /
data PHhothalo(2,:,    176    ) / 8.267E+0_WP,    9.223E-1_WP,    8.590E-2_WP /
data PHhothalo(2,:,    177    ) / 4.316E+1_WP,    4.800E+0_WP,    4.352E-1_WP /
data PHhothalo(2,:,    178    ) / 4.316E+1_WP,    4.800E+0_WP,    4.352E-1_WP /
data PHhothalo(2,:,    179    ) / 6.787E+1_WP,    7.544E+0_WP,    6.812E-1_WP /
data PHhothalo(2,:,    180    ) / 6.787E+1_WP,    7.544E+0_WP,    6.812E-1_WP /
data PHhothalo(2,:,    181    ) / 6.205E+0_WP,    6.924E-1_WP,    6.456E-2_WP /
data PHhothalo(2,:,    182    ) / 4.154E+0_WP,    4.636E-1_WP,    4.320E-2_WP /
data PHhothalo(2,:,    183    ) / 4.154E+0_WP,    4.636E-1_WP,    4.320E-2_WP /
data PHhothalo(2,:,    184    ) / 4.154E+0_WP,    4.636E-1_WP,    4.320E-2_WP /
data PHhothalo(2,:,    185    ) / 8.280E+0_WP,    9.235E-1_WP,    8.561E-2_WP /
data PHhothalo(2,:,    186    ) / 3.552E+0_WP,    3.973E-1_WP,    3.751E-2_WP /
data PHhothalo(2,:,    187    ) / 3.005E+0_WP,    3.364E-1_WP,    3.211E-2_WP /
data PHhothalo(2,:,    188    ) / 3.005E+0_WP,    3.364E-1_WP,    3.211E-2_WP /
data PHhothalo(2,:,    189    ) / 4.177E+1_WP,    4.645E+0_WP,    4.207E-1_WP /
data PHhothalo(2,:,    190    ) / 4.177E+1_WP,    4.645E+0_WP,    4.207E-1_WP /
data PHhothalo(2,:,    191    ) / 9.745E-1_WP,    1.094E-1_WP,    1.057E-2_WP /
data PHhothalo(2,:,    192    ) / 9.745E-1_WP,    1.094E-1_WP,    1.057E-2_WP /
data PHhothalo(2,:,    193    ) / 6.537E-1_WP,    7.351E-2_WP,    7.178E-3_WP /
data PHhothalo(2,:,    194    ) / 6.537E-1_WP,    7.351E-2_WP,    7.178E-3_WP /
data PHhothalo(2,:,    195    ) / 2.129E+0_WP,    2.483E-1_WP,    3.016E-2_WP /
data PHhothalo(2,:,    196    ) / 2.129E+0_WP,    2.483E-1_WP,    3.016E-2_WP /
data PHhothalo(2,:,    197    ) / 5.462E-1_WP,    6.147E-2_WP,    6.051E-3_WP /
data PHhothalo(2,:,    198    ) / 5.462E-1_WP,    6.147E-2_WP,    6.051E-3_WP /
data PHhothalo(2,:,    199    ) / 7.248E-1_WP,    1.156E-1_WP,    2.733E-2_WP /
data PHhothalo(2,:,    200    ) / 7.248E-1_WP,    1.156E-1_WP,    2.733E-2_WP /
data PHhothalo(2,:,    201    ) / 3.162E-1_WP,    4.855E-2_WP,    1.039E-2_WP /
data PHhothalo(2,:,    202    ) / 3.162E-1_WP,    4.855E-2_WP,    1.039E-2_WP /
data PHhothalo(2,:,    203    ) / 1.978E-1_WP,    2.714E-2_WP,    4.988E-3_WP /
data PHhothalo(2,:,    204    ) / 1.978E-1_WP,    2.714E-2_WP,    4.988E-3_WP /
data PHhothalo(2,:,    205    ) / 3.469E-1_WP,    8.303E-2_WP,    2.863E-2_WP /
data PHhothalo(2,:,    206    ) / 3.469E-1_WP,    8.303E-2_WP,    2.863E-2_WP /
data PHhothalo(2,:,    207    ) / 9.331E-2_WP,    1.885E-2_WP,    4.876E-3_WP /
data PHhothalo(2,:,    208    ) / 9.331E-2_WP,    1.885E-2_WP,    4.876E-3_WP /
data PHhothalo(2,:,    209    ) / 9.331E-2_WP,    1.885E-2_WP,    4.876E-3_WP /
data PHhothalo(2,:,    210    ) / 7.520E-2_WP,    1.520E-2_WP,    4.900E-3_WP /
data PHhothalo(2,:,    211    ) / 7.520E-2_WP,    1.520E-2_WP,    4.900E-3_WP /
data PHhothalo(2,:,    212    ) / 7.520E-2_WP,    1.520E-2_WP,    4.900E-3_WP /
data PHhothalo(2,:,    213    ) / 2.579E-2_WP,    4.427E-3_WP,    1.027E-3_WP /
data PHhothalo(2,:,    214    ) / 2.579E-2_WP,    4.427E-3_WP,    1.027E-3_WP /
data PHhothalo(2,:,    215    ) / 2.298E-2_WP,    5.175E-3_WP,    1.578E-3_WP /
data PHhothalo(2,:,    216    ) / 2.298E-2_WP,    5.175E-3_WP,    1.578E-3_WP /
data PHhothalo(2,:,    217    ) / 2.298E-2_WP,    5.175E-3_WP,    1.578E-3_WP /
data PHhothalo(2,:,    218    ) / 6.569E-2_WP,    1.723E-2_WP,    6.185E-3_WP /
data PHhothalo(2,:,    219    ) / 6.569E-2_WP,    1.723E-2_WP,    6.185E-3_WP /
data PHhothalo(2,:,    220    ) / 1.917E-2_WP,    5.895E-3_WP,    2.012E-3_WP /
data PHhothalo(2,:,    221    ) / 1.917E-2_WP,    5.895E-3_WP,    2.012E-3_WP /
data PHhothalo(2,:,    222    ) / 1.917E-2_WP,    5.895E-3_WP,    2.012E-3_WP /
data PHhothalo(2,:,    223    ) / 1.970E-2_WP,    5.747E-3_WP,    2.109E-3_WP /
data PHhothalo(2,:,    224    ) / 1.970E-2_WP,    5.747E-3_WP,    2.109E-3_WP /
data PHhothalo(2,:,    225    ) / 4.176E-3_WP,    9.519E-4_WP,    3.338E-4_WP /
data PHhothalo(2,:,    226    ) / 4.176E-3_WP,    9.519E-4_WP,    3.338E-4_WP /
data PHhothalo(2,:,    227    ) / 7.660E-3_WP,    2.302E-3_WP,    8.454E-4_WP /
data PHhothalo(2,:,    228    ) / 7.660E-3_WP,    2.302E-3_WP,    8.454E-4_WP /
data PHhothalo(2,:,    229    ) / 4.401E-3_WP,    1.126E-3_WP,    4.059E-4_WP /
data PHhothalo(2,:,    230    ) / 4.401E-3_WP,    1.126E-3_WP,    4.059E-4_WP /
data PHhothalo(2,:,    231    ) / 7.010E-3_WP,    1.551E-3_WP,    5.617E-4_WP /
data PHhothalo(2,:,    232    ) / 8.936E-4_WP,    2.375E-4_WP,    9.213E-5_WP /
data PHhothalo(2,:,    233    ) / 3.488E-3_WP,    9.261E-4_WP,    3.010E-4_WP /
data PHhothalo(2,:,    234    ) / 3.488E-3_WP,    9.261E-4_WP,    3.010E-4_WP /
data PHhothalo(2,:,    235    ) / 1.635E-3_WP,    4.173E-4_WP,    1.516E-4_WP /
data PHhothalo(2,:,    236    ) / 1.635E-3_WP,    4.173E-4_WP,    1.516E-4_WP /
data PHhothalo(2,:,    237    ) / 5.371E-4_WP,    1.457E-4_WP,    5.576E-5_WP /
data PHhothalo(2,:,    238    ) / 9.953E-4_WP,    2.532E-4_WP,    8.888E-5_WP /
data PHhothalo(2,:,    239    ) / 1.590E-3_WP,    4.038E-4_WP,    1.342E-4_WP /
data PHhothalo(2,:,    240    ) / 1.590E-3_WP,    4.038E-4_WP,    1.342E-4_WP /
data PHhothalo(2,:,    241    ) / 1.590E-3_WP,    4.038E-4_WP,    1.342E-4_WP /
data PHhothalo(2,:,    242    ) / 3.533E-3_WP,    8.895E-4_WP,    2.873E-4_WP /
data PHhothalo(2,:,    243    ) / 5.891E-4_WP,    1.868E-4_WP,    5.830E-5_WP /
data PHhothalo(2,:,    244    ) / 5.891E-4_WP,    1.868E-4_WP,    5.830E-5_WP /
data PHhothalo(2,:,    245    ) / 5.891E-4_WP,    1.868E-4_WP,    5.830E-5_WP /
data PHhothalo(2,:,    246    ) / 5.891E-4_WP,    1.868E-4_WP,    5.830E-5_WP /
data PHhothalo(2,:,    247    ) / 2.321E-4_WP,    6.497E-5_WP,    2.343E-5_WP /
data PHhothalo(2,:,    248    ) / 2.321E-4_WP,    6.497E-5_WP,    2.343E-5_WP /
data PHhothalo(2,:,    249    ) / 5.338E-4_WP,    1.427E-4_WP,    4.689E-5_WP /
data PHhothalo(2,:,    250    ) / 2.507E-4_WP,    6.900E-5_WP,    2.335E-5_WP /
data PHhothalo(2,:,    251    ) / 2.507E-4_WP,    6.900E-5_WP,    2.335E-5_WP /
data PHhothalo(2,:,    252    ) / 2.788E-4_WP,    7.887E-5_WP,    2.565E-5_WP /
data PHhothalo(2,:,    253    ) / 7.692E-4_WP,    2.166E-4_WP,    6.635E-5_WP /
data PHhothalo(2,:,    254    ) / 7.692E-4_WP,    2.166E-4_WP,    6.635E-5_WP /
data PHhothalo(2,:,    255    ) / 7.692E-4_WP,    2.166E-4_WP,    6.635E-5_WP /
data PHhothalo(2,:,    256    ) / 9.981E-5_WP,    2.997E-5_WP,    9.900E-6_WP /
data PHhothalo(2,:,    257    ) / 9.981E-5_WP,    2.997E-5_WP,    9.900E-6_WP /
data PHhothalo(2,:,    258    ) / 7.607E-5_WP,    2.230E-5_WP,    7.530E-6_WP /
data PHhothalo(2,:,    259    ) / 1.065E-4_WP,    3.157E-5_WP,    1.002E-5_WP /
data PHhothalo(2,:,    260    ) / 1.065E-4_WP,    3.157E-5_WP,    1.002E-5_WP /
data PHhothalo(2,:,    261    ) / 1.065E-4_WP,    3.157E-5_WP,    1.002E-5_WP /
data PHhothalo(2,:,    262    ) / 4.790E-5_WP,    1.434E-5_WP,    4.769E-6_WP /
data PHhothalo(2,:,    263    ) / 4.790E-5_WP,    1.434E-5_WP,    4.769E-6_WP /
data PHhothalo(2,:,    264    ) / 8.344E-5_WP,    2.633E-5_WP,    7.972E-6_WP /
data PHhothalo(2,:,    265    ) / 8.344E-5_WP,    2.633E-5_WP,    7.972E-6_WP /
data PHhothalo(2,:,    266    ) / 2.975E-5_WP,    9.092E-6_WP,    2.976E-6_WP /
data PHhothalo(2,:,    267    ) / 2.975E-5_WP,    9.092E-6_WP,    2.976E-6_WP /
data PHhothalo(2,:,    268    ) / 2.284E-5_WP,    7.062E-6_WP,    2.296E-6_WP /
data PHhothalo(2,:,    269    ) / 2.284E-5_WP,    7.062E-6_WP,    2.296E-6_WP /
data PHhothalo(2,:,    270    ) / 1.737E-5_WP,    5.433E-6_WP,    1.754E-6_WP /
data PHhothalo(2,:,    271    ) / 1.737E-5_WP,    5.433E-6_WP,    1.754E-6_WP /
data PHhothalo(2,:,    272    ) / 1.545E-5_WP,    4.953E-6_WP,    1.553E-6_WP /
data PHhothalo(2,:,    273    ) / 1.545E-5_WP,    4.953E-6_WP,    1.553E-6_WP /
data PHhothalo(2,:,    274    ) / 1.151E-5_WP,    3.894E-6_WP,    1.170E-6_WP /
data PHhothalo(2,:,    275    ) / 7.273E-6_WP,    2.351E-6_WP,    7.449E-7_WP /
data PHhothalo(2,:,    276    ) / 5.348E-6_WP,    1.746E-6_WP,    5.499E-7_WP /
data PHhothalo(2,:,    277    ) / 3.897E-6_WP,    1.284E-6_WP,    4.021E-7_WP /
data PHhothalo(2,:,    278    ) / 2.812E-6_WP,    9.349E-7_WP,    2.910E-7_WP /
data PHhothalo(2,:,    279    ) / 2.009E-6_WP,    6.732E-7_WP,    2.084E-7_WP /
data PHhothalo(2,:,    280    ) / 1.420E-6_WP,    4.792E-7_WP,    1.475E-7_WP /
data PHhothalo(2,:,    281    ) / 6.859E-7_WP,    2.340E-7_WP,    7.135E-8_WP /
data PHhothalo(2,:,    282    ) / 4.680E-7_WP,    1.604E-7_WP,    4.867E-8_WP /
data PHhothalo(2,:,    283    ) / 3.151E-7_WP,    1.084E-7_WP,    3.274E-8_WP /
data PHhothalo(2,:,    284    ) / 2.092E-7_WP,    7.215E-8_WP,    2.171E-8_WP /

real(WP) :: energy    ! in eV

Fhothalo = 0.0_WP
energy = h*c/lambda/q*10.0_WP    ! in eV
if (energy>=Energies(1) .and. energy<=Energies(nenergies)) then
    Fhothalo = energy*h/c*10.0_WP**LInterpol(log10(Energies),&
                        & log10(PHhothalo(t,r,:)),log10(energy))    ! in erg/cm^3/Hz
end if

end function Fhothalo

end module FieldIntensities
